/*
 * cp40frame.c --
 *
 *      Implement management of photon counting frames and pre-processing
 *      of CP40 data.
 *
 * Copyright (c) 1995-2000 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id: cp40frame.c,v 2.2 2005/06/29 20:26:03 eric Exp $
 *	$Log: cp40frame.c,v $
 *	Revision 2.2  2005/06/29 20:26:03  eric
 *	 - Saved prior to making CP40 as a standalone Yorick package.
 *
 *	Revision 2.1  2003/06/25 14:08:23  eric
 *	 - Version saved before major change.
 *
 *	Revision 2.0  2000/01/20 07:58:01  eric
 *	*** empty log message ***
 *
 *      Revision 1.5  1999/05/31 15:03:03  eric
 *       - Fixed a bug that causes frames with many more or many less photons
 *         than there really are.
 *       - Too many changes to remember.
 *
 *      Revision 1.4  1998/06/08 18:08:02  eric
 *       - Mainly cosmetic changes.
 *       - CP40 Version ~ 4.2 saved before version 4.3.
 *
 *      Revision 1.3  1997/12/05 19:29:07  eric
 *       - Fix error raising when exposure is set when not in read mode.
 *         Now exposure can always be set but its effective value should
 *         be 1 in write/update mode regardless the value of
 *         this->exposure.
 *
 *      Revision 1.2  1997/12/05 19:18:31  eric
 *       - Make some mofifications to avoid problems in case one of these
 *         routines get interrupted.
 *
 *      Revision 1.1  1997/10/31 10:31:40  eric
 *      Initial revision
 *-----------------------------------------------------------------------------
 */

#include <time.h>
#include <math.h>
#include "cp40.h"

/*---------------------------------------------------------------------------*/
/* PRIVATE ROUTINES */

static cp40_select_op_t spurious_leave;
/* Copy raw-photons from FRAME->owner->buffer to FRAME->buffer.  Only copy
   raw-photons that belong to channel FRAME->channel and have a non-null
   weight.  Also perform raw pixel conversion.  This routine must be called
   before any other pre-processing functions.  Return standard result: GPL_OK
   or GPL_ERROR.  In case of error, a message is left that can be retrieved by
   a call to gpl_error_info(). */

static int spurious_delete(cp40_frame_t *frame);
static int spurious_centre(cp40_frame_t *frame);
static int spurious_single(cp40_frame_t *frame);
/* Select photons from FRAME->channel and deal with spurious photons; one of:
    - delete multiple events
    - replace multiple events by their photo-centre
    - keep the first one of multiple events
   Must be called after spurious_leave. */


static int correction_none(cp40_frame_t *frame);
/* Perform no distortion correction, and select photons from their
   positions (i.e. apply "window" settings). */

static int correction_simple(cp40_frame_t *frame);
/* Perform simple distortion correction, and select photons from their
   positions (i.e. apply "window" settings). */

static int correction_random(cp40_frame_t *frame);
/* Perform "random subpixel" distortion correction, and select photons from
   their positions (i.e. apply "window" settings). */

static int correction_weighted(cp40_frame_t *frame);
/* Perform weighted distortion correction, and select photons from
   their positions (i.e. apply "window" settings). */

/*---------------------------------------------------------------------------*/

cp40_spurious_t cp40_get_spurious(cp40_instance_t *this)
{
  if (this->spurious == spurious_delete) return CP40_SPURIOUS_DELETE;
  if (this->spurious == spurious_centre) return CP40_SPURIOUS_CENTRE;
  if (this->spurious == spurious_single) return CP40_SPURIOUS_SINGLE;
  this->spurious= spurious_leave;
  return CP40_SPURIOUS_LEAVE;
}

int cp40_set_spurious(cp40_instance_t *this, cp40_spurious_t value)
{
  switch (value) {
  case CP40_SPURIOUS_LEAVE:  this->spurious = spurious_leave;  break;
  case CP40_SPURIOUS_DELETE: this->spurious = spurious_delete; break;
  case CP40_SPURIOUS_CENTRE: this->spurious = spurious_centre; break;
  case CP40_SPURIOUS_SINGLE: this->spurious = spurious_single; break;
  default: return gpl_error("bad value for \"spurious\"");
  }
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

cp40_correction_t cp40_get_correction(cp40_instance_t *this)
{
  if (this->correction == correction_simple)   return CP40_CORRECTION_SIMPLE;
  if (this->correction == correction_random)   return CP40_CORRECTION_RANDOM;
  if (this->correction == correction_weighted) return CP40_CORRECTION_WEIGHTED;
  if (this->correction == correction_none)     return CP40_CORRECTION_NONE;
  cp40_set_correction(this, CP40_CORRECTION_NONE);
  return CP40_CORRECTION_NONE;
}

int cp40_set_correction(cp40_instance_t *this, cp40_correction_t value)
{
  cp40_select_op_t *func;

  if (this->task) {
    return gpl_error("cannot set \"correction\": scheduled tasks exist");
  }
  switch (value) {
  case CP40_CORRECTION_NONE:     func = correction_none;     break;
  case CP40_CORRECTION_SIMPLE:   func = correction_simple;   break;
  case CP40_CORRECTION_RANDOM:   func = correction_random;   break;
  case CP40_CORRECTION_WEIGHTED: func = correction_weighted; break;
  default: return gpl_error("bad value for \"correction\"");
  }
  if (func != this->correction) {
    this->correction= func;
    _cp40_invalidate_cache(this);
  }
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

cp40_raw_frame_t *cp40_new_raw_frame(void)
{
  cp40_raw_frame_t *frame;
  if (! (frame= gpl_alloc(sizeof *frame))) {
    gpl_error("insufficient memory for new raw frame");
    return NULL;
  }
  frame->count= 0;
  frame->date= -1;
  frame->size= CP40_RAW_FRAME_STATIC_BUFFER_SIZE;
  frame->buffer= frame->static_buffer;
  return frame;
}

void cp40_delete_raw_frame(cp40_raw_frame_t *frame)
{
  if (frame) {
    if (frame->buffer && frame->buffer!=frame->static_buffer) {
      gpl_free(frame->buffer);
    }
    gpl_free(frame);
  }
}

void cp40_reset_raw_frame(cp40_raw_frame_t *frame)
{
  if (frame) {
    frame->count= 0;
    frame->date= -1;
  }
}

int cp40_resize_raw_frame(cp40_raw_frame_t *frame, int new_size)
{
  void *new_buffer, *old_buffer;
  unsigned old_size= frame->size;
  if (new_size < frame->count)
    return gpl_error("cannot resize raw frame below current count");
  if (new_size <= CP40_RAW_FRAME_STATIC_BUFFER_SIZE) {
    new_buffer= frame->static_buffer;
    new_size= CP40_RAW_FRAME_STATIC_BUFFER_SIZE;
  } else if (new_size == old_size) {
    return GPL_OK;
  } else {
    if (! (new_buffer= gpl_alloc(new_size * sizeof *frame->buffer))) {
      return gpl_error("insufficient memory to augment raw photon buffer");
    }
  }
  old_buffer= frame->buffer;
  if (old_buffer != new_buffer) {
    if (new_size < old_size) frame->size= new_size;
    memcpy(new_buffer, old_buffer, frame->count * sizeof *frame->buffer);
    frame->buffer= new_buffer;
    if (old_buffer && old_buffer!=frame->static_buffer) gpl_free(old_buffer);
  }
  frame->size= new_size;
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

void _cp40_init_frame(cp40_frame_t *frame, cp40_instance_t *owner,
		      int channel)
{
  if (frame) {
    (void)memset(frame, 0, sizeof(*frame));
    frame->date= -1;
    frame->channel= channel;
    frame->owner= owner;
  }
}

int _cp40_resize_frame(cp40_frame_t *frame, int size)
{
  cp40_photon_t *new, *old;

  if (size < 0 || size < frame->count) {
    return gpl_error("bag new frame size (must be >= FRAME->count)");
  }
  old = frame->buffer;
  if (size == 0) {
    new= NULL;
  } else {
    if (! (new = GPL_ALLOC_MANY(cp40_photon_t, size))) {
      return gpl_sys_error("cannot (re)alloc frame buffer");
    }
    if (frame->count) {
      memcpy(new, old, frame->count * sizeof(cp40_photon_t));
    }
  }
  frame->buffer = new;
  frame->size = size;
  if (old) gpl_free(old);
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

cp40_frame_t *cp40_get_frame(cp40_instance_t *this, int flags)
{
  int n;
  double w, x, y, sx, sy, sw, s2;
  cp40_photon_t *photon;
  cp40_raw_frame_t *raw_frame = this->raw_frame;
  int channel = flags&3;
  cp40_frame_t *frame = &this->frame[channel];

  /* (0) Minimum checking. */
  if (this->notify) {
    gpl_error("bug in cp40_get_frame: cp40_update_cache() should have been called before");
    frame->status = -1;
    return NULL;
  }

  /* (1) Maybe nothing to do (assume same frame selection rules) */
  if (! (flags & CP40_FORCE_REPROCESSING)
      && frame->date == this->raw_frame->date) {
    return frame;
  }

  /* (2) Check frame parity. */
  if (! (this->parity & ((raw_frame->date & 1) ? CP40_PARITY_ODD
			     : CP40_PARITY_EVEN))) {
    /* No data selected but nevertheless instanciate structure. */
    frame->date = raw_frame->date;
    frame->status = 0;
    frame->count = 0;
    frame->flux = 0.0;
    frame->x = 0.0;
    frame->y = 0.0;
    frame->spread = 0.0;
    frame->distance = 0.0;
    frame->ratio = 0.0;
    return frame;
  }

  /* (3-4) Photon selection and preprocessing. */
  if (this->spurious(frame) != GPL_OK || this->correction(frame) != GPL_OK) {
    frame->status = -1;
    return NULL;
  }

  /* (5) Compute frame statistics. */
  n = frame->count;
  photon = frame->buffer;
  sx = sy = sw = s2 = 0.0;
  while (n--) {
    sw += (w = photon->w);
    sx += w*(x = photon->x);
    sy += w*(y = photon->y);
    s2 += w*(x*x + y*y);
    ++photon;
  }
  if (sw <= 0.0) {
    frame->count = 0; /* shoudn't be needed, but... */
    frame->flux = 0.0;
    frame->x = 0.0;
    frame->y = 0.0;
    frame->spread = 0.0;
    frame->distance = 0.0;
    frame->ratio = 0.0;
  } else {
    frame->flux = sw;
    frame->x = (x = sx/sw);
    frame->y = (y = sy/sw);
    s2= s2/sw - (x*x + y*y);
    frame->distance = x - frame->xmin;
#undef MIN_DISTANCE
#define MIN_DISTANCE(expr) if ((sx=expr)<frame->distance) frame->distance=sx
    MIN_DISTANCE(frame->xmax - x);
    MIN_DISTANCE(y - frame->xmin);
    MIN_DISTANCE(frame->ymax - y);
    if (s2 > 0.0) {
      frame->ratio = frame->distance / (frame->spread = sqrt(s2));
    } else {
      /* I'm not sure this will ever happend... */
      frame->spread = 0.0;
      frame->ratio = 0.0;
    }
  }

  /* (6) Apply frame selection rules. */
#undef OUT_OF_RANGE
#define OUT_OF_RANGE(MEMBER)						      \
  (((this->MEMBER[channel].flags & CP40_MIN_SET) 			      \
    && frame->MEMBER < this->MEMBER[channel].min)			      \
   || ((this->MEMBER[channel].flags & CP40_MAX_SET)			      \
       && frame->MEMBER > this->MEMBER[channel].max))
  if (OUT_OF_RANGE(flux) || OUT_OF_RANGE(distance) || OUT_OF_RANGE(spread)
      || OUT_OF_RANGE(ratio)) {
    frame->status= 0;
  }
  return frame;
}

/*---------------------------------------------------------------------------*/

#define PHOTON_CMSK 0x00800800U /* Bit mask for channel. */
#define PHOTON_WMSK 0xFF000000U /* Bit mask for weight. */

#define PhotonWeight(p)	 ((p)>>24U & 0x0FFU)    /* Weight, shifted. */
#define PhotonX0(p)	 ((p)      & 0x7FFU)	/* Abscissa, 1/4 CCD pixel. */
#define PhotonY0(p)	 ((p)>>12U & 0x7FFU)	/* Ordinate, 1/4 CCD pixel. */
#define PhotonX1(p)	 ((p)>> 1U & 0x3FFU)	/* Abscissa, 1/2 CCD pixel. */
#define PhotonY1(p)	 ((p)>>13U & 0x3FFU)	/* Ordinate, 1/2 CCD pixel. */
#define PhotonX2(p)	 ((p)>> 2U & 0x1FFU)	/* Abscissa, CCD pixel. */
#define PhotonY2(p)	 ((p)>>14U & 0x1FFU)	/* Ordinate, CCD pixel. */

static int spurious_leave(cp40_frame_t *frame)
{
  cp40_photon_t *dst;
  cp40_raw_photon_t *src, cmsk, photon;
  cp40_raw_frame_t *raw_frame = frame->owner->raw_frame;
  unsigned int width, height;
  int count, pixel;

  /* Make sure that the frame buffer is large enough (usually much larger
   * since this->count is the number of photons for *all* channels). */
  if ((count = raw_frame->count) < 0) goto corrupted;
  if (frame->size < count && _cp40_resize_frame(frame, count) != GPL_OK) {
    return GPL_ERROR;
  }

  /* Make channel mask to select only raw-photons from
   * the channel of interest. */
  cmsk = frame->channel;
  cmsk = (cmsk << 22U & 0x00800000U) | (cmsk << 11U & 0x00000800U);

  /* Select photons with non-zero weight and from channel frame->channel. */
  dst = frame->buffer;
  src = raw_frame->buffer;
  pixel = frame->owner->pixel;
  width = CP40_WIDTH(pixel);
  height = CP40_HEIGHT(pixel);
  if (pixel == CP40_PIXEL_PHYSICAL) {
    while (count--) {
      photon = *src++;
      if ((photon & PHOTON_WMSK) && (photon & PHOTON_CMSK) == cmsk) {
	if ((unsigned int)(dst->y = PhotonY2(photon)) >= height ||
	    (unsigned int)(dst->x = PhotonX2(photon)) >= width) goto corrupted;
	dst->w = PhotonWeight(photon);
	dst++;
      }
    }
  } else if (pixel == CP40_PIXEL_HALF_PHYSICAL) {
    while (count--) {
      photon = *src++;
      if ((photon & PHOTON_WMSK) && (photon & PHOTON_CMSK) == cmsk) {
	if ((unsigned int)(dst->y = PhotonY1(photon)) >= height ||
	    (unsigned int)(dst->x = PhotonX1(photon)) >= width) goto corrupted;
	dst->w = PhotonWeight(photon);
	dst++;
      }
    }
  } else if (pixel == CP40_PIXEL_LOGICAL) {
    while (count--) {
      photon = *src++;
      if ((photon & PHOTON_WMSK) && (photon & PHOTON_CMSK) == cmsk) {
	if ((unsigned int)(dst->y = PhotonY0(photon)) >= height ||
	    (unsigned int)(dst->x = PhotonX0(photon)) >= width) goto corrupted;
	dst->w = PhotonWeight(photon);
	dst++;
      }
    }
  } else {
    return gpl_error("unexpected value for \"pixel\" attribute");
  }
  frame->count = dst - frame->buffer;
  frame->date = raw_frame->date;
  return GPL_OK;

 corrupted:
  return gpl_error("corrupted raw photon frame");
}

static int spurious_delete(cp40_frame_t *frame)
{
  cp40_photon_t *last, *p1, *p2;
  int x1, y1, d, dmin;

  /* Select photons with non-zero weight and from channel frame->channel
   * (also make sure that frame->buffer is large enough). */
  if (spurious_leave(frame) != GPL_OK) return GPL_ERROR;

  /* Search for spurious photons and set their weight to zero. */
  dmin = 8 >> frame->owner->pixel;
  p1 = frame->buffer - 1;
  last = p1 + frame->count;
  while (++p1 < last) {
    y1 = p1->y;
    x1 = p1->x;
    p2 = p1;
    while (++p2 <= last) {
      if (((d = p2->y) >= y1 ? d - y1 : y1 - d) < dmin &&
	  ((d = p2->x) >= x1 ? d - x1 : x1 - d) < dmin) {
	p1->w = p2->w = 0;
      }
    }
  }

  /* Delete photons with zero weight, set number of photons and return. */
  p1 = frame->buffer - 1;
  p2 = frame->buffer;
  while (++p1 <= last) {
    if (p1->w) *p2++ = *p1;
  }
  frame->count = p2 - frame->buffer;
  return GPL_OK;
}

static int spurious_centre(cp40_frame_t *frame)
{
  cp40_photon_t *last, *p1, *p2;
  int d, dmin, x1, y1, count;

  /* Select photons with non-zero weight and from channel frame->channel
   * (also make sure that frame->buffer is large enough). */
  if (spurious_leave(frame) != GPL_OK) return GPL_ERROR;

  /* Search for spurious photons. */
  dmin = 8 >> frame->owner->pixel;
  p1 = frame->buffer - 1;
  last = p1 + frame->count;
  while (++p1 < last) {
    double sw = p1->w;
    if (sw) {
      double sy = sw*(y1 = p1->y);
      double sx = sw*(x1 = p1->x);
      p2 = p1;
      count = 1;
      while (++p2 <= last) {
	if (((d = p2->y) >= y1 ? d - y1 : y1 - d) < dmin &&
	    ((d = p2->x) >= x1 ? d - x1 : x1 - d) < dmin) {
	  double w2 = p2->w;
	  if (w2) {
	    sy += w2*p2->y;
	    sx += w2*p2->x;
	    sw += w2;
	    p2->w = 0;
	    ++count;
	  }
	}
      }
      if (count > 1) {
	double scl = 1.0/sw;
	p1->y = floor(scl*sx + 0.5);
	p1->x = floor(scl*sy + 0.5);
	p1->w = sw/count;
      }
    }
  }

  /* Delete photons with zero weight, set number of photons and return. */
  p1 = frame->buffer - 1;
  p2 = frame->buffer;
  while (++p1 <= last) {
    if (p1->w) *p2++ = *p1;
  }
  frame->count = p2 - frame->buffer;
  return GPL_OK;
}

static int spurious_single(cp40_frame_t *frame)
{
  cp40_photon_t *last, *p1, *p2;
  int d, dmin, x1, y1;

  /* Select photons with non-zero weight and from channel frame->channel
   * (also make sure that frame->buffer is large enough). */
  if (spurious_leave(frame) != GPL_OK) {
    return GPL_ERROR;
  }
  /* Search for spurious photons. */
  dmin = 8 >> frame->owner->pixel;
  p1 = frame->buffer - 1;
  last = p1 + frame->count;
  while (++p1 < last) {
    y1 = p1->y;
    x1 = p1->x;
    p2 = p1;
    while (++p2 <= last) {
      if (((d = p2->y) >= y1 ? d - y1 : y1 - d) < dmin &&
	  ((d = p2->x) >= x1 ? d - x1 : x1 - d) < dmin) {
	p2->w = 0;
      }
    }
  }

  /* Delete photons with zero weight, set number of photons and return. */
  p1 = frame->buffer - 1;
  p2 = frame->buffer;
  while (++p1 <= last) {
    if (p1->w) *p2++ = *p1;
  }
  frame->count = p2 - frame->buffer;
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

static int correction_none(cp40_frame_t *frame)
{
  cp40_photon_t *p = frame->buffer;
  int i, j, n = frame->count;
  int xmin= frame->xmin;
  int xmax= frame->xmax;
  int ymin= frame->ymin;
  int ymax= frame->ymax;
  int x, y;

  for (i=j=0 ; i<n ; ++i) {
    if (p[i].w &&
	(x = p[i].x) >= xmin && x <= xmax &&
	(y = p[i].y) >= ymin && y <= ymax) {
      p[j++] = p[i];
    }
  }
  frame->count = j;
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

static int correction_simple(cp40_frame_t *frame)
{
  const double *coefs = frame->owner->coefs + (2*CP40_NCOEFS)*frame->channel;
  cp40_photon_t *p = frame->buffer;
  int i, j, n = frame->count;
  int xmin = frame->xmin;
  int xmax = frame->xmax;
  int ymin = frame->ymin;
  int ymax = frame->ymax;
  int x, y;

  for (i=j=0 ; i<n ; ++i) {
    if (p[i].w) {
      double xx = p[i].x;
      double yy = p[i].y;
      if ((y = CP40_YCOR(coefs, xx, yy)) >= ymin && y <= ymax &&
	  (x = CP40_XCOR(coefs, xx, yy)) >= xmin && x <= xmax) {
	p[j].w = p[i].w;
	p[j].x = x;
	p[j].y = y;
	++j;
      }
    }
  }
  frame->count = j;
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

#define RAN_M           1771875
#define RAN_A              2416
#define RAN_C            374441
#define RAN_INIT        (time(0)%RAN_M)
#define RAN0(i)         (i=(RAN_A*i+RAN_C)%RAN_M)
#define RAN1(TYPE,i)    ((TYPE)(i=(RAN_A*i+RAN_C)%RAN_M)/(TYPE)RAN_M)
#define RAN2(TYPE,i)    ((TYPE)(i=(RAN_A*i+RAN_C)%RAN_M)/(TYPE)RAN_M-(TYPE).5)
#define RANDOM(i)       RAN2(double, i)

static int correction_random(cp40_frame_t *frame)
{
  static int init = 0;
  static unsigned long seed;
  const double *coefs = frame->owner->coefs + (2*CP40_NCOEFS)*frame->channel;
  cp40_photon_t *p = frame->buffer;
  int i, j, n = frame->count;
  int xmin = frame->xmin;
  int xmax = frame->xmax;
  int ymin = frame->ymin;
  int ymax = frame->ymax;
  int x, y;

  if (! init) {
    seed = RAN_INIT;
    init = 1;
  }
  for (i=j=0 ; i<n ; ++i) {
    if (p[i].w) {
      double xx = p[i].x + RANDOM(seed);
      double yy = p[i].y + RANDOM(seed);
      if ((y = CP40_YCOR(coefs, xx, yy)) >= ymin && y <= ymax &&
	  (x = CP40_XCOR(coefs, xx, yy)) >= xmin && x <= xmax) {
	p[j].w = p[i].w;
	p[j].x = x;
	p[j].y = y;
	++j;
      }
    }
  }
  frame->count = j;
  return GPL_OK;
}
#undef RAN_M
#undef RAN_A
#undef RAN_C
#undef RAN_INIT
#undef RAN0
#undef RAN1
#undef RAN2
#undef RANDOM

/*---------------------------------------------------------------------------*/
/* WEIGHTED DISTORTION CORRECTION */

typedef struct _wght_table   wght_table_t;
typedef struct _wght_element wght_element_t;

#ifndef WGHT_NDIV
# define WGHT_NDIV 15 /* Number of raw pixel subdivisions along each axis. */
#endif
#define WGHT_SPREAD  1 /* Maximum allowed spread in undistorted pixel units. */
#define WGHT_STRIDE (2*WGHT_SPREAD + 1)

static wght_table_t *wght_new(cp40_instance_t *this, int channel);
static void          wght_delete(void *wght);

struct _wght_table {
  /* This structure is used  to store data for weighted distortion correction.
     In this case, each raw-event is replaced by, possibly several, sub-events
     with   different  undistorted   coordinates  and   weights.   Let  I0  be
     INDEX[raw_x + WIDTH*raw_y], then:
       ELEM[i0 + i].w      weight   of i-th sub-event
       ELEM[i0 + i].x      abscissa of i-th sub-event
       ELEM[i0 + i].y      ordinate of i-th sub-event
     for i=0...NSUB-1, with NSUB the number of sub-events given by:
       NSUB = INDEX[raw_x+WIDTH*raw_y + 1] - INDEX[raw_x+WIDTH*raw_y] */
  int width, height;    /* Raw CCD dimension must match current raw "pixel"
			   settings. */
  int ndiv;             /* Number of raw pixel subdivisions along each axis. */
  int nout;             /* Number of too large raw pixels. */
  int reweighted;       /* Has been re-weighted? */
  cp40_pixel_t pixel;   /* Raw pixel resolution. */
  wght_element_t *elem; /* Correction data: position and weights. */
  int  index[1];        /* A WIDTH*HEIGHT+1 table of offsets in member
			   array DATA (*MUST* be last member, since
			   real size is not 1 but WIDTH*HEIGHT+1). */
};

struct _wght_element {
  double w;
  int x, y;
};

static int correction_weighted(cp40_frame_t *frame)
{
  /* Static buffer. */
  static cp40_photon_t *buffer = NULL;
  static int size = 0;

  /* Other variables. */
  cp40_photon_t *dst;
  const cp40_photon_t *src;
  const wght_table_t *wght;
  const wght_element_t *wght_elem;
  const int *wght_index;
  int k;
  int wght_width;
  int new_size;
  int pending; /* number of remaining events to process */
  
  /* Get number of events to process. */
  pending = frame->count;
  frame->count = 0;
  if (pending <= 0) return GPL_OK; /* Nothing to do. */

  /* Make sure the buffer is large enough (a maximum of WGHT_STRIDE**2
     sub-event per event is expected). */
  if (! buffer || size <= 0) {
    /* Cope with interrupts (i.e. it is better to never
       free memory than to free memory twice). */
    buffer = NULL;
    size = 0;
  }
  if (size < (new_size = (WGHT_STRIDE*WGHT_STRIDE)*pending)) {
    size = 0;
    if (buffer) gpl_free(buffer);
    if ((buffer = gpl_alloc(new_size*sizeof(*buffer))) == NULL) {
      return gpl_error("not enough memory to apply weighted correction");
    }
    size = new_size;
    memset(buffer, 0, new_size*sizeof(*buffer));
  }
  
  /* Swap buffers -- I don't even try to be safe against interrupts,
     since optimizer would probably break this attempt ;-( */
  {
    int            tmp_size;
    cp40_photon_t *tmp_buffer;
#define ROLL(a,b,c) a=b; b=c; c=a
    ROLL(tmp_size,   frame->size,   size);
    ROLL(tmp_buffer, frame->buffer, buffer);
#undef ROLL
  }

  /* Get table of correction. */
  if ((wght = (wght_table_t *)frame->owner->wght[frame->channel]) == NULL &&
      (wght = wght_new(frame->owner, frame->channel)) == NULL) {
    return GPL_ERROR;
  }
  wght_elem = wght->elem;
  wght_index = wght->index;
  wght_width = wght->width; /* stride */
  
  /* Apply correction. */
  dst = frame->buffer;
  src = buffer;
  for (k=0 ; k<pending ; ++k) {
    double w = src[k].w;
    if (w) {
      int i, j = src[k].x + wght_width*src[k].y;
      int i0 = wght_index[j];
      int i1 = wght_index[j+1];
      for (i=i0 ; i<i1 ; ++i) {
	dst->w = w*wght_elem[i].w;
	dst->x = wght_elem[i].x;
	dst->y = wght_elem[i].y;
	++dst;
      }
    }
  }

  /* Update frame information and return. */
  frame->count = dst - frame->buffer;
  return GPL_OK;
}

static wght_table_t *wght_new(cp40_instance_t *this, int channel)
{
  int *index;
  wght_table_t *wght= NULL;
  wght_element_t  *elem;
  int elem_count, elem_index;
  size_t size;
  int width =  CP40_WIDTH(this->pixel);
  int height = CP40_HEIGHT(this->pixel);
  int i, j;
  const int ndiv = WGHT_NDIV, stride = WGHT_STRIDE;
  const double *coef = this->coefs + channel*(2*CP40_NCOEFS);
  const cp40_frame_t *frame = &this->frame[channel];
  double x0, y0, div[WGHT_NDIV];
  int sum_buf[WGHT_STRIDE*WGHT_STRIDE];

  if (this->notify) {
    gpl_error("bug in wght_new: cp40_update_cache() should have been called before");
    return NULL;
  }

  /* Allocates memory for the table of correction and initializes it
     (note: the wght->data is purposely allocated at maximimum size). */
  if ((wght = this->wght[frame->channel]) != NULL) {
    this->wght[frame->channel] = NULL;
    wght_delete(wght);
  }
  size= GPL_OFFSET_OF(wght_table_t, index)
    + (width*height + 1)*sizeof(*wght->index);
  if ((wght= gpl_alloc(size)) == NULL) goto not_enough_memory;
  memset(wght, 0, size);
  wght->width = width;
  wght->height = height;
  wght->pixel = this->pixel;
  wght->ndiv = ndiv;
  wght->nout = 0;
  wght->reweighted = 0;
  size = (elem_count = stride*stride*width*height)*sizeof(*elem);
  if ((wght->elem = gpl_alloc(size)) == NULL) goto not_enough_memory;
  memset(wght->elem, 0, size);

  /* Instanciates subdivision list. */
  for (i=0 ; i<ndiv ; ++i) {
    div[i] = (double)((1 - ndiv) + 2*i)/(double)(2*ndiv);
  }

  /* Clear sum. */
  memset(sum_buf, 0, sizeof(sum_buf));

  /* Loops over pixels. */
  index = wght->index;
  elem = wght->elem;
  *index++ = (elem_index = 0);
  for (y0=0 ; y0<height ; ++y0) {
    for (x0=0 ; x0<width ; ++x0) {
      /* Compute undistorted coordinates of centre of raw pixel. */
      int xp = CP40_XCOR(coef, x0, y0);
      int yp = CP40_YCOR(coef, x0, y0);

      /* Compute bounding box of undistorted pixel. */
      int xmin = xp - WGHT_SPREAD;
      int xmax = xp + WGHT_SPREAD;
      int ymin = yp - WGHT_SPREAD;
      int ymax = yp + WGHT_SPREAD;
      
      /* Compute bounding of intersection of undistorted raw pixel and
	 window. */
      if (xmin < frame->xmin) xmin = frame->xmin;
      if (xmax > frame->xmax) xmax = frame->xmax;
      if (ymin < frame->ymin) ymin = frame->ymin;
      if (ymax > frame->ymax) ymax = frame->ymax;

      /* Skip this pixel if it is completly outside window. */
      if (xmin <= xmax && ymin <= ymax) {
	/* Integrate image of undistorted raw pixel within the bounding box.
	   Note: real->integer conversion behaves differently than floor()
	   around value zero, in order to avoid this bias, I directly work
	   with coordinates, i.e. not with offsets with respect to
	   (XP0,YP0). */
	const double nrm = 1.0/(WGHT_NDIV*WGHT_NDIV);
	int nout = 0;
	int *sum = sum_buf - xmin - stride*ymin;

	/* Integrate raw pixel into undistorted pixel grid. */
	for (j=0 ; j<ndiv ; ++j) {
	  /* Compute factorized coefficients. */
#if (CP40_NCOEFS != 10)
# error "bad value for CP40_NCOEFS (must be 10)"
#endif
	  double y = y0 + div[j];
	  double cx0 = ((coef[9]*y + coef[5])*y + coef[2])*y + coef[0];
	  double cx1 = (coef[8]*y + coef[4])*y + coef[1];
	  double cx2 = coef[7]*y + coef[3];
	  double cx3 = coef[6];
	  double cy0 = ((coef[19]*y + coef[15])*y + coef[12])*y + coef[10];
	  double cy1 = (coef[18]*y + coef[14])*y + coef[11];
	  double cy2 = coef[17]*y + coef[13];
	  double cy3 = coef[16];
	  for (i=0 ; i<ndiv ; ++i) {
	    double x = x0 + div[i];
	    if ((xp = ((cx3*x + cx2)*x + cx1)*x + cx0) >= xmin && xp <= xmax &&
		(yp = ((cy3*x + cy2)*x + cy1)*x + cy0) >= ymin && yp <= ymax) {
	      ++sum[xp + stride*yp];
	    } else {
	      ++nout;
	    }
	  }
	}
	if (nout) ++wght->nout;	  

	/* Instanciate correction for this raw pixel. */
	for (yp=ymin ; yp<=ymax ; ++yp) {
	  for (xp=xmin ; xp<=xmax ; ++xp) {
	    i = xp + stride*yp;
	    if (sum[i]) {
	      elem[elem_index].w = nrm*sum[i];
	      elem[elem_index].x = xp;
	      elem[elem_index].y = yp;
	      ++elem_index;
	      sum[i] = 0;
	    }
	  }
	}
      }
      *index++ = elem_index;
    }
  }
  if (wght->nout) {
    int ntot = (frame->xmax - frame->xmin + 1)*(frame->ymax - frame->ymin + 1);
    gpl_warning("%d clipped pixels out of %d pixels for channel %d",
		wght->nout, ntot, channel);
  }

  /* Resize table if it is too large by more than 25% (ignore alloc error). */ 
  if (elem_count > 1.25*elem_index) {
    wght_element_t *old = wght->elem;
    if (elem_index > 0) {
      size = elem_index*sizeof(*elem);
      if ((elem = gpl_alloc(size)) != NULL) {
	memcpy(elem, old, size);
	wght->elem = elem;
      }
    } else {
      gpl_warning("no undistorted pixels!");
      wght->elem = NULL;
    }
    if (old) gpl_free(old);
  }
  return (this->wght[frame->channel]= wght);

 not_enough_memory:
  wght_delete(wght);
  gpl_error("insuficient memory to store weighted correction table");
  return NULL;
}

static void wght_delete(void *ptr)
{
  if (ptr) {
    GPL_FREE_MAYBE(((wght_table_t *)ptr)->elem);
    gpl_free(ptr);
  }
}

int cp40_reweight_table(cp40_instance_t *this, int channel,
			const double weight[],
			int width, int height)
{
  const double zero = 0.0, one = 1.0;
  wght_table_t *wght;
  wght_element_t *elem;
  int *index;
  double scl;
  int i, n, j, j1, j0;

  /* Make sure coefficients are up to date. */
  if (this->notify) cp40_update_cache(this);

  /* Check arguments. */
  if (this->correction != correction_weighted)
    return gpl_error("distortion correction method must be \"weighted\"");
  if (cp40_check_channel(channel) != GPL_OK) return GPL_ERROR;
  if (width != CP40_WIDTH(this->pixel) || height != CP40_HEIGHT(this->pixel))
    return gpl_error("bad dimensions for array of weights");
  n = width*height;
  for (i=0 ; i<n ; ++i) {
    if (weight[i] < zero) return gpl_error("weights must be non-negative");
  }

  /* Get table of correction (create a new one if already re-weighted). */
  wght = (wght_table_t *)this->wght[channel];
  if (! wght || wght->reweighted) {
    wght = wght_new(this, channel);
    if (! wght) return GPL_ERROR;
  }
  elem = wght->elem;
  index = wght->index;

  /* Re-weight the table. */
  j0 = index[0];
  for (i=0 ; i<n ; ++i) {
    j1 = index[i+1];
    if ((scl = weight[i]) != one) {
      for (j=j0 ; j<j1 ; ++j) {
	elem[j].w *= scl;
      }
    }
    j0 = j1;
  }
  wght->reweighted = 1;
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

void _cp40_invalidate_cache(cp40_instance_t *this)
{
  int channel;

  for (channel=0 ; channel<CP40_NCHANNELS ; ++channel) {
    void *wght = this->wght[channel];
    if (wght) {
      this->wght[channel] = NULL;
      wght_delete(wght);
    }
  }
  this->notify = 1;
}

/*---------------------------------------------------------------------------*/
