/*
 * cp40task.c --
 *
 *	Management of real-time integration tasks for CP40.
 *
 * Copyright (c) 1999-2000 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id: cp40task.c,v 1.2 2005/06/29 20:31:48 eric Exp $
 *	$Log: cp40task.c,v $
 *	Revision 1.2  2005/06/29 20:31:48  eric
 *	 - Saved prior to making CP40 as a standalone Yorick package.
 *
 *	Revision 1.1  2003/06/25 14:08:35  eric
 *	Initial revision
 *
 *-----------------------------------------------------------------------------
 */

#include <unistd.h>
#include "cp40.h"

/* Loop over each CP40 channels. */
#undef EACH_CHANNEL
#define EACH_CHANNEL(var) (var=0 ; var<CP40_NCHANNELS ; var++)

/*---------------------------------------------------------------------------*/
/* PRIVATE ROUTINES */
static int task_fov(cp40_task_t *task);
static int task_it(cp40_task_t *task);
static int task_ac(cp40_task_t *task);
static int task_ic(cp40_task_t *task);
/*	Implement photon-counting integration tasks: `it' (integration),
	`ac' (autocorrelation), or `ic' (intercorrelation).  Return standard
	result: GPL_OK or GPL_ERROR (which could only be due to a bug). */

static void incr_fov(cp40_task_pixel_t *fov, cp40_raw_frame_t *raw_frame,
		     int sign);
/*	Add (SIGN>=0) or subtract (SIGN<0) valid photons from RAW_FRAME in
	field-of-view image FOV. */

static cp40_task_t *get_task(cp40_instance_t *instance, const char *token);
/*	Same as public version cp40_get_task except that no error message
	are stored. */

static const char *format_token(const char *token);
/*	Return static string (maximum length is 31 characters) with:
          <NULL>  if TOKEN==NULL;
	  "TOK[...]" if TOKEN is too long;
	  "TOKEN" otherwise.
	This routine is intended for printing error message. */

/*---------------------------------------------------------------------------*/

void cp40_clear_task(cp40_task_t *task)
{
  if (task) {
    memset(task->data, 0,
	   task->width*task->height*sizeof(cp40_task_pixel_t));
  }
}
void cp40_clear_all_tasks(cp40_instance_t *instance)
{
  if (instance) {
    cp40_task_t *task= instance->task;
    while (task) {
      cp40_clear_task(task);
      task= task->next;
    }
  }
}

void cp40_stop_task(cp40_task_t *task)  { if (task) task->run = 0; }

void cp40_stop_all_tasks(cp40_instance_t *instance)
{
  if (instance) {
    cp40_task_t *task= instance->task;
    while (task) {
      task->run= 0;
      task= task->next;
    }
  }
}

void cp40_start_task(cp40_task_t *task) { if (task) task->run = 1; }

void cp40_start_all_tasks(cp40_instance_t *instance)
{
  if (instance) {
    cp40_task_t *task= instance->task;
    while (task) {
      task->run = 1;
      task= task->next;
    }
  }
}

void cp40_cancel_all_tasks(cp40_instance_t *instance)
{
  if (instance) {
    cp40_task_t *task = instance->task;
    while (task) {
      cp40_task_t *next = task->next;
      cp40_cancel_task(task);
      task = next;
    }
  }
}

void cp40_cancel_task(cp40_task_t *task)
{
  if (task) {
    cp40_instance_t *owner= task->owner;

    /* Remove task from the chain of scheduled tasks. */
    if (owner->task==task) {
      owner->task= task->next;
    } else {
      cp40_task_t *prev= owner->task;
      while (prev->next != task) {
	prev= prev->next;
      }
      prev->next= task->next;
    }

    /* Free task resources. */
    gpl_free_2d_index(task->pixel, sizeof(task->pixel[0][0]),
		    task->xmin, task->xmax, task->ymin, task->ymax);
    gpl_free(task);
  }
}

static cp40_task_t *get_task(cp40_instance_t *instance, const char *token)
{
  if (instance && token) {
    cp40_task_t *task= instance->task;
    int c= token[0];
    while (task) {
      if (task->token[0]==c && strcmp(task->token, token)==0) return task;
      task= task->next;
    }
  }
  return NULL;
}

cp40_task_t *cp40_get_task(cp40_instance_t *instance, const char *token)
{
  cp40_task_t *task= get_task(instance, token);
  if (task) return task;
  gpl_error("no such task %s", format_token(token));
  return NULL;
}

int cp40_save_task(cp40_task_t *task, const char *name, int flags)
{
  return gpl_error("cp40_save_task: not yet implemented");
#if 0
  int result= GPL_OK;
  FILE *file= NULL;
  gpl_dstr_t dstr;
  
  gpl_dstr_init(&dstr);
  if (flags & CP40_USE_NAME_AS_PREFIX) {
    if (gpl_dstr_append(&dstr, name, -1) != GPL_OK
	|| gpl_dstr_append_char(&dstr, '-') != GPL_OK
	|| gpl_dstr_append(&dstr, task->token, -1) != GPL_OK
	|| gpl_dstr_append(&dstr, ".fits", -1) != GPL_OK) {
      result= GPL_ERROR;
      goto done;
    }
    name= gpl_dstr_value(&dstr);
  }
  if (! (flags & CP40_MAY_OVERWRITE) && access(name, F_OK)==0) {
    result= gpl_error("file already exists");
    goto done;
  }
  if (! (file= fopen(name, "w"))) {
    result= gpl_error("error opening output file");
    goto done;
  }
  result= tim_fits_write_image(file, task->image);
 done:
  if (file) fclose(file);
  gpl_dstr_free(&dstr);
  return result;
#endif
}

static const char *format_token(const char *token)
{
  static char buf[32];
  int len;
  if (! token) return "<NULL>";
  buf[0]= '"';
  if ((len= strlen(token)) <= sizeof(buf)-3) {
    memcpy(buf+1, token, len);
    buf[len+1]= '"';
    buf[len+2]= 0;
  } else {
    len= sizeof(buf)-8;
    memcpy(buf+1, token, len);
    memcpy(buf+1+len, "[...]\"", 7);
  }
  return buf;
}
  

/*---------------------------------------------------------------------------*/

cp40_task_t *cp40_schedule_task(cp40_instance_t *owner, const char *token)
{
  cp40_task_t *task= NULL;
  cp40_task_op_t *op= NULL;
  int chn1, chn2;
  int size, width, height, xmin, xmax, ymin, ymax;
  cp40_frame_t *frame1, *frame2;

  /* If the task already exists, just (re)start it. */
  if ((task= get_task(owner, token))!=NULL) goto start;

  /* Check task token and existence of readable stream. */
  if (! (op= cp40_parse_task_token(token, &chn1, &chn2))) return NULL;
  if (! cp40_readable(owner)) {
    gpl_error("no stream to read photon-counting data");
    return NULL;
  }
  
  /* If no task are already scheduled, update preprocessing settings NOW
     (since this update will be forbidden after). */
  if (! owner->task) {
    CP40_UPDATE_CACHE_MAYBE(owner);
  }
  
  /* Get input frame and result limits for the task. */
  frame1= chn1>=0 ? &owner->frame[chn1] : NULL;
  frame2= chn2>=0 ? &owner->frame[chn2] : NULL;
  if (op == task_it) {
    xmin= frame1->xmin;
    xmax= frame1->xmax;
    ymin= frame1->ymin;
    ymax= frame1->ymax;
  } else if (op == task_fov) {
    xmin= 0;
    xmax= CP40_FOV_WIDTH-1;
    ymin= 0;
    ymax= CP40_FOV_HEIGHT-1;
  } else if (op == task_ac) {
    xmin= frame1->xmin - frame1->xmax;
    xmax= frame1->xmax - frame1->xmin;
    ymin= frame1->ymin - frame1->ymax;
    ymax= frame1->ymax - frame1->ymin;
  } else if (op == task_ic) {
    xmin= frame2->xmin - frame1->xmax;
    xmax= frame2->xmax - frame1->xmin;
    ymin= frame2->ymin - frame1->ymax;
    ymax= frame2->ymax - frame1->ymin;
  } else {
    gpl_error("(BUG) unexpected task operation");
    return NULL;
  }

    
  /* Allocate and instanciate task structure. */ 
  width= xmax-xmin+1;
  height= ymax-ymin+1;
#if 0 /* REAL_TIME_DEBUG */
  fprintf(stderr, "scheduling task \"%s\" (width=%d, height=%d)\n",
	 token, width, height);
#endif
  size= GPL_OFFSET_OF(cp40_task_t, data)
    + width*height*sizeof(cp40_task_pixel_t);
  if ((task= cp40_alloc_clear(size, "task structure")) == NULL) {
    return NULL;
  }
  if (gpl_alloc_2d_index(&task->pixel, sizeof(task->pixel[0][0]),
			 task->data, xmin, xmax, ymin, ymax) != GPL_OK) {
    gpl_free(task);
    gpl_error("insufficient memory for task (2D index)");
    return NULL;
  }
  task->owner= owner;
  strcpy(task->token, token);
  task->op= op;
  task->run= 0;
  task->chn1= chn1;
  task->chn2= chn2;
  task->width= width;
  task->height= height;
  task->xmin= xmin;
  task->xmax= xmax;
  task->ymin= ymin;
  task->ymax= ymax;

  /* Insert newly created task into chain of scheduled tasks. */
  task->next= owner->task;
  owner->task= task;

  /* (Re)start the scheduled task. */
 start:
  cp40_start_task(task);
  return task;
}

cp40_task_op_t *cp40_parse_task_token(const char *token, int *chn1, int *chn2)
{
  int c1, c2;

  if (chn1) *chn1= -1;
  if (chn2) *chn2= -1;
  if (token) {
    if (token[0]=='a') {
      if (token[1]=='c' && (unsigned)(c1= token[2]-'0')<4 && token[3]==0) {
	if (chn1) *chn1= c1;
	return task_ac;
      }
    } else if (token[0]=='i') {
      if (token[1]=='t') {
	if ((unsigned)(c1= token[2]-'0')<4 && token[3]==0) {
	  if (chn1) *chn1= c1;
	  return task_it;
	}
      } else if (token[1]=='c') {
	if ((unsigned)(c1= token[2]-'0')<4 && (unsigned)(c2= token[3]-'0')<4
	    && token[4]==0) {
	  if (chn1) *chn1= c1;
	  if (chn2) *chn2= c2;
	  return task_ic;
	}
      }
    } else if (token[0]=='f') {
      if (token[1]=='o' && token[2]=='v' && token[3]==0) {
	return task_fov;
      }
    }
  }
  gpl_error("bad task token %s", format_token(token));
  return NULL;
}

char *cp40_print_task_token(char *token, const char *task, int chn1, int chn2)
{
  if ((token[0]= task[0])!=0 && (token[1]= task[1])!=0 && task[2]==0) {
    if (token[0]=='a') {
      if (token[1]=='c' && (unsigned)chn1<4) {
	token[2]= chn1+'0';
	token[3]= 0;
	return token;
      }
    } else if (token[1]=='c') {
      if ((unsigned)chn1<4 && (unsigned)chn2<4) {
	token[2]= chn1+'0';
	token[3]= chn2+'0';
	token[4]= 0;
	return token;
      }
    } else if (token[1]=='t') {
      if ((unsigned)chn1<4) {
	token[2]= chn1+'0';
	token[3]= 0;
	return token;
      } else {
	token[2]= 0;
	return token;
      }
    }
  }
  gpl_error("bad task or channel(s)");
  return NULL;
}

/* BUG: All channel get pre-processed not only the ones required
        This should be avoided if no statistics are required. */
int cp40_run_tasks(cp40_instance_t *instance, int count)
{
  int nframes= 0, any, chn, need[4];
  cp40_frame_t *frame;
  cp40_task_t *task;
  cp40_stat_t *stat;

  if (count==0) return 0;

  if (instance->stat_size <= 0 || ! instance->stat) {
    instance->stat= NULL;
    instance->stat_size= 0;
    instance->stat_index= 0;
    stat= NULL;
  } else {
    if ((unsigned)instance->stat_index >= instance->stat_size) {
      instance->stat_index= 0;
    }
    stat= &instance->stat[instance->stat_index];
  }
  
  /* Figure out which channels are needed. */
  if (stat) {
    /* All channels will be preprocessed. */
    for EACH_CHANNEL(chn) {
      need[chn]= 1;
    }
  } else {
    /* Only channels required by running tasks will be preprocessed. */
    for EACH_CHANNEL(chn) {
      need[chn]= 0;
    }
    for (task=instance->task ; task!=NULL ; task=task->next) {
      if (! task->run) continue;
      if ((chn= task->chn1) >= 0) need[chn]= 1;
      if ((chn= task->chn2) >= 0) need[chn]= 1;
    }
  }

  for (;;) {
    /* Read next frame and pre-process it for each channel. */
    if (cp40_read_frame(instance) <= 0) {
      return nframes; /* No more frame or error reading data: return number
			 of frames processed so far. */
    }

    any= 0;
    if (stat) {
      /* Roll statistics. */
      if (++instance->stat_index >= instance->stat_size) {
	instance->stat_index= 0;
      }
      stat= &instance->stat[instance->stat_index];

      /* Pre-process every channel and integrate real-time statistics. */
      for EACH_CHANNEL(chn) {
	if (! (frame= cp40_get_frame(instance, chn))) goto io_error;
	if (frame->flux > 0.0) {
	  any= 1;
	  stat->flux[chn]= frame->flux;
	  stat->x[chn]= frame->x;
	  stat->y[chn]= frame->y;
	  stat->spread[chn]= frame->spread;
	} else {
	  stat->flux[chn]= 0.0;
	  stat->x[chn]= 0.0;
	  stat->y[chn]= 0.0;
	  stat->spread[chn]= 0.0;
	}
      }
    } else if (instance->task) {
      /* Only pre-process needed channels. */
      for EACH_CHANNEL(chn) {
	if (! need[chn]) continue;
	if (! (frame= cp40_get_frame(instance, chn))) goto io_error;
	if (frame->flux > 0.0) any= 1;
      }
    }

    /* Execute every registered real-time operations. */
    if (any) {
      for (task=instance->task ; task!=NULL ; task=task->next) {
	if (! task->run) continue;
	if (task->op(task) != GPL_OK) goto io_error;
      }
      ++nframes;
    }

    /* Exchange raw frame buffers (WARNING: Not safe against interrupts but
       try to ignore errors of minor importance). */
    if (stat && (stat->raw_frame || (stat->raw_frame= cp40_new_raw_frame()))) {
      cp40_raw_frame_t *swap= stat->raw_frame;
      stat->raw_frame= instance->raw_frame;
      instance->raw_frame= swap;
      swap->count= 0;
      swap->date= -1;
    }

    /* Check for normal termination. */
    if (count>=0 && nframes>=count) return nframes;
  }

 io_error:
  /* Consider that an input/output error occurred and return number
     of frames processed so far. */
  instance->stream.error= 1;
  return nframes;
}

int cp40_average_stat(cp40_instance_t *instance, cp40_avg_stat_t avg[4])
{
  int result= 0, index, channel;

  if (! instance || ! instance->stat || instance->stat_size<=0) {
    memset(avg, 0, sizeof avg);
    return 0;
  }
  for EACH_CHANNEL(channel) {
    double w, sw= 0.0, swx= 0.0, swy= 0.0, sws= 0.0;
    int sample_size= 0;
    for (index=0 ; index<instance->stat_size ; index++) {
      cp40_stat_t *stat= &instance->stat[index];
      if ((w= stat->flux[channel]) > 0.0) {
	sw+= w;
	swx+= w * stat->x[channel];
	swy+= w * stat->y[channel];
	sws+= w * stat->spread[channel];
	sample_size+= 1;
      }
    }
    if (sample_size>0) {
      avg[channel].nframes= sample_size;
      avg[channel].mean_flux= sw/sample_size;
      avg[channel].mean_x= swx/sw;
      avg[channel].mean_y= swy/sw;
      avg[channel].mean_spread= sws/sw;
      if (result < sample_size) result= sample_size;
    } else {
      avg[channel].nframes= 0;
      avg[channel].mean_flux= 0.0;
      avg[channel].mean_x= 0.0;
      avg[channel].mean_y= 0.0;
      avg[channel].mean_spread= 0.0;
    }
  }
  return result;
}

int cp40_set_stat_duration(cp40_instance_t *instance, int size)
{
  cp40_task_t *fov;

  if (instance->stat_size == size) return GPL_OK;

  /* Free all statistics resources. */
  instance->stat_index= 0;
  if (instance->stat) {
    int i, stat_size= instance->stat_size;
    cp40_stat_t *stat= instance->stat;
    instance->stat= NULL;
    instance->stat_size= 0;
    for (i=0 ; i<stat_size ; i++)
      cp40_delete_raw_frame(stat[i].raw_frame);
    gpl_free(stat);
  } else {
    instance->stat_size= 0;
  }

  if (size<=0) {
    return GPL_OK;
  }

  /* Allocate new resources and try to schedule "field-of-view" task. */
  if (! (instance->stat= cp40_alloc_clear(size*sizeof(*instance->stat),
					  "real-time statistics"))) {
    return GPL_ERROR;
  }
  instance->stat_size= size;
  if (! (fov= cp40_schedule_task(instance, "fov"))) {
    gpl_warning("RUNTIME ERROR: %s\n", gpl_error_info(1));
  } else {
    cp40_clear_task(fov);
    cp40_start_task(fov);
  }
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

static int task_it(cp40_task_t *task)
{
  int count;
  cp40_frame_t *frame;
  
  if (! (frame= cp40_get_frame(task->owner, task->chn1))) {
    return GPL_ERROR;
  }
  if ((count= frame->count)>0) {
    cp40_photon_t *photon= frame->buffer;
    cp40_task_pixel_t **pixel= task->pixel;
    while (count--) {
      pixel[photon->y][photon->x] += photon->w;
      ++photon;
    }
  }
  return GPL_OK;
}

static int task_ac(cp40_task_t *task)
{
  cp40_frame_t *frame;
  int i1, i2, n;
  if (! (frame = cp40_get_frame(task->owner, task->chn1))) {
    return GPL_ERROR;
  }
  if ((n = frame->count - 1) > 0) {
    cp40_photon_t *p = frame->buffer;
    cp40_task_pixel_t **pixel = task->pixel;
    for (i1=0 ; i1<n ; ++i1) {
      double w1 = p[i1].w;
      int y1 = p[i1].y;
      int x1 = p[i1].x;
      for (i2=i1+1 ; i2<=n ; ++i2) {
	double w1w2 = w1*p[i2].w;
	int x = p[i2].x - x1;
	int y = p[i2].y - y1;
	pixel[y][x] += w1w2;
	pixel[-y][-x] += w1w2;
      }
    }
  }
  return GPL_OK;
}

static int task_ic(cp40_task_t *task)
{
  cp40_frame_t *frame1, *frame2;
  int i1, count1, i2, count2;

  if (! (frame1 = cp40_get_frame(task->owner, task->chn1)) ||
      ! (frame2 = cp40_get_frame(task->owner, task->chn2))) {
    return GPL_ERROR;
  }
  if ((count1 = frame1->count) > 0 && (count2 = frame2->count) > 0) {
    cp40_photon_t *p1 = frame1->buffer;
    cp40_photon_t *p2 = frame2->buffer;
    cp40_task_pixel_t **pixel = task->pixel;
    for (i1=0 ; i1<count1 ; ++i1) {
      double w1 = p1[i1].w;
      int y1 = p1[i1].y;
      int x1 = p1[i1].x;
      for (i2=0 ; i2<count2 ; ++i2) {
	pixel[p2[i2].y - y1][p2[i2].x - x1] += w1*p2[i2].w;
      }
    }
  }
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/

#define FOV_WIDTH       CP40_FOV_WIDTH
#define FOV_HEIGHT      CP40_FOV_HEIGHT
#define FOV_HALF_WIDTH  (FOV_WIDTH/2)
#define FOV_HALF_HEIGHT (FOV_HEIGHT/2)
#define FOV_CHN_MSK     0x00800800             /* Bit mask for channel. */
#define FOV_CHN_0       0x00000000
#define FOV_CHN_1       0x00000800
#define FOV_CHN_2       0x00800000
#define FOV_CHN_3       0x00800800
#define FOV_X(p)        (((p) & 0x000007ff) >> CP40_FOV_REDUCTION)
#define FOV_Y(p)        (((p) & 0x007ff000) >> (12+CP40_FOV_REDUCTION))

static void incr_fov(cp40_task_pixel_t *fov, cp40_raw_frame_t *raw_frame,
		     int sign)
{
  int i, x, y, nphotons = raw_frame->count;
  int wght = sign<0 ? -1 : 1;
  cp40_raw_photon_t p, *photons = raw_frame->buffer;

#define FOV_ROTATE_0 0
#define FOV_ROTATE_1 0
#define FOV_ROTATE_2 1
#define FOV_ROTATE_3 1
  for (i=0 ; i<nphotons ; i++) {
    p = photons[i];
    x = FOV_X(p);
    y = FOV_Y(p);
    if (x >= FOV_HALF_WIDTH || y >= FOV_HALF_HEIGHT) continue;
    switch ((p & FOV_CHN_MSK)) {
    case FOV_CHN_0:
#if FOV_ROTATE_0
      x = (FOV_HALF_WIDTH-1) - x;
      y = (FOV_HALF_HEIGHT-1) - y;
#endif
      break;
    case FOV_CHN_1:
#if FOV_ROTATE_1
      x = (FOV_WIDTH-1) - x;
      y = (FOV_HALF_HEIGHT-1) - y;
#else
      x += FOV_HALF_WIDTH;
#endif
      break;
    case FOV_CHN_2:
#if FOV_ROTATE_2
      x= (FOV_HALF_WIDTH-1) - x;
      y= (FOV_HEIGHT-1) - y;
#else
      y+= FOV_HALF_HEIGHT;
#endif
      break;
    case FOV_CHN_3:
#if FOV_ROTATE_3
      x = (FOV_WIDTH-1) - x;
      y = (FOV_HEIGHT-1) - y;
#else
      x += FOV_HALF_WIDTH;
      y += FOV_HALF_HEIGHT;
#endif
      break;
    }
    fov[x + FOV_WIDTH*y] += wght;
  }
}

/* Update field-of-view image: subtract events from oldest
   frame and then add events from newest frame. */
static int task_fov(cp40_task_t *task)
{
  void *data;
  cp40_raw_frame_t *raw_frame;

#if CP40_DEBUG
  if (task==NULL || task->op!=task_fov || task->frame1!=NULL
      || task->frame2!=NULL) {
    return gpl_error("corrupted data for task_fov()");
  }
#endif

  /* Subtract older frame (if any) from FOV or clear FOV (if none) and
     add new frame to FOV. */
  data= task->data;
  if (task->owner->stat_size > 1) {
    raw_frame= task->owner->stat[task->owner->stat_index].raw_frame;
    if (raw_frame) incr_fov(data, raw_frame, -1);
  } else {
    memset(data, 0, CP40_FOV_NBYTES);
  }
  raw_frame= task->owner->raw_frame;
  if (raw_frame) incr_fov(data, raw_frame, 1);
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/
