
#include "ytclInt.h"
#include <sys/time.h>
#include <unistd.h>





#undef REAL_TIME_DEBUG

/* Type definitions. */
typedef struct RealTimeData     kpc_master_t;
typedef struct RealTimeTaskData RealTimeTaskData;
typedef int RealTimeTaskProc(Tcl_Interp *interp, RealTimeTaskData *taskPtr);

static int		GetAttrFileAndFlags(const char **filenamePtr,
				int *flagsPtr, Tcl_Interp *interp, int argc,
				char *argv[]);
static int		get_channel(Tcl_Interp *interp, char *str, int *chnPtr);
static void		RealTimeCallback(ClientData client_data, int mask);
static void		FreeRealTimeData(void *client_data);
static kpc_master_t *	GetRealTimeData(cp40_instance_t *instance,
				Tcl_Interp *interp);
static char *		Cancel(ClientData client_data, Tcl_Interp *interp,
				char *name1, char *name2, int flags);
static int		SetStatValue(Tcl_Interp *interp, char *name1,
				char *format2, int channel, double value);



/* Real time routines. */
static RealTimeTaskProc	RealTimeAc;
static RealTimeTaskProc	RealTimeIc;
static RealTimeTaskProc	RealTimeIt;

/*  API:

    $master $cmd ...
    $master assoc ...     associate to photon-counting data stream
    $master close
    $master start $task   start a new task or re-start a stopped task
    $master stop $task    stop a running task (its result is still visible)
    $master clear $task   erase task result
    $master cancel $task  stop a task and release all related ressources
*/


/*_____________________________________________________________________________
 *
 * Real-time data processing:
 *	The idea is to schedule a number of real-time tasks to perform every
 *	time a new frame can be read.  Each of these tasks is linked with a
 *	global Tcl variable that also stores the reference of the TiM matrix
 *	that holds the result of the operation.
 *
 *	The following rules apply:
 *	  - Only CP40 data instances that are connected to a raw-photon
 *	    server can undergo real-time tasks.  When such a data instance
 *	    is destroyed, all real time tasks get canceled (but the related
 *	    TiM matrices are left unchanged).
 *	  - Each task can be canceled (that does not destroy the related TiM
 *	    matrix).
 *	  - If the global Tcl variable is deleted, the related TiM matrix is
 *	    destroyed (that's done by TiM) and the related real time task is
 *	    canceled (that's done by the trace Cancel() in this file).
 *
 *	Restrictions:
 *	  - Once a data instance is connected to a raw-photon server, its
 *	    attributes cannot be changed (a smarter behavior would be to
 *	    forbid setting of attributes only when there are running tasks).
 */


/* One kpc_master_t structure is associated with each CP40 data instance that
   is connected to a photon server. */

struct kpc_master_t {
    Tcl_Interp       *interp;    /* Tcl interpreter. */
    char             *tclCommand;/* Tcl command to evaluate (at global
				    level) after a new data frame has been
				    read and before processing the data by
				    the scheduled real-time tasks. */
    Tcl_File          tclFile;    /* Tcl file handle associated with the
				     socket file descriptor from which come
				     the raw data.  This handle is needed to
				     make Tcl event loop calls the real-time
				     tasks when some data arrive. */
    cp40_frame_t       *frame_ptr[4];/* One frame for each channel. */
    RealTimeTaskData *firstTask;  /* First task in a chained list of tasks
				     to perform (the last one scheduled is
				     the first one executed). */
    
    /* Statistical information (provides crude estimates of telescope pointing
       errors and turbulence strength).  Note: there is certainly a more
       clever way to store this information (e.g. faster access), than the
       following but I have no time to imporve this... */
#define STAT_AVG_MAX 500          /* Maximum number of frame statistics to
				     average (500 = 10 seconds). */
    int index;                    /* Frame number modulo STAT_AVG_MAX. */
    double w[4][STAT_AVG_MAX];    /* Frame weights (for every channel). */
    double wx[4][STAT_AVG_MAX];   /* Store: weight * mean_abscissa. */
    double wy[4][STAT_AVG_MAX];   /* Store: weight * mean_ordinate. */
    double ws[4][STAT_AVG_MAX];   /* Store: weight * standard_deviation. */
    struct timeval start;         /* Date of connection. */
};

typedef struct kpc_master_s kpc_master_t;
typedef struct kpc_master_s kpc_master_t;
typedef struct kpc_task_s   kpc_task_t;



#define ROUND(x,y) ((((x)+((y)-1))/(y))*(y))
#define VARIABLE_FLAGS (TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY)
#define TRACE_FLAGS  (TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY)

/*_____________________________________________________________________________
 *
 * schedule_method --
 *
 *	Implement `schedule` method: schedule a new real-time operation to
 *	perform.  Syntax:
 *	  dataInstance schedule varName operation channel1 ?channel2?
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static int schedule_method(cp40_instance_t *instance, Tcl_Interp *interp,
	int argc, char *argv[])
{
    char *varName, *taskName, *refString;
    int i, chn1, chn2= -1;
    Tim_Matrix       *matrixPtr= NULL;
    kpc_master_t     *master= NULL;
    RealTimeTaskData *taskPtr= NULL;
    RealTimeTaskProc *Proc;
    int *base, **pixelPtr, width, height, offset, xmin, xmax, ymin, ymax;

    /* Check and parse arguments. */
    if (argc < 5 || argc > 6) {
syntax_error:
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "syntax: ", argv[0], " ", argv[1],
		" varName task chn1 ?chn2?", (char *)NULL);
	return TCL_ERROR;
    }
    if ((master= GetRealTimeData(instance, interp))==NULL) {
	return TCL_ERROR;
    }
    varName= argv[2];
    if (varName==NULL || varName[0]==0 || strlen(varName) >= KPC_TOKEN_SIZE) {
	return report_error(interp, "bad variable name");
    }
    taskName= argv[3];
    if (strcmp(taskName, "it") == 0) {
	if (argc != 5) goto syntax_error;
	if (get_channel(interp, argv[4], &chn1) != TCL_OK) return TCL_ERROR;
	Proc= RealTimeIt;
	xmin= master->frame[chn1]->xmin;
	xmax= master->frame[chn1]->xmax;
	ymin= master->frame[chn1]->ymin;
	ymax= master->frame[chn1]->ymax;
    } else if (strcmp(taskName, "ac") == 0) {
	if (argc != 5) goto syntax_error;
	if (get_channel(interp, argv[4], &chn1) != TCL_OK) return TCL_ERROR;
	Proc= RealTimeAc;
	xmin= master->frame[chn1]->xmin - master->frame[chn1]->xmax;
	xmax= master->frame[chn1]->xmax - master->frame[chn1]->xmin;
	ymin= master->frame[chn1]->ymin - master->frame[chn1]->ymax;
	ymax= master->frame[chn1]->ymax - master->frame[chn1]->ymin;
    } else if (strcmp(taskName, "ic") == 0) {
	if (argc != 6) goto syntax_error;
	if (get_channel(interp, argv[4], &chn1) != TCL_OK
		|| get_channel(interp, argv[5], &chn2) != TCL_OK)
	    return TCL_ERROR;
	Proc= RealTimeIc;
	xmin= master->frame[chn2]->xmin - master->frame[chn1]->xmax;
	xmax= master->frame[chn2]->xmax - master->frame[chn1]->xmin;
	ymin= master->frame[chn2]->ymin - master->frame[chn1]->ymax;
	ymax= master->frame[chn2]->ymax - master->frame[chn1]->ymin;
    } else {
	return report_error(interp,
		"real-time task must be one of: \"it\", \"ic\", or \"ac\"");
    }
    
    /* Get dimensions of result. */
    width= xmax-xmin+1;
    height= ymax-ymin+1;
#ifdef REAL_TIME_DEBUG    
    printf("xmin=%d, xmax=%d, width=%d\nymin=%d, ymax=%d, height=%d\n",
	    xmin, xmax, width, ymin, ymax, height);
#endif
    
    /* Possibly create a new TiM matrix to store the result and link it
       with a Tcl global variable with the task name.   Reuse existing
       TiM matrix if possible. */
    if ((refString= Tcl_GetVar(interp, varName, TCL_GLOBAL_ONLY))==NULL
	    || (matrixPtr= Tim_GetMatrixByReference(refString))==NULL
	    || TimType(matrixPtr) != TIM_LONG || TimNCol(matrixPtr) != width
	    || TimNRow(matrixPtr) != height) {
	/* Allocate new matrix to store result. */
	if (refString!=NULL) Tcl_UnsetVar(interp, varName, TCL_GLOBAL_ONLY);
	if (TimIsTemporary(matrixPtr)) Tim_Free(matrixPtr);
	if ((matrixPtr= Tim_Alloc(TIM_LONG, width, height, TIM_CLEAR))==NULL) {
	    return Tim_Tclreport_error(interp, TIM_ALL_ERRORS);
	}
	if (Tim_SetMatrix(interp, varName, matrixPtr,
		TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) != TCL_OK) {
	    return TCL_ERROR;
	}
    } else if (Tcl_VarTraceInfo(interp, varName, TRACE_FLAGS, Cancel,
	    (ClientData)NULL /* To get the first trace. */) != NULL) {
	/* Cannot reuse former matrix: already in use by some other
	   real-time task. */
	return report_error(interp, "at most one real-time task by Tcl variable");
    }
    
    /* Allocate new RealTimeTaskData structure plus some extra memory to
       store the array of row pointers so that it is sufficient to free the
       RealTimeTaskData pointer. */
    offset= ROUND(sizeof(RealTimeTaskData), sizeof(long *));
    taskPtr= (RealTimeTaskData *)gpl_alloc(offset+height*sizeof(long *));
    if (taskPtr == NULL) {
	Tcl_UnsetVar(interp, varName, VARIABLE_FLAGS);
	Tcl_SetResult(interp, "memory allocation error", TCL_STATIC);
	return TCL_ERROR;
    }
    
    /* Clear result matrix and initialize array of pointers. */
    base= (long *)TimBase(matrixPtr);
    pixelPtr= (long **)((char *)taskPtr + offset);
    base -= xmin + width;
    for (i=0; i<height; i++)
	pixelPtr[i]= (base += width);
    
    /* Initialize real-time task structure. */
    taskPtr->master= master;
    strcpy(taskPtr->name, varName);
    taskPtr->Proc= Proc;
    taskPtr->next= master->firstTask;
    taskPtr->frame1= master->frame[chn1];
    taskPtr->frame2= chn2==-1 ? NULL :  master->frame[chn2];
    taskPtr->pixelPtr= pixelPtr - ymin;
    
    /* Set callback and insert task in the chain. */
    if (Tcl_TraceVar(interp, varName, TRACE_FLAGS,
	    Cancel, (ClientData)taskPtr) != TCL_OK) {
	Tcl_UnsetVar(interp, varName, VARIABLE_FLAGS);
	return TCL_ERROR;
    }
    master->firstTask= taskPtr;
    return TCL_OK;
}

/*_____________________________________________________________________________
 *
 * Cancel --
 *
 *	Cancel a real-time operation linked to a TiM matrix.  Cancel trace
 *	and free real time data.
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static char *Cancel(ClientData client_data, Tcl_Interp *interp, char *name1,
	char *name2, int flags)
{
    RealTimeTaskData *taskPtr= (RealTimeTaskData *)client_data;

    if (! (flags & (TCL_TRACE_DESTROYED /* | TCL_INTERP_DESTROYED */))) {
	/* Remove the trace. */
	Tcl_UntraceVar(interp, taskPtr->name, TRACE_FLAGS, Cancel,
		(ClientData)taskPtr);
    }
    taskPtr->master->firstTask= taskPtr->next;
    gpl_free(taskPtr);
    return NULL;
}


/*_____________________________________________________________________________
 *
 * stat_method --
 *
 *	Implement `stat` method: compute average of real-time statistics.
 *	Syntax:
 *	  dataInstance avgstat count arrayName
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static int stat_method(cp40_instance_t *instance, Tcl_Interp *interp,
		       int argc, char *argv[])
{
    char string[TCL_DOUBLE_SPACE];
    int chn, i, n, nok, count;
    char *name1;
    double w, x, y, s;
    kpc_master_t *master;
    struct timeval now;

    /* Check and parse arguments. */
    if (argc != 4) {
	Tcl_AppendResult(interp, "syntax: ", argv[0], " ", argv[1],
		" count arrayName", (char *)NULL);
	return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[2], &count) != TCL_OK || count <= 0) {
	return report_error(interp, "bad number of photon-counting frames");
    } else if (count > STAT_AVG_MAX) {
	return report_error(interp,
		"can't remember that number of photon-counting frames!");
    }
    name1= argv[3];

    /* Get real-time (also check that instance is ok for real-time
       processing). */
    if ((master= GetRealTimeData(instance, interp))==NULL) {
	return TCL_ERROR;
    }

    /* Set the time value (in seconds). */
    gettimeofday(&now, NULL);
    Tcl_PrintDouble(interp, now.tv_sec - master->start.tv_sec
	    + 1e-6 * (now.tv_usec - master->start.tv_usec), string);
    if (Tcl_SetVar2(interp, name1, "time", string, TCL_LEAVE_ERR_MSG)==NULL) {
	return TCL_ERROR;
    }
    
    for (chn=0; chn<4; chn++) {
	w= x= y= s= 0.0;
	nok= 0;
	n= count;
	i= master->index;
	while (n--) {
	    if (master->w[chn][i] > 0.0) {
		w += master->w[chn][i];
		x += master->wx[chn][i];
		y += master->wy[chn][i];
		s += master->ws[chn][i];
		nok += 1;
	    }
	    if (--i < 0) {
		/* Wrap index. */
		i= STAT_AVG_MAX-1;
	    }
	}
	if (w > 0.0) {
	    x /= w;
	    y /= w;
	    s /= w;
	    w /= nok;
	} else {
	    w= x= y= s= 0.0;
	}
	if (SetStatValue(interp, name1, "weight,%d", chn, w)!=TCL_OK
		|| SetStatValue(interp, name1, "x,%d", chn, x)!=TCL_OK
		|| SetStatValue(interp, name1, "y,%d", chn, y)!=TCL_OK
		|| SetStatValue(interp, name1, "spread,%d", chn, s)!=TCL_OK) {
	    return TCL_ERROR;
	}
    }
    
    return TCL_OK;
}

/*_____________________________________________________________________________
 *
 * SetStatValue --
 *
 *	Set Tcl array value with statistics computed by stat_method().
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static int SetStatValue(Tcl_Interp *interp, char *name1, char *format2,
	int channel, double value)
{
    char name2[20];
    char string[TCL_DOUBLE_SPACE];

    sprintf(name2, format2, channel);
    Tcl_PrintDouble(interp, value, string);
    return Tcl_SetVar2(interp, name1, name2, string, TCL_LEAVE_ERR_MSG)
	? TCL_OK : TCL_ERROR;
}

/*_____________________________________________________________________________
 *
 * GetRealTimeData --
 *
 *	Get real-time data of a CP40 data instance.
 *
 * Return:
 *	Valid address or NULL in case of error (a message is left in interp).
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static RealTimeData *GetRealTimeData(cp40_instance_t *instance, Tcl_Interp *interp)
{
    /* Check that instance is ok for real-time processing. */
    if (instance->delete_client_data!=FreeRealTimeData || instance->client_data==NULL) {
	report_error(interp, "real-time operation impossible on that data instance");
	return NULL;
    }
    return (kpc_master_t *)instance->client_data;
}


/*_____________________________________________________________________________
 *
 * RealTimeCallback --
 *
 *	Process all real time tasks a CP40 instance associated with a photon
 *	server.
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description; in case of error, use Tcl_Backgroundreport_error() to
 *	report the error.
 *_____________________________________________________________________________
 */

static void RealTimeCallback(ClientData client_data, int mask)
{
    int end, status, chn;
    double w;
    kpc_master_t     *master= (kpc_master_t *)client_data;
    Tcl_Interp       *interp= master->interp;
    RealTimeTaskData *taskPtr;

    /* Read next frame and pre-process it for each channel. */
    if ((status= cp40_select_frame(master->frame[0], &end)) != 1) {
	if (status == 0) {
	    if (end) return;
	} else {
	    goto cp40_error;
	}
    }
    if (cp40_select_frame(master->frame[1], NULL) == -1
	    || cp40_select_frame(master->frame[2], NULL) == -1
	    || cp40_select_frame(master->frame[3], NULL) == -1) {
	goto cp40_error;
    }

    /* Integrate statistics. */
    if (++master->index >= STAT_AVG_MAX) {
	master->index= 0;
    }
    for (chn=0; chn<4; chn++) {
	if ((w= master->frame[chn]->flux) > 0.0) {
	    master->w[chn][master->index]= w;
	    master->wx[chn][master->index]= w * master->frame[chn]->x;
	    master->wy[chn][master->index]= w * master->frame[chn]->y;
	    master->ws[chn][master->index]= w * master->frame[chn]->spread;
	} else {
	    master->w[chn][master->index]= 0.0;
	    master->wx[chn][master->index]= 0.0;
	    master->wy[chn][master->index]= 0.0;
	    master->ws[chn][master->index]= 0.0;
	}
    }
	
    /* Evaluate Tcl command. */
    if (master->tclCommand != NULL) {
	status= Tcl_GlobalEval(interp, master->tclCommand);
	if (status != TCL_OK) {
	    /* Status is one of: TCL_ERROR, TCL_BREAK, TCL_CONTINUE, or
               TCL_RETURN. */
	    if (status == TCL_ERROR)
		goto bg_error;
	    if (status == TCL_BREAK)
		return;
	}
    }
    
    /* Execute every scheduled real-time operations. */
    for (taskPtr= master->firstTask; taskPtr != NULL; taskPtr= taskPtr->next) {
	if (taskPtr->Proc(interp, taskPtr) != TCL_OK) {
	    goto bg_error;
	}
    }
    return;
    
cp40_error:
    report_error(interp, NULL);
bg_error:
    Tcl_Backgroundreport_error(interp);
    return;
}

/*_____________________________________________________________________________
 *
 * attr_method --
 *
 *	Implement CP40 `attr` method:
 *	  dataInstance attr ?options?              get attribute name(s);
 *	  dataInstance attr ?options? name         get attribute value;
 *	  dataInstance attr ?options? name value   set attribute value.
 *
	Return standard Tcl result.
 *
 * Side effects:
 *	In case of error, a message is left in interp->result.
 *_____________________________________________________________________________
 */

static int attr_method(cp40_instance_t *instance, Tcl_Interp *interp, int argc,
		       char *argv[])
{
  int i;
  int extra= 0;
  int standard= 1;
  const char *arg;
  
  for (i=2; i<argc; i++) {
    arg= argv[i];
    if (arg[0] != '-') {
      /* Not an option. */
      break;
    }
    if (arg[1] == '-' && arg[2] == 0) {
      /* End of options. */
      i += 1;
      break;
    }
    if (arg[1] == 'e' && strcmp(arg, "-extra") == 0) {
      extra= 1;
      standard= 0;
      continue;
    }
    if (arg[1] == 'a' && strcmp(arg, "-all") == 0) {
      extra= 1;
      standard= 1;
      continue;
    }
    return report_error(interp, "unknown option");
  }
  
  if (i >= argc) {
    
    /* Get attribute names. */
    if (standard) {
      for (i=0; std_attributes[i] != NULL; i++) {
	Tcl_AppendElement(interp, (char *)std_attributes[i]);
      }
    }
    if (extra) {
      int index, status;
      const char *name, *value;
      for (index=-1;;) {
	status= cp40_attr_get_next_extra(instance, &index, &name, &value);
	if (status == 0) {
	  break;
	} else if (status == 1) {
	  Tcl_AppendElement(interp, (char *)name);
	} else {
	  return report_error(interp, NULL);
	}
      }
    }
    return TCL_OK;
    
  } else if (i == argc - 1) {
    
    /* Get attribute value. */
    int count;
    char **name= NULL;
    const char *value;
    if (Tcl_SplitList(interp, argv[argc-1], &count, &name) != TCL_OK)
      return TCL_ERROR;
    for (i=0; i<count; i++) {
      if ((value= cp40_attr_get(instance, name[i])) == NULL) {
	(void)gpl_error_info(0);	/* Forget error. */
	value= "";
      }
      Tcl_AppendElement(interp, (char *)value);
    }
    Tcl_Free((char *)name);
    return TCL_OK;
    
  } else if (i == argc - 2) {

    /* Set attribute value (one at a time). */
    const char *value;

    if (instance->delete_client_data==FreeRealTimeData) {
      return report_error(interp,
			  "cannot set attribute(s): real-time operation running");
    }
    if (cp40_attr_set(instance, argv[argc-2], argv[argc-1]) != GPL_OK)
      return report_error(interp, NULL);
    if ((value= cp40_attr_get(instance, argv[argc-2])) == NULL) {
      (void)gpl_error_info(0);	/* Forget error. */
      value= "";
    }
    Tcl_SetResult(interp, (char *)value, TCL_VOLATILE);
    return TCL_OK;
  }
  Tcl_AppendResult(interp, "syntax: ", argv[0], " ", argv[1],
		   " ?options -- name value?", (char *)NULL);
  return TCL_ERROR;
}

/*_____________________________________________________________________________
 *
 * save_attr_method --
 *
 *	Implement CP40 `save` method that saves attribute(s) to file:
 *	  dataInstance save ?options? fileName
 *
	Return standard Tcl result.
 *
 * Side effects:
 *	In case of error, a message is left in interp->result.
 *_____________________________________________________________________________
 */


/*_____________________________________________________________________________
 *
 * restore_attr_method --
 *
 *	Implement CP40 `restore` method that restores attribute(s) from file:
 *	  dataInstance restore ?options? fileName
 *
	Return standard Tcl result.
 *
 * Side effects:
 *	In case of error, a message is left in interp->result.
 *_____________________________________________________________________________
 */

/*_____________________________________________________________________________
 *
 * FreeRealTimeData --
 *
 *	delete real time data of a CP40 instance associated with a photon
 *	server.
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

static void FreeRealTimeData(void *client_data)
{
  int chn;
  kpc_master_t *master= (kpc_master_t *)client_data;
  
  if (master != NULL) {
    /* Delete/free Tcl file handle. */
    Tcl_DeleteFileHandler(master->tclFile);
    Tcl_FreeFile(master->tclFile);
    
    /* Cancel all real-time tasks. */
    while (master->firstTask != NULL) {
      Cancel((ClientData)master->firstTask, master->interp,
	     master->firstTask->name, NULL, TRACE_FLAGS);
    }
    
    /* delete Photon frames. */
    for (chn=0; chn<4; chn++)
      cp40_delete_frame(master->frame[chn]);
  }
}
