/*
 * cp40tcl.c --
 *
 *	Provides interface of CP40 with Tcl.
 *
 * Copyright (c) 1999, Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#include "cp40tcl.h"
#if WITH_TIM
# include "tim.h"
#endif

/*---------------------------------------------------------------------------*/
/* MACROS */

/*---------------------------------------------------------------------------*/
/* CP40 INSTANCE METHODS */

/* Methods understood by a CP40-Tcl instance.   All methods return standard
   result: GPL_OK or GPL_ERROR. */
typedef struct method_def_s method_def_t;
typedef int method_op_t(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			int argc, char *argv[]);
static method_op_t assoc_method;
static method_op_t config_method;
static method_op_t cget_method;
static method_op_t command_method;
static method_op_t cancel_method;
static method_op_t close_method;
/*	Implement the `close` method: close data stream associated with a
	CP40 instance.  Return standard result: GPL_OK or GPL_ERROR. */

static method_op_t delete_method;
/*	Implement the `delete` method: delete a CP40 instance. Return
	standard result: GPL_OK or GPL_ERROR. */

static method_op_t link_method;
static method_op_t schedule_method;
static method_op_t restore_config_method;
static method_op_t save_config_method;
static method_op_t save_image_method;
static method_op_t stat_method;
static method_op_t start_method;
static method_op_t stop_method;
static method_op_t run_method;
static method_op_t rewind_method;
static method_op_t read_header_method;
static method_op_t write_header_method;

static method_op_t tasks_method;
/*	Implement `tasks` method: return a list of tasks scheduled for this
	data instance. Return standard TCL result: TCL_OK or TCL_ERROR.
	Syntax:
		$instance tasks           ;# all scheduled tasks
		$instance tasks all       ;# all scheduled tasks
		$instance tasks running   ;# only running tasks
		$instance tasks idle      ;# only idle tasks */

struct method_def_s {
  const char *name;
  method_op_t *method;
};

static method_def_t method_table[] = {
  {"assoc", assoc_method},
  {"cancel", cancel_method},
  {"cget", cget_method},
  {"close", close_method},
  {"command", command_method},
  {"config", config_method},
  {"configure", config_method},
  {"delete", delete_method},
  {"link", link_method},
  {"read_header", read_header_method},
  {"restore_config", restore_config_method},
  {"rewind", rewind_method},
  {"run", run_method},
  {"save_config", save_config_method},
  {"save_image", save_image_method},
  {"schedule", schedule_method},
  {"start", start_method},
  {"stat", stat_method},
  {"stop", stop_method},
  {"tasks", tasks_method},
  {"write_header", write_header_method},
  {NULL, NULL}
};

/*---------------------------------------------------------------------------*/
/* PRIVATE ROUTINES */

static int main_cp40_command(ClientData dummy, Tcl_Interp *interp,
			     int argc, char *argv[]);
/*	Implement the "cp40" Tcl command:
	  Create a Tcl command related to a new CP40 data instance.  Return
	standard result: GPL_OK or GPL_ERROR.  Allocate a new CP40 data
	structure possibly associated with a stream.
          cp40 new NAME ... 
	  cp40 delete INSTANCE
	  cp40 exists INSTANCE
	  cp40 instances
*/

static const char *cget(cp40tcl_instance_data_t *data, const char *option);
/*	Return TEMPORARY string with the current value of OPTION for
	CP40 data INSTANCE.  In case of error, return NULL and left an
	error message that can be retrieved wit gpl_error_info. */

static int config(Tcl_Interp *interp, cp40tcl_instance_data_t *data,
		  const char *option, const char *value);
/*	Set OPTION with VALUE for CP40 data INSTANCE. Return standard Tcl
	result TCL_OK or TCL_ERROR (error message is left in INTERP's
	result). */

static int report_error(Tcl_Interp *interp, const char *msg);
/*	Store given error message (or GPLib error message from gpl_error_info
	if MSG is NULL) in interpretor INTERP.  Return TCL_ERROR. */

static int instance_command(ClientData client_data, Tcl_Interp *interp,
			    int argc, char *argv[]);
/*	Implement methods understood by CP40 photon counting data instances.
	Return standard Tcl result. In case of error, a message is left in
	INTERP->result. */

/*---------------------------------------------------------------------------*/
/* PRIVATE DATA */

/* Name of standard options. */
static const char *std_options[]= {
  "-angle",
  "-coefs",
  "-correction",
  "-distance",
  "-exposure",
  "-extreme",
  "-flux",
  "-frame_count",
  "-frame_date",
  "-nframes",
  "-nphotons",
  "-nsaturations",
  "-parity",
  "-pixel",
  "-ratio",
  "-raw_coefs",
  "-scale",
  "-spread",
  "-spurious",
  "-window",
  NULL
};

static gpl_hash_table_t method_htable;
static gpl_hash_table_t instance_htable;

/*---------------------------------------------------------------------------*/

int Cp40tcl_Init(Tcl_Interp *interp)
{
  static int done= 0;
  gpl_hash_entry_t *entry;
  int i, new;

  /* (Re)create CP40 commands and initialize instance/method hash tables. */  
  Tcl_CreateCommand(interp, "cp40", main_cp40_command,
		    (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL);
  if (! done) {
    gpl_hash_init_table(&instance_htable, TCL_STRING_KEYS);
    gpl_hash_init_table(&method_htable, TCL_STRING_KEYS);
    for (i=0 ; method_table[i].name != NULL ; i++) {
      entry= gpl_hash_create_entry(&method_htable, method_table[i].name, &new);
      gpl_hash_set_value(entry, &method_table[i], GPL_HASH_STATIC);
    }
    done= 1;
  }
  return TCL_OK;
}

/*---------------------------------------------------------------------------*/
/* DELETE CALLBACK */

/* These routines get called when a CP40-Tcl data instance / command / entry
   is destroyed to free all resources of the instance.  They do not update
   associated file stream if any.  Since there are several deletion callback,
   to avoid conflicts, each callback clear members of the structure
   cp40tcl_instance_data_t to avoid deleting it again (see comments in source
   code). */

static void delete_instance_data(void *client_data);
/*	Callback routine called by cp40_delete_instance. */

static void delete_instance_entry(void *client_data);
/*	Callback routine called by gpl_hash_delete_entry. */

static void delete_instance_command(ClientData client_data);
/*	Callback routine called by Tcl_DeleteCommand. */

/*---------------------------------------------------------------------------*/

cp40tcl_instance_data_t *cp40tcl_new_instance_data(Tcl_Interp *interp,
						   const char *name)
{
  int new;
  char buffer[32];
  cp40tcl_instance_data_t *data= NULL;

  /* Possibly create unique name. */
  if (strcmp(name, "#auto") == 0) {
    gpl_hash_entry_t *entry;
    Tcl_CmdInfo cmd_info;
    int i= 0;
    do {
      sprintf(buffer, "cp40.%d", i++);
    } while (Tcl_GetCommandInfo(interp, buffer, &cmd_info)
	     || gpl_hash_find_entry(&instance_htable, buffer));
    name= buffer;
    entry= NULL;
  }

  /* Create instance record (order of operations is important to avoid
     conflicts in case of error). */
  data= cp40_alloc_clear(sizeof *data, "structure cp40tcl_instance_data_t");
  if (! data) goto error_1;
  data->interp= interp;
  data->connection= -1;
  data->name= gpl_str_clone(name);
  if (! data->name) goto error_1;
  data->owner= cp40_new_instance();
  if (! data->owner) goto error_1;
  data->owner->client_data= data;
  data->owner->delete_client_data= delete_instance_data;
  data->entry= gpl_hash_create_entry(&instance_htable, name, &new);
  if (! data->entry) goto error_1;
  gpl_hash_set_value(data->entry, data, delete_instance_entry);
  if (! new) {
    Tcl_AppendResult(interp, "(BUG) unexpected result in \"", __FILE__,
		     "\" at line ", GPL_STRINGIFY(__LINE__), NULL);
    goto error_2;
  }
  data->command= Tcl_CreateCommand(interp, data->name, instance_command,
				   (ClientData *)data,
				   delete_instance_command);
  if (! data->command) goto error_2;

  return data;

  /* Store error message in interpretor. */
 error_1:
  report_error(interp, NULL);

  /* Free resources. */
 error_2:
  cp40tcl_delete_instance_data(data, 0 /* no update */);
  return NULL;
}

void cp40tcl_delete_instance_data(cp40tcl_instance_data_t *data, int update)
{
  if (data) {
    if (data->owner) {
      /* This will also trigger the deletion of DATA by
	 delete_instance_data(). */
      cp40_delete_instance(data->owner, update);
    } else {
      /* We must explicitly call delete_instance_data() to delete DATA. */
      delete_instance_data(data);
    }
  }
}

static void delete_instance_data(void *client_data)
{
  cp40tcl_instance_data_t *data= (cp40tcl_instance_data_t *)client_data;
  if (data) {
    data->owner= NULL; /* to avoid delete by other delete callbacks */
    if (data->connection>=0) {
      Tcl_DeleteFileHandler(data->connection);
      data->connection= -1;
    }
    if (data->entry) {
      /* Juste deleteting the instance hash entry will trigger all necessary
	 cleanup (it will eventually recall delete_instance_data but with
	 DATA->entry set to NULL). */
      gpl_hash_delete_entry(data->entry);
    } else if (data->command) {
      /* Juste deleteting the instance command will trigger all necessary
	 cleanup (it will eventually recall delete_instance_data but with
	 DATA->command set to NULL). */
      Tcl_DeleteCommand(data->interp,
			Tcl_GetCommandName(data->interp, data->command));
    } else {
      /* Perform its own cleanup (no other callbacks do this). */
      GPL_FREE_MAYBE(data->name);
      GPL_FREE_MAYBE(data->script);
      gpl_free(data);
    }
  }
}

static void delete_instance_entry(void *client_data)
{
  cp40tcl_instance_data_t *data= (cp40tcl_instance_data_t *)client_data;
  if (data) {
    data->entry= NULL; /* to avoid delete by further delete callbacks */
    cp40tcl_delete_instance_data(data, 0 /* no update */);
  }
}

static void delete_instance_command(ClientData client_data)
{
  cp40tcl_instance_data_t *data= (cp40tcl_instance_data_t *)client_data;
  if (data) {
    data->command= NULL; /* to avoid delete by further delete callbacks */
    cp40tcl_delete_instance_data(data, 0 /* no update */);
  }
}

/*---------------------------------------------------------------------------*/

cp40tcl_instance_data_t *cp40tcl_get_instance_data(const char *name)
{
  gpl_hash_entry_t *entry= gpl_hash_find_entry(&instance_htable, name);
  if (entry) return (cp40tcl_instance_data_t *)gpl_hash_get_value(entry);
  gpl_error("no such CP40 instance");
  return NULL;
}

cp40_instance_t *cp40tcl_get_instance_owner(const char *name)
{
  cp40tcl_instance_data_t *data= cp40tcl_get_instance_data(name);
  return data ? data->owner : NULL;
}

cp40_task_t *cp40tcl_get_task(const char *name)
{
  char buffer[100], *ptr;
  unsigned len;
  cp40tcl_instance_data_t *data;

  /* Parse DATA->TASK token. */
  if ((len= strlen(name)) >= sizeof(buffer)) {
    gpl_error("CP40 instance + task token too long");
    return NULL;
  }
  memcpy(buffer, name, len+1);
  if ((ptr= strrchr(buffer, '>')) && ptr>buffer && ptr[-1]=='-') {
    ptr[-1]= 0;
    ptr++;
  } else {
    gpl_error("bad CP40 instance task token, should be: DATA->TASK");
    return NULL;
  }
  if (! (data= cp40tcl_get_instance_data(buffer))) {
    return NULL;
  }
  return cp40_get_task(data->owner, ptr);
}

/*---------------------------------------------------------------------------*/
static int main_cp40_command(ClientData dummy, Tcl_Interp *interp, int argc,
			     char *argv[])
{
  int i, c;
  const char *option;
  cp40tcl_instance_data_t *data;
  
  if (argc<2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " option ?...?\"", NULL);
    return TCL_ERROR;
  }
  c= (option= argv[1])[0];
  if (c=='d' && strcmp(option, "delete")==0) {

    /* Destroy an instance. */
    if (argc != 3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], " instance\"", NULL);
      return TCL_ERROR;
    }
    if (! (data= cp40tcl_get_instance_data(argv[2]))
	|| cp40_delete_instance(data->owner, 0 /* update */) != GPL_OK) {
      return report_error(interp, NULL);
    }

  } else if (c=='e' && strcmp(option, "exists")==0) {

    /* Check existence of an instance. */
    int result;
    if (argc != 3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], " instance\"", NULL);
      return TCL_ERROR;
    }
    result= gpl_hash_find_entry(&instance_htable, argv[2]) != NULL;
    Tcl_SetObjResult(interp, Tcl_NewIntObj(result));

  } else if (c=='i' && strcmp(option, "instances")==0) {

    /* List existing instances. */
    gpl_hash_table_t *table= &instance_htable;
    gpl_hash_entry_t *entry;
    gpl_hash_search_t search;
    if (argc != 2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], NULL);
      return TCL_ERROR;
    }
    for (entry=gpl_hash_first_entry(table, &search); entry!=NULL;
	 entry=gpl_hash_next_entry(&search)) {
      Tcl_AppendElement(interp, gpl_hash_get_key(entry));
    }

  } else if (c=='n' && strcmp(option, "new")==0) {

    /* Create a new instance. */
    if (argc<3 || argc%2!=1) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], " name ?-option value ...?", NULL);
      return TCL_ERROR;
    }
    if (! (data= cp40tcl_new_instance_data(interp, argv[2]))) {
      return TCL_ERROR;
    }
    for (i=3 ; i<argc ; i+=2) {
      if (config(interp, data, argv[i], argv[i+1]) != TCL_OK) {
	cp40_delete_instance(data->owner, 0);
	return TCL_ERROR;
      }
    }
    Tcl_SetResult(interp, (char *)data->name, TCL_VOLATILE);

  } else {
    Tcl_AppendResult(interp, "unknown command for ", argv[0],
		     " should be one of: delete exists instances new", NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

static int instance_command(ClientData client_data, Tcl_Interp *interp,
			    int argc, char *argv[])
{
  gpl_hash_entry_t *entry;
  int i;
  cp40tcl_instance_data_t *data= (cp40tcl_instance_data_t *)client_data;

  if (data == NULL) {
    return report_error(interp, "corrupted CP40 data instance");
  }
  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " method ?...?\"", NULL);
    return TCL_ERROR;
  }
  if (! (entry= gpl_hash_find_entry(&method_htable, argv[1]))) {
    Tcl_AppendResult(interp, "bad CP40 method, should be one of:", NULL);
    for (i=0 ; method_table[i].name != NULL ; i++) {
      Tcl_AppendResult(interp, " ", method_table[i].name, NULL);
    }
    return TCL_ERROR;
  }
  return ((method_def_t *)gpl_hash_get_value(entry))->method(data, interp,
						       argc, argv);
}

/*---------------------------------------------------------------------------*/
/* METHODS */

static int link_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
		       int argc, char *argv[])
{
#if WITH_TIM
  cp40_task_t *task;
  tim_array_t *array;
  tim_object_t *object;
  if (argc!=4) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " varName task\"", NULL);
    return TCL_ERROR;
  }
  if ((task= cp40_get_task(data->owner, argv[3])) == NULL) {
    return report_error(interp, NULL);
  }
  array= task->image->array;
  object= (tim_object_t *)array;
#if 0
  printf("array= 0x%x\n", (unsigned)array);
  printf("array->refs= %d\n", array->refs);
  printf("array->id= %d\n", array->id);
  printf("array->data= 0x%x\n", (unsigned)array->data);
  printf("array->number= %d\n", array->number);
  printf("array->dim1= %d\n", array->dim1);
  printf("array->dim2= %d\n", array->dim2);
  printf("object= 0x%x\n", (unsigned)object);
  printf("object->refs= %d\n", object->refs);
  printf("object->id= %d\n", object->id);
#endif
  if (tim_store(interp, argv[2], object, TCL_LEAVE_ERR_MSG) != TCL_OK) {
    return TCL_ERROR;
  }
  return TCL_OK;
#else
  return report_error(interp, "compiled witout TiM support");
#endif
}

/*	Implement the `assoc` method (may be also called by the `cp40new`
	command): associate a CP40 instance with a file stream or a photon
	server.  Return standard result: GPL_OK or GPL_ERROR. */

static void real_time_callback(ClientData client_data, int mask);


static int assoc_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			int argc, char *argv[])
{
  if (argc==2) {
    /* Without arguments, return current stream settings. */
    int io_type= cp40_get_io_type(data->owner);
    int io_mode= cp40_get_io_mode(data->owner);
    if (io_type == CP40_IO_FILE) {
      Tcl_AppendElement(interp, "-file");
      Tcl_AppendElement(interp, (char *)cp40_get_io_filename(data->owner));
      if (io_mode == CP40_IO_MODE_READ) {
	Tcl_AppendElement(interp, "read");
      } else if (io_mode == CP40_IO_MODE_WRITE) {
	Tcl_AppendElement(interp, "write");
      } else if (io_mode == CP40_IO_MODE_UPDATE) {
	Tcl_AppendElement(interp, "update");
      } else {
	gpl_error("(BUG) unexpected file stream mode %d", io_mode);
	goto lib_error;
      }
    } else if (io_type == CP40_IO_SOCKET) {
      char buffer[32];
      if (io_mode != CP40_IO_MODE_READ) {
	gpl_error("(BUG) unexpected socket stream mode %d", io_mode);
	goto lib_error;
      }
      Tcl_AppendElement(interp, "-socket");
      Tcl_AppendElement(interp, (char *)cp40_get_io_hostname(data->owner));
      sprintf(buffer, "%d", cp40_get_io_port(data->owner));
      Tcl_AppendElement(interp, buffer);
    } else if (io_type != CP40_IO_NONE) {
      gpl_error("(BUG) unexpected stream type %d", io_type);
      goto lib_error;
    }

  } else if ((argc==4 || argc==5) && strcmp(argv[2], "-file")==0) {
    /* Associate CP40 instance with a file stream. */
    char *path= argv[3];
    int   mode= CP40_IO_MODE_READ;
    if (argc==5) {
      char *str= argv[4];
      if (       str[0] == 'r' && (str[1]==0 || strcmp(str, "read")==0)) {
	mode= CP40_IO_MODE_READ;
      } else if (str[0] == 'w' && (str[1]==0 || strcmp(str, "write")==0)) {
	mode= CP40_IO_MODE_WRITE;
      } else if (str[0] == 'u' && (str[1]==0 || strcmp(str, "update")==0)) {
	mode= CP40_IO_MODE_UPDATE;
      } else {
	return report_error(interp, "bad mode");
      }
    }
    if (cp40_assoc_file(data->owner, path, mode) != GPL_OK) {
      return report_error(interp, NULL);
    }
    
  } else if (argc==5 && strcmp(argv[2], "-socket")==0) {
    /* Associate CP40 instance with a photon server. */
    char *host= argv[3];
    int   port;
    if (Tcl_GetInt(interp, argv[4], &port)!=TCL_OK || (port&0xffff)!=port) {
      return report_error(interp, "bad port number");
    }
    if (cp40_assoc_socket(data->owner, host, port) != GPL_OK) {
      return report_error(interp, NULL);
    }
    gettimeofday(&data->start, NULL);
    data->nframes= 0;
    data->connection= data->owner->stream.io.socket.handle;
    Tcl_CreateFileHandler(data->connection, TCL_READABLE,
			  real_time_callback, (ClientData)data);
  } else {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " ?-file path ?mode? | -socket host port?\"",
		     NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
lib_error:
  Tcl_ResetResult(interp);
  report_error(interp, NULL);
  return TCL_ERROR;
}

static void real_time_callback(ClientData client_data, int mask)
{
  cp40tcl_instance_data_t *data= (cp40tcl_instance_data_t *)client_data;
  int nframes= cp40_run_tasks(data->owner, -1); /* process next pending
						   frames */
  if (nframes>0) {
    if (data->nframes==0) gettimeofday(&data->start, NULL);
    data->nframes+= nframes;
  
    /* Evaluate Tcl script (return value is one of: TCL_ERROR, TCL_BREAK,
       TCL_CONTINUE, or TCL_RETURN). */
    if (data->script) {
      int status= Tcl_GlobalEval(data->interp, data->script);
      if (status == TCL_ERROR) {
	Tcl_BackgroundError(data->interp);
      }
    }
  }
}

/*---------------------------------------------------------------------------*/
/* CONFIGURATION METHODS */

static const char *cget(cp40tcl_instance_data_t *data, const char *option)
{
  static gpl_dstr_t dstr;
  static int need_init= 1;

  if (need_init) {
    gpl_dstr_init(&dstr);
    need_init= 0;
  } else {
    gpl_dstr_reset(&dstr);
  }

  if (option[0] != '-') {
    gpl_dstr_reset(&dstr);
    gpl_dstr_append(&dstr, "unknown option \"", -1);
    gpl_dstr_append(&dstr, option, -1);
    gpl_dstr_append_char(&dstr, '"');
    return NULL;
  }

  if (cp40_get(data->owner, option+1, &dstr) != GPL_OK) {
    /* Forget error: a non-existing attribute is the same as an empty one. */
    (void)gpl_error_info(1);
  }

  return gpl_dstr_value(&dstr);
}

static int config(Tcl_Interp *interp, cp40tcl_instance_data_t *data,
		  const char *option, const char *value)
{
  if (option[0] != '-') {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "unknown option \"", option, "\"", NULL);
    return TCL_ERROR;
  }
  if (cp40_set(data->owner, option+1, value) != GPL_OK) {
    report_error(interp, NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

static int cget_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
		       int argc, char *argv[])
{
  const char *value;
  if (argc!=3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " -option\"", NULL);
    return TCL_ERROR;
  }
  if (! (value= cget(data, argv[2]))) {
    return report_error(interp, NULL);
  }
  Tcl_SetResult(interp, (char *)value, TCL_VOLATILE);
  return TCL_OK;
}

static int config_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			int argc, char *argv[])
{
  static gpl_dstr_t dstr;
  static int need_init= 1;
  int i, index, status;
  const char *value, *name, *option;

  if (need_init) {
    gpl_dstr_init(&dstr);
    need_init= 0;
  } else {
    gpl_dstr_reset(&dstr);
  }

  if (argc%2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " -option value ?...?\"", NULL);
    return TCL_ERROR;
  }

  if (argc==2) {
    /* Return all attributes. */
    for (i=0 ; (option=std_options[i]) != NULL ; i++) {
      if (! (value= cget(data, option))) {
	return report_error(interp, NULL);
      }
      Tcl_AppendElement(interp, (char *)option);
      Tcl_AppendElement(interp, (char *)value);
    }
    for (index=-1 ; ; ) {
      status= cp40_get_next_extra(data->owner, &index, &name, &value);
      if (status == 0) {
	break;
      } else if (status == 1) {
	gpl_dstr_reset(&dstr);
	if (gpl_dstr_append_char(&dstr, '-') != GPL_OK
	    || gpl_dstr_append(&dstr, name, -1) != GPL_OK) {
	  return report_error(interp, NULL);
	}
	Tcl_AppendElement(interp, (char *)gpl_dstr_value(&dstr));
	Tcl_AppendElement(interp, (char *)value);
      } else {
	return report_error(interp, NULL);
      }
    }
    return TCL_OK;
  }

  for (i=2 ; i<argc ; i+=2) {
    if (config(interp, data, argv[i], argv[i+1]) != TCL_OK) {
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

/*---------------------------------------------------------------------------*/
/* TASK METHODS */

static int save_image_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			     int argc, char *argv[])
{
  cp40_task_t *task;
  const char *option, *token, *name;
  int i, c, flags= 0;

  for (i=2 ; i<argc ; i++) {
    if ((option= argv[i])[0] != '-') {
      break;
    }
    if ((c= option[1])=='-') {
      if (option[2]=='\0') ++i; /* skip end_of-option mark */
      break;
    }
    if (c=='p' && strcmp(argv[i], "-prefix")==0) {
      flags|= CP40_USE_NAME_AS_PREFIX;
    } else if (c=='o' && strcmp(argv[i], "-overwrite")==0) {
      flags|= CP40_MAY_OVERWRITE;
    } else {
      break;
    }
  }
  if ((argc-i) != 2) {
    Tcl_AppendResult(interp, "wrong # args, should be: \"", argv[0], " ",
		     argv[1], " ?options? task name\"", NULL);
    return TCL_ERROR;
  }
  token= argv[i];
  name= argv[i+1];
  if (strcmp(token, "all")==0) {
    for (task=data->owner->task ; task ; task=task->next) {
      if (cp40_save_task(task, name, flags) != GPL_OK) {
	return report_error(interp, NULL);
      }
    }
  } else {
    if (! (task= cp40_get_task(data->owner, token))
	|| cp40_save_task(task, name, flags) != GPL_OK) {
      return report_error(interp, NULL);
    }
  }
  return TCL_OK;
#if 0
duplicate_option:
  Tcl_AppendResult(interp, "duplicate option: ", option, NULL);
  return TCL_ERROR;
missing_argument:
  Tcl_AppendResult(interp, "missing argument for option ", option, NULL);
  return TCL_ERROR;
#endif
}

#undef METHOD
#define METHOD(NAME, ALL, ONE)						      \
static int NAME(cp40tcl_instance_data_t *data, Tcl_Interp *interp,	      \
		int argc, char *argv[])					      \
{									      \
  if (argc != 3) {							      \
    Tcl_AppendResult(interp, "wrong # args, should be: \"", argv[0], " ",     \
		     argv[1], " task|all\"", NULL);			      \
    return TCL_ERROR;							      \
  }									      \
  if (strcmp(argv[2], "all")==0) {					      \
    ALL(data->owner);							      \
  } else { 								      \
    cp40_task_t *task= cp40_get_task(data->owner, argv[2]);		      \
    if (!task) return report_error(interp, NULL);			      \
    ONE(task);								      \
  }									      \
  return TCL_OK;							      \
}
METHOD(cancel_method, cp40_cancel_all_tasks, cp40_cancel_task)
METHOD(start_method,  cp40_start_all_tasks,  cp40_start_task)
METHOD(stop_method,   cp40_stop_all_tasks,   cp40_stop_task)

#undef METHOD
#define METHOD(NAME,OP)							      \
static int NAME(cp40tcl_instance_data_t *data, Tcl_Interp *interp,	      \
		int argc, char *argv[])					      \
{									      \
  cp40_instance_t *instance= data->owner;				      \
  if (instance->stream.ops->OP(instance) != GPL_OK) {			      \
    return report_error(interp, NULL);					      \
  }									      \
  return TCL_OK;							      \
}
METHOD(rewind_method, rewind)
METHOD(read_header_method, read_header)
METHOD(write_header_method, write_header)

static int run_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
		      int argc, char *argv[])
{
  int nframes= -1;
  if (argc!=2 && (argc!=3 || Tcl_GetInt(interp, argv[2], &nframes)!=TCL_OK)) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " ?nframes?\"", NULL);
    return TCL_ERROR;
  }
  nframes= cp40_run_tasks(data->owner, nframes);
  Tcl_SetObjResult(interp, Tcl_NewIntObj(nframes));
  return TCL_OK;
}

static int schedule_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			   int argc, char *argv[])
{
  int i;
  if (argc<=2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " task ?...?\"", NULL);
    return TCL_ERROR;
  }

  /* Make sure all task tokens are valid. */
  for (i=2 ; i<argc ; i++) {
    if (! cp40_parse_task_token(argv[i], NULL, NULL)) {
      return report_error(interp, NULL);
    }
  }

  /* Effectively schedule tasks. */
  for (i=2 ; i<argc ; i++) {
    if (! cp40_schedule_task(data->owner, argv[i])) {
      return report_error(interp, NULL);
    }
  }
  return TCL_OK;
}

static int tasks_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			int argc, char *argv[])
{
  cp40_task_t *task;
  cp40_instance_t *instance;
  int what= 3;   /* default is to list all tasks */

  if (argc == 3) {
    if (strcmp(argv[2], "running")==0) what= 1;
    else if (strcmp(argv[2], "idle")==0) what= 2;
    else if (strcmp(argv[2], "all")==0) what= 3;
    else goto syntax_error;
  } else if (argc != 2) {
  syntax_error:
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " ?all|running|idle?\"", NULL);
    return TCL_ERROR;
  }
  instance= data->owner;
  for (task=instance->task ; task!=NULL ; task=task->next) {
    if (task->run ? (what&1) : (what&2)) {
      Tcl_AppendElement(interp, task->token);
    }
  }
  return TCL_OK;
}

static int command_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			  int argc, char *argv[])
{
  if (argc==2) {
    if (data->script) {
      Tcl_SetResult(interp, data->script, TCL_VOLATILE);
    }
  } else if (argc==3) {
    if (data->script) {
      gpl_free(data->script);
    }
    if (! (data->script= gpl_str_clone(argv[2]))) {
      return report_error(interp, NULL);
    }
  } else {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " ?script?\"", NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

static char *set_var2_i(Tcl_Interp *interp, char *name1, char *name2,
			int value, int flags);
static char *set_var2_d(Tcl_Interp *interp, char *name1, char *name2,
			double value, int flags);

static char *set_var2_i(Tcl_Interp *interp, char *name1, char *name2,
			int value, int flags)
{
  char string[TCL_DOUBLE_SPACE];
  sprintf(string, "%d", value);
  return Tcl_SetVar2(interp, name1, name2, string, flags);
}

static char *set_var2_d(Tcl_Interp *interp, char *name1, char *name2,
		      double value, int flags)
{
  char string[TCL_DOUBLE_SPACE];
  Tcl_PrintDouble(interp, value, string);
  return Tcl_SetVar2(interp, name1, name2, string, flags);
}

static int stat_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
		       int argc, char *argv[])
{
  int c;
  char *option;
  if (argc<3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " duration|average ?...?\"", NULL);
    return TCL_ERROR;
  }
  option= argv[2];
  c= option[0];
  if (c=='a' && strcmp(option, "average")==0) {
    /* Average real time statistics. */
    cp40_avg_stat_t avg[4];
    int chn, flags= TCL_LEAVE_ERR_MSG;
    char *name1, name2[32];
    if (argc!=4) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], " ", argv[2], " varname\"", NULL);
      return TCL_ERROR;
    }
    name1= argv[3];
    cp40_average_stat(data->owner, avg);
    for (chn=0 ; chn<4 ; chn++) {
      sprintf(name2, "nframes,%d", chn);
      if (! set_var2_i(interp, name1, name2, avg[chn].nframes, flags)) {
	return TCL_ERROR;
      }
      sprintf(name2, "flux,%d", chn);
      if (! set_var2_d(interp, name1, name2, avg[chn].mean_flux, flags)) {
	return TCL_ERROR;
      }
      sprintf(name2, "x,%d", chn);
      if (! set_var2_d(interp, name1, name2, avg[chn].mean_x, flags)) {
	return TCL_ERROR;
      }
      sprintf(name2, "y,%d", chn);
      if (! set_var2_d(interp, name1, name2, avg[chn].mean_y, flags)) {
	return TCL_ERROR;
      }
      sprintf(name2, "spread,%d", chn);
      if (! set_var2_d(interp, name1, name2, avg[chn].mean_spread, flags)) {
	return TCL_ERROR;
      }
    }
  } else if (c=='d' && strcmp(option, "duration")==0) {
    /* Get/set duration for real time statistics. */
    int nframes;
    if (argc==4) {
      if (Tcl_GetInt(interp, argv[3], &nframes) != TCL_OK) {
	return TCL_ERROR;
      }
      if (cp40_set_stat_duration(data->owner, nframes) != GPL_OK) {
	return report_error(interp, NULL);
      }
    } else if (argc!=3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], " ", argv[2], " ?nframes?\"", NULL);
      return TCL_ERROR;
    }
    nframes= data->owner->stat_size;
    Tcl_SetObjResult(interp, Tcl_NewIntObj(nframes));
  } else if (c=='e' && strcmp(option, "elapsed")==0) {
    /* Get elapsed time or number of frames since connect. */
    char *unit;
    if (argc!=4) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], " ", argv[2], " unit\"", NULL);
      return TCL_ERROR;
    }
    c= (unit= argv[3])[0];
    if (c=='f' && strcmp(unit, "frames")==0) {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(data->nframes));
    } else {
      double elapsed;
      struct timeval now;
      if (data->nframes<=0) {
	elapsed= 0.0;
      } else {
	gettimeofday(&now, NULL);
	elapsed= (now.tv_usec - data->start.tv_usec) / 1e9
	  + (now.tv_sec - data->start.tv_sec) ;
      }
      if (c=='s' && strcmp(unit, "seconds")==0) {
	;
      } else if (c=='m' && strcmp(unit, "milliseconds")==0) {
	elapsed*= 1e6;
      } else if (c=='m' && strcmp(unit, "microseconds")==0) {
	elapsed*= 1e9;
      } else if (c=='m' && strcmp(unit, "minutes")==0) {
	elapsed/= 60.0;
      } else if (c=='h' && strcmp(unit, "hours")==0) {
	elapsed/= 3600.0;
      } else {
	Tcl_AppendResult(interp, "bad unit, should be one of:",
			 "frames hours minutes seconds milliseconds ",
			 "microseconds", NULL);
	return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, Tcl_NewDoubleObj(elapsed));
    }
  } else {
    Tcl_AppendResult(interp, "bad option for \"", argv[1], "\" method, ",
		     "should be one of: average duration elapsed", NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/*---------------------------------------------------------------------------*/
/* SAVE/RESTORE CONFIG METHODS */

static int parse_io_config_args(const char **filename_ptr, int *flags_ptr,
				Tcl_Interp *interp, int argc, char *argv[]);

static int save_config_method(cp40tcl_instance_data_t *data,
			      Tcl_Interp *interp, int argc, char *argv[])
{
  const char *filename;
  int flags;
  
  if (parse_io_config_args(&filename, &flags, interp, argc, argv) != TCL_OK) {
    return TCL_ERROR;
  }
  if (cp40_save_attr(data->owner, filename, flags)!=GPL_OK) {
    return report_error(interp, NULL);
  }
  return TCL_OK;
}

static int restore_config_method(cp40tcl_instance_data_t *data,
				 Tcl_Interp *interp, int argc, char *argv[])
{
  const char *filename;
  int flags;
  
  if (parse_io_config_args(&filename, &flags, interp, argc, argv) != TCL_OK) {
    return TCL_ERROR;
  }
  if (cp40_restore_attr(data->owner, filename, flags) != GPL_OK) {
    return report_error(interp, NULL);
  }
  return TCL_OK;
}

static int parse_io_config_args(const char **filename_ptr, int *flags_ptr,
				  Tcl_Interp *interp, int argc, char *argv[])
{
  int i, flags= 0;
  char *arg;
    
  for (i=2 ; i<argc ; i++) {
    arg= argv[i];
    if (arg[0] != '-') {
      /* Not an option. */
      break;
    }
    if (arg[1] == '-' && arg[2] == 0) {
      /* End of options. */
      i += 1;
      break;
    }
    if (arg[1] == 'a' && strcmp(arg, "-all") == 0) {
      flags |= CP40_ATTR_ALL;
      continue;
    }
    if (arg[1] == 'e' && strcmp(arg, "-extra") == 0) {
      flags |= CP40_ATTR_EXTRA;
      continue;
    }
    if (arg[1] == 'p' && strcmp(arg, "-preprocessing") == 0) {
      flags |= CP40_ATTR_PREPROCESSING;
      continue;
    }
    if (arg[1] == 's' && strcmp(arg, "-standard") == 0) {
      flags |= CP40_ATTR_STANDARD;
      continue;
    }
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "unknown option, must be one of: ",
		     "-all -extra -preprocessing -standard", (char *)NULL);
    return TCL_ERROR;
  }
  if (i != argc-1) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1],  " ?options --? filename\"", (char *)NULL);
    return TCL_ERROR;
  }
  *filename_ptr= argv[i];
  *flags_ptr= flags ? flags : CP40_ATTR_PREPROCESSING;
  return TCL_OK;
}

/*---------------------------------------------------------------------------*/

typedef int close_or_delete_op_t(cp40_instance_t *instance, int update);

static int close_or_delete(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			   int argc, char *argv[], close_or_delete_op_t *op);

static int close_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			int argc, char *argv[])
{ return close_or_delete(data, interp, argc, argv, cp40_close_stream); }

static int delete_method(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			int argc, char *argv[])
{ return close_or_delete(data, interp, argc, argv, cp40_delete_instance); }

static int close_or_delete(cp40tcl_instance_data_t *data, Tcl_Interp *interp,
			   int argc, char *argv[], close_or_delete_op_t *op)
{
  int update;
  if (argc==3 && strcmp(argv[2], "-noupdate")==0) {
    update= 0;
  } else if (argc==2) {
    update= cp40_updatable(data->owner);
  } else {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		     argv[1], " ?-noupdate?\"", NULL);
    return TCL_ERROR;
  }
  if (data->connection>=0) {
    Tcl_DeleteFileHandler(data->connection);
    data->connection= -1;
  }
  if (op(data->owner, update) != GPL_OK) {
    return report_error(interp, NULL);
  }
  return TCL_OK;
}

/*---------------------------------------------------------------------------*/
/* UTILITIES */

static int report_error(Tcl_Interp *interp, const char *msg)
{
    Tcl_SetResult(interp, (char *)(msg ? msg : gpl_error_info(0)), TCL_VOLATILE);
    return TCL_ERROR;
}

#if 0
static int get_channel(Tcl_Interp *interp, char *str, int *chn_ptr);
/*	Get CP40 channel number from string STR and store it into CHN_PTR.
	Return standard Tcl result.  In case of error, a message is left
	in INTERP->result. */

static int get_channel(Tcl_Interp *interp, char *str, int *chn_ptr)
{
  char buf[TCL_DOUBLE_SPACE];
  if (Tcl_GetInt(interp, str, chn_ptr) != TCL_OK
      || (unsigned)*chn_ptr >= CP40_NCHANNELS) {
    sprintf(buf, "%d", *chn_ptr);
    Tcl_AppendResult(interp, "bad channel number ", buf, NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}
#endif

/*---------------------------------------------------------------------------*/
