/*
 * cp40yeti.c --
 *
 *	C part of the Yorick interface with the CP40 library.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1997-2004 Eric Thibaut.
 *
 *	This file is part of Yeti.
 *	
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *	
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *	
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: cp40yeti.c,v 1.1 2005/06/29 20:35:32 eric Exp $
 *	$Log: cp40yeti.c,v $
 *	Revision 1.1  2005/06/29 20:35:32  eric
 *	Initial revision
 *
 */

#include "yeti.h"
#include "cp40.h"
#include "yio.h"

/*---------------------------------------------------------------------------*/
/* report_error: call YError with error message from GPLib. */
static void report_error(int clear) __attribute__ ((noreturn));

static void report_error(int clear)
{
  YError(gpl_error_info(clear));
}

static void at_least_n_args(const char *name, int n)
     __attribute__ ((noreturn));
static void exactly_n_args(const char *name, int n)
     __attribute__ ((noreturn));

#define SMALL_WORKSPACE_SIZE 128
static char small_workspace[SMALL_WORKSPACE_SIZE];

static void at_least_n_args(const char *name, int n)
{
  sprintf(small_workspace, "builtin %s takes at least %d argument%s",
	  name, n, n>1?"s":"");
  YError(small_workspace);
}

static void exactly_n_args(const char *name, int n)
{
  sprintf(small_workspace, "builtin %s takes exactly %d argument%s",
	  name, n, n>1?"s":"");
  YError(small_workspace);
}

/*---------------------------------------------------------------------------*/
/* Implementation of an opaque CP40 instance as a foreign Yorick
   data type (see examples in graph.c or ydrat.c). */

static void _y_cp40_delete(void *data);
static void _y_cp40_print(void *data);
static const yeti_opaque_class_t _y_cp40_class= {
  "CP40", _y_cp40_delete, _y_cp40_print
};

/* _y_cp40_delete is automatically called by Yorick to delete a CP40 instance
   that is no more referenced. */
static void _y_cp40_delete(void *data)
{
  cp40_instance_t *instance= (cp40_instance_t *)data;
  if (cp40_delete_instance(instance, 1) != GPL_OK) {
    report_error(1);
  }
}

/* _y_cp40_print is used by Yorick's info command. */
static void _y_cp40_print(void *data)
{
  cp40_instance_t *instance= (cp40_instance_t *)data;

  ForceNewline();
  PrintFunc("CP40 instance: ");
  if (! instance) {
    PrintFunc("*** CORRUPTED ***");
  } else if (cp40_get_io_mode(instance) == CP40_IO_MODE_NONE) {
    PrintFunc("no stream");
  } else {
    const char *mode_str;
    switch (cp40_get_io_mode(instance)) {
    case CP40_IO_MODE_READ:  mode_str= "READ"; break;
    case CP40_IO_MODE_WRITE: mode_str= "WRITE"; break;
    case CP40_IO_MODE_UPDATE: mode_str= "UPDATE"; break;
    default: mode_str= "*** CORRUPTED ***"; break;
    }
    if (cp40_get_io_type(instance)==CP40_IO_FILE) {
      PrintFunc("file=\"");
      PrintFunc(cp40_get_io_filename(instance));
      PrintFunc("\", mode=");
      PrintFunc(mode_str);
    } else if (cp40_get_io_type(instance)==CP40_IO_SOCKET) {
      char buffer[20];
      PrintFunc("socket= { port=");
      sprintf(buffer, "%d", cp40_get_io_port(instance));
      PrintFunc(", host=\"");
      PrintFunc(cp40_get_io_hostname(instance));
      PrintFunc("\"}, mode= ");
      PrintFunc(mode_str);
    } else {
      PrintFunc("*** UNKNOWN IO TYPE ***");
    }
  }
  ForceNewline();
}

static cp40_instance_t *_y_cp40_get(Symbol *s);
static cp40_instance_t *_y_cp40_get(Symbol *s)
{
  return (cp40_instance_t *)yeti_get_opaque_data(s, &_y_cp40_class);
}

/*---------------------------------------------------------------------------*/
/* assoc_file -- associate CP40 instance with a file stream */

static int assoc_file(cp40_instance_t *instance, Symbol *pathSym,
		      Symbol *modeSym);

static int assoc_file(cp40_instance_t *instance, Symbol *pathSym,
		      Symbol *modeSym)
{
  int mode, status;
  char *path;

  if (modeSym == NULL) {
    mode= CP40_IO_MODE_READ;
  } else {
    char *modeStr= YGetString(modeSym);
    if (modeStr[0] == 'r'
	&& (modeStr[1] == 0 || strcmp(modeStr, "read") == 0)) {
      mode= CP40_IO_MODE_READ;
    } else if (modeStr[0] == 'w'
	       && (modeStr[1] == 0 || strcmp(modeStr, "write") == 0)) {
      mode= CP40_IO_MODE_WRITE;
    } else if (modeStr[0] == 'u'
	       && (modeStr[1] == 0 || strcmp(modeStr, "update") == 0)) {
      mode= CP40_IO_MODE_UPDATE;
    } else {
      return gpl_error("bad mode");
    }
  }
  path = YExpandName(YGetString(pathSym));
  status = cp40_assoc_file(instance, path, mode);
  p_strfree(path);
  return status;
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_new -- Create a new CP40 data instance possibly associated
 *               with a file stream. */

extern BuiltIn Y_cp40_new;

void Y_cp40_new(int argc)
{
  cp40_instance_t *instance;
  Symbol *path, *mode, *stack=sp;

  if (argc > 2) YError("cp40_new takes at most 2 arguments");
  if (CalledAsSubroutine()) YError("cp40_new must be called as a function");
  mode = (argc >= 2 ? stack-- : NULL);
  path = (argc >= 1 ? stack-- : NULL);
  if (yeti_is_nil(path)) path = NULL; /* Must be done since argc=1 when
					 routine called with no argument. */
  if (! (instance = cp40_new_instance())) report_error(1);
  PushDataBlock(yeti_new_opaque(instance, &_y_cp40_class));
  if (path && assoc_file(instance, path, mode) != GPL_OK) report_error(1);
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_assoc -- Built-in function to associate a CP40 instance with a file.
 */

extern BuiltIn Y_cp40_assoc;

void Y_cp40_assoc(int argc)
{
  cp40_instance_t *instance;
  Symbol *mode=NULL, *path, *stack=sp;
  
  if (argc == 3) {
    mode = stack--;
  } else if (argc != 2) {
    YError("cp40_assoc takes 2 or 3 arguments");
  }
  path = stack--;
  instance = _y_cp40_get(stack);
  if (assoc_file(instance, path, mode) != GPL_OK) report_error(1);
  Drop(argc-1); /* left CP40 instance (1st argument) on top of the stack */
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_close -- Close data stream of an existing CP40 data instance (coded
 *                 after the Y_close built-in Yorick routine). */

extern BuiltIn Y_cp40_close;

void Y_cp40_close(int argc)
{
  int update= 1; /* default is to update header */
  int i;
  cp40_instance_t *instance= NULL;

  /* Parse arguments from first to last one. */
  for (i=1-argc; i<=0; i++) {
    if (sp[i].ops) {
      /* Normal argument. */
      if (instance == NULL) {
	instance= _y_cp40_get(sp+i);
      } else {
	goto bad_nargs;
      }
    } else {
      /* Must be a keyword: sp[i] is for keyword and sp[i+1] for its value. */
      const char *glob_name= globalTable.names[sp[i++].index];
      if (strcmp(glob_name, "update") == 0) {
	update= YGetInteger(sp+i);
      } else {
	YError("unrecognized keyword in builtin function call");
      }
    }
  }
  if (instance == NULL) {
  bad_nargs:
    YError("cp40_close takes exactly one non-keyword argument");
  }
  if (cp40_close_stream(instance, update) != GPL_OK) {
    report_error(1);
  }
  Drop(argc-1); /* left CP40 instance (1st argument) on top of the stack */
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_rewind -- Go to the beginning of CP40 data stream. */

extern BuiltIn Y_cp40_rewind;

void Y_cp40_rewind(int argc)
{
  if (argc == 1) {
    /* Note: we can use cp40_rewindable(INSTANCE) before... */
    cp40_instance_t *this = _y_cp40_get(sp);
    if (this->stream.ops->rewind(this) != GPL_OK) report_error(1);
  } else {
    if (argc != 1) YError("cp40_rewind takes exactly one argument");
  }
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_readable
   Y_cp40_rewindable
   Y_cp40_updatable
   Y_cp40_writable   -- Check various capabilities of CP40 instance. */

extern BuiltIn Y_cp40_readable;
extern BuiltIn Y_cp40_rewindable;
extern BuiltIn Y_cp40_updatable;
extern BuiltIn Y_cp40_writable;

void Y_cp40_readable(int argc)
{
  if (argc != 1) YError("cp40_readable takes exactly one argument");
  PushIntValue(cp40_readable(_y_cp40_get(sp)));
}

void Y_cp40_rewindable(int argc)
{
  if (argc != 1) YError("cp40_rewindable takes exactly one argument");
  PushIntValue(cp40_rewindable(_y_cp40_get(sp)));
}

void Y_cp40_updatable(int argc)
{
  if (argc != 1) YError("cp40_updatable takes exactly one argument");
  PushIntValue(cp40_updatable(_y_cp40_get(sp)));
}

void Y_cp40_writable(int argc)
{
  if (argc != 1) YError("cp40_writable takes exactly one argument");
  PushIntValue(cp40_writable(_y_cp40_get(sp)));
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_check -- Built-in function to check validity of a CP40 data file. */

extern BuiltIn Y_cp40_check;

void Y_cp40_check(int argc)
{
  int result;
  char *name;
    
  if (argc != 1) YError("cp40_check takes exactly one argument");
  name = YExpandName(YGetString(sp));
  result = cp40_get_file_version(name);
  p_strfree(name);
  if (result < 0) {
    (void)gpl_error_info(1);	/* Clear error. */
    result = 0;
  } else {
    result = 1;
  }
  PushIntValue(result);
}

/*---------------------------------------------------------------------------*/
/* GLOBAL DYNAMIC STRING */

static gpl_dstr_t *global_dstr(void);
static gpl_dstr_t *global_dstr(void)
{
  static gpl_dstr_t dstr;
  static int init = 0;
  if (init) {
    gpl_dstr_reset(&dstr);
  } else {
    gpl_dstr_init(&dstr);
    init = 1;
  }
  return &dstr;
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_get,
 * Y_cp40_set -- Built-in functions to get/set CP40 attribute value(s). */

extern BuiltIn Y_cp40_get;
extern BuiltIn Y_cp40_set;

void Y_cp40_get(int argc)
{
  cp40_instance_t *instance;
  Dimension *dims;
  long i, number;
  char **input, **output;
  gpl_dstr_t *dstr=global_dstr();
  Symbol *argv=sp+1-argc;

  if (argc != 2) YError("cp40_get takes exactly two arguments");
  instance = _y_cp40_get(&argv[0]);
  input = YGet_Q(&argv[1], 1, &dims);
  if (input) {
    number = TotalNumber(dims);
    output = yeti_push_new_string_array(dims)->value.q;
    for (i=0; i<number; i++) {
      if (input[i]) {
	if (cp40_get(instance, input[i], dstr) != GPL_OK) {
	  /* Most probably because attribute is not set but may also be due
	     to some "real" error. In any case, the value will be returned
	     as a nil string. */
	  (void)gpl_error_info(0);	/* Forget error. */
	  output[i] = 0;
	} else {
	  output[i] = p_strcpy(gpl_dstr_value(dstr));
	}
      } else {
	output[i] = 0;
      }
    }
  } else {
    PushDataBlock(RefNC(&nilDB));
  }
  gpl_dstr_free(dstr);
}

void Y_cp40_set(int argc)
{
  cp40_instance_t *instance;
  Dimension *nameDims, *valueDims;
  char **name, **value;
  long i, number;
  int am_subroutine= CalledAsSubroutine();
  Symbol *argv= sp+1-argc;
  
  if (argc != 3) YError("cp40_set takes exactly three arguments");
  instance = _y_cp40_get(&argv[0]);
  name = YGet_Q(&argv[1], 0, &nameDims);
  value = YGet_Q(&argv[2], 0, &valueDims);
  number = yeti_total_number_2(nameDims, valueDims);
  for (i=0; i<number; i++) {
    if (name[i] && cp40_set(instance, name[i], value[i]) != GPL_OK) {
      report_error(1);
    }
  }
  if (! am_subroutine) PushDataBlock(RefNC(&nilDB));
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_extra -- Query name(s) of CP40 non standard attributes. */

extern BuiltIn Y_cp40_extra;

void Y_cp40_extra(int argc)
{
  cp40_instance_t *instance;
  long count, i;
  int index, status;
  const char *name, *value;
  char **output;

  if (argc != 1) YError("cp40_extra takes exactly one argument");
  instance= _y_cp40_get(sp);

  /* Count the number of attributes defined. */
  count= 0L;
  index= -1;
  for (;;) {
    status= cp40_get_next_extra(instance, &index, &name, &value);
    if (status == 0) {
      break;
    } else if (status == 1) {
      count+= 1;
    } else {
      report_error(1);
    }
  }
    
  if (count) {
    /* Make dimension list for a 1D array */
    yeti_reset_dims();	/* MANDATORY!!! */
    tmpDims= NewDimension(count, 1L, (Dimension *)0);
    output= yeti_push_new_string_array(tmpDims)->value.q;
    
    /* Store strings. */
    for (index = -1, i = 0; i<count; i++) {
      if (cp40_get_next_extra(instance, &index, &name, &value) != 1)
	YError("unexpected error in cp40_extra");
      output[i]= p_strcpy(name);
    }
  } else {
    PushDataBlock(RefNC(&nilDB));
  }
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_copy -- Built-in function to copy attributes between data instances
 *                or from/to data instance to/from file. */

extern BuiltIn Y_cp40_copy;
static void get_string_or_instance(Symbol *s, char **string_ptr,
				   cp40_instance_t **instance_ptr);

static void get_string_or_instance(Symbol *s, char **string_ptr,
				   cp40_instance_t **instance_ptr)
{
  Operand op;
  yeti_opaque_t *obj;

  if ((obj= yeti_get_opaque(s, &_y_cp40_class, 0)) != NULL) {
    *string_ptr= NULL;
    *instance_ptr= obj->data;
    return;
  }
  if (!s->ops) YError("unexpected keyword argument");
  s->ops->FormOperand(s, &op);
  if (op.ops==&stringOps && !op.type.dims) {
    *string_ptr= *(char **)op.value;
    *instance_ptr= NULL;
    return;
  }
  YError("expecting CP40 instance argument or scalar string");
}

void Y_cp40_copy(int argc)
{
  char *src_filename= NULL;
  char *dst_filename= NULL;
  cp40_instance_t *src_instance= NULL;
  cp40_instance_t *dst_instance= NULL;
  int i, n= 0, flags= 0, status= GPL_ERROR;
  int am_subroutine= CalledAsSubroutine();

  for (i=1-argc; i<=0; i++) {
    if (sp[i].ops) {
      /* Normal argument. */
      if (++n == 1) {
	get_string_or_instance(sp+i, &src_filename, &src_instance);
      } else if (n == 2) {
	get_string_or_instance(sp+i, &dst_filename, &dst_instance);
      } else {
	goto bad_argc;
      }
    } else {
      /* Must be a keyword: sp[i] is for keyword and sp[i+1] for
       * related value. */
      const char *glob_name= globalTable.names[sp[i++].index];
      if (strcmp(glob_name, "all") == 0) {
	if (YGetInteger(sp+i)) {
	  flags |= CP40_ATTR_ALL;
	}
      } else if (strcmp(glob_name, "extra") == 0) {
	if (YGetInteger(sp+i)) {
	  flags |= CP40_ATTR_EXTRA;
	}
      } else if (strcmp(glob_name, "preprocessing") == 0) {
	if (YGetInteger(sp+i)) {
	  flags |= CP40_ATTR_PREPROCESSING;
	}
      } else if (strcmp(glob_name, "standard") == 0) {
	if (YGetInteger(sp+i)) {
	  flags |= CP40_ATTR_STANDARD;
	}
      } else {
	YError("unrecognized keyword in builtin function call");
      }
    }
  }
  if (n != 2) {
  bad_argc:
    YError("cp40_copy takes exactly 2 non-keyword arguments");
  }
  if (flags == 0)
    flags= CP40_ATTR_PREPROCESSING;
  if (src_instance) {
    if (dst_instance) {
      status= cp40_copy_attr(dst_instance, src_instance, flags);
    } else {
      /* dst_filename is checked by cp40_save_attr(). */
      status= cp40_save_attr(src_instance, dst_filename, flags);
    }
  } else if (dst_instance) {
    /* src_filename is checked by cp40_restore_attr(). */
    status= cp40_restore_attr(dst_instance, src_filename, flags);
  } else {
    YError("one of SRC or DST must be a CP40 instance");
  }
  if (status != GPL_OK)
    report_error(1);
  if (! am_subroutine) {
    Drop(argc);
    PushDataBlock(RefNC(&nilDB));
  }
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_coefs -- Built-in function to set/get distortion coefficients. */

extern BuiltIn Y_cp40_coefs;

void Y_cp40_coefs(int argc)
{
  Dimension *dims= NULL;
  cp40_instance_t *instance= NULL;
  double *coefs= NULL;
  int i, n=0, scaled= 0;
  int am_subroutine= CalledAsSubroutine();

  for (i=1-argc; i<=0; i++) {
    if (sp[i].ops) {
      /* Normal argument. */
      if (n == 0) {
	instance= _y_cp40_get(sp+i);
      } else if (n == 1) {
	/* Get array of raw distortion correction coefficients. */
	coefs= YGet_D(sp, 0, &dims);
	if (!dims || dims->number != CP40_NCHANNELS || !dims->next
	    || dims->next->number != 2 || !dims->next->next
	    || dims->next->next->number != 10
	    || dims->next->next->next) {
	  YError("expecting a (10,2,4) array of coefficients");
	}
      } else {
	goto bad_argc;
      }
      n++;
    } else {
      /* Must be a keyword: sp[i] is for keyword and sp[i+1] for
       * related value. */
      const char *glob_name= globalTable.names[sp[i++].index];
      if (strcmp(glob_name, "scaled") == 0) {
	int *intPtr= YGet_I(sp+i, 1, NULL);
	scaled= (intPtr && *intPtr) ? 1 : 0;
      } else {
	YError("unrecognized keyword in builtin function call");
      }
    }
  }
  if (instance == NULL) {
  bad_argc:
    YError("cp40_coefs takes 1 or 2 non-keyword arguments");
  }
  if (coefs && cp40_set_raw_coefs(instance, coefs) != GPL_OK) {
    report_error(1);
  }
  if (! am_subroutine) {
    if (dims) {
      coefs= yeti_push_new_double_array(dims)->value.d;
    } else {
      /* Prepare dimension list for output array. */
      yeti_reset_dims();	/* MANDATORY!!! */
      tmpDims= NewDimension(10L, 1L, (Dimension *)0);
      tmpDims= NewDimension(2L, 1L, tmpDims);
      tmpDims= NewDimension(CP40_NCHANNELS, 1L, tmpDims);
      coefs= yeti_push_new_double_array(tmpDims)->value.d;
    }
    if (scaled) {
      cp40_get_coefs(instance, coefs);
    } else {
      cp40_get_raw_coefs(instance, coefs);
    }
  }
}

static double *get_2d_d(Symbol *s, long *width, long *height);
extern BuiltIn Y___cp40_reweight;
void Y___cp40_reweight(int argc)
{
  cp40_instance_t *this = NULL;
  int channel;
  double *weight;
  long width, height;

  if (argc != 3) YError("__cp40_reweight takes exactly 3 arguments");
  this = _y_cp40_get(sp - 2);
  channel = YGetInteger(sp - 1);
  weight = get_2d_d(sp, &width, &height);
  if (cp40_reweight_table(this, channel, weight, width, height) != GPL_OK) {
    report_error(1);
  }
  Drop(2); /* left instance on top of the stack */
}

static double *get_2d_d(Symbol *s, long *width, long *height)
{
  Operand op;
  Dimension *dims;
  if (! s->ops) YError("unexpected keyword");
  s->ops->FormOperand(s, &op);
  dims = op.type.dims;
  if (dims) {
    *height = dims->number;
    dims = dims->next;
    if (dims) {
      *width = dims->number;
      if (dims->next) YError("expecting 2D array");
    } else {
      *width = 1;
    }
  } else {
    *width = 1;
    *height = 1;
  }
  op.ops->ToDouble(&op);
  return op.value;  
}

/*---------------------------------------------------------------------------*/
/* Y_cp40_scale,
 * Y_cp40_angle -- Built-in function to set/get double valued attributes. */

static void scalar_attr_d(int argc, int (*set)(cp40_instance_t *, double),
			  double (*get)(cp40_instance_t *));

static void scalar_attr_d(int argc, int (*set)(cp40_instance_t *, double),
			  double (*get)(cp40_instance_t *))
{
  cp40_instance_t *instance;
  Symbol *argv= sp-argc+1;

  if (argc < 1 || argc > 2) YError("expecting 1 or 2 arguments");
  instance= _y_cp40_get(&argv[0]);
  if (argc == 2) {
    /* Set value. */
    if (set == NULL)
      YError("cannot set read-only attribute");
    if ((*set)(instance, YGetReal(&argv[1])) != GPL_OK)
      report_error(1);
  }
  /* Get current value. */
  PushDoubleValue((*get)(instance));
}

extern BuiltIn Y_cp40_angle, Y_cp40_scale;

void Y_cp40_angle(int argc)
{ scalar_attr_d(argc, cp40_set_angle, cp40_get_angle); }

void Y_cp40_scale(int argc)
{ scalar_attr_d(argc, cp40_set_scale, cp40_get_scale); }

/*---------------------------------------------------------------------------*/
/* Y_cp40_nframes,
 * Y_cp40_nsaturations,
 * Y_cp40_frame_count,
 * Y_cp40_frame_date,
 * Y_cp40_exposure -- Built-in functions to set/get long valued attributes. */

static void scalar_attr_i(int argc, int (*set)(cp40_instance_t *, int),
			  int (*get)(cp40_instance_t *));

static void scalar_attr_i(int argc, int (*set)(cp40_instance_t *, int),
			  int (*get)(cp40_instance_t *))
{
  cp40_instance_t *instance;
  Symbol *argv= sp-argc+1;
    
  if (argc < 1 || argc > 2) YError("expecting 1 or 2 arguments");
  instance= _y_cp40_get(&argv[0]);
  if (argc == 2) {
    /* Set value. */
    if (set == NULL)
      YError("cannot set read-only attribute");
    if ((*set)(instance, YGetInteger(&argv[1])) != GPL_OK)
      report_error(1);
  }
  /* Get current value. */
  PushLongValue((*get)(instance));
}

extern BuiltIn Y_cp40_nframes, Y_cp40_nsaturations;
extern BuiltIn Y_cp40_frame_count, Y_cp40_frame_date, Y_cp40_exposure;

void Y_cp40_nframes(int argc)
{ scalar_attr_i(argc, NULL, cp40_get_nframes); }

void Y_cp40_nsaturations(int argc)
{ scalar_attr_i(argc, NULL, cp40_get_nsaturations); }

void Y_cp40_frame_count(int argc)
{ scalar_attr_i(argc, NULL, cp40_get_frame_count); }

void Y_cp40_frame_date(int argc)
{ scalar_attr_i(argc, NULL, cp40_get_frame_date); }

void Y_cp40_exposure(int argc)
{ scalar_attr_i(argc, cp40_set_exposure, cp40_get_exposure); }

/*---------------------------------------------------------------------------*/
/* Y_cp40_nphotons -- Built-in function to get total number of photons. */

extern BuiltIn Y_cp40_nphotons;

void Y_cp40_nphotons(int argc)
{
  cp40_instance_t *instance;
  int chn;
  long *nphotons;

  if (argc != 1) YError("cp40_nphotons takes exactly one argument");
  instance= _y_cp40_get(sp);
  nphotons=
    yeti_push_new_long_array(yeti_first_dimension(CP40_NCHANNELS, 1L))->value.l;
  for (chn=0; chn<CP40_NCHANNELS; chn++) {
    nphotons[chn]= cp40_get_nphotons(instance, chn);
    if (nphotons[chn] < 0) report_error(1);
  }
}

/*---------------------------------------------------------------------------*/

extern BuiltIn Y_cp40_schedule;

void Y_cp40_schedule(int argc)
{
  Symbol *argv= sp+1-argc;
  Dimension *dims;
  int i;
  long j, number;
  cp40_instance_t *instance;
  char **token;

  if (argc<2) YError("cp40_schedule takes at least 2 arguments");
  instance= _y_cp40_get(&argv[0]);
  for (i=1 ; i<argc ; ++i) {
    token= YGet_Q(&argv[i], 1, &dims);
    if (! token) continue;
    number= TotalNumber(dims);    
    for (j=0 ; j<number ; ++j) {
      if (token[j] && ! cp40_schedule_task(instance, token[j])) {
	report_error(1);
      }
    }
  }
  Drop(argc-1); /* left CP40 instance (1st argument) on top of the stack */
}

/*---------------------------------------------------------------------------*/

extern BuiltIn Y_cp40_tasks;

void Y_cp40_tasks(int argc)
{
  int i;
  long number;
  cp40_instance_t *instance;
  cp40_task_t *task;
  char **list;

  if (argc!=1) YError("cp40_tasks takes exactly 1 argument");
  instance = _y_cp40_get(sp);
  task = instance->task;
  number = 0;
  while (task) {
    ++number;
    task = task->next;
  }
  if (number < 1) {
    PushDataBlock(RefNC(&nilDB));
  } else {
    list = yeti_push_new_string_array(yeti_first_dimension(number,1))->value.q;
    task = instance->task;
    i= 0;
    while (task) {
      list[i++] = p_strcpy(task->token);
      task = task->next;
    }
  }
}

/*---------------------------------------------------------------------------*/
extern BuiltIn Y_cp40_clear;
extern BuiltIn Y_cp40_cancel;
extern BuiltIn Y_cp40_start;
extern BuiltIn Y_cp40_stop;

static void manage_task(int argc, const char *name,
			void (*manage_one)(cp40_task_t *task),
			void (*manage_all)(cp40_instance_t *instance));

static void manage_task(int argc, const char *name,
			void (*manage_one)(cp40_task_t *task),
			void (*manage_all)(cp40_instance_t *instance))
{
  Symbol *argv= sp+1-argc;
  Dimension *dims;
  int i;
  long j, number;
  cp40_instance_t *instance;
  char **token;
  cp40_task_t *task;

  if (argc<2) at_least_n_args(name, 2);
  instance = _y_cp40_get(&argv[0]);
  for (i=1 ; i<argc ; ++i) {
    token= YGet_Q(&argv[i], 1, &dims);
    if (! token) continue;
    number= TotalNumber(dims);    
    for (j=0 ; j<number ; ++j) {
      if (! token[j]) continue;
      if (strcmp(token[j], "*") == 0) {
	manage_all(instance);
      } else if ((task= cp40_get_task(instance, token[j])) != NULL) {
	manage_one(task);
      } else {
	report_error(1);
      }
    }
  }
  Drop(argc-1); /* left CP40 instance (1st argument) on top of the stack */
}

void Y_cp40_start(int argc)
{
  manage_task(argc, "cp40_start", cp40_start_task, cp40_start_all_tasks);
}

void Y_cp40_stop(int argc)
{
  manage_task(argc, "cp40_stop", cp40_stop_task, cp40_stop_all_tasks);
}

void Y_cp40_cancel(int argc)
{
  manage_task(argc, "cp40_cancel", cp40_cancel_task, cp40_cancel_all_tasks);
}

void Y_cp40_clear(int argc)
{
  manage_task(argc, "cp40_clear", cp40_clear_task, cp40_clear_all_tasks);
}

/*---------------------------------------------------------------------------*/
extern BuiltIn Y_cp40_run;
extern BuiltIn Y_cp40_task_data;
extern BuiltIn Y_cp40_task_limits;

void Y_cp40_run(int argc)
{
  Symbol *argv= sp+1-argc;
  cp40_instance_t *instance;
  int count;

  if (argc!=1 && argc!=2) {
    YError("builtin cp40_run takes 1 or 2 arguments");
  }
  instance= _y_cp40_get(&argv[0]);
  if (argc>=2) {
    count= YGetInteger(&argv[1]);
    if (count < 0) YError("COUNT must be >= 0");
  } else {
    count= -1;
  }
  count= cp40_run_tasks(instance, count);
  if (instance->stream.error) {
    YError("input/output error");
  }
  PushLongValue(count);
}

void Y_cp40_task_data(int argc)
{
  extern Dimension *tmpDims;
  cp40_task_t *task;
  StructDef *struct_def;

  if (argc!=2) exactly_n_args("cp40_task_data", 2);
  if ((task= cp40_get_task(_y_cp40_get(sp-1), YGetString(sp))) == NULL) {
    report_error(1);
  }
#if 0
  struct_def = &longStruct; /* was "long" */
#else
  struct_def = &doubleStruct; /* now "double" */
#endif
  tmpDims= NewDimension(task->height, task->ymin,
			yeti_first_dimension(task->width, task->xmin));
  memcpy(((Array *)PushDataBlock(NewArray(struct_def, tmpDims)))->value.l,
	 task->data, task->width * task->height * struct_def->size);
}

void Y_cp40_task_limits(int argc)
{
  cp40_task_t *task;
  long *limits;

  if (argc!=2) exactly_n_args("cp40_task_limits", 2);
  if ((task= cp40_get_task(_y_cp40_get(sp-1), YGetString(sp))) == NULL) {
    report_error(1);
  }
  limits= yeti_push_new_long_array(yeti_first_dimension(4, 1))->value.l;
  limits[0]= task->xmin;
  limits[1]= task->xmax;
  limits[2]= task->ymin;
  limits[3]= task->ymax;
}

/*---------------------------------------------------------------------------*/

BuiltIn Y_cp40_stat;

#define STAT_DIM1 7
#define STAT_DIM2 (CP40_NCHANNELS)
#define STAT_STRIDE (STAT_DIM1*STAT_DIM2)

void Y_cp40_stat(int argc)
{
  double *buffer;
  cp40_instance_t *instance;
  int status, chn;
  long count, number, length;	/* Buffer has 7*length elements. */
  double *ptr;

  /* Parse arguments and make sure we have at least 2 scratch stack
     elements that will be used to temporarily store worspace arrays. */
  if (argc == 2) {
    if ((count = YGetInteger(sp)) < 0) YError("bad number of frames");
    Drop(1);
  } else if (argc == 1) {
    count = -1;
    CheckStack(2);
  } else {
    YError("cp40_stat takes 1 or 2 arguments");
  }

  /* Get CP40 instance and allocate+configure its internal frames. */
  instance = _y_cp40_get(sp);
  cp40_update_cache(instance);

  /* Loop over frames. */
  buffer = NULL;
  length = 0;
  number = 0;
  status = 0;
  do {
    if (count) {
      /* Read next frame. */
      if (count>0 && number>=count) break;
      status = cp40_read_frame(instance);
      if (status != 1) {
	if (status == 0) break;
	report_error(1);
      }
    }
    if (number >= length) {
      /* Augment size of buffer (yeti_push_workspace always return a valid
         address or call YError). */
      double *old_buffer = buffer;
      length = (length ? 2*length : 1000);
      buffer = yeti_push_workspace(length*STAT_STRIDE*sizeof(double));
      if (old_buffer) {
	memcpy(buffer, old_buffer, number*STAT_STRIDE*sizeof(double));
	PopTo(sp-1); /* this delete the former workspace buffer and left
                        only the new workspace on top of the stack */
      }
    }

    /* Copy statistics information. */
    ptr = buffer + STAT_STRIDE*number;
    for (chn=0 ; chn<CP40_NCHANNELS ; ++chn) {
      cp40_frame_t *frame = cp40_get_frame(instance, chn);
      if (! frame) report_error(1);
      ptr[0] = frame->date;
      ptr[1] = frame->flux;
      ptr[2] = frame->x;
      ptr[3] = frame->y;
      ptr[4] = frame->spread;
      ptr[5] = frame->distance;
      ptr[6] = frame->ratio;
      ptr += 7;
    }
    ++number;
  } while (count);

  /* Store statistics into Yorick's array. */
  if (number < 1) {
    PushDataBlock(RefNC(&nilDB));
  } else {
    extern Dimension *tmpDims;
    yeti_reset_dims();
    tmpDims = NewDimension(STAT_DIM1, 1, tmpDims);
    tmpDims = NewDimension(STAT_DIM2, 1, tmpDims);
    tmpDims = NewDimension(number, 1, tmpDims);
    ptr = yeti_push_new_double_array(tmpDims)->value.d;
    memcpy(ptr, buffer, number*STAT_STRIDE*sizeof(double));
  }
}

#undef STAT_DIM1
#undef STAT_DIM2
#undef STAT_STRIDE

/*---------------------------------------------------------------------------*/

extern BuiltIn Y_cp40_shift;

void Y_cp40_shift(int argc)
{
  Symbol *argv= sp+1-argc;
  cp40_instance_t *instance;
  int channel;
  double dx, dy;
    
  /* Parse arguments. */
  if (argc!=4) YError("cp40_shift takes exactly 4 arguments");
  instance= _y_cp40_get(&argv[0]);
  channel= YGetInteger(&argv[1]);
  dx= YGetReal(&argv[2]);
  dy= YGetReal(&argv[3]);

  if (cp40_shift(instance, channel, dx, dy) != GPL_OK) report_error(1);

  Drop(argc-1); /* left CP40 instance (1st argument) on top of the stack */
}

/*---------------------------------------------------------------------------*/
