/*
 * gpl_fft.c --
 *
 *	Multivariate Fast Fourier Transform with mixed-radix algorithm by
 *	R. C. Singleton.  Real FFT in 2D.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	The original FTMR algorithm is from R. C. Singleton (Stanford
 *	Research Institute, Sept. 1968, see Ref. [1]).  This implementation
 *	removes the limitations of the original routine and is written in
 *	C (intead of FORTRAN).
 *
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * References:
 *	[1] R.C. Singleton, `An algorithm for computing the mixed radix fast
 *	    Fourier transform`, IEEE Trans., Audio and Electroacoustics,
 *	    AU-17, No. 2,  pp. 93-103, June 1969.
 *
 * History:
 *	$Id$
 *	$Log$
 */

#ifndef _GPL_FFT_C
#define _GPL_FFT_C 1
#include "gpl_fft.h"
#include <math.h>

#define PI     3.14159265358979323846264338327950
#define RADIAN 6.28318530717958647692528676655901   /* 2 pi */

/*---------------------------------------------------------------------------*/
/* COMPLEX IN-PLACE FFT */

int gpl_fft_c(gpl_fcomplex_t *z, int ntot, int n, int nspan,
	    int sign, void *ws)
{
  return gpl_fft_f(&z->re, &z->im, ntot, n, nspan, (sign>=0?2:-2), ws);
}

int gpl_fft_z(gpl_dcomplex_t *z, int ntot, int n, int nspan,
	    int sign, void *ws)
{
  return gpl_fft_d(&z->re, &z->im, ntot, n, nspan, (sign>=0?2:-2), ws);
}

/*---------------------------------------------------------------------------*/
/* UTILITIES */
int gpl_fft_best_dim(int dim)
{
  int try2, try3, try5;
  int best= 2*dim;

  for (try5=1 ; try5<=dim ; try5*=5) {
    for (try3=try5 ; try3<=dim ; try3*=3) {
      for (try2=try3 ; ; try2*= 2) {
	if (try2>=dim) {
	  if (try2==dim) return dim;
	  if (try2<best) best= try2;
	  break;
	}
      }
    }
  }
  return best;
}

#undef ROLL
#define ROLL(S, type_t)							      \
void GPL_JOIN(gpl_roll_,S)(type_t a[], const type_t b[],			      \
		       int stride, int len, int number, int offset)	      \
{									      \
  type_t *ap;								      \
  const type_t *bp;							      \
  int i, j, k, l;							      \
  int ns= stride*len; /* spacing along `last' dimension */		      \
									      \
  if (number<=0 || len<=0 || stride<=0 || number%ns!=0 || b==a) return;	      \
  if (offset>=0) offset%= len;						      \
  else offset= len + offset%len;					      \
  if (offset==0) {							      \
    memcpy(a, b, sizeof(a[0])*number);					      \
    return;								      \
  }									      \
  if (stride==1) {							      \
    ap= a+offset;							      \
    bp= b+len-offset;							      \
    for (l=0 ; l<number ; l+=len) {					      \
      for (i=0, j=l ; i<offset ; ++i, ++j)    a[j]= bp[j];		      \
      for (i=offset, j=l ; i<len ; ++i, ++j) ap[j]= b[j];		      \
    }									      \
  } else {								      \
   int j1= stride*offset, j2= ns-j1;					      \
   for (l=0 ; l<number ; l+=ns) {					      \
      for (i=0, j=l ; i<offset ; ++i, j+=stride) {			      \
	for (k=0, ap=a+j, bp=b+(j+j2) ; k<stride ; ++k) ap[k]= bp[k];	      \
      }									      \
      for (i=offset, j=l ; i<len ; ++i, j+=stride) {			      \
	for (k=0, ap=a+(j+j1), bp=b+j ; k<stride ; ++k) ap[k]= bp[k];	      \
      }									      \
    }									      \
  }									      \
}
ROLL(b, gpl_byte_t)
ROLL(s, short)
ROLL(i, int)
ROLL(l, long)
ROLL(f, float)
ROLL(d, double)

void gpl_roll_c(gpl_fcomplex_t dst[], const gpl_fcomplex_t src[],
	      int stride, int len, int number, int offset)
{
  gpl_roll_f(&dst->re, &src->re, stride, len*2, number*2, offset*2);
}

void gpl_roll_z(gpl_dcomplex_t dst[], const gpl_dcomplex_t src[],
	      int stride, int len, int number, int offset)
{
  gpl_roll_d(&dst->re, &src->re, stride, len*2, number*2, offset*2);
}

/*---------------------------------------------------------------------------*/
/* FACTORIZATION AND WORKSPACE */

/* WS_N -> number of REAL elements to store given number of integers */
#define WS_N(number_of_integers, size_of_real) \
     (((number_of_integers)*sizeof(int)+(size_of_real)-1)/(size_of_real))

static void *fft_ws= NULL; /* previously allocated internal workspace */
static void *rfft2_ws= NULL;


void gpl_fft_cleanup(void)
{
  void *ws= fft_ws;
  fft_ws= NULL;
  if (ws) free(ws);
}

/* Factorize N in primes for FTMR algorithm. */
void gpl_ftmr_factorize(int n, int factor[])
{
  int j, jj, k=n, kt, m=0, *fac=factor+5;

  /* Factors of 4^2 */
  while (k%16 == 0) {
    fac[m++]= 4;
    k/= 16;
  }

  /* Factors of 3^2, 5^2, 7^2, ... */
  for (j=3 ; (jj= j*j) <= k ; j+=2) {
    while (k%jj == 0) {
      fac[m++]= j;
      k/= jj;
    }
  }

  /* Factors of 4, 2, 3, ... */
  if (k > 4) {
    if (k%4 == 0) {
      fac[m++]= 2;
      k/= 4;
    }
    kt= m;
    j= 2;
    do {
      if (k%j == 0) {
	fac[m++]= j;
	k/= j;
      }
      j= ((j + 1) / 2 << 1) + 1;
    } while (j <= k);
  } else {
    kt= m;
    fac[m]= k;
    if (k != 1) ++m;
  }
  j= kt;
  while (j) {
    fac[m++]= fac[--j];
  }

  /* Search maximum factor MAXF. */
  for (jj=0, j=0 ; j<m ; j++) {
    if (fac[j]>jj) jj= fac[j];
  }
  factor[0]= n;    /* length of dimension */
  factor[1]= m;    /* number of factors */
  factor[2]= kt;
  factor[3]= jj;   /* MAXF */
  factor[4]= n-1;  /* This is a secure guess for MAXP: I'm too lazzy to
		      figure out a better (i.e. smaller) estimate for MAXP. */
}

void *gpl_fft_setup(int n, size_t size_of_real)
{
  int *ws;
  int factor[GPL_FTMR_NFACTORS];
  int i, ni, n0, n1, n2;
  size_t size;

  /* Factorize N in primes. */
  gpl_ftmr_factorize(n, factor);

  /* Allocate workspace. */
  /* N0 = number of REAL elements to store FACTOR + 1 integer
     N1 = number of REAL elements for AT and BT  or SK and CK
     N2 = number of REAL elements for NP */
  n0= WS_N(GPL_FTMR_NFACTORS+1, size_of_real);
  n1= 2 * factor[3];                 /* factor[3]=MAXF */
  n2= WS_N(factor[4], size_of_real); /* factor[4]=MAXP */
  size= (n0 + n1 + (n1 >= n2 ? n1 : n2)) * size_of_real;
  ws= gpl_alloc(size);
  if (ws==NULL) {
    gpl_error("insufficient memory for FTMR workspace");
    return NULL;
  }
  ws[0]= size;
  ni= factor[1]+5; /* factor[1]=M */
  for (i=0 ; i<ni ; i++) {
    ws[i+1]= factor[i];
  }
  return ws;
}

/*---------------------------------------------------------------------------*/
/* INCLUDE THIS FILE TWICE FOR FLOAT AND DOUBLE TYPES */
#define DOUBLE_PRECISION 0
#include __FILE__

#define DOUBLE_PRECISION 1
#include __FILE__

#else /* _GPL_FFT_C defined ---------------------------------------------------*/
/* Some constants. */
#ifndef DOUBLE_PRECISION
# define DOUBLE_PRECISION 0
#endif
#if DOUBLE_PRECISION
# define real_t    double
# define complex_t gpl_dcomplex_t
# define FFT       gpl_fft_d
# define FTMR      gpl_ftmr_d
# define F_RFFT2   gpl_forward_rfft2_d
# define B_RFFT2   gpl_backward_rfft2_d
# define ZERO      0.0
# define ONE       1.0
# define TWO       2.0
# define RAD       6.28318530717958647692528676655901   /* 2 pi */
# define C72       0.30901699437494742410229341718282   /* cos(72 deg.) */
# define S72       0.95105651629515357211643933337938   /* sin(72 deg.) */
# define C120     -0.5                                  /* cos(120 deg.) */
# define S120      0.86602540378443864676372317075294   /* sin(120 deg.) */
# define C144     -0.80901699437494742410229341718282   /* cos(144 deg.) */
# define S144      0.58778525229247312916870595463907   /* sin(144 deg.) */
#else
# define real_t    float
# define complex_t gpl_fcomplex_t
# define FFT       gpl_fft_f
# define FTMR      gpl_ftmr_f
# define F_RFFT2   gpl_forward_rfft2_f
# define B_RFFT2   gpl_backward_rfft2_f
# define ZERO      0.0F
# define ONE       1.0F
# define TWO       2.0F
# define RAD       6.28318530717958647692528676655901F  /* 2 pi */
# define C72       0.30901699437494742410229341718282F  /* cos(72 deg.) */
# define S72       0.95105651629515357211643933337938F  /* sin(72 deg.) */
# define C120     -0.5F                                 /* cos(120 deg.) */
# define S120      0.86602540378443864676372317075294F  /* sin(120 deg.) */
# define C144     -0.80901699437494742410229341718282F  /* cos(144 deg.) */
# define S144      0.58778525229247312916870595463907F  /* sin(144 deg.) */
#endif

int FFT(real_t *a, real_t *b, int ntot, int n, 
	int nspan, int isn, void *workspace)
{
  /* N0 = number of REAL elements to store FACTOR + 1 integer
     N1 = number of REAL elements for AT and BT  or SK and CK
     N2 = number of REAL elements for NP */
  int *ws= workspace;
  int n0= WS_N(GPL_FTMR_NFACTORS+1, sizeof(real_t));

  if (ws!=NULL) {
    /* Check input workspace. */
    int n1= 2 * ws[4];                   /* ws[4]=factor[3]=MAXF */
    int n2= WS_N(ws[5], sizeof(real_t)); /* ws[5]=factor[4]=MAXP */
    if (ws[1]!=n) {
      return gpl_error("bad FTMR workspace");
    }
    if (ws[0]< (n0 + n1 + (n1>=n2 ? n1 : n2)) * sizeof(real_t)) {
      return gpl_error("too small FTMR workspace");
    }
  } else {
    /* Allocate new workspace or re-use previous one. */
    ws= fft_ws;
    if (ws!=NULL) {
      if (ws[1]!=n) {
	ws= NULL;
      } else {
	int n1= 2 * ws[4];                   /* ws[4]=factor[3]=MAXF */
	int n2= WS_N(ws[5], sizeof(real_t)); /* ws[5]=factor[4]=MAXP */
	if (ws[0]< (n0 + n1 + (n1>=n2 ? n1 : n2)) * sizeof(real_t)) {
	  ws= NULL;
	}
      }
    }
    if (ws==NULL) {
      gpl_fft_cleanup();
      fft_ws= gpl_fft_setup(n, sizeof(real_t));
      if (fft_ws==NULL) return GPL_ERROR;
      ws= fft_ws;
    }
  }
  return FTMR(a, b, ntot, n, nspan, isn,
	      (int *)ws+1, (real_t *)ws+n0);
}

#undef SWAP
#define SWAP(tmp, a, b) tmp=a; a= b; b= tmp

int FTMR(real_t *a, real_t *b, int ntot, int n,
	 int nspan, int isn, int *factor, real_t *at)
{
  /* Builtin functions */
  extern double sin(double), cos(double);
  
  /* Local variables */
  int nfac[GPL_FTMR_NFACTORS-5];
  real_t radf;
  int j, k, kspan;
  int kspnn, k1, k2, k3, k4;
  real_t aa, bb, cd, aj;
  int jc;
  real_t ak, bk;
  int jf;
  int ii;
  real_t bj;
  int jj;
  real_t sd;
  int kk;
  int nn;
  int ks, nt;
  real_t rad, ajm, akm;
  int inc;
  real_t ajp, akp, bkp, bkm, bjp, bjm;
  int m, kt, maxf, maxp, *np;
  real_t *bt, *sk, *ck;
  
  /* Simple check. */
  if (n < 2) return GPL_OK; /* Nothing to do. */
  if (n != factor[0]) {
    return gpl_error("invalid FACTOR array in FTMR");
  }

  /* Parameter adjustments (FORTRAN -> C). */
  --b;
  --a;

  /* Extract factorization information (we must make a private copy of
     the prime factor decomposition because it is overwritten). */
  m= factor[1];
  kt= factor[2];
  maxf= factor[3];
  maxp= factor[4];
  for (j=0, factor+=5 ; j<m ; j++) nfac[j]= factor[j];

  /* Workspace arrays. */
  bt= at+maxf;
  sk= bt+maxf;
  ck= sk+maxf;
  np= (int *)sk;

  /* Other initialization. */
  if (isn >= 0) {
    inc= isn;
    rad= RAD;
  } else {
    inc= -isn;
    rad= -RAD;
  }
  nt= inc * ntot;
  ks= inc * nspan;
  kspan= ks;
  nn= nt - inc;
  jc= ks / n;
  radf= rad * (real_t) jc / TWO;
  ii= 0;
  jf= 0;

  /* Compute Fourier transform. */
  for (;;) {
    sd= radf / (real_t) kspan;
    cd= sin(sd);
    cd= cd * cd * TWO;
    sd= sin(sd + sd);
    kk= 1;
    k= nfac[ii++];
    if (k == 2) {
      /* Transform for factor of 2 (including rotation factor). */
      kspan/= 2;
      k1= kspan + 2;
      do {
	do {
	  k2= kk + kspan;
	  ak= a[k2];
	  bk= b[k2];
	  a[k2]= a[kk] - ak;
	  b[k2]= b[kk] - bk;
	  a[kk]+= ak;
	  b[kk]+= bk;
	  kk= k2 + kspan;
	} while (kk <= nn);
	kk-= nn;
      } while (kk <= jc);
      if (kk > kspan) {
	break;
      }
      do {
	real_t c1= ONE - cd;
	real_t s1= sd;
	do {
	  do {
	    do {
	      k2= kk + kspan;
	      ak= a[kk] - a[k2];
	      bk= b[kk] - b[k2];
	      a[kk]+= a[k2];
	      b[kk]+= b[k2];
	      a[k2]= c1 * ak - s1 * bk;
	      b[k2]= s1 * ak + c1 * bk;
	      kk= k2 + kspan;
	    } while (kk < nt);
	    k2= kk - nt;
	    c1= -c1;
	    kk= k1 - k2;
	  } while (kk > k2);
	  ak= c1 - (cd * c1 + sd * s1);
	  s1= sd * c1 - cd * s1 + s1;
	  c1= TWO - (ak*ak + s1*s1);
	  s1= c1 * s1;
	  c1*= ak;
	  kk+= jc;
	} while (kk < k2);
	k1= k1 + inc + inc;
	kk= (k1 - kspan) / 2 + jc;
      } while (kk <= jc + jc);

    } else if (k == 4) {

      /* Transform for factor of 4. */
      kspnn= kspan;
      kspan/= 4;
      do {
	real_t c1= ONE, s1= ZERO;
	real_t c2= ONE, s2= ZERO;/* (initialized just to avoid warnings) */
	real_t c3= ONE, s3= ZERO;/* (initialized just to avoid warnings) */
	for (;;) {
	  k1= kk + kspan;
	  k2= k1 + kspan;
	  k3= k2 + kspan;
	  akp= a[kk] + a[k2];
	  akm= a[kk] - a[k2];
	  ajp= a[k1] + a[k3];
	  ajm= a[k1] - a[k3];
	  a[kk]= akp + ajp;
	  ajp= akp - ajp;
	  bkp= b[kk] + b[k2];
	  bkm= b[kk] - b[k2];
	  bjp= b[k1] + b[k3];
	  bjm= b[k1] - b[k3];
	  b[kk]= bkp + bjp;
	  bjp= bkp - bjp;
	  if (isn < 0) {
	    akp= akm + bjm;
	    akm-= bjm;
	    bkp= bkm - ajm;
	    bkm+= ajm;
	  } else {
	    akp= akm - bjm;
	    akm+= bjm;
	    bkp= bkm + ajm;
	    bkm-= ajm;
	  }
	  if (s1 == ZERO) {
	    a[k1]= akp;
	    b[k1]= bkp;
	    a[k2]= ajp;
	    b[k2]= bjp;
	    a[k3]= akm;
	    b[k3]= bkm;
	  } else {
	    a[k1]= akp*c1 - bkp*s1;
	    b[k1]= akp*s1 + bkp*c1;
	    a[k2]= ajp*c2 - bjp*s2;
	    b[k2]= ajp*s2 + bjp*c2;
	    a[k3]= akm*c3 - bkm*s3;
	    b[k3]= akm*s3 + bkm*c3;
	  }
	  kk= k3 + kspan;
	  if (kk > nt) {
	    kk+= jc-nt;
	    if (kk > kspan) break;
	    c2= c1 - (cd*c1 + sd*s1);
	    s1= sd*c1 - cd*s1 + s1;
	    c1= TWO - (c2*c2 + s1*s1);
	    s1*= c1;
	    c1*= c2;
	    c2= c1*c1 - s1*s1;
	    s2= c1*TWO*s1;
	    c3= c2*c1 - s2*s1;
	    s3= c2*s1 + s2*c1;
	  }
	}
	kk+= inc - kspan;
      } while (kk <= jc);

      if (kspan == jc) {
	break;
      }

    } else {

      /*  Transform for odd factors. */

      kspnn= kspan;
      kspan/= k;
      if (k == 3) {
	/* Transform for factor of 3 (optional code). */
	real_t c120= C120;
	real_t s120= (isn>=0 ? S120 : -S120);
	do {
	  do {
	    k1= kk + kspan;
	    k2= k1 + kspan;
	    ak= a[kk];
	    bk= b[kk];
	    aj= a[k1] + a[k2];
	    bj= b[k1] + b[k2];
	    a[kk]= ak + aj;
	    b[kk]= bk + bj;
	    ak+= aj * c120;
	    bk+= bj * c120;
	    aj= (a[k1] - a[k2]) * s120;
	    bj= (b[k1] - b[k2]) * s120;
	    a[k1]= ak - bj;
	    b[k1]= bk + aj;
	    a[k2]= ak + bj;
	    b[k2]= bk - aj;
	    kk= k2 + kspan;
	  } while (kk < nn);
	  kk-= nn;
	} while (kk <= kspan);
      } else if (k == 5) {
	/* Transform for factor of 5 (optional code). */
	real_t c72= C72;
	real_t s72= (isn>=0 ? S72 : -S72);
	real_t c2= C144;
	real_t s2= (isn>=0 ? S144 : -S144);
	do {
	  do {
	    k1= kk + kspan;
	    k2= k1 + kspan;
	    k3= k2 + kspan;
	    k4= k3 + kspan;
	    akp= a[k1] + a[k4];
	    akm= a[k1] - a[k4];
	    bkp= b[k1] + b[k4];
	    bkm= b[k1] - b[k4];
	    ajp= a[k2] + a[k3];
	    ajm= a[k2] - a[k3];
	    bjp= b[k2] + b[k3];
	    bjm= b[k2] - b[k3];
	    aa= a[kk];
	    bb= b[kk];
	    a[kk]= aa + akp + ajp;
	    b[kk]= bb + bkp + bjp;
	    ak= akp * c72 + ajp * c2 + aa;
	    bk= bkp * c72 + bjp * c2 + bb;
	    aj= akm * s72 + ajm * s2;
	    bj= bkm * s72 + bjm * s2;
	    a[k1]= ak - bj;
	    a[k4]= ak + bj;
	    b[k1]= bk + aj;
	    b[k4]= bk - aj;
	    ak= akp * c2 + ajp * c72 + aa;
	    bk= bkp * c2 + bjp * c72 + bb;
	    aj= akm * s2 - ajm * s72;
	    bj= bkm * s2 - bjm * s72;
	    a[k2]= ak - bj;
	    a[k3]= ak + bj;
	    b[k2]= bk + aj;
	    b[k3]= bk - aj;
	    kk= k4 + kspan;
	  } while (kk < nn);
	  kk-= nn;
	} while (kk <= kspan);
      } else {
	/* Transform for any other odd factors. */
	if (k != jf) {
	  real_t c1, s1= rad / (real_t) k;
	  c1= cos(s1);
	  s1= sin(s1);
	  if ((jf= k) > maxf) return gpl_error("MAXF too small in FTMR");
	  ck[jf-1]= ONE;
	  sk[jf-1]= ZERO;
	  j= 1;
	  do {
	    --k;
	    ck[j-1]= ck[k] * c1 + sk[k] * s1;
	    sk[j-1]= ck[k] * s1 - sk[k] * c1;
	    ck[k-1]= ck[j-1];
	    sk[k-1]= -sk[j-1];
	    ++j;
	  } while (j < k);
	}
	do {
	  do {
	    k1= kk;
	    k2= kk + kspnn;
	    aa= a[kk];
	    bb= b[kk];
	    ak= aa;
	    bk= bb;
	    j= 1;
	    k1+= kspan;
	    do {
	      k2-= kspan;
	      ak+= (at[j]= a[k1] + a[k2]);
	      bk+= (bt[j]= b[k1] + b[k2]);
	      ++j;
	      at[j]= a[k1] - a[k2];
	      bt[j]= b[k1] - b[k2];
	      ++j;
	      k1+= kspan;
	    } while (k1 < k2);
	    a[kk]= ak;
	    b[kk]= bk;
	    k1= kk;
	    k2= kk + kspnn;
	    j= 1;
	    do {
	      k1+= kspan;
	      k2-= kspan;
	      jj= j;
	      ak= aa;
	      bk= bb;
	      aj= ZERO;
	      bj= ZERO;
	      k= 1;
	      do {
		ak= at[k] * ck[jj-1] + ak;
		bk= bt[k] * ck[jj-1] + bk;
		++k;
		aj= at[k] * sk[jj-1] + aj;
		bj= bt[k] * sk[jj-1] + bj;
		++k;
		jj+= j;
		if (jj > jf) {
		  jj-= jf;
		}
	      } while (k < jf);
	      k= jf - j;
	      a[k1]= ak - bj;
	      b[k1]= bk + aj;
	      a[k2]= ak + bj;
	      b[k2]= bk - aj;
	      ++j;
	    } while (j < k);
	    kk+= kspnn;
	  } while (kk <= nn);
	  kk-= nn;
	} while (kk <= kspan);
      }

      /* Multiply by rotation factor (except for factors of 2 and 4) */
      if (ii == m) {
	break;
      }
      kk= jc + 1;
      do {
	real_t c2= ONE-cd;
	real_t s1= sd;
	do {
	  real_t c1= c2;
	  real_t s2= s1;
	  kk+= kspan;
	  do {
	    do {
	      ak= a[kk];
	      a[kk]= c2 * ak - s2 * b[kk];
	      b[kk]= s2 * ak + c2 * b[kk];
	      kk+= kspnn;
	    } while (kk <= nt);
	    ak= s1 * s2;
	    s2= s1 * c2 + c1 * s2;
	    c2= c1 * c2 - ak;
	    kk= kk - nt + kspan;
	  } while (kk <= kspnn);
	  c2= c1 - (cd * c1 + sd * s1);
	  s1+= sd * c1 - cd * s1;
	  c1= TWO - (c2*c2 + s1*s1);
	  s1*= c1;
	  c2*= c1;
	  kk= kk - kspnn + jc;
	} while (kk <= kspan);
	kk= kk - kspan + jc + inc;
      } while (kk <= jc + jc);
    }
  }



  /* Permute the results to normal order---done in two stages. */

  /* STAGE 1: Permutation for square factors of N. */
  np[0]= ks;
  if (kt != 0) {
    k= kt + kt + 1;
    if (m < k) {
      --k;
    }
    j= 1;
    np[k]= jc;
    do {
      np[j]= np[j-1] / nfac[j-1];
      np[k-1]= np[k] * nfac[j-1];
    } while (++j < --k);
    k3= np[k];
    kspan= np[1];
    kk= jc + 1;
    k2= kspan + 1;
    j= 1;
    if (n == ntot) {
      /* Permutation for single-variate transform (optional code) */
    L820:
      ak= a[kk];
      a[kk]= a[k2];
      a[k2]= ak;
      bk= b[kk];
      b[kk]= b[k2];
      b[k2]= bk;
      kk+= inc;
      k2= kspan + k2;
      if (k2 < ks) {
	goto L820;
      }
    L830:
      k2-= np[j - 1];
      ++j;
      k2= np[j] + k2;
      if (k2 > np[j - 1]) {
	goto L830;
      }
      j= 1;
    L840:
      if (kk < k2) {
	goto L820;
      }
      kk+= inc;
      k2= kspan + k2;
      if (k2 < ks) {
	goto L840;
      }
      if (kk < ks) {
	goto L830;
      }
    } else {
      /* Permutation for multivariate transform */
    L850:
      k= kk + jc;
    L860:
      ak= a[kk];
      a[kk]= a[k2];
      a[k2]= ak;
      bk= b[kk];
      b[kk]= b[k2];
      b[k2]= bk;
      kk+= inc;
      k2+= inc;
      if (kk < k) {
	goto L860;
      }
      kk= kk + ks - jc;
      k2= k2 + ks - jc;
      if (kk < nt) {
	goto L850;
      }
      k2= k2 - nt + kspan;
      kk= kk - nt + jc;
      if (k2 < ks) {
	goto L850;
      }
    L870:
      k2-= np[j - 1];
      ++j;
      k2= np[j] + k2;
      if (k2 > np[j - 1]) {
	goto L870;
      }
      j= 1;
    L880:
      if (kk < k2) {
	goto L850;
      }
      kk+= jc;
      k2= kspan + k2;
      if (k2 < ks) {
	goto L880;
      }
      if (kk < ks) {
	goto L870;
      }
    }
    jc= k3;
  }
  if ((kt << 1) + 1 >= m) {
    return 0;
  }
  kspnn= np[kt];


  /* STAGE 2: Permutation for square-free factors of N. */

  j= m - kt;
  nfac[j]= 1;
  do {
    nfac[j-1]*= nfac[j];
  } while (--j != kt);
  nn= nfac[kt++] - 1;
  if (nn > maxp) {
    fprintf(stderr, "MAXP=%d too small (>=%d)\n", maxp, nn);
    return gpl_error("MAXP too small in FTMR");
  }
  jj= 0;
  j= 0;
  for (;;) {
    k2= nfac[kt-1];
    kk= nfac[kt];
    k= kt + 1;
    if (++j > nn) break;
    while ((jj+= kk) >= k2) {
      jj-= k2;
      k2= kk;
      kk= nfac[k++];
    }
    np[j-1]= jj;
  }

  /* Determine the permutation cycles of length greater than 1. */
  j= 0;
  k3= kk; /* (just to avoid compiler warning) */
  for (;;) {
    do {
      kk= np[j++];
    } while (kk < 0);
    if (kk != j) {
      do {
	k= kk;
	kk= np[k-1];
	np[k-1]= -kk;
      } while (kk != j);
      k3= kk;
    } else {
      np[j-1]= -j;
      if (j == nn) break;
    }
  }
  maxf*= inc;

  /* Reorder A and B, following the permutation cycles. */
  while ((nt-= kspnn) >= 0) {
    j= k3 + 1;
    ii= nt - inc + 1;
    do {
      do {
	--j;
      } while (np[j-1] < 0);
      jj= jc;
      do {
	kspan= jj;
	if (jj > maxf) {
	  kspan= maxf;
	}
	jj-= kspan;
	k= np[j - 1];
	kk= jc * k + ii + jj;
	k1= kk + kspan;
	k2= 0;
	do {
	  at[k2]= a[k1];
	  bt[k2]= b[k1];
	  ++k2;
	  k1-= inc;
	} while (k1 != kk);
	do {
	  k1= kk + kspan;
	  k2= k1 - jc * (k + np[k-1]);
	  k= -np[k-1];
	  do {
	    a[k1]= a[k2];
	    b[k1]= b[k2];
	    k1-= inc;
	    k2-= inc;
	  } while (k1 != kk);
	  kk= k2;
	} while (k != j);
	k1= kk + kspan;
	k2= 0;
	do {
	  a[k1]= at[k2];
	  b[k1]= bt[k2];
	  ++k2;
	  k1-= inc;
	} while (k1 != kk);
      } while (jj != 0);
    } while (j != 1);
  }
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/
/* 2D REAL FFT */

int F_RFFT2(complex_t *zdst, const real_t *src, int nx, int ny, real_t *ws)
{
  real_t *dst= (real_t *)zdst;
  double r1, i1, r2, i2;
  double a, b, wa, wb;
  int x, y;
  int k1, k2;
  int n1, n2;
  int nx2= 2*nx;

  if (nx%2) return gpl_error("expecting even 1st dimension in RFFT2");

  /* Copy input in temporary array. */
  if (ws==NULL) {
    ws= rfft2_ws;
    rfft2_ws= NULL;
    if (ws) gpl_free(ws);
    if (!(rfft2_ws= gpl_alloc(nx*ny*sizeof(real_t)))) {
      return gpl_error("insufficient memory for RFFT2");
    }
    ws= rfft2_ws;
  }
  if (src!=ws) {
    memcpy(ws, src, nx*ny*sizeof(real_t));
  }

  /* Compute "in-place" 2D FFT of WS. */
  if (FFT(ws, ws+1, ny*(nx/2), (nx/2), (nx/2), -2, NULL)!=GPL_OK ||
      FFT(ws, ws+1, ny*(nx/2), ny,  ny*(nx/2), -2, NULL)!=GPL_OK) {
    return GPL_ERROR;
  }

  /* Convert WS to DST. */

  /* X=0, Y=0 */
  r1= ws[0]; i1= ws[1];
  dst[0]= r1+i1;
  dst[1]= 0;
  dst[nx]= r1-i1;
  dst[nx+1]= 0;

  /* X=0, Y>0 */
  k1= 0;
  k2= nx*ny;
  n1= 0;
  n2= nx;
  for (y=1 ; y<ny ; y++) {
    r1= ws[k1+=nx]; i1= ws[k1+1];
    r2= ws[k2-=nx]; i2= ws[k2+1];
    dst[n1+=nx2]= 0.5*(r1+r2+i1+i2);
    dst[n1+1]= 0.5*(-r1+r2+i1-i2);
    dst[n2+=nx2]= 0.5*(r1+r2-i1-i2);
    dst[n2+1]= 0.5*(r1-r2+i1-i2);
  }

  /* X > 0 */
  wa=  cos(RADIAN/(double)nx);
  wb= -sin(RADIAN/(double)nx);
  a= 1.0; b= 0.0;
  for (x=1 ; x<(nx/2) ; x++) {
    r1= a*wa-b*wb;		/* A= cos(X*THETA) */
    b= a*wb+b*wa; a= r1;	/* B= sin(X*THETA) */
    k1= 2*x;			/* X>0, Y=0 */
    k2= nx-k1;
    n1= k1;
    n2= nx2-n1;
    r1= ws[k1]; i1= ws[k1+1];
    r2= ws[k2]; i2= ws[k2+1];
    dst[n2]= dst[n1]= 0.5*(r1+r2+(i1+i2)*a+(r1-r2)*b);
    dst[n2+1]= -(dst[n1+1]= 0.5*(i1-i2-(r1-r2)*a+(i1+i2)*b));
    k2= nx*(ny+1)-k1;	        /* X>0, Y>0 */
    n2= nx2*(ny+1)-n1;
    for (y= 1 ; y < ny ; y++) {
      r1= ws[k1+=nx]; i1= ws[k1+1];
      r2= ws[k2-=nx]; i2= ws[k2+1];
      dst[n2-=nx2]= dst[n1+=nx2]= 0.5*(r1+r2+(i1+i2)*a+(r1-r2)*b);
      dst[n2+1]= -(dst[n1+1]= 0.5*(i1-i2-(r1-r2)*a+(i1+i2)*b));
    }
  }

  /* Free memory and return. */
  if (rfft2_ws!=NULL) {
    ws= rfft2_ws;
    rfft2_ws= NULL;
    gpl_free(ws);
  }
  return GPL_OK;
}

int B_RFFT2(real_t *dst, const complex_t *zsrc, int nx, int ny)
{
  const real_t *src= (const real_t *)zsrc;
  double r1, i1, r2, i2;
  double a, b, wa, wb;
  int x, y;
  int k1;
  int n1, n2;
  int nx2= 2*nx;

  if (nx%2) return gpl_error("expecting even 1st dimension in RFFT2");
  
  /* SRC -> DST */
  /* X=0, Y=0 */
  r1= src[0]; r2= src[nx];
  dst[0]= r1+r2;
  dst[1]= r1-r2;
  
  /* X=0, Y>0 */
  k1= 0;
  n1= 0;
  n2= nx*(2*ny+1);
  for (y=1 ; y<ny ; y++) {
    r1= src[n1+=nx2]; i1= src[n1+1];
    r2= src[n2-=nx2]; i2= src[n2+1];
    dst[k1+=nx]= r1+r2-i1-i2;
    dst[k1+1]= r1-r2+i1-i2;
  }

  /* X > 0 */
  wa= cos(RADIAN/(double)nx);
  wb= sin(RADIAN/(double)nx);
  a= 1.0; b= 0.0;
  for (x=1 ; x < (nx/2) ; x++) {
    /* A=cos(X*THETA), B=sin(X*THETA) */
    r1= a*wa-b*wb; b= a*wb+b*wa; a= r1;
    
    /* Y=0 */
    k1= 2*x;
    n1= k1;
    n2= nx-n1;
    r1= src[n1]; i1= src[n1+1];
    r2= src[n2]; i2= src[n2+1];
    dst[k1]= (r1+r2-(i1+i2)*a-(r1-r2)*b);
    dst[k1+1]= (i1-i2+(r1-r2)*a-(i1+i2)*b);
    
    /* Y>0 */
    n2= nx*(2*ny+1)-n1;
    for (y= 1 ; y < ny ; y++) {
      r1= src[n1+=nx2]; i1= src[n1+1];
      r2= src[n2-=nx2]; i2= src[n2+1];
      dst[k1+=nx]= (r1+r2-(i1+i2)*a-(r1-r2)*b);
      dst[k1+1]= (i1-i2+(r1-r2)*a-(i1+i2)*b);
    }
  }
  
  /* Compute "in-place" 2D FFT of DST. */
  if (FFT(dst, dst+1, ny*(nx/2), (nx/2), (nx/2), 2, NULL)!=GPL_OK ||
      FFT(dst, dst+1, ny*(nx/2), ny,  ny*(nx/2), 2, NULL)!=GPL_OK) {
    return GPL_ERROR;
  }

  /* Normalize. */
  n1= nx*ny;
  i1= 1.0/(double)n1;
  for (k1=0 ; k1<n1 ; k1++) {
    dst[k1]*= i1;
  }
  
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/
/* Undefine macros to allow for re-include. */
#undef DOUBLE_PRECISION
#undef real_t
#undef complex_t
#undef FFT
#undef FTMR
#undef F_RFFT2
#undef B_RFFT2
#undef ZERO
#undef ONE
#undef TWO
#undef RAD
#undef C72
#undef S72
#undef C120
#undef S120
#undef C144
#undef S144
#endif /* _GPL_FFT_C ----------------------------------------------------------*/
