/*
 * gpl_hash.h --
 *
 *	Hash tables and editable hash files for GPLib library.
 *
 *	Implementation of in-memory hash tables is mostly based on
 *	implementation made in Tcl with some small improvements (take care
 *	of memory allocation failures, const char keys, use of INLINE
 *	declaration, ...).  Look at the file `doc/tcl_hash.txt' distributed
 *	with the sources of the GPLib library for a full
 *	documentation/tutorial about the use of hash tables (within Tcl).
 *
 * Copyright (c) 1991-1993 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 *	See the file "Tcl.license.terms" for information on usage and
 *	redistribution of Tcl sources and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef GPL_HASH_H
#define GPL_HASH_H 1

#include <stdio.h>
#include "gpl_base.h"

GPL_BEGIN_DECLS

/*---------------------------------------------------------------------------*/

/* Types defined for hash tables. */
typedef struct _gpl_hash_table  gpl_hash_table_t;
typedef struct _gpl_hash_entry  gpl_hash_entry_t;
typedef struct _gpl_hash_search gpl_hash_search_t;
typedef struct _gpl_hash_ops    gpl_hash_ops_t;

/* Exported procedures / macros to manage hash tables: */

#define GPL_HASH_STRING_KEYS     0
#define GPL_HASH_ONE_WORD_KEYS   1
extern void gpl_hash_init_table(gpl_hash_table_t *table, int key_type);
/*	Creates a new hash table and stores  information  about  the  table at
	*TABLE.  KEY_TYPE is either  GPL_HASH_STRING_KEYS,  GPL_HASH_ONE_WORD_KEYS, or
	an integer greater than 1 */

extern void gpl_hash_delete_table(gpl_hash_table_t *table);
/*	Deletes all the entries in the hash table and frees up related
	storage. */

#define gpl_hash_create_entry(table, key, new) (table)->ops->create(table, key, new)
/*	Returns a pointer to the entry in TABLE whose key is KEY,
	creating a new entry if needed.  *NEW is set to 1 if a new entry
	was created or 0 if the entry already existed. */

extern void gpl_hash_delete_entry(gpl_hash_entry_t *entry);
/*	Deletes ENTRY from its hash table.  ENTRY should never again be used
	by the caller. Call ENTRY->delete to free ENTRY->client_data if such a
	callback was registered, otherwise it is up to the caller to free the
	ENTRY->client_data member of the entry, if that is relevant. */

extern void gpl_hash_set_value(gpl_hash_entry_t *entry, void *client_data,
			 void (*delete)(void *client_data));
/*	Sets the value associated with a hash table entry as well as its
	`delete' method.   Which can be a user routine or one of: */
#define GPL_HASH_DYNAMIC gpl_mem_free	/* Delete if non-null. */
#define GPL_HASH_STATIC  NULL	        /* Do nothing. */

#define gpl_hash_get_value(h) ((h)->client_data)
/*	Returns the value associated with a hash table entry H. */

#define gpl_hash_get_key(h)   \
    ((char *) (((h)->table->key_type == GPL_HASH_ONE_WORD_KEYS) \
	    ? (h)->key.one_word_value : (h)->key.string))
/* Returns the key associated with a hash table entry H. */

#define gpl_hash_find_entry(table, key)  (table)->ops->find(table, key)
/*	Returns a pointer to the entry in TABLE whose key is KEY,  or NULL
	if no such entry exists. */

extern gpl_hash_entry_t *gpl_hash_first_entry(gpl_hash_table_t *table, gpl_hash_search_t *search);
/*	Starts a search through all  the  elements  of  a  hash  table. Stores
	information about  the  search  at  *search_ptr  and  returns  the hash
	table's first entry or NULL if it has no entries. */

extern gpl_hash_entry_t *gpl_hash_next_entry(gpl_hash_search_t *search);
/*	Returns the next entry in the search identified by SEARCH or NULL
	if all entries in the table have been returned. */



extern char *gpl_hash_stats(gpl_hash_table_t *table);
/*	Returns a string giving usage statistics for TABLE.  The  string is
	dynamically allocated and must be freed by the caller. */

/* Structure definition for an entry in a hash table.  No-one outside the
   GPLib routines should access any of these fields directly; use the defined
   macros instead. */
struct _gpl_hash_entry {
  gpl_hash_entry_t *next_ptr;         /* Pointer to next entry in this hash bucket,
				   or NULL for end of chain. */
  gpl_hash_table_t *table;        /* Pointer to table containing entry. */
  gpl_hash_entry_t **bucket_ptr;      /* Pointer to bucket that points to first
				   entry in this entry's chain: used for
				   deleting the entry. */
  void *client_data;            /* Application stores something here with
				   gpl_hash_set_value. */
  void (*delete)(void *);       /* Method called to free client_data. */
  union { /* Key has one of these forms: */
    const char *one_word_value; /* One-word value for key. */
    int words[1];	        /* Multiple integer words for key.  The actual
				   size will be as large as necessary for this
				   table's keys. */
    char string[4];	        /* String for key.  The actual size will be as
				   large as needed to hold the key. */
  } key;		        /* MUST BE LAST FIELD IN RECORD!! */
};

/* Structure definition for a hash table.  Must be in this header file so
   clients can allocate space for these structures, but clients should
   never access any fields in this structure.  */
struct _gpl_hash_table {
  gpl_hash_entry_t **buckets; /* Pointer to bucket array.  Each element points to
			   first entry in bucket's hash chain, or NULL. */
#define GPL_HASH_SMALL_TABLE 4
  gpl_hash_entry_t *static_buckets[GPL_HASH_SMALL_TABLE];
  /* Bucket array used for small tables (to avoid
     mallocs and frees). */
  int num_buckets;	/* Total number of buckets allocated at **buckets. */
  int num_entries;	/* Total number of entries present in table. */
  int rebuild_size;	/* Enlarge table when num_entries gets to be this
			   large. */
  int down_shift;	/* Shift count used in hashing function.  Designed
			   to use high-order bits of randomized keys. */
  int mask;		/* Mask value used in hashing function. */
  int key_type;	        /* Type of keys used in this table.  It's either
			   GPL_HASH_STRING_KEYS, GPL_HASH_ONE_WORD_KEYS, or an integer
			   giving the number of ints that is the size of
			   the key. */
  const gpl_hash_ops_t *ops;  /* Hash operations table. */
};

/* Hash operations table. */
struct _gpl_hash_ops {
  gpl_hash_entry_t *(*find)(gpl_hash_table_t *table, const char *key);
  gpl_hash_entry_t *(*create)(gpl_hash_table_t *table, const char *key, int *new);
};

/* Structure definition for information used to keep track of searches through
   hash tables: */
struct _gpl_hash_search {
  gpl_hash_table_t *table;      /* Table being searched. */
  int next_index;	  /* Index of next bucket to be enumerated
			    after present one. */
  gpl_hash_entry_t *next_entry; /* Next entry to be enumerated in the current
				 bucket. */
};

/*---------------------------------------------------------------------------*/

/* SIMPLE EDITABLE DATABASE FILES
  ==============================

  GPLib library implements a simple mechanism to save hash tables into
  editable database files.  As far as the hash table contents is pure text
  (both keys and values), the database files are editable by hand (i.e.,
  with your favorite text editor) since they follow a very simple syntax:
    - A line that begins with a hash '#' (after any number of spaces) is
      considered as being a comment (treated as a blank line).
    - Hash entry key and value are specified by:
        KEY=VALUE
      where KEY is made of any characters (except '#', '>', '=' or space).
      There may be any number of spaces before and/or after KEY or the
      equal '=': VALUE starts with the first non-space character after the
      '=' and up to the end-of-line.  A too long value can be continued on
      the next line by starting it with a '>'.  If there are no non-space
      character after the '>', a newline '\n' is appended to VALUE;
      otherwise every character, including spaces, after the '>' is
      considered to be part of VALUE up to the end of line.  In order to
      keep track of previous settings, the value of the current entry can
      be reset by a line starting with a '=' as in:
        KEY = OLD_VALUE
            = NEW_VALUE
            = NEWER_VALUE
      There must be no blank/comment lines between successive new values for
      a card entry.
    - Blank lines are ignored.
  In order to handle newlines in VALUE, the convention is to convert newline
  into empty continuation line (and reciprocally).
  
  Another improvement with respect to actual implementation of simple hash
  tables is that I tried to do the necessary housekeeping to avoid problems if
  some of these routines get interrupted, e.g. by user typing ^C.  This may
  seem too paranoid but is really needed if the GPLib library is linked with
  some interactive language, e.g. Yorick or IDL, which allow such
  interruptions. */

extern int gpl_hash_restore(gpl_hash_table_t *table, const char *filename);
/*	Reads editable database file and store its contents in hash table
	pointed by TABLE which is assumed to have been correctly
	initialized -- i.e., with gpl_hash_init_table(TABLE, GPL_HASH_STRING_KEYS) --
	and of which the contents may be freed before overwriten -- i.e.,
	all entries that already exit in the table must have a dynamically
	allocated string value (or NULL) that is referenced nowhere else.
	Returns standard result: GPL_OK or GPL_ERROR.  Beware that the restored
	values are dynamically malloc'ed and that value of existing entries
	are ckfree'd before being overwritten. */

extern int gpl_hash_save(gpl_hash_table_t *tabl, const char *filename, int overwrite);
/*	Saves contents of hash table pointed by TABLE to editable database
	file FILENAME and returns standard result: GPL_OK or GPL_ERROR.  If
	OVERWRITE is zero, FILENAME must not exists.  This routine is a
	very simple driver, you may prefer using the following routines to
	add your own comments. */

extern int gpl_hash_write_entry(FILE *file, const char *key, const char *value,
			  int comment);
/*	Print out a hash entry key-value pair taking care of splitting
	value if it is too long (if COMMENT is non-zero the resulting line
	get commented).  String pointer NAME can be NULL (same as
	empty/blank string).  Returns standard result: GPL_OK or GPL_ERROR. */

extern int gpl_hash_write_comment(FILE *file, const char *string);
/*	Print out a comment line in a database file (STRING must not
	contain any newline; this is not checked).  Returns standard
	result: GPL_OK or GPL_ERROR. */

GPL_END_DECLS
#endif /* GPL_HASH_H ----------------------------------------------------------*/
