/*
 * gpl_mem.c --
 *
 *	Memory allocation routines for GPLib.
 *
 * Copyright and warranty:
 *	Copyright (c) 1999-2001 Eric THIEBAUT.
 *	
 *	This program is  free software; you can redistribute  it and/or modify
 *	it under the  terms of the GNU General Public  License as published by
 *	the delete  Software Foundation; either  version 2 of the  License, or
 *	(at your option) any later version.
 *	
 *	This program  is distributed in the  hope that it will  be useful, but
 *	WITHOUT   ANY  WARRANTY;   without  even   the  implied   warranty  of
 *	MERCHANTABILITY  or FITNESS  FOR A  PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You  should have received  a copy  of the  GNU General  Public License
 *	along with this program (see  file ``LICENSE'' at the top directory of
 *	this distribution);  if not, write to the  delete Software Foundation,
 *	Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#include <string.h>
#include "gpl_base.h"

/* Undef'ine macro version of the routines provided by this file:*/
#undef gpl_mem_free

extern void *gpl_mem_alloc(size_t size);
extern void *gpl_mem_alloc_clear(size_t size);
extern void *gpl_mem_realloc(void *ptr, size_t size);
extern void  gpl_mem_free(void *ptr);

extern void  gpl_mem_free_all_units(gpl_mem_block_t *block);
extern void  gpl_mem_free_unit(gpl_mem_block_t *block, void *unit);
extern void *gpl_mem_next_unit(gpl_mem_block_t *block);

extern void  gpl_dbuf_init(gpl_dbuf_t *dbuf);
extern void  gpl_dbuf_free(gpl_dbuf_t *dbuf);
extern int   gpl_dbuf_resize(gpl_dbuf_t *dbuf, size_t size);
extern int   gpl_dbuf_append(gpl_dbuf_t *dbuf, void *data, size_t nbytes);
extern void  gpl_dbuf_clear(gpl_dbuf_t *dbuf);

extern int gpl_swap_bytes(void *dst, const void *src, int elsize, int number);

/*---------------------------------------------------------------------------*/

static void alloc_error(size_t size);
static void realloc_error(size_t size);

void gpl_mem_free(void *ptr) { if (ptr) gpl_free(ptr); }

void *gpl_mem_alloc(size_t size)
{
  void *ptr= gpl_alloc(size);
  if (! ptr) alloc_error(size);
  return ptr;
}

void *gpl_mem_alloc_clear(size_t size)
{
  void *ptr= gpl_alloc(size);
  if (! ptr) alloc_error(size);
  else memset(ptr, 0, size);
  return ptr;
}

void *gpl_mem_realloc(void *ptr, size_t size)
{
  if (ptr) {
    if (! (ptr= gpl_realloc(ptr, size))) realloc_error(size);
  } else {
    if (! (ptr= gpl_alloc(size))) alloc_error(size);
  }
  return ptr;
}

static void alloc_error(size_t size)
{
  gpl_sys_error("couldn't allocate %lu bytes", (unsigned long)size);
}

static void realloc_error(size_t size)
{
  gpl_sys_error("couldn't reallocate %lu bytes", (unsigned long)size);
}

/*---------------------------------------------------------------------------*/

static void *new_block(gpl_mem_block_t *mb);

void gpl_mem_init_unit(gpl_mem_block_t *mb, size_t unit_size,
		       size_t units_per_block)
{
  mb->unit_list= NULL;
  mb->block_list= NULL;
  mb->unit_size= GPL_MAX(unit_size, 1);
  mb->units_per_block= GPL_MAX(units_per_block, 1);
}

void *gpl_mem_next_unit(gpl_mem_block_t *mb)
{
  void *unit= mb->unit_list;
  
  if (unit) {
    /* (void**)<->(void*) typecast in next line is equivalent to
       list->next if list is a struct List {struct List *next;}... */
    mb->unit_list= *(void **)unit;
    return unit;
  }
  return new_block(mb);
}

void gpl_mem_free_unit(gpl_mem_block_t *mb, void *unit)
{
  *(void **)unit= mb->unit_list;
  mb->unit_list= unit;
}

void gpl_mem_free_all_units(gpl_mem_block_t *mb)
{
  void *block, *next= mb->block_list;
  while ((block= next) != NULL) {
    next= *(void **)block;
    gpl_free(block);
  }
}

#undef BLOCK_HEADER_SIZE
#define BLOCK_HEADER_SIZE GPL_ROUND_UP(sizeof(void *), GPL_ALIGN)

static void *new_block(gpl_mem_block_t *mb)
{
  void *block;
  int block_size, unit_size, units_per_block;

  /* Allocate a new block and insert it in the chain of allocated blocks. */
  if ((unit_size= GPL_ROUND_UP(mb->unit_size, GPL_ALIGN)) < sizeof(void *)
      || (units_per_block= mb->units_per_block) < 1) {
    gpl_error("bad memory block definition");
    return NULL;
  }
  block_size= BLOCK_HEADER_SIZE + unit_size * units_per_block;
  if (! (block= gpl_alloc(block_size))) {
    gpl_error("not enough memory for a new block of units");
    return NULL;
  }
  *(void **)block= mb->block_list;
  mb->block_list= block;

  /* Adavnce to 1st unit in new block. */
  block= (void *)((char *)block + BLOCK_HEADER_SIZE);

  /* Chain other allocated units. */
  if (units_per_block > 1) {
    char *next= (char *)block + unit_size;
    char *last= (char *)block + unit_size*(units_per_block-1);
    char *unit;
    while ((unit= next) < last) {
      next+= unit_size;
      *(void **)unit= next;
    }
    *(void **)last= mb->unit_list;
    mb->unit_list= (char *)block + unit_size;
  }
  return block;
}

/*---------------------------------------------------------------------------*/
/* DYNAMIC BUFFERS */

void gpl_dbuf_init(gpl_dbuf_t *dbuf)
{
  dbuf->size= 0;
  dbuf->max_size= 0;
  dbuf->base= NULL;
}

void gpl_dbuf_free(gpl_dbuf_t *dbuf)
{
  if (dbuf) {
    void *tmp= dbuf->base;
    gpl_dbuf_init(dbuf);
    if (tmp) gpl_free(tmp);
  }
}

int gpl_dbuf_resize(gpl_dbuf_t *dbuf, size_t new_max_size)
{
  void *new, *old;

  /* Make size in multiple of GPL_DBUF_GRAIN. */
  new_max_size= GPL_ROUND_UP(new_max_size, GPL_DBUF_GRAIN);

  if (new_max_size != dbuf->max_size) {
    if (new_max_size < dbuf->size) {
      return gpl_error("cannot truncate dynamic buffer below its actual size");
    }

    /* Allocate new buffer and copy old contents into fresh buffer. */
    if (! (new= gpl_alloc(new_max_size))) {
      return gpl_sys_error("insufficient memory for dynamic buffer");
    }
    old= dbuf->base;
    if (dbuf->size) memcpy(new, old, dbuf->size);

    /* Take care of updating buffer address before size in case we
       get interrupted by some signal. */  
    dbuf->base= new;
    dbuf->max_size= new_max_size;

    /* Free old buffer. */
    if (old) gpl_free(old);
  }
  return GPL_OK;
}

int gpl_dbuf_append(gpl_dbuf_t *dbuf, void *data, size_t nbytes)
{
  if (nbytes>0) {
    if (GPL_DBUF_AVAILABLE(dbuf) < nbytes
	&& gpl_dbuf_resize(dbuf, dbuf->size+nbytes) != GPL_OK) {
      return GPL_ERROR;
    }
    memcpy(GPL_DBUF_TAIL(dbuf), data, nbytes);
    dbuf->size+= nbytes;
  }
  return GPL_OK;
}

void gpl_dbuf_clear(gpl_dbuf_t *dbuf)
{
  if (dbuf->base && dbuf->size) {
    memset(dbuf->base, 0, dbuf->size);
  }
}

/*---------------------------------------------------------------------------*/

static int alloc_2d(void *index_ptr, unsigned elsize, void *ptr,
		   int xmin, int xmax, int ymin, int ymax);

static int alloc_2d(void *index_ptr, unsigned elsize, void *ptr,
		   int xmin, int xmax, int ymin, int ymax)
{
  int width= xmax-xmin+1;
  int height= ymax-ymin+1;
  int pitch= width*elsize;
  int y;
  char **index, *array= ptr;

  /* Check arguments. */
  if (elsize<=0) {
    return gpl_error("bad element size");
  }
  if (width<=0 || height<=0) {
    return gpl_error("bad boundaries");
  }

  /* Allocate memory. */
  GPL_ALLOC_AT(index, height);
  if (array) {
    if (! index) {
      gpl_error("insufficient memory for 2D index");
    }
  } else {
    GPL_ALLOC_AT(array, height*pitch);
    if (! array || ! index) {
      GPL_FREE_MAYBE(index);
      GPL_FREE_MAYBE(array);
      return gpl_error("insufficient memory for indexed matrix");
    }
    memset(array, 0, height*pitch);
  }

  /* Instanciate index. */
  index-= ymin;
  index[ymin]= array - xmin*elsize;
  for (y=ymin ; y<ymax ; y++) {
    index[y+1]= index[y] + pitch;
  }
  *(char ***)index_ptr= index;
  return GPL_OK;
}

#undef FUNCTION
#define FUNCTION(NAME, CTYPE)						      \
int NAME(CTYPE ***index_ptr, CTYPE *array,				      \
	 int xmin, int xmax, int ymin, int ymax)			      \
{ return gpl_alloc_2d_index(index_ptr, sizeof(CTYPE), array,		      \
			  xmin, xmax, ymin, ymax); }
FUNCTION(gpl_alloc_2d_index_b, gpl_byte_t)
FUNCTION(gpl_alloc_2d_index_s, short)
FUNCTION(gpl_alloc_2d_index_i, int)
FUNCTION(gpl_alloc_2d_index_l, long)
FUNCTION(gpl_alloc_2d_index_f, float)
FUNCTION(gpl_alloc_2d_index_d, double)
FUNCTION(gpl_alloc_2d_index_c, gpl_fcomplex_t)
FUNCTION(gpl_alloc_2d_index_z, gpl_dcomplex_t)

#undef FUNCTION
#define FUNCTION(NAME, CTYPE)						      \
void NAME(CTYPE **index, int xmin, int xmax, int ymin, int ymax)	      \
{ gpl_free_2d_index(index, sizeof(CTYPE), xmin, xmax, ymin, ymax); }
FUNCTION(gpl_free_2d_index_b, gpl_byte_t)
FUNCTION(gpl_free_2d_index_s, short)
FUNCTION(gpl_free_2d_index_i, int)
FUNCTION(gpl_free_2d_index_l, long)
FUNCTION(gpl_free_2d_index_f, float)
FUNCTION(gpl_free_2d_index_d, double)
FUNCTION(gpl_free_2d_index_c, gpl_fcomplex_t)
FUNCTION(gpl_free_2d_index_z, gpl_dcomplex_t)

int gpl_alloc_2d_index(void *index_ptr, unsigned elsize, void *array,
		     int xmin, int xmax, int ymin, int ymax)
{
  if (! array) return gpl_error("bad array address");
  return alloc_2d(index_ptr, elsize, array, xmin, xmax, ymin, ymax);
}

void gpl_free_2d_index(void *index, unsigned elsize,
		     int xmin, int xmax, int ymin, int ymax)
{
  if (elsize>0 && xmax>=xmin && ymax>=ymin) {
    char **row= index;
    gpl_free(&row[ymin]);
  }
}

int gpl_alloc_matrix(void *index_ptr, unsigned elsize,
		   int xmin, int xmax, int ymin, int ymax)
{
  return alloc_2d(index_ptr, elsize, NULL, xmin, xmax, ymin, ymax);
}

void gpl_free_matrix(void *index, unsigned elsize,
		     int xmin, int xmax, int ymin, int ymax)
{
  if (elsize>0 && xmax>=xmin && ymax>=ymin) {
    char **row= index;
    gpl_free(&row[ymin][xmin*elsize]);
    gpl_free(&row[ymin]);
  }
}

/*---------------------------------------------------------------------------*/

int gpl_swap_bytes(void *dst, const void *src, int elsize, int number)
{
  int i;
  
#undef SWAP
#define SWAP(i,j) c=p[i];p[i]=p[j];p[j]=c

  if (dst!=src) memcpy(dst, src, elsize*number);
  if (elsize==1) return GPL_OK;
  if (elsize == 2) {
    gpl_uint16_t *p= dst;
    for (i=0 ; i<number ; ++i) {
      p[i]= GPL_BYTESWAP_16(p[i]);
    }
    return GPL_OK;
  }
  if (elsize == 4) {
    gpl_uint32_t *p= dst;
    for (i=0 ; i<number ; ++i) {
      p[i]= GPL_BYTESWAP_32(p[i]);
    }
    return GPL_OK;
  }
  if (elsize == 8) {
    char c, *p= dst;
    for (i=0 ; i<number ; ++i, p+=8) {
      SWAP(0,7);
      SWAP(1,6);
      SWAP(2,5);
      SWAP(3,4);
    }
    return GPL_OK;
  }
  if (elsize == 16) {
    char c, *p= dst;
    for (i=0 ; i<number ; ++i, p+=8) {
      SWAP(0,15);
      SWAP(1,14);
      SWAP(2,13);
      SWAP(3,12);
      SWAP(4,11);
      SWAP(5,10);
      SWAP(6, 9);
      SWAP(7, 8);
    }
    return GPL_OK;
  }
  return gpl_error("cannot swap %d byte data", elsize);
}

/*---------------------------------------------------------------------------*/
