/*
 * gpl_num_blas.c --
 *
 *	Basic Linear Algebra Subroutines in GPLib.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef _GPL_NUM_BLAS_C
# define _GPL_NUM_BLAS_C 1
#include <math.h>
#include "gpl_num.h"

#ifndef UNROLL_LOOPS
# define UNROLL_LOOPS 0 /* I do not see the gain in using `by-hand' loop
			   unrolling with modern compilers. */
#endif


#define GPL_NUM_SINGLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_SINGLE_PRECISION

#define GPL_NUM_DOUBLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_DOUBLE_PRECISION

#else /* _GPL_NUM_BLAS_C defined --------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* DRIVER (HIGH LEVEL) ROUTINES */

#if 0
int SPOFS(real_t a[], int lda, int n, real_t v[], int itask,
	     int *ind, real_t work[])
{
  int info;
  real_t rcond;

  /* Check arguments. */
  if (lda < n) {
    *ind= -1;
    return gpl_error("*** (%s) LDA=%d is less than N=%d", GPL_STRINGIFY(SPOFS),
		   lda, n);
  }
  if (n <= 0) {
    *ind= -2;
    return gpl_error("*** (%s) N=%d is less than 1", GPL_STRINGIFY(SPOFS), n);
  }
  if (itask < 1) {
    *ind= -3;
    return gpl_error("*** (%s) ITASK=%d is less than 1", GPL_STRINGIFY(SPOFS),
		   itask);
  }
  if (itask == 1) {
    /* Factor matrix A into R. */
    SPOCO(a, lda, n, &rcond, work, &info);

    /* Check for positive definite matrix. */
    if (info != 0) {
      *ind = -4;
      return gpl_error("*** (%s) singular or not positive matrix, no solution",
		     GPL_STRINGIFY(SPOFS));
    }

    /* Compute IND (estimate of No. of significant digits) and check
       for IND greater than zero. */
    *ind= log10(rcond / R1MACH(4));
    if (*ind <= 0) {
      *ind = -10;
      return gpl_error("*** (%s) solution may have no significance",
		     GPL_STRINGIFY(SPOFS));
    }
  }

  /* Solve after factoring. */
  SPOSL(a, lda, n, v);
  return GPL_OK;
}
#endif

/*---------------------------------------------------------------------------*/

void SPOFA(real_t *a, int lda, int n, int *info)
{
  int j, k;
  real_t s, t;

  /* Definitions and adjustements to mimics FORTRAN indexing. */
#undef a_
#define a_(a_1,a_2) a[(a_2)*lda + a_1]
  a-= lda + 1;

  /* Function Body */
  for (j=1 ; j<=n ; ++j) {
    *info= j;
    s= ZERO;
    for (k=1 ; k<j ; ++k) {
      t= (a_(k,j) - SDOT(k-1, &a_(1,k), 1, &a_(1,j), 1)) / a_(k,k);
      a_(k,j)= t;
      s+= t * t;
    }
    s= a_(j,j) - s;
    if (s <= ZERO) return; /* the matrix is not positive definite */
    a_(j,j)= SQRT(s);
  }
  *info= 0;
  return;
}

void SPOSL(const real_t *a, int lda, int n, real_t *b)
{
  /* Local variables */
  int k;
  real_t t;

  /* Definitions and adjustements to mimics FORTRAN indexing. */
#undef a_
#define a_(a_1,a_2) a[(a_2)*lda + a_1]
  a-= lda + 1;
  --b;

  /* SOLVE TRANSPOSE(R)*Y = B */
  for (k=1 ; k<=n ; ++k) {
    t= SDOT(k-1, &a_(1, k), 1, &b[1], 1);
    b[k]= (b[k] - t) / a_(k, k);
  }

  /* SOLVE R*X = Y */
  for (k=n ; k>=1 ; --k) {
    t= -(b[k]/= a_(k, k));
    SAXPY(k-1, t, &a_(1, k), 1, &b[1], 1);
  }
}

void SPOCO(real_t *a, int lda, int n, real_t *rcond, real_t *z, int *info)
{
  int i, j, k;
  real_t s, t, anorm, ynorm, ek, sm, wk, wkm;
  int kp1, a_offset;


  /* Definitions and adjustements to mimics FORTRAN indexing. */
#undef a_
#define a_(a_1,a_2) a[(a_2)*lda + a_1]
  a_offset= lda+1;
  a-= a_offset;
  --z;

  /* Find norm of A using only upper half. */
  for (j=1 ; j <= n; ++j) {
    z[j]= SASUM(j, &a_(1, j), 1);
    for (i=1 ; i<j ; ++i) {
      z[i]+= ABS(a_(i, j));
    }
  }


  /* Compute maximum norm. */
  anorm= ZERO;
  for (j=1 ; j<=n; ++j) {
    if (anorm < z[j]) anorm= z[j];
  }

  /* Factor */
  SPOFA(&a[a_offset], lda, n, info);
  if (*info != 0) return;


  /* RCOND = 1/(NORM(A)*(ESTIMATE OF NORM(INVERSE(A))))
     ESTIMATE = NORM(Z)/NORM(Y) where  A*Z = Y  and  A*Y = E.
     The components of E are chosen to cause maximum local
     growth in the elements of W where TRANS(R)*W = E.
     The vectors are frequently rescaled to avoid overflow. */

  /* Solve TRANS(R)*W = E */
  ek= ONE;
  for (j=1 ; j<=n ; ++j) {
    z[j]= ZERO;
  }
  for (k=1 ; k<=n ; ++k) {
    if (z[k] != ZERO) {
      ek= SIGN(ek, -z[k]);
    }
    if ((s= ABS(ek - z[k])) > a_(k, k)) {
      s= a_(k, k) / s;
      SSCAL(n, s, &z[1], 1);
      ek*= s;
    }
    wk= ek - z[k];
    wkm= -ek - z[k];
    s= ABS(wk);
    sm= ABS(wkm);
    wk/= a_(k, k);
    wkm/= a_(k, k);
    kp1 = k + 1;
    if (kp1 <= n) {
      for (j=kp1 ; j<=n ; ++j) {
	sm+= ABS(z[j] + wkm * a_(k, j));
	z[j]+= wk * a_(k, j);
	s+= ABS(z[j]);
      }
      if (s < sm) {
	t= wkm - wk;
	wk= wkm;
	for (j=kp1 ; j<=n ; ++j) {
	  z[j]+= t * a_(k, j);
	}
      }
    }
    z[k]= wk;
  }
  s= ONE / SASUM(n, &z[1], 1);
  SSCAL(n, s, &z[1], 1);

  /* Solve R*Y = W */
  for (k=n ; k>=1 ; --k) {
    if ((s= ABS(z[k])) > a_(k, k)) {
      s= a_(k, k) / s;
      SSCAL(n, s, &z[1], 1);
    }
    z[k]/= a_(k, k);
    t= -z[k];
    SAXPY(k-1, t, &a_(1, k), 1, &z[1], 1);
  }
  s= ONE / SASUM(n, &z[1], 1);
  SSCAL(n, s, &z[1], 1);

  ynorm= ONE;

  /* Solve TRANS(R)*V = Y */
  for (k = 1; k <= n; ++k) {
    z[k]-= SDOT(k-1, &a_(1, k), 1, &z[1], 1);
    if ((s= ABS(z[k])) > a_(k, k)) {
      ynorm*= (s= a_(k, k) / s);
      SSCAL(n, s, &z[1], 1);
    }
    z[k]/= a_(k, k);
  }
  ynorm*= (s= ONE / SASUM(n, &z[1], 1));
  SSCAL(n, s, &z[1], 1);

  /* Solve R*Z = V */
  for (k=n ; k>=1 ; --k) {
    if ((s= ABS(z[k])) > a_(k, k)) {
      ynorm*= (s= a_(k, k) / s);
      SSCAL(n, s, &z[1], 1);
    }
    z[k]/= a_(k, k);
    SAXPY(k-1, -z[k], &a_(1, k), 1, &z[1], 1);
  }

  /* Make ZNORM = 1.0 */
  ynorm*= (s= ONE / SASUM(n, &z[1], 1));
  SSCAL(n, s, &z[1], 1);
  *rcond= anorm!=ZERO ? ynorm/anorm : ZERO;
}

/*---------------------------------------------------------------------------*/
/* Solve tridiagonal system. */
void SGTSL(int n, real_t *c, real_t *d, real_t *e, real_t *b, int *info)
{
  real_t c0, c1, sw;
  int k, kp1, nm1;

  /* Function Body */
  *info= 0;
  c[0]= d[0];
  if ((nm1= n-1) >= 1) {
    d[0]= e[0];
    e[0]= ZERO;
    e[nm1]= ZERO;
    for (k=0, kp1=1 ; k<nm1 ; ++k, ++kp1) {
      /* Find the largest of the two rows. */
      if ((c0= c[k], ABS(c0)) < (c1= c[kp1], ABS(c1))) {
	/* Interchange rows. */
	c[k]= c1;
	/* c[kp1]= c0;  not needed because c[kp1] is set below */
	SWAP3(sw, c0,     c1);
	SWAP3(sw, d[kp1], d[k]);
	SWAP3(sw, e[kp1], e[k]);
	SWAP3(sw, b[kp1], b[k]);
      }

      /* Check for zero elements on the diagonal. */
      if (! c0 /* i.e., c[k]==ZERO */) {
	*info= kp1;
	return;
      }
      c1/= c0; /* i.e., c1= c[kp1] / c[k]; */
      c[kp1]= d[kp1] - c1 * d[k];
      d[kp1]= e[kp1] - c1 * e[k];
      e[kp1]= ZERO;
      b[kp1]-= c1 * b[k];
    }
  }
  if (c[nm1] == ZERO) {
    *info= n;
    return;
  }

  /* Back solve. */
  b[nm1]/= c[nm1];
  if (n > 1) {
    int nm2= n-2;
    b[nm2]= (b[nm2] - d[nm2] * b[nm1]) / c[nm2];
    for (k=n-3 ; k>=0 ; --k) {
      b[k]= (b[k] - d[k]*b[k+1] - e[k]*b[k+2]) / c[k];
    }
  }
}

/*---------------------------------------------------------------------------*/
/* BLAS ROUTINES */

real_t SASUM(integer_t n, const real_t x[], integer_t incr)
{
  real_t result= ZERO, tmp;
  integer_t i;

  if (n <= 0) return result;
  if (incr == 1) {
    /* Code for increment equal to 1. */
#if UNROLL_LOOPS
    /* Clean-up loop so remaining vector length is a multiple of 6. */
    int m= n%6;
    for (i=0 ; i<m ; ++i) {
      result+= ABS_OF_EXPR(tmp, x[i]);
    }
    for (i=m ; i<n ; i+=6) {
      result+= ABS_OF_EXPR(tmp, x[i]);
      result+= ABS_OF_EXPR(tmp, x[i+1]);
      result+= ABS_OF_EXPR(tmp, x[i+2]);
      result+= ABS_OF_EXPR(tmp, x[i+3]);
      result+= ABS_OF_EXPR(tmp, x[i+4]);
      result+= ABS_OF_EXPR(tmp, x[i+5]);
    }
#else
    for (i=0 ; i<n ; ++i) {
      result+= ABS_OF_EXPR(tmp, x[i]);
    }
#endif
  } else {
    /* Code for increment not equal to 1. */
    int ix= incr<0 ? (1-n) * incr : 0;
    for (i=1 ; i<n ; ++i, ix+=incr) {
      result += ABS_OF_EXPR(tmp, x[ix]);
    }
  }
  return result;
}

real_t SCASUM(integer_t n, const complex_t z[], integer_t incr)
{
  const real_t *x;
  real_t result= ZERO, tmp;
  integer_t i;

  if (n <= 0) return result;
  x= (const real_t *)z;
  if (incr == 1) {
    /* Code for increment equal to 1. */
    n*= 2;
    for (i=0 ; i<n ; ++i) {
      result+= ABS_OF_EXPR(tmp, x[i]);
    }
  } else {
    /* Code for increment not equal to 1. */
    int ix= (incr*=2)<0 ? (1-n) * incr : 0;
    for (i=1 ; i<n ; ++i, ix+=incr) {
      result+= ABS_OF_EXPR(tmp, x[ix]);
      result+= ABS_OF_EXPR(tmp, x[ix+1]);
    }
  }
  return result;
}

integer_t ISAMAX(integer_t n, const real_t *x, integer_t incx)
{
  real_t am, a;
  integer_t i, im;

  if (n<=0) return 0;
  --x; /* FORTRAN indices start at 1. */
  im= 1;
  if (incx == 1) {
    /* Code for increments equal to 1. */
    am= ABS(x[1]);
    for (i=2 ; i<=n ; ++i) {
      if ((a= ABS(x[i])) > am) {
	am= a;
	im= i;
      }
    }
  } else {
    /* Code for increment not equal to 1. */
    int ix= incx<0 ? (1-n)*incx+1 : 1;
    am= ABS(x[ix]);
    ix+= incx;
    for (i=2 ; i<=n ; ++i, ix+=incx) {
      if ((a= ABS(x[ix])) > am) {
	am= a;
	im= i;
      }
    }
  }
  return im;
}

real_t SDOT(int n, const real_t *x, int incx, const real_t *y, int incy)
{
  real_t result= ZERO;
  int i;

  if (n <= 0) return result;
  if (incx>0 && incx==incy) {
    if (incx == 1) {
      /* Code for both increments equal to 1. */
#if UNROLL_LOOPS
      int m= n%5;
      for (i=0 ; i<m; ++i) {
	result+= x[i] * y[i];
      }
      for (i=m ; i<n ; i+=5) {
	result+= x[i]*y[i] + x[i+1]*y[i+1] + x[i+2]*y[i+2]
	  + x[i+3]*y[i+3] + x[i+4]*y[i+4];
      }
#else
      for (i=0 ; i<n ; ++i) {
	result+= x[i] * y[i];
      }
#endif
    } else {
      /* Code for equal, positive, non-unit increments. */
      int ns= n * incx;
      for (i=0 ; i<ns; i+=incx) {
	result+= x[i] * y[i];
      }
    }
  } else {
    /* Code for unequal or nonpositive increments. */
    int ix= (incx>0) ? 0 : (1-n)*incx;
    int iy= (incy>0) ? 0 : (1-n)*incy;
    for (i=0 ; i<n ; ++i, ix+=incx, iy+=incy) {
      result+= x[ix] * y[iy];
    }
  }
  return result;
}

#ifdef GPL_NUM_SINGLE_PRECISION
double DSDOT(int n, const real_t *x, int incx, const real_t *y, int incy)
{
  double result= 0.0;
  int i;

  if (n <= 0) return result;
  if (incx>0 && incx==incy) {
    if (incx == 1) {
      /* Code for both increments equal to 1. */
      for (i=0 ; i<n ; ++i) result+= (double)x[i] * (double)y[i];
    } else {
      /* Code for equal, positive, non-unit increments. */
      int ns= n*incx;
      for (i=0 ; i<ns; i+=incx) result+= (double)x[i] * (double)y[i];
    }
  } else {
    /* Code for unequal or nonpositive increments. */
    int ix= (incx>0) ? 0 : (1-n)*incx;
    int iy= (incy>0) ? 0 : (1-n)*incy;
    for (i=0 ; i<n ; ++i, ix+=incx, iy+=incy) {
      result+= (double)x[ix] * (double)y[iy];
    }
  }
  return result;
}
#endif /* GPL_NUM_SINGLE_PRECISION */

void SAXPY(int n, real_t a, const real_t x[], int incx, real_t y[],
	      int incy)
{
  int i;

  if (n<=0 || a==ZERO) return;
  if (incx>0 && incx==incy) {
    if (incx == 1) {
      /* Code for both increments equal to 1. */
#if UNROLL_LOOPS
      int m= n%4;
      for (i=0 ; i<m; ++i) {
	y[i]+= a * x[i];
      }
      for (i=m ; i<n ; i+=4) {
	y[i]+= a * x[i];
	y[i+1]+= a * x[i+1];
	y[i+2]+= a * x[i+2];
	y[i+3]+= a * x[i+3];
      }
#else
      for (i=0 ; i<n ; ++i) {
	y[i]+= a * x[i];
      }
#endif
    } else {
      /* Code for equal, positive, non-unit increments. */
      int ns= n * incx;
      for (i=0 ; i<ns ; i+=incx) {
	y[i]+= a * x[i];
      }
    }
  } else {
    /* Code for unequal or nonpositive increments. */
    int ix= (incx>0) ? 0 : (1-n)*incx;
    int iy= (incy>0) ? 0 : (1-n)*incy;
    for (i=0 ; i<n ; ++i, ix+=incx, iy+=incy) {
      y[iy]+= a * x[ix];
    }
  }
}

void SSCAL(int n, real_t a, real_t x[], int incx)
{
  int i;

  if (n <= 0) return;
  if (incx == 1) {
    /* Code for increment equal to 1. */
#if UNROLL_LOOPS
    int m= n%5;
    for (i=0 ; i<m; ++i) x[i]*= a;
    for (i=m ; i<n ; i+=5) {
      x[i]*= a;
      x[i+1]*= a;
      x[i+2]*= a;
      x[i+3]*= a;
      x[i+4]*= a;
    }
#else
    for (i=0 ; i<n ; ++i) x[i]*= a;
#endif
  } else {
    /* Code for increment not equal to 1. */
    int ix= (incx>0) ? 0 : (1-n)*incx;
    for (i=0 ; i<n ; ++i, ix+=incx) x[ix]*= a;
  }
}

/* Euclidian norm (any increment). */
real_t SNRM2(int n, const real_t x[], int incx)
{
  real_t axm, axi, sx2;
  int i;

  /* Shortcuts. */
  if (n < 1 || incx < 1) return ZERO;
  if (n == 1) return ABS(x[0]);

  /* Loop over all elements, avoiding underflows and overflows. */
  axm= sx2= ZERO;
  n*= incx;
  for (i=0 ; i<n ; i+=incx) {
    if ((axi= x[i])<0) axi= -axi;
    if (axi > axm) {
      axm/= axi;
      sx2= sx2*axm*axm + ONE;
      axm= axi;
    } else if (axi) {
      axi/= axm;
      sx2+= axi*axi;
    }
  }
  return axm * SQRT(sx2);
}

/* Euclidian norm (unit increment). */
real_t ENORM(integer_t n, const real_t x[])
{
  double sx2= 0.0, axi, axm= 0.0;
  int i;
  for (i=0 ; i<n ; ++i) {
    if ((axi= fabs(x[i])) > axm) {
      axm/= axi;
      sx2= axm*axm*sx2 + 1.0;
      axm= axi;
    } else if (axi) {
      axi/= axm;
      sx2+= axi*axi;
    }
  }
  return (real_t)(axm * sqrt(sx2));
}

/* Vector copy. */
void SCOPY(int n, const real_t x[], int incx, real_t y[], int incy)
{
  int i;

  if (n <= 0) return;
  if (incx>0 && incx==incy) {
    if (incx == 1) {
      /* Code for both increments equal to 1. */
#if UNROLL_LOOPS
      int m= n%7;
      for (i=0 ; i<m; ++i) {
	y[i]= x[i];
      }
      for (i=m ; i<n ; i+=7) {
	y[i]= x[i];
	y[i+1]= x[i+1];
	y[i+2]= x[i+2];
	y[i+3]= x[i+3];
	y[i+4]= x[i+4];
	y[i+5]= x[i+5];
	y[i+6]= x[i+6];
      }
#else
      for (i=0 ; i<n ; ++i) {
	y[i]= x[i];
      }
#endif
    } else {
      /* Code for equal, positive, non-unit increments. */
      int ns= n * incx;
      for (i=0 ; i<ns; i+=incx) {
	y[i]= x[i];
      }
    }
  } else {
    /* Code for unequal or nonpositive increments. */
    int ix= (incx>0) ? 0 : (1-n)*incx;
    int iy= (incy>0) ? 0 : (1-n)*incy;
    for (i=0 ; i<n ; ++i, ix+=incx, iy+=incy) {
      y[iy]= x[ix];
    }
  }
}

/* Swap vector elements. */
void SSWAP(int n, real_t x[], int incx, real_t y[], int incy)
{
  real_t tmp0;
  int i;

  if (n<=0) return;
  if (incx>0 && incx==incy) {
    if (incx == 1) {
      /* Code for both increments equal to 1. */
#if UNROLL_LOOPS
      real_t tmp1, tmp2;
      int m= n%3;
      for (i=0 ; i<m; ++i) {
	tmp0= x[i]; x[i]= y[i]; y[i]= tmp0;
      }
      for (i=m ; i<n ; i+=3) {
	tmp0= x[i];
	tmp1= x[i+1];
	tmp2= x[i+2];
	x[i]= y[i];
	x[i+1]= y[i+1];
	x[i+2]= y[i+2];
	y[i]= tmp0;
	y[i+1]= tmp1;
	y[i+2]= tmp2;
      }
#else
      for (i=0 ; i<n ; ++i) {
	tmp0= x[i]; x[i]= y[i]; y[i]= tmp0;
      }
#endif
    } else {
      /* Code for equal, positive, non-unit increments. */
      int ns= n*incx;
      for (i=0 ; i<ns ; i+=incx) {
	tmp0= x[i]; x[i]= y[i]; y[i]= tmp0;
      }
    }
  } else {
    /* Code for unequal or nonpositive increments. */
    int ix= (incx>0) ? 0 : (1-n)*incx;
    int iy= (incy>0) ? 0 : (1-n)*incy;
    for (i=0 ; i<n ; ++i, ix+=incx, iy+=incy) {
      tmp0= x[ix]; x[ix]= y[iy]; y[iy]= tmp0;
    }
  }
}

/* Apply a plane Givens rotation. */
void SROT(int n, real_t x[], int incx, real_t y[], int incy,
	     real_t cs, real_t sn)
{
  real_t xp, yp;
  int i;

  if (n<=0 || (cs==ONE && sn==ZERO)) return;
  if (incx>0 && incx==incy) {
    if (incx == 1) {
      /* Code for both increments equal to 1. */
      for (i=0 ; i<n ; ++i) {
	xp= x[i]; yp= y[i];
	x[i]= cs*xp + sn*yp;
	y[i]= cs*yp - sn*xp;
      }
    } else {
      /* Code for equal, positive, non-unit increments. */
      int ns= n*incx;
      for (i=0 ; i<ns ; i+=incx) {
	xp= x[i]; yp= y[i];
	x[i]= cs*xp + sn*yp;
	y[i]= cs*yp - sn*xp;
      }
    }
  } else {
    /* Code for unequal or nonpositive increments. */
    int ix= (incx>0) ? 0 : (1-n)*incx;
    int iy= (incy>0) ? 0 : (1-n)*incy;
    for (i=0 ; i<n ; ++i, ix+=incx, iy+=incy) {
      xp= x[ix]; yp= y[iy];
      x[ix]= cs*xp + sn*yp;
      y[iy]= cs*yp - sn*xp;
    }
  }
}

/* Construct a plane Givens rotation. */
void SROTG(real_t *a, real_t *b, real_t *cs, real_t *sn)
{
#if 0
  real_t u, v, r;
  if (ABS(a[0]) > ABS(b[0])) {
    /* Notes:
       - U and R have the sign of A.
       - CS is positive. */
    u= TWO *  a[0];
    v= b[0] / u;
#ifdef GPL_NUM_SINGLE_PRECISION
    r= SQRT(0.25F + v*v) * u;
#else
    r= SQRT(0.25 + v*v) * u;
#endif
    cs[0]= a[0] / r;
    sn[0]= TWO * v * cs[0];
    b[0]= sn[0];
    a[0]= r;
  } else /* ABS(A) <= ABS(B) */ {
    if (b[0] == ZERO) {
      /* A = B = 0.0 */
      cs[0]= ONE;
      sn[0]= ZERO;
    } else {
      /* Notes:
	 - U and R have the sign of B;
         - R could be immediately stored in A;
	 - SN is positive. */
      u= TWO * b[0];
      v= a[0] / u;
      r= SQRT(ONE_4TH + v*v) * u;
      sn[0]= b[0] / r;
      cs[0]= TWO * v * sn[0];
      b[0]= cs[0] ? ONE/cs[0] : ONE;
      a[0]= r;
    }
  }
#else
  real_t v, r;
  if (ABS(a[0]) > ABS(b[0])) {
    v= b[0] / a[0];
    r= sqrt(ONE + v*v);
    cs[0]= ONE / r;
    sn[0]= v * cs[0];
    b[0]= sn[0];
    a[0]*= r;
  } else /* ABS(A) <= ABS(B) */ {
    if (b[0] == ZERO) {
      /* A = B = 0.0 */
      cs[0]= ONE;
      sn[0]= ZERO;
    } else {
      v= a[0] / b[0];
      r= SQRT(ONE + v*v);
      sn[0]= ONE / r;
      cs[0]= v * sn[0];
      a[0]= b[0] * r;
      b[0]= cs[0] ? ONE/cs[0] : ONE;
    }
  }
#endif
}

#endif /* _GPL_NUM_BLAS_C -----------------------------------------------------*/
