/*
 * gpl_num_lu.c --
 *
 *	Solve linear system of linear equations by LU factorization.
 *	Original FORTRAN routines are from SLATEC and LINPACK.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef _GPL_NUM_LU_C
# define _GPL_NUM_LU_C 1
#include <math.h>
#include "gpl_num.h"

#define GPL_NUM_SINGLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_SINGLE_PRECISION

#define GPL_NUM_DOUBLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_DOUBLE_PRECISION

#else /* _GPL_NUM_LU_C defined ------------------------------------------------*/

/* SGEFA Factor a matrix using Gaussian elimination. */

void SGEFA(real_t *a, integer_t lda, integer_t n, integer_t *ipvt, 
	   integer_t *info)
{
  integer_t j, k, l;
  real_t t;

  /* Parameter adjustments and macro to mimic FORTRAN indexing. */
#undef a_
#define a_(a1,a2) a[(a2)*lda + (a1)]
  a-= lda+1;
  --ipvt;

  /* Gaussian elimination with partial pivoting. */
  *info= 0;
  for (k=1 ; k<n ; ++k) {

    /* Find L = pivot index */
    l= ISAMAX(n-k+1, &a_(k,k), 1) + k-1;
    ipvt[k]= l;
    
    /* ZERO pivot implies this column already triangularized. */
    if (a_(l,k) == ZERO) {
      *info= k;
      continue;
    }

    /* Interchange if necessary. */
    if (l!= k) {
      t= a_(l,k);
      a_(l,k)= a_(k,k);
      a_(k,k)= t;
    }

    /* Compute multipliers. */
    SSCAL(n-k, -ONE/a_(k,k), &a_(k+1,k), 1);
    
    /* Row elimination with column indexing. */
    for (j=k+1 ; j<=n ; ++j) {
      t= a_(l,j);
      if (l!=k) {
	a_(l,j)= a_(k,j);
	a_(k,j)= t;
      }
      SAXPY(n-k, t, &a_(k+1,k), 1, &a_(k+1,j), 1);
    }
  }
  ipvt[n]= n;
  if (a_(n,n) == ZERO) *info= n;
}
/*---------------------------------------------------------------------------*/
/* SGECO Factor a matrix using Gaussian elimination and estimate
 *       the condition number of the matrix. */

void SGECO(real_t *a, integer_t lda, integer_t n, integer_t *ipvt, 
	   real_t *rcond, real_t *z)
{
  integer_t info, j, k, l;
  real_t s, t, anorm, ynorm, ek, sm, wk, wkm;
 
  /* Parameter adjustments and macro to mimic FORTRAN indexing. */
#undef a_
#define a_(a1,a2) a[(a2)*lda + (a1)]
  a-= lda+1;
  --ipvt;
  --z;

  /* Compute 1-norm of A. */
  anorm= ZERO;
  for (j=1 ; j<=n ; ++j) {
    t= SASUM(n, &a_(1,j), 1);
    if (anorm<t) anorm= t;
  }

  /* Factor. */
  SGEFA(&a_(1,1), lda, n, &ipvt[1], &info);

  /* RCOND = 1/(NORM(A)*(ESTIMATE OF NORM(INVERSE(A)))) .
   * where:
   *   ESTIMATE= NORM(Z)/NORM(Y)  where  A*Z = Y  and  TRANS(A)*Y = E .
   *   TRANS(A)  is the transpose of A .  The components of  E  are chosen
   *   to cause maximum local growth in the elements of W  where
   *   TRANS(U)*W = E .  The vectors are frequently rescaled to avoid
   *   overflow. */
  
  /* Solve TRANS(U)*W = E */
  ek= ONE;
  for (j=1 ; j<=n ; ++j) {
    z[j]= ZERO;
  }
  for (k=1 ; k<=n ; ++k) {
    if (z[k] != ZERO) ek= SIGN(ek, -z[k]);
    if ((t= ek-z[k]) < ZERO) t= -t; /* T = ABS(EK-Z(K)) */
    if ((s= a_(k,k)) < ZERO) s= -s; /* S = ABS(A(K,K)) */
    if (t > s) {
      s/= t;
      SSCAL(n, s, &z[1], 1);
      ek*= s;
    }
    wk= ek - z[k];
    wkm= -ek - z[k];
    s= ABS(wk);
    sm= ABS(wkm);
    if ((t= a_(k,k)) /* != ZERO */) {
      wk/= t;
      wkm/= t;
    } else {
      wk= ONE;
      wkm= ONE;
    }
    if (k < n) {
      for (j=k+1 ; j<=n ; ++j) {
	sm+= ABS_OF_EXPR(t, z[j] + wkm*a_(k,j));
	z[j]+= wk*a_(k,j);
	s+= ABS_OF_EXPR(t, z[j]);
      }
      if (s < sm) {
	t= wkm - wk;
	wk= wkm;
	for (j=k+1 ; j<=n ; ++j) {
	  z[j]+= t * a_(k,j);
	}
      }
    }
    z[k]= wk;    
  }
  s= ONE / SASUM(n, &z[1], 1);
  SSCAL(n, s, &z[1], 1);

  /* Solve TRANS(L)*Y = W */
  for (k=n ; k>=1 ; --k) {
    if (k < n) z[k]+= SDOT(n-k, &a_(k+1,k), 1, &z[k+1], 1);
    if ((t= z[k]) < ZERO) t= -t; /* T = ABS(Z(K)) */
    if (t > ONE) SSCAL(n, ONE/t, &z[1], 1);
    l= ipvt[k];
    if (l!=k) {
      SWAP3(t, z[l], z[k]);
    }
  }
  s= ONE / SASUM(n, &z[1], 1);
  SSCAL(n, s, &z[1], 1);
  
  ynorm= ONE;

  /* Solve L*V = Y */
  for (k=1 ; k<=n ; ++k) {
    l= ipvt[k];
    t= z[l];
    if (l!=k) {
      z[l]= z[k];
      z[k]= t;
    }
    if (k < n) SAXPY(n-k, t, &a_(k+1,k), 1, &z[k+1], 1);
    if ((s= z[k]) < ZERO) s= -s; /* S = ABS(Z(K)) */
    if (s > ONE) {
      s= ONE/s; /* S = 1.0/ABS(Z(K)) */
      SSCAL(n, s, &z[1], 1);
      ynorm*= s;
    }
  }
  s= ONE / SASUM(n, &z[1], 1);
  SSCAL(n, s, &z[1], 1);
  ynorm*= s;
  
  /* Solve  U*Z = V */
  for (k=n ; k>=1 ; --k) {
    if ((t= z[k])    < ZERO) t= -t; /* T = ABS(Z(K)) */
    if ((s= a_(k,k)) < ZERO) s= -s; /* S = ABS(A(K,K)) */
    if (t > s) {
      s/= t;
      SSCAL(n, s, &z[1], 1);
      ynorm*= s;
    }
    if ((t= a_(k,k)) /* != ZERO */) {
      z[k]/= t;
    } else {
      z[k]= ONE;
    }
    SAXPY(k-1, -z[k], &a_(1,k), 1, &z[1], 1);
  }
  
  /* Make ZNORM = 1.0 */
  s= ONE / SASUM(n, &z[1], 1);
  SSCAL(n, s, &z[1], 1);
  ynorm*= s;
  *rcond= anorm ? ynorm/anorm : ZERO;
}
/*---------------------------------------------------------------------------*/
/* SGEDI Compute the determinant and inverse of a matrix using the
 *       factors computed by SGECO or SGEFA. */

void SGEDI(real_t *a, integer_t lda, integer_t n, integer_t *ipvt, 
	   real_t *det, real_t *work, integer_t job)
{
  integer_t i, j, k, l;
  real_t t;

  /* Parameter adjustments and macro to mimic FORTRAN indexing. */
#undef a_
#define a_(a1,a2) a[(a2)*lda + a1]
  a-= lda+1;
  --ipvt;
  --work;

  /* Compute determinant. */
  if ((job/10) != 0) {
    real_t dt= ONE, ten= 10;
    int pw= 0;

    for (i=1 ; i<=n ; ++i) {
      if (ipvt[i] != i) dt= -dt;
      dt*= a_(i,i);
      if (dt == ZERO) {
	pw= 0;
	break;
      }
      while (ABS(dt) < ONE) {
	dt*= ten;
	--pw;
      }
      while (ABS(dt) >= ten) {
	dt/= ten;
	++pw;
      }
    }
    det[0]= dt;
    det[1]= pw;
  }

  if ((job%10) != 0) {
    /* Compute INVERSE(U) */
    for (k=1 ; k<=n ; ++k) {
      a_(k,k)= ONE/a_(k,k);
      SSCAL(k-1, -a_(k,k), &a_(1,k), 1);
      for (j=k+1 ; j<=n ; ++j) {
	t= a_(k,j);
	a_(k,j)= ZERO;
	SAXPY(k, t, &a_(1,k), 1, &a_(1,j), 1);
      }
    }

    /* Form INVERSE(U)*INVERSE(L) */
    for (k=n-1 ; k>=1 ; --k) {
      for (i=k+1 ; i<=n ; ++i) {
	work[i]= a_(i,k);
	a_(i,k)= ZERO;
      }
      for (j=k+1 ; j<=n ; ++j) {
	SAXPY(n, work[j], &a_(1,j), 1, &a_(1,k), 1);
      }
      if ((l= ipvt[k]) != k) SSWAP(n, &a_(1,k), 1, &a_(1,l), 1);
    }
  }
}
/*---------------------------------------------------------------------------*/
/* SGESL Solve the real system A*X=B or TRANS(A)*X=B using the
 *          factors of SGECO or SGEFA. */

void SGESL(const real_t *a, integer_t lda, integer_t n,
	   const integer_t *ipvt, real_t *b, integer_t job)
{
  integer_t k, l;
  real_t t;

  /* Parameter adjustments and macro to mimic FORTRAN indexing. */
#undef a_
#define a_(a1,a2) a[(a2)*lda + (a1)]
  a-= lda+1;
  --ipvt;
  --b;

  if (job==0) {
    /* JOB = 0, Solve  A * X = B */
    /*     First solve  L*Y = B */
    for (k=1 ; k<n ; ++k) {
      l= ipvt[k];
      t= b[l];
      if (l!=k) {
	b[l]= b[k];
	b[k]= t;
      }
      SAXPY(n-k, t, &a_(k+1,k), 1, &b[k+1], 1);
    }
    
    /*     Now solve  U*X = Y */
    for (k=n ; k>=1 ; --k) {
      b[k]/= a_(k,k);
      SAXPY(k-1, -b[k], &a_(1,k), 1, &b[1], 1);
    }
  } else {
    /* JOB = NONZERO, Solve  TRANS(A) * X = B */
    /*     First solve  TRANS(U)*Y = B */
    for (k=1 ; k<=n ; ++k) {
      b[k]= (b[k] - SDOT(k-1, &a_(1,k), 1, &b[1], 1)) / a_(k,k);
    }

    /*     Now solve TRANS(L)*X = Y */
    for (k=n-1 ; k>=1 ; --k) {
      b[k]+= SDOT(n-k, &a_(k+1,k), 1, &b[k+1], 1);
      l= ipvt[k];
      if (l!=k) {
	t= b[l];
	b[l]= b[k];
	b[k]= t;
      }
    }
  }
}
/*---------------------------------------------------------------------------*/
/* DRIVER ROUTINE */

/* SGEFS Solve a general system of linear equations. */

int SGEFS(real_t a[], integer_t lda, integer_t n, real_t v[], 
	  integer_t task, integer_t *ind, real_t work[],
	  integer_t ipvt[])
{
  real_t rcond;

  if (lda < n) {
    *ind= -1;
    return gpl_error("%s: LDA=%d is less than N=%d", __FUNCTION__,
		   lda, n);
  }
  if (n <= 0) {
    *ind= -2;
    return gpl_error("%s: N=%d is less than 1", __FUNCTION__, n);
  }
  if (task < 1) {
    *ind= -3;
    return gpl_error("%s: TASK=%d is less than 1", __FUNCTION__, task);
  }
  if (task == 1) {
    /* Factor matrix A into LU */
    SGECO(a, lda, n, ipvt, &rcond, work);

    /* Check for computationally singular matrix. */
    if (rcond == ZERO) {
      *ind= -4;
      return gpl_error("%s: singular matrix A - no solution",
		     __FUNCTION__);
    }

    /* Compute IND (estimate of No. of significant digits) and check
       for IND greater than zero. */
    if ((*ind= -LOG10(R1MACH(4)/rcond)) <= 0) {
      *ind= -10;
      return gpl_error("%s: solution may have no significance",
		     __FUNCTION__);
    }
  }

  /* Solve after factoring. */
  SGESL(a, lda, n, ipvt, v, 0);
  return GPL_OK;
}

/*---------------------------------------------------------------------------*/
/* DRIVER ROUTINE */

/* SGEIR Solve a general system of linear equations.  Iterative
 *       refinement is used to obtain an error estimate. */

int SGEIR(real_t *a, integer_t lda, integer_t n, real_t *v, 
	  integer_t task, integer_t *ind, real_t *work, integer_t *ipvt)
{
  real_t r1, r2;  
  integer_t j, info;
  real_t dnorm, xnorm;

  /* Parameter adjustments and macro to mimic FORTRAN indexing. */
#undef work_
#define work_(a1,a2) work[(a2)*n + a1]
#undef a_
#define a_(a1,a2) a[(a2)*lda + a1]
  a-= lda+1;
  work -= n+1;

  if (lda < n) {
    *ind= -1;
    return gpl_error("%s: LDA=%d is less than N=%d", __FUNCTION__, lda, n);
  }
  if (n <= 0) {
    *ind= -2;
    return gpl_error("%s: N=%d is less than 1", __FUNCTION__, n);
  }
  if (task < 1) {
    *ind= -3;
    return gpl_error("%s: TASK=%d is less than 1", __FUNCTION__, task);
  }
  if (task == 1) {
    /* Move matrix A to WORK. */
    for (j=1 ; j<=n ; ++j) {
      SCOPY(n, &a_(1,j), 1, &work_(1,j), 1);
    }

    /* Factor matrix A into LU */
    SGEFA(&work_(1,1), n, n, ipvt, &info);

    /* Check FOR COMPUTATIONALLY SINGULAR MATRIX */
    if (info != 0) {
      *ind= -4;
      return gpl_error("%s: singular matrix A - no solution", __FUNCTION__);
    }
  }

  /* Solve when factoring complete */
  /* Move vector B to WORK */
  SCOPY(n, v, 1, &work_(1,n+1), 1);
  SGESL(&work_(1,1), n, n, ipvt, v, 0);
  
  /* Form norm of X0. */
  xnorm= SASUM(n, v, 1);
  if (xnorm == ZERO) {
    *ind= 75;
    return GPL_OK;
  }
  
  /* Compute residual R. */
  for (j=1 ; j<=n ; ++j) {
    work_(j,n+1)= DSDOT(n, &a_(j,1), lda, v, 1) - work_(j,n+1);
  }
  
  /* Solve A*DELTA=R */ 
  SGESL(&work_(1,1), n, n, ipvt, &work_(1,n+1), 0);
  
  /* Form norm of DELTA */
  dnorm= SASUM(n, &work_(1,n+1), 1);
  
  /* Compute IND (estimate of No. of significant digits) and check
     for IND greater than zero. */
  r1= R1MACH(4);
  r2= dnorm / xnorm;
  *ind= -LOG10(MAX(r1,r2));
  if (*ind <= 0) {
    *ind= -10;
    return gpl_error("%s: solution may have no significance", __FUNCTION__);
  }
  return GPL_OK;
}
#endif /* _GPL_NUM_LU_C -------------------------------------------------------*/
