/*
 * gpl_ops1.h --
 *
 *	Definitions for vectorized operations for GPLib library.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */
#ifndef GPL_OPS1_H
#define GPL_OPS1_H 1

#include "gpl_base.h"

#ifdef __cplusplus
extern "C" {
#endif
/*-----------------------------------------------------------------------------
 * Suffixes:
 *     _b == `gpl_byte_t'
 *     _s == `short'
 *     _i == `int'
 *     _l == `long'
 *     _f == `float'
 *     _d == `double'
 *     _c == `gpl_fcomplex_t'
 *     _z == `gpl_dcomplex_t'
 *
 * Arguments:
 *     DST    -- Destination array with NUMBER elements.
 *     SRC    -- Source array with NUMBER elements, always `const'.
 *     LEFT   -- Left operand array (NUMBER elements) or scalar (see FLAGS),
 *               always `const'.
 *     RIGHT  -- Right operand array (NUMBER elements) or scalar (see FLAGS),
 *               always `const'.
 *     FLAGS  -- 1 if left operand is scalar, 2 if right operand is scalar,
 *               otherwise both operand have NUMBER elements.
 *     NUMBER -- Number of elements of arrays.
 *
 * Unary operations:
 *     Prototype: unary_op(dst, src, number);
 *     copy       -- copy operation (if DST==SRC, nothing is done).
 *     to_TYPE    -- conversion operations:      DST[i]= (TYPE)SRC[i];
 *     negate     -- unary minus:                DST[i]= -SRC[i];
 *     complement -- two-complement:             DST[i]= ~SRC[i];
 *     not        -- logical unary not:          DST[i]= !SRC[i];
 *     true       -- logical unary true:         DST[i]= !!SRC[i];
 *     square     -- compute square of a vector: DST[i]= SRC[i]*SRC[i];
 *     cube       -- compute cube of a vector:   DST[i]= SRC[i]*SRC[i]*SRC[i];
 *
 * Binary operation:
 *     Prototype: binary_op(dst, left, right, number, flags);
 *     add         -- DST[i]= LEFT[i] + RIGHT[i];
 *     subtract    -- DST[i]= LEFT[i] - RIGHT[i];
 *     multiply    -- DST[i]= LEFT[i] * RIGHT[i];
 *     divide      -- DST[i]= LEFT[i] / RIGHT[i];
 *                    (check for integer division by zero).
 *     modulo      -- DST[i]= LEFT[i] % RIGHT[i];
 *                    (check for integer division by zero).
 *     power       -- DST[i]= pow(LEFT[i], RIGHT[i]);
 *     bitwise_or  -- DST[i]= LEFT[i] | RIGHT[i];
 *     bitwise_and -- DST[i]= LEFT[i] & RIGHT[i];
 *     bitwise_xor -- DST[i]= LEFT[i] ^ RIGHT[i];
 *     shift_left  -- DST[i]= LEFT[i] << RIGHT[i];
 *     shift_right -- DST[i]= LEFT[i] >> RIGHT[i];
 *     equal       -- DST[i]= LEFT[i] == RIGHT[i];
 *     not_equal   -- DST[i]= LEFT[i] != RIGHT[i];
 *     greater     -- DST[i]= LEFT[i] >  RIGHT[i];
 *     greater_eq  -- DST[i]= LEFT[i] >= RIGHT[i];
 *     less        -- DST[i]= LEFT[i] <  RIGHT[i];
 *     less_eq     -- DST[i]= LEFT[i] <= RIGHT[i];
 *     logical_and -- DST[i]= LEFT[i] && RIGHT[i];
 *     logical_or  -- DST[i]= LEFT[i] || RIGHT[i];
 *     min2        -- DST[i]= min(LEFT[i], RIGHT[i]);
 *     max2        -- DST[i]= max(LEFT[i], RIGHT[i]);
 *
 * Common features:
 *     (1) Operations are `element-wise'.
 *     (2) Operations can be performed `in-place', e.g. with DST==SRC or
 *         DST==LEFT and/or DST==RIGHT.
 *
 * Return value:
 *     The result DST (this is to allow fo chaining of operations).
 *
 * Restrictions:
 *     These operators should be considered as low-level ones: they do not
 *     check the validity of their arguments.
 *-----------------------------------------------------------------------------
 */

/* Prototype of unary operators. */
#define _GPL_UNARY_OP_PROTO(OP, TYPE) \
TYPE *OP(TYPE *dst, const TYPE *src, int number)

/* Prototype of unary logical operators. */
#define _GPL_LOGICAL_PROTO(OP, TYPE) \
int *OP(int *dst, const TYPE *src, int number)

extern _GPL_UNARY_OP_PROTO(gpl_copy_b, gpl_byte_t);
extern _GPL_UNARY_OP_PROTO(gpl_copy_s, short);
extern _GPL_UNARY_OP_PROTO(gpl_copy_i, int);
extern _GPL_UNARY_OP_PROTO(gpl_copy_l, long);
extern _GPL_UNARY_OP_PROTO(gpl_copy_f, float);
extern _GPL_UNARY_OP_PROTO(gpl_copy_d, double);
extern _GPL_UNARY_OP_PROTO(gpl_copy_c, gpl_fcomplex_t);
extern _GPL_UNARY_OP_PROTO(gpl_copy_z, gpl_dcomplex_t);

extern _GPL_UNARY_OP_PROTO(gpl_negate_b, gpl_byte_t);
extern _GPL_UNARY_OP_PROTO(gpl_negate_s, short);
extern _GPL_UNARY_OP_PROTO(gpl_negate_i, int);
extern _GPL_UNARY_OP_PROTO(gpl_negate_l, long);
extern _GPL_UNARY_OP_PROTO(gpl_negate_f, float);
extern _GPL_UNARY_OP_PROTO(gpl_negate_d, double);
extern _GPL_UNARY_OP_PROTO(gpl_negate_c, gpl_fcomplex_t);
extern _GPL_UNARY_OP_PROTO(gpl_negate_z, gpl_dcomplex_t);

extern _GPL_UNARY_OP_PROTO(gpl_complement_b, gpl_byte_t);
extern _GPL_UNARY_OP_PROTO(gpl_complement_s, short);
extern _GPL_UNARY_OP_PROTO(gpl_complement_i, int);
extern _GPL_UNARY_OP_PROTO(gpl_complement_l, long);

extern _GPL_LOGICAL_PROTO(gpl_true_b, gpl_byte_t);
extern _GPL_LOGICAL_PROTO(gpl_true_s, short);
extern _GPL_LOGICAL_PROTO(gpl_true_i, int);
extern _GPL_LOGICAL_PROTO(gpl_true_l, long);
extern _GPL_LOGICAL_PROTO(gpl_true_f, float);
extern _GPL_LOGICAL_PROTO(gpl_true_d, double);
extern _GPL_LOGICAL_PROTO(gpl_true_c, gpl_fcomplex_t);
extern _GPL_LOGICAL_PROTO(gpl_true_z, gpl_dcomplex_t);

extern _GPL_LOGICAL_PROTO(gpl_not_b, gpl_byte_t);
extern _GPL_LOGICAL_PROTO(gpl_not_s, short);
extern _GPL_LOGICAL_PROTO(gpl_not_i, int);
extern _GPL_LOGICAL_PROTO(gpl_not_l, long);
extern _GPL_LOGICAL_PROTO(gpl_not_f, float);
extern _GPL_LOGICAL_PROTO(gpl_not_d, double);
extern _GPL_LOGICAL_PROTO(gpl_not_c, gpl_fcomplex_t);
extern _GPL_LOGICAL_PROTO(gpl_not_z, gpl_dcomplex_t);

extern _GPL_UNARY_OP_PROTO(gpl_square_b, gpl_byte_t);
extern _GPL_UNARY_OP_PROTO(gpl_square_s, short);
extern _GPL_UNARY_OP_PROTO(gpl_square_i, int);
extern _GPL_UNARY_OP_PROTO(gpl_square_l, long);
extern _GPL_UNARY_OP_PROTO(gpl_square_f, float);
extern _GPL_UNARY_OP_PROTO(gpl_square_d, double);
extern _GPL_UNARY_OP_PROTO(gpl_square_c, gpl_fcomplex_t);
extern _GPL_UNARY_OP_PROTO(gpl_square_z, gpl_dcomplex_t);

extern _GPL_UNARY_OP_PROTO(gpl_cube_b, gpl_byte_t);
extern _GPL_UNARY_OP_PROTO(gpl_cube_s, short);
extern _GPL_UNARY_OP_PROTO(gpl_cube_i, int);
extern _GPL_UNARY_OP_PROTO(gpl_cube_l, long);
extern _GPL_UNARY_OP_PROTO(gpl_cube_f, float);
extern _GPL_UNARY_OP_PROTO(gpl_cube_d, double);
extern _GPL_UNARY_OP_PROTO(gpl_cube_c, gpl_fcomplex_t);
extern _GPL_UNARY_OP_PROTO(gpl_cube_z, gpl_dcomplex_t);

/*---------------------------------------------------------------------------*/
/* Prototype of conversion operators. */
#define _GPL_CONVERSION_PROTO(OP, DST_TYPE, SRC_TYPE) \
DST_TYPE *OP(DST_TYPE *dst, const SRC_TYPE *src, int number)

extern _GPL_CONVERSION_PROTO(gpl_to_byte_b, gpl_byte_t, gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_byte_s, gpl_byte_t, short);
extern _GPL_CONVERSION_PROTO(gpl_to_byte_i, gpl_byte_t, int);
extern _GPL_CONVERSION_PROTO(gpl_to_byte_l, gpl_byte_t, long);
extern _GPL_CONVERSION_PROTO(gpl_to_byte_f, gpl_byte_t, float);
extern _GPL_CONVERSION_PROTO(gpl_to_byte_d, gpl_byte_t, double);
extern _GPL_CONVERSION_PROTO(gpl_to_byte_c, gpl_byte_t, gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_byte_z, gpl_byte_t, gpl_dcomplex_t);

extern _GPL_CONVERSION_PROTO(gpl_to_short_b, short, gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_short_s, short, short);
extern _GPL_CONVERSION_PROTO(gpl_to_short_i, short, int);
extern _GPL_CONVERSION_PROTO(gpl_to_short_l, short, long);
extern _GPL_CONVERSION_PROTO(gpl_to_short_f, short, float);
extern _GPL_CONVERSION_PROTO(gpl_to_short_d, short, double);
extern _GPL_CONVERSION_PROTO(gpl_to_short_c, short, gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_short_z, short, gpl_dcomplex_t);

extern _GPL_CONVERSION_PROTO(gpl_to_int_b, int, gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_int_s, int, short);
extern _GPL_CONVERSION_PROTO(gpl_to_int_i, int, int);
extern _GPL_CONVERSION_PROTO(gpl_to_int_l, int, long);
extern _GPL_CONVERSION_PROTO(gpl_to_int_f, int, float);
extern _GPL_CONVERSION_PROTO(gpl_to_int_d, int, double);
extern _GPL_CONVERSION_PROTO(gpl_to_int_c, int, gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_int_z, int, gpl_dcomplex_t);

extern _GPL_CONVERSION_PROTO(gpl_to_long_b, long, gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_long_s, long, short);
extern _GPL_CONVERSION_PROTO(gpl_to_long_i, long, int);
extern _GPL_CONVERSION_PROTO(gpl_to_long_l, long, long);
extern _GPL_CONVERSION_PROTO(gpl_to_long_f, long, float);
extern _GPL_CONVERSION_PROTO(gpl_to_long_d, long, double);
extern _GPL_CONVERSION_PROTO(gpl_to_long_c, long, gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_long_z, long, gpl_dcomplex_t);

extern _GPL_CONVERSION_PROTO(gpl_to_float_b, float, gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_float_s, float, short);
extern _GPL_CONVERSION_PROTO(gpl_to_float_i, float, int);
extern _GPL_CONVERSION_PROTO(gpl_to_float_l, float, long);
extern _GPL_CONVERSION_PROTO(gpl_to_float_f, float, float);
extern _GPL_CONVERSION_PROTO(gpl_to_float_d, float, double);
extern _GPL_CONVERSION_PROTO(gpl_to_float_c, float, gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_float_z, float, gpl_dcomplex_t);

extern _GPL_CONVERSION_PROTO(gpl_to_double_b, double, gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_double_s, double, short);
extern _GPL_CONVERSION_PROTO(gpl_to_double_i, double, int);
extern _GPL_CONVERSION_PROTO(gpl_to_double_l, double, long);
extern _GPL_CONVERSION_PROTO(gpl_to_double_f, double, float);
extern _GPL_CONVERSION_PROTO(gpl_to_double_d, double, double);
extern _GPL_CONVERSION_PROTO(gpl_to_double_c, double, gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_double_z, double, gpl_dcomplex_t);

extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_b, gpl_fcomplex_t, gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_s, gpl_fcomplex_t, short);
extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_i, gpl_fcomplex_t, int);
extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_l, gpl_fcomplex_t, long);
extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_f, gpl_fcomplex_t, float);
extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_d, gpl_fcomplex_t, double);
extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_c, gpl_fcomplex_t,
			   gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_fcomplex_z, gpl_fcomplex_t,
			   gpl_dcomplex_t);

extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_b, gpl_dcomplex_t,
			   gpl_byte_t);
extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_s, gpl_dcomplex_t, short);
extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_i, gpl_dcomplex_t, int);
extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_l, gpl_dcomplex_t, long);
extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_f, gpl_dcomplex_t, float);
extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_d, gpl_dcomplex_t, double);
extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_c, gpl_dcomplex_t,
			   gpl_fcomplex_t);
extern _GPL_CONVERSION_PROTO(gpl_to_dcomplex_z, gpl_dcomplex_t,
			   gpl_dcomplex_t);

/*---------------------------------------------------------------------------*/
/* Prototype of binary operator. */
#define _GPL_BINARY_OP_PROTO(OP, TYPE) \
TYPE *OP(TYPE *dst, const TYPE *left, const TYPE *right, int number, int flag)

extern _GPL_BINARY_OP_PROTO(gpl_add_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_add_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_add_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_add_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_add_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_add_d, double);
extern _GPL_BINARY_OP_PROTO(gpl_add_c, gpl_fcomplex_t);
extern _GPL_BINARY_OP_PROTO(gpl_add_z, gpl_dcomplex_t);

extern _GPL_BINARY_OP_PROTO(gpl_subtract_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_subtract_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_subtract_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_subtract_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_subtract_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_subtract_d, double);
extern _GPL_BINARY_OP_PROTO(gpl_subtract_c, gpl_fcomplex_t);
extern _GPL_BINARY_OP_PROTO(gpl_subtract_z, gpl_dcomplex_t);

extern _GPL_BINARY_OP_PROTO(gpl_multiply_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_multiply_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_multiply_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_multiply_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_multiply_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_multiply_d, double);
extern _GPL_BINARY_OP_PROTO(gpl_multiply_c, gpl_fcomplex_t);
extern _GPL_BINARY_OP_PROTO(gpl_multiply_z, gpl_dcomplex_t);

extern _GPL_BINARY_OP_PROTO(gpl_divide_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_divide_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_divide_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_divide_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_divide_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_divide_d, double);
extern _GPL_BINARY_OP_PROTO(gpl_divide_c, gpl_fcomplex_t);
extern _GPL_BINARY_OP_PROTO(gpl_divide_z, gpl_dcomplex_t);

extern _GPL_BINARY_OP_PROTO(gpl_modulo_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_modulo_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_modulo_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_modulo_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_modulo_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_modulo_d, double);
extern _GPL_BINARY_OP_PROTO(gpl_modulo_c, gpl_fcomplex_t);
extern _GPL_BINARY_OP_PROTO(gpl_modulo_z, gpl_dcomplex_t);

extern _GPL_BINARY_OP_PROTO(gpl_power_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_power_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_power_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_power_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_power_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_power_d, double);
extern _GPL_BINARY_OP_PROTO(gpl_power_c, gpl_fcomplex_t);
extern _GPL_BINARY_OP_PROTO(gpl_power_z, gpl_dcomplex_t);

extern _GPL_BINARY_OP_PROTO(gpl_bitwise_or_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_or_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_or_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_or_l, long);

extern _GPL_BINARY_OP_PROTO(gpl_bitwise_and_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_and_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_and_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_and_l, long);

extern _GPL_BINARY_OP_PROTO(gpl_bitwise_xor_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_xor_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_xor_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_bitwise_xor_l, long);

extern _GPL_BINARY_OP_PROTO(gpl_shift_left_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_shift_left_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_shift_left_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_shift_left_l, long);

extern _GPL_BINARY_OP_PROTO(gpl_shift_right_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_shift_right_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_shift_right_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_shift_right_l, long);

extern _GPL_BINARY_OP_PROTO(gpl_min2_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_min2_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_min2_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_min2_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_min2_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_min2_d, double);

extern _GPL_BINARY_OP_PROTO(gpl_max2_b, gpl_byte_t);
extern _GPL_BINARY_OP_PROTO(gpl_max2_s, short);
extern _GPL_BINARY_OP_PROTO(gpl_max2_i, int);
extern _GPL_BINARY_OP_PROTO(gpl_max2_l, long);
extern _GPL_BINARY_OP_PROTO(gpl_max2_f, float);
extern _GPL_BINARY_OP_PROTO(gpl_max2_d, double);

/*---------------------------------------------------------------------------*/
/* Prototype of comparison operator. */
#define _GPL_COMPARISON_PROTO(OP, TYPE) \
int *OP(int *dst, const TYPE *left, const TYPE *right, int number, int flag)

extern _GPL_COMPARISON_PROTO(gpl_equal_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_equal_s, short);
extern _GPL_COMPARISON_PROTO(gpl_equal_i, int);
extern _GPL_COMPARISON_PROTO(gpl_equal_l, long);
extern _GPL_COMPARISON_PROTO(gpl_equal_f, float);
extern _GPL_COMPARISON_PROTO(gpl_equal_d, double);
extern _GPL_COMPARISON_PROTO(gpl_equal_c, gpl_fcomplex_t);
extern _GPL_COMPARISON_PROTO(gpl_equal_z, gpl_dcomplex_t);

extern _GPL_COMPARISON_PROTO(gpl_not_equal_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_not_equal_s, short);
extern _GPL_COMPARISON_PROTO(gpl_not_equal_i, int);
extern _GPL_COMPARISON_PROTO(gpl_not_equal_l, long);
extern _GPL_COMPARISON_PROTO(gpl_not_equal_f, float);
extern _GPL_COMPARISON_PROTO(gpl_not_equal_d, double);
extern _GPL_COMPARISON_PROTO(gpl_not_equal_c, gpl_fcomplex_t);
extern _GPL_COMPARISON_PROTO(gpl_not_equal_z, gpl_dcomplex_t);

extern _GPL_COMPARISON_PROTO(gpl_greater_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_greater_s, short);
extern _GPL_COMPARISON_PROTO(gpl_greater_i, int);
extern _GPL_COMPARISON_PROTO(gpl_greater_l, long);
extern _GPL_COMPARISON_PROTO(gpl_greater_f, float);
extern _GPL_COMPARISON_PROTO(gpl_greater_d, double);

extern _GPL_COMPARISON_PROTO(gpl_greater_eq_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_greater_eq_s, short);
extern _GPL_COMPARISON_PROTO(gpl_greater_eq_i, int);
extern _GPL_COMPARISON_PROTO(gpl_greater_eq_l, long);
extern _GPL_COMPARISON_PROTO(gpl_greater_eq_f, float);
extern _GPL_COMPARISON_PROTO(gpl_greater_eq_d, double);

extern _GPL_COMPARISON_PROTO(gpl_less_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_less_s, short);
extern _GPL_COMPARISON_PROTO(gpl_less_i, int);
extern _GPL_COMPARISON_PROTO(gpl_less_l, long);
extern _GPL_COMPARISON_PROTO(gpl_less_f, float);
extern _GPL_COMPARISON_PROTO(gpl_less_d, double);

extern _GPL_COMPARISON_PROTO(gpl_less_eq_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_less_eq_s, short);
extern _GPL_COMPARISON_PROTO(gpl_less_eq_i, int);
extern _GPL_COMPARISON_PROTO(gpl_less_eq_l, long);
extern _GPL_COMPARISON_PROTO(gpl_less_eq_f, float);
extern _GPL_COMPARISON_PROTO(gpl_less_eq_d, double);

extern _GPL_COMPARISON_PROTO(gpl_logical_or_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_logical_or_s, short);
extern _GPL_COMPARISON_PROTO(gpl_logical_or_i, int);
extern _GPL_COMPARISON_PROTO(gpl_logical_or_l, long);
extern _GPL_COMPARISON_PROTO(gpl_logical_or_f, float);
extern _GPL_COMPARISON_PROTO(gpl_logical_or_d, double);
extern _GPL_COMPARISON_PROTO(gpl_logical_or_c, gpl_fcomplex_t);
extern _GPL_COMPARISON_PROTO(gpl_logical_or_z, gpl_dcomplex_t);

extern _GPL_COMPARISON_PROTO(gpl_logical_and_b, gpl_byte_t);
extern _GPL_COMPARISON_PROTO(gpl_logical_and_s, short);
extern _GPL_COMPARISON_PROTO(gpl_logical_and_i, int);
extern _GPL_COMPARISON_PROTO(gpl_logical_and_l, long);
extern _GPL_COMPARISON_PROTO(gpl_logical_and_f, float);
extern _GPL_COMPARISON_PROTO(gpl_logical_and_d, double);
extern _GPL_COMPARISON_PROTO(gpl_logical_and_c, gpl_fcomplex_t);
extern _GPL_COMPARISON_PROTO(gpl_logical_and_z, gpl_dcomplex_t);

/*---------------------------------------------------------------------------*/
#ifdef __cplusplus
} /* C++ */
#endif
#endif /* GPL_OPS1_H ----------------------------------------------------------*/
