/*
 * gpl_smooth.c --
 *
 *	Convolution and wavelet smoothing along one dimension.
 *
 * Copyright (c) 1997, 1999, 2000 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *----------------------------------------------------------------------------
 */

#ifndef _GPL_SMOOTH_C
#define _GPL_SMOOTH_C 1
#include "gpl_smooth.h"

/*---------------------------------------------------------------------------*/
/* OPERATIONS FOR COMPLEX DATA TYPE */

#undef ZOP
#define ZOP(name, op, complex_t, real_t)				      \
void name(complex_t dst[], const complex_t src[], int stride, int n,	      \
	  int nafter, const real_t ker[], int w, int scale, int border,	      \
	  real_t ws[])							      \
{									      \
  op(&dst[0].re, &src[0].re, 2*stride, n, nafter, ker, w, scale, border, ws); \
  op(&dst[0].im, &src[0].im, 2*stride, n, nafter, ker, w, scale, border, ws); \
}
ZOP(gpl_smooth_c, gpl_smooth_f, gpl_fcomplex_t, float)
ZOP(gpl_smooth_z, gpl_smooth_d, gpl_dcomplex_t, double)

/*---------------------------------------------------------------------------*/
/* OPERATIONS FOR FLOATING POINT DATA TYPE */
#define type_t           float
#define ZERO             0.0f
#define SMOOTH           gpl_smooth_f
#define SMOOTH_1         smooth_f
#include __FILE__

#define type_t           double
#define ZERO             0.0
#define SMOOTH           gpl_smooth_d
#define SMOOTH_1         smooth_d
#include __FILE__

#else /* _GPL_SMOOTH_C defined. ---------------------------------------------*/

/* Private routines, data and definitions used in this file. */

#ifdef SMOOTH_1
static void SMOOTH_1(type_t dst[], const type_t src[], int n,
		     const type_t ker[], int w, int scale, int border);
#endif

#ifdef SMOOTH
void SMOOTH(type_t *dst, const type_t *src, int stride, int n,
	    int nafter, const type_t *ker, int w, int scale,
	    int border, type_t *ws)
{
  int i, j, k, l;

  ker+= w;
  if (stride == 1) {
    if (dst == src) {
      for (k=l=0; l<nafter; l++, k+=n) {
	for (j=0, i=k; j<n; j++, i++) {
	  ws[j]= src[i];
	}
	SMOOTH_1(dst+k, ws, n, ker, w, scale, border);
      }
    } else {
      for (k=l=0; l<nafter; l++, k+=n) {
	SMOOTH_1(dst+k, src+k, n, ker, w, scale, border);
      }
    }
  } else {
    type_t *wp= ws+n;
    for (l=0; l<nafter ; l++) {
      for (k=0; k<stride; k++) {
	for (j=0, i=k+stride*n*l; j<n; j++, i+=stride) {
	  ws[j]= src[i];
	}
	SMOOTH_1(wp, ws, n, ker, w, scale, border);
	for (j=0, i=k+stride*n*l; j<n; j++, i+=stride) {
	  dst[i]= wp[j];
	}
      }
    }
  }
}
#endif /* SMOOTH */

#ifdef SMOOTH_1
static void SMOOTH_1(type_t dst[], const type_t src[], int n,
			   const type_t ker[], int w, int scale, int border)
{
  int i, j, k;
  type_t sum, xl, xr;

  if (scale>1) {

    /* WAVELET SMOOTHING */

    int ws= w*scale;
    switch (border) {
    case 0: /* Extrapolate missing left values by the leftmost one
	       and missing right values by the rightmost one. */
      xl= src[0];
      xr= src[n-1];
      for (i=0 ; i<n ; ++i) {
	for (sum=ZERO, j=-w, k=i-ws ; j<=w ; ++j, k+=scale) {
	  sum+= ker[j] * (k>=0 ? (k<n ? src[k] : xl): xr);
	}
	dst[i]= sum;
      }
      break;
    case 1: /* Extrapolate missing left values by zero and
	       missing right values by the rightmost one. */
      xr= src[n-1];
      for (i=0 ; i<n ; ++i) {
	for (sum=ZERO, j=-w, k=i-ws ; j<=w ; ++j, k+=scale) {
	  if (k>=n) {
	    sum+= ker[j] * xr;
	  } else if (k>=0) {
	    sum+= ker[j] * src[k];
	  }
	}
	dst[i]= sum;
      }
      break;
    case 2: /* Extrapolate missing left values by the leftmost one
	       and missing right values by zero. */
      xl= src[0];
      for (i=0 ; i<n ; ++i) {
	for (sum=ZERO, j=-w, k=i-ws ; j<=w ; ++j, k+=scale) {
	  if (k<0) {
	    sum+= ker[j] * xl;
	  } else if (k<n) {
	    sum+= ker[j] * src[k];
	  } else {
	    break;
	  }
	}
	dst[i]= sum;
      }
      break;
    case 3: /* Extrapolate missing values by zero. */
      for (i=0 ; i<n ; ++i) {
	for (sum=ZERO, j=-w, k=i-ws ; j<=w && k<n ; ++j, k+=scale) {
	  if (k>=0) sum+= ker[j] * src[k];
	}
	dst[i]= sum;
      }
      break;
    case 4: /* Periodic conditions. */
      for (i=0 ; i<n ; ++i) {
	if ((k= i-(ws%n)) < 0) k+= n;
	for (sum=ZERO, j=-w ; j<=w ; ++j, k+=scale) {
	  sum+= ker[j] * src[k%n];
	}
	dst[i]= sum;
      }
      break;
    default: /* Do not extrapolate missing values but normalize convolution
		product by sum of kernel weights taken into account (assuming
		they are all positive). */
      for (i=0 ; i<n ; ++i) {
	for (xl=sum=ZERO, j=-w, k=i-ws ; j<=w && k<n ; ++j, k+=scale) {
	  if (k>=0) {
	    sum+= (xr= ker[j]) * src[k];
	    xl+= xr;
	  }
	}
	dst[i]= xl ? sum/xl : ZERO;
      }
    }

  } else {

    /* NORMAL CONVOLUTION */
    int jl, jr, wp1=w+1;

    switch (border) {
    case 0: /* Extrapolate missing left values by the leftmost one
	       and missing right values by the rightmost one. */
      xl= src[0];
      xr= src[n-1];
      for (i=0 ; i<n ; ++i) {
	sum= ZERO;
	jl= -GPL_MIN(i,w);              /* limit of left border */
	if ((jr= n-i) > wp1) jr= wp1; /* limit of right border */
	for (j=-w ; j<jl ; ++j) {
	  sum+= ker[j] * xl;
	}
	for (j=jl, k=i+j ; j<jr ; ++j, ++k) {
	  sum+= ker[j] * src[k];
	}
	for (j=jr ; j<=w ; ++j) {
	  sum+= ker[j] * xr;
	}
	dst[i]= sum;
      }
      break;
    case 1: /* Extrapolate missing left values by zero and
	       missing right values by the rightmost one. */
      xr= src[n-1];
      for (i=0 ; i<n ; ++i) {
	sum= ZERO;
	jl= -GPL_MIN(i,w);              /* limit of left border */
	if ((jr= n-i) > wp1) jr= wp1; /* limit of right border */
	for (j=jl, k=i+j ; j<jr ; ++j, ++k) {
	  sum+= ker[j] * src[k];
	}
	for (j=jr ; j<=w ; ++j) {
	  sum+= ker[j] * xr;
	}
	dst[i]= sum;
      }
      break;
    case 2: /* Extrapolate missing left values by the leftmost one
	       and missing right values by zero. */
      xl= src[0];
      for (i=0 ; i<n ; ++i) {
	sum= ZERO;
	jl= -GPL_MIN(i,w);              /* limit of left border */
	if ((jr= n-i) > wp1) jr= wp1; /* limit of right border */
	for (j=-w ; j<jl ; ++j) {
	  sum+= ker[j] * xl;
	}
	for (j=jl, k=i+j ; j<jr ; ++j, ++k) {
	  sum+= ker[j] * src[k];
	}
	dst[i]= sum;
      }
      break;
    case 3: /* Extrapolate missing values by zero. */
      for (i=0 ; i<n ; ++i) {
	sum= ZERO;
	jl= -GPL_MIN(i,w);              /* limit of left border */
	if ((jr= n-i) > wp1) jr= wp1; /* limit of right border */
	for (j=jl, k=i+j ; j<jr ; ++j, ++k) {
	  sum+= ker[j] * src[k];
	}
	dst[i]= sum;
      }
      break;
    case 4: /* Periodic conditions. */
      for (i=0 ; i<n ; ++i) {
	sum= ZERO;
	if ((k= i-(w%n)) < 0) k+= n;
	for (j=-w ; j<=w ; ++j, ++k) {
	  if (k>=n) k= 0;
	  sum+= ker[j] * src[k];
	}
	dst[i]= sum;
      }
      break;
    default: /* Do not extrapolate missing values but normalize convolution
		product by sum of kernel weights taken into account (assuming
		they are all positive). */
      for (i=0 ; i<n ; ++i) {
	for (xl=sum=ZERO, j=-w, k=i-w ; j<=w && k<n ; ++j, ++k) {
	  if (k>=0) {
	    sum+= (xr= ker[j]) * src[k];
	    xl+= xr;
	  }
	}
	dst[i]= xl ? sum/xl : ZERO;
      }
    }
  }

}
#endif /* SMOOTH_1 */
/*---------------------------------------------------------------------------*/
#undef type_t
#undef ZERO
#undef SMOOTH
#undef SMOOTH_1
#endif /* _GPL_SMOOTH_C defined. ----------------------------------------------*/
