/*
 * tclsrv_cmds.c --
 *
 *	Implements Tcl commands for TclSrv.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2002 Eric THIEBAUT.
 *
 *	This file is part of TclSrv.
 *
 *	TclSrv is free  software; you can redistribute it  and/or modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	TclSrv  is distributed  in the  hope that  it will  be  useful, but
 *	WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 *	MERCHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU
 *	General Public License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with TclSrv (file "COPYING"  in the top source directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330, Boston, MA 02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <tcl.h>
#include "tclsrv_private.h"

#ifndef DEBUG
#  define DEBUG 0
#endif
#if DEBUG
#  define DEBUG_MSG(format, args...) fprintf(stderr, "server: "format, ## args)
#else
#  define DEBUG_MSG(format, args...)
#endif

static Tcl_FileProc server_handler;
static Tcl_FileProc client_handler;
static Tcl_ExitProc shutdown_server;

static int WrongArgs _ANSI_ARGS_((Tcl_Interp *interp, Tcl_Obj *objCmd,
				  CONST char *syntax));
static int PosixError _ANSI_ARGS_((Tcl_Interp *interp, CONST char *begin,
				   CONST char *end));

/* Allocate a single item of a given type. */
#define ALLOC(nbytes)     ((VOID *)(ckalloc((nbytes))))
#define FREE(ptr)         ckfree(((VOID *)(ptr)))
#define ALLOC_ITEM(type)  ((type *)(ckalloc((sizeof(type)))))

/*-----------------------------------------------------------------------------
 * Tclsrv_Init --
 *     Initializes Tclsrv extension in Tcl interpreter.
 *
 * Results:
 *      Standard TCL result.
 *-----------------------------------------------------------------------------
 */
int
Tclsrv_Init(interp)
     Tcl_Interp *interp;
{
  static int first_time = 1;
  if (first_time) {
    Tcl_CreateExitHandler(shutdown_server, NULL);
    first_time = 0;
  }
  Tcl_CreateObjCommand(interp, "::tclsrv::pipe",
		       Tclsrv_PipeObjCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "::tclsrv::cmdsrv",
		       Tclsrv_CommandServerObjCmd, NULL, NULL);
  return TCL_OK;
}

/*-----------------------------------------------------------------------------
 * Tclsrv_PipeObjCmd --
 *     Implements the pipe TCL command:
 *         pipe ?fileId_var_r fileId_var_w?
 *
 * Results:
 *      Standard TCL result.
 *-----------------------------------------------------------------------------
 */
int
Tclsrv_PipeObjCmd(clientData, interp, objc, objv)
     ClientData  clientData;
     Tcl_Interp *interp;
     int         objc;
     Tcl_Obj    *CONST objv[];
{
  Tcl_Channel  channels[2];
  Tcl_Obj     *channelObjects[2];
  int          fd[2];

  if (!((objc == 1) || (objc == 3))) {
    return WrongArgs(interp, objv[0], "?fileId_var_r fileId_var_w?");
  }
  if (pipe(fd) < 0) {
    return PosixError(interp, "pipe creation failed: ", NULL);
  }

  channels[0] = Tcl_MakeFileChannel((ClientData) fd[0], TCL_READABLE);
  Tcl_RegisterChannel(interp, channels[0]);
  channelObjects[0] = Tcl_NewStringObj(Tcl_GetChannelName(channels[0]), -1);

  channels[1] = Tcl_MakeFileChannel((ClientData) fd[1], TCL_WRITABLE);
  Tcl_RegisterChannel(interp, channels [1]);
  channelObjects[1] = Tcl_NewStringObj(Tcl_GetChannelName(channels[1]), -1);

  if (objc == 1) {
    Tcl_SetObjResult(interp, Tcl_NewListObj(2, channelObjects));
  } else if (Tcl_ObjSetVar2(interp, objv[1], NULL, channelObjects[0],
			    TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG) == NULL ||
	     Tcl_ObjSetVar2(interp, objv[2], NULL, channelObjects[1],
			    TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG) == NULL) {
    Tcl_Close(NULL, channels[0]);
    Tcl_Close(NULL, channels[1]);
    return TCL_ERROR;
  }
  return TCL_OK;  
}
/*-----------------------------------------------------------------------------
 * Tclsrv_CommandServerObjCmd --
 *     Implements the command_server TCL command:
 *         cmdsrv ?-mode mode? create|shutdown path
 *
 * Results:
 *      Standard TCL result.
 *-----------------------------------------------------------------------------
 */

typedef struct _cmd_server_t cmd_server_t;
typedef struct _cmd_client_t cmd_client_t;
struct _cmd_server_t {
  cmd_server_t *next;        /* Next server in the list. */
  cmd_client_t *client;      /* First connection to client. */
  Tcl_Interp   *interp;      /* Current interpreter. */
  int           fd;          /* File descriptor of listen socket. */
  int           single;      /* Only one connected client at a time. */
  char          path[1];     /* Local socket name (MUST BE THE LAST MEMBER). */
};
struct _cmd_client_t {
  cmd_server_t *server;  /* Parent server. */
  cmd_client_t *next;    /* Next connection in the list. */
  int           fd;      /* File descriptor of connection socket. */
  int           allowed; /* Clent allowed to evaluate script on this server. */
  char         *buf;     /* Buffer to store input/output data. */
  size_t        bufsize; /* Size of buffer. */
};

static cmd_server_t *server_list = NULL;
int Tclsrv_CreateCommandServer _ANSI_ARGS_((Tcl_Interp *interp,
					    const char *path, int mode));
int Tclsrv_ShutdownCommandServer _ANSI_ARGS_((Tcl_Interp *interp,
					      const char *path));

static int  close_server _ANSI_ARGS_((Tcl_Interp *interp,
				      cmd_server_t *server));
static void close_client _ANSI_ARGS_((cmd_client_t *client));
static int  alloc_client_buffer _ANSI_ARGS_((cmd_client_t *client,
					     size_t size));
static int  write_client _ANSI_ARGS_((cmd_client_t *client, size_t  pending));
static int  read_client _ANSI_ARGS_((cmd_client_t *client, size_t  pending));

int
Tclsrv_CommandServerObjCmd(notUsed, interp, objc, objv)
     ClientData notUsed;	/* Not used. */
     Tcl_Interp *interp;	/* Current interpreter. */
     int objc;			/* Number of arguments. */
     Tcl_Obj *CONST objv[];	/* Argument objects. */
{
  char *path, *option;

  if (objc != 3) {
    return WrongArgs(interp, objv[0], "create|shutdown path");
  }
  option = Tcl_GetString(objv[1]);
  path = Tcl_GetString(objv[2]);

  if (! strcmp(option, "create")) {
    /* Make a new socket server. */
    int mode = TCLSRV_MODE_PRIVATE;
    return Tclsrv_CreateCommandServer(interp, path, mode);
  }

  if (! strcmp(option, "shutdown")) {
    /* Shutdown server. */
    return Tclsrv_ShutdownCommandServer(interp, path);
  }

  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp, "bad option \"", option,
		   "\": must be create, or delete", NULL);
  return TCL_ERROR;
}

int Tclsrv_CreateCommandServer(interp, path, mode)
     Tcl_Interp *interp;
     const char *path;
     int mode;
{
  cmd_server_t *server;
  size_t len, size;
  struct sockaddr_un sa;
  mode_t mask;

  /* Set up umask for access mode. */
  mask = 0777;
  if (mode & TCLSRV_MODE_OWNER) mask &= 0077; /* set user access */
  if (mode & TCLSRV_MODE_GROUP) mask &= 0707; /* set group access */
  if (mode & TCLSRV_MODE_OTHER) mask &= 0770; /* set others access */
  
  if (! path) path = "";
  len = strlen(path);
  if (len <= 0 || len >= sizeof(sa.sun_path)) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "invalid socket name \"", path, "\"", NULL);
    return TCL_ERROR;
  }

  /* Create a new socket instance. */
  size = STRUCT_OFFSET(cmd_server_t, path) + len + 1;
  if (! (server = ALLOC(size))) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "insufficient memory", NULL);
    return TCL_ERROR;
  }
  memset(server, 0, size);
  server->interp = interp;
  server->single = ((mode & TCLSRV_MODE_SINGLE) ? 1 : 0);
  if (len) memcpy(server->path, path, len);
  server->path[len] = 0;
  server->fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (server->fd < 0) {
    (void)PosixError(interp, "failed to create socket: ", NULL);
    FREE(server);
    return TCL_ERROR;
  }
  
  /* Bind a name to the socket.  Note: the size of the address is the offset
     of the start of the filename, plus its length, plus one for the final
     null byte.  Alternatively you can just do: size = SUN_LEN(&name); */
  sa.sun_family = AF_LOCAL;
  if (len) memcpy(sa.sun_path, path, len);
  sa.sun_path[len] = 0;
  size = STRUCT_OFFSET(struct sockaddr_un, sun_path) + len + 1;
  mask = umask(mask); /* restrict access */
  if (bind(server->fd, (struct sockaddr *)&sa, size) < 0) {
    (void)PosixError(interp, "failed to bind socket: ", NULL);
    (void)umask(mask); /* restore previous umask */
    (void)close(server->fd);
    FREE(server);
    return TCL_ERROR;
  }
  (void)umask(mask); /* restore previous umask */

  /* Start listening for connections. */
  if (listen(server->fd, 5) == -1) {
    (void)PosixError(interp, "failed to listen socket: ", NULL);
    (void)unlink(server->path);
    (void)close(server->fd);
    FREE(server);
    return TCL_ERROR;
  }
 
  /* Set callback to accept connection. */
  Tcl_CreateFileHandler(server->fd, TCL_READABLE|TCL_EXCEPTION,
			server_handler, (ClientData)server);

  /* Insert new server in list of server. */
  server->next = server_list;
  server_list = server;
  return TCL_OK;
}

int Tclsrv_ShutdownCommandServer(interp, path)
     Tcl_Interp *interp;
     CONST char *path;
{
  int c = path[0];
  cmd_server_t *server = server_list;
  while (server) {
    if (interp == server->interp &&
	c == server->path[0] && ! strcmp(path, server->path)) {
      return close_server(interp, server);
    }
    server = server->next;
  }
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp, "no server named \"", path,
		   "\" on this interpreter", NULL);
  return TCL_ERROR;
}

static void shutdown_server(context)
     ClientData context;
{
  Tcl_Interp *interp = (Tcl_Interp *)context; /* Target interpreter or NULL. */
  if (interp) {
    /* Shutdown all command servers running on this interpreter. */
    cmd_server_t *server = server_list;
    while (server) {
      if (server->interp == interp) {
	close_server(interp, server);
	server = server_list;
      } else {
	server = server->next;
      }
    }
  } else {
    /* Shutdown all command servers regardless on which
       interpreter they run. */
    while (server_list) close_server(NULL, server_list);
  }
}

static int close_server(interp, server)
     Tcl_Interp *interp;	/* Tcl interpreter, only used to report error
				   and if it matches server->interp. */
     cmd_server_t *server;	/* Socket server to close. */
{
  cmd_server_t *other = server_list;
  cmd_server_t *prev = NULL;
  int status = TCL_OK;

  /* Tcl interpreter is only used to report errors if it is non-NULL and
     matches server->interp. */
  if (interp != server->interp) interp = NULL;

  /* Find server in list of servers. */
  while (other) {
    if (other == server) {
      /* Remove server from list of servers. */
      if (prev) prev->next  = server->next;
      else      server_list = server->next;

      /* Close server (listen) socket and unlink file counterpart. */
      if (server->fd >= 0) {
	Tcl_DeleteFileHandler(server->fd);
	if (unlink(server->path)) {
	  if (interp) (void)PosixError(interp,
				       "failed to unlink server socket: ",
				       NULL);
	  status = TCL_ERROR;
	}
	if (close(server->fd)) {
	  if (interp) (void)PosixError(interp,
				       "failed to close server socket: ",
				       NULL);
	  status = TCL_ERROR;
	}
      }

      /* Close every peer connection. */
      server->single = 0; /* avoid client to restore file handle */
      while (server->client) close_client(server->client);

      /* Free memory and return. */
      FREE(server);
      return status;
    }
    prev = server;
    server = server->next; 
  }
  Tcl_Panic("corrupted server list or bad server address");
  return TCL_ERROR;  
}

static void server_handler(clientData, mask)
     ClientData clientData;
     int mask;
{
  cmd_server_t *server = (cmd_server_t *)clientData;
  cmd_client_t *client;
  struct sockaddr_un sa;
  socklen_t len = sizeof(sa);
  int flags, s;

  if (mask&TCL_READABLE) {
    /* Accept a new connection (create socket connect to client). */
    if ((s = accept(server->fd, (struct sockaddr *)&sa, &len)) < 0) {
      DEBUG_MSG("accept failed for \"%s\"\n", server->path);
      return;
    }

    /* Set socket i/o buffers. */
#   if defined(SOCKET_BUFSIZE) && SOCKET_BUFSIZE > 0
    {
      int value;
      socklen_t value_length;
      value_length = sizeof(value);
      (void)getsockopt(s, SOL_SOCKET, SO_SNDBUF, &value, &value_length);
      if (value < SOCKET_BUFSIZE) {
	value = SOCKET_BUFSIZE;
	(void)setsockopt(s, SOL_SOCKET, SO_SNDBUF, &value,
			 sizeof(value));
      }
      value_length = sizeof(value);
      (void)getsockopt(s, SOL_SOCKET, SO_RCVBUF, &value, &value_length);
      if (value < SOCKET_BUFSIZE) {
	value = SOCKET_BUFSIZE;
	(void)setsockopt(s, SOL_SOCKET, SO_RCVBUF, &value,
			 sizeof(value));
      }
    }
#   endif /* SOCKET_BUFSIZE */

    /* Set blocking mode. */
    flags = fcntl(s, F_GETFL);
    if (flags == -1 || ((flags & (O_NONBLOCK)) && 
			fcntl(s, F_SETFL, (flags & ~(O_NONBLOCK))))) {
      DEBUG_MSG("set blocking mode failed for \"%s\"\n", server->path);
      close(s);
      return;
    }
    
    /* Allocate instance. */
    if (! (client = ALLOC_ITEM(cmd_client_t))) {
      DEBUG_MSG("insufficient memory for client of \"%s\"\n", server->path);
      close(s);
      return;
    }
    memset(client, 0, sizeof(cmd_client_t));
    client->server = server;
    client->next = server->client;
    client->allowed = (server->single ? 0 : 1);
    client->fd = s;
    server->client = client;
    Tcl_CreateFileHandler(s, TCL_READABLE|TCL_EXCEPTION,
			  client_handler, (ClientData)client);
    DEBUG_MSG("connection established for \"%s\".\n", client->server->path);
  }
  if (mask&TCL_WRITABLE) {
    DEBUG_MSG("listen socket for \"%s\" is writable\n", server->path);
  }
  if (mask&TCL_EXCEPTION) {
    DEBUG_MSG("exception occured for \"%s\"\n", server->path);
  }
}

static void close_client(client)
     cmd_client_t *client;
{
  cmd_server_t *server = client->server;
  cmd_client_t *other = server->client;
  cmd_client_t *prev = NULL;
  while (other) {
    if (other == client) {
      if (prev) prev->next     = client->next;
      else      server->client = client->next;
      if (client->fd >= 0) {
	Tcl_DeleteFileHandler(client->fd);
	(void)close(client->fd);
      }
      if (client->buf) FREE(client->buf);
      FREE(client);
      DEBUG_MSG("connection closed on \"%s\".\n", server->path);      
      return;
    }
  }
  Tcl_Panic("corrupted server/peer list");
}

/* Handle connection with client: receive a script, evaluate it and send
   back the result of the script. */
static void client_handler(context, mask)
     ClientData context;
     int mask;
{
  cmd_client_t *client = (cmd_client_t *)context;
  cmd_server_t *server = client->server;

  if (mask&TCL_EXCEPTION) {
    DEBUG_MSG("exception occured for peer on \"%s\"\n", server->path);
    goto shutdown;
  }

  if (mask&TCL_READABLE) {
    char *result = "connection refused (too many connected clients)";
    int status = TCL_ERROR;
    size_t size;
    tclsrv_uint_t length;
    unsigned char *hdr;
    
    /* Receive data from the client.
     * Notes:
     *	 1/ if no data at all can be read, that means that the client has
     *	    closed its side of the connction
     *	 2/ if incomplete data is read from the connection, there is no
     *	    reliable way to re-synchronize the connection
     *	==> the connection has to be closed in these two cases.
     */
 
    /* Read message header, check magic number and compute length of
       message. */
    if (read_client(client, 8) != TCL_OK) goto shutdown;
    hdr = (unsigned char *)client->buf;
    if (BAD_MAGIC(hdr)) {
      DEBUG_MSG("bad magic number received from \"%s\"\n",
		server->path);
      goto shutdown;
    }
    length = GET_LENGTH(hdr);
    DEBUG_MSG("expecting %u bytes message from \"%s\"\n",
	      length, server->path);

    /* Read message and evaluate script. */
    if (length > 0 && read_client(client, length) != TCL_OK) goto shutdown;
    client->buf[length] = 0;
    if (client->allowed) {
    eval_script:
      DEBUG_MSG("eval script \"%s\" from \"%s\"\n",
		client->buf, server->path);
      status = Tcl_GlobalEval(server->interp, client->buf);
      result = Tcl_GetStringResult(server->interp);
    } else {
      /* First time a script is received from this client and
	 server->single is true: refuse connection if one other client is
	 already allowed to send scripts. */
      cmd_client_t *other = server->client;
      for (;;) {
	if (! other) {
	  client->allowed = 1;
	  goto eval_script;
	}
	if (other->allowed) {
	  DEBUG_MSG("refuse connection of new client on \"%s\"\n",
		    server->path);
	  break;
	}
	other = other->next;
      }
    }
    length = result ? strlen(result)+1 : 1; /* result length + 1 byte status */

    /* Make sure the buffer is large enough for the reply message. */
    size = length + 8;
    if (size > client->bufsize &&
	alloc_client_buffer(client, size) != TCL_OK) {
      goto shutdown;
    }
    hdr	= (unsigned char *)client->buf;
    SET_MAGIC(hdr);
    SET_LENGTH(hdr, length);
    hdr[8] = status;
    if (length > 1) memcpy(hdr + 9, result, length - 1);
    if (write_client(client, size) != TCL_OK || ! client->allowed) {
      goto shutdown;
    }
  }

  return;

 shutdown:
  close_client(client);
}

/* Send reply.  For a sockect/pipe channel, the SIGPIPE signal will be
   received if the peer has closed the reading end of the channel and an
   attempt is made to write data to the peer.  Ignoring the signal causes
   the write operation to returns an error with errno set to EPIPE.  We
   therefore disable the SIGPIPE signal prior to writing and restore
   previous signal handler after the write. */
static int write_client(client, pending)
     cmd_client_t     *client;    /* Connection to client. */
     size_t          pending; /* Number of bytes to send. */
{
  void (*handler) _ANSI_ARGS_((int));
  ssize_t nbytes;
  char *address = client->buf;
  int status = TCL_OK;
  
  handler = signal(SIGPIPE, SIG_IGN);
  for (;;) {
    /* Attemp to write data. */
    nbytes = write(client->fd, address, pending);
    if (nbytes == pending) break;
    if (nbytes >= 0) {
      /* Partial write: update number of pending bytes and buffer offset. */
      pending -= nbytes;
      address += nbytes;
      continue;
    }
#if DEBUG
    if (errno == EPIPE) {
      DEBUG_MSG("peer has closed connection to \"%s\"\n",
		client->server->path);
    } else if (errno == EAGAIN) {
      DEBUG_MSG("write to \"%s\" would block (unexpected in blocking mode)\n",
		client->server->path);
    } else {
      DEBUG_MSG("write error to peer of \"%s\"\n", client->server->path);
    }
#endif
    status = TCL_ERROR;
    break;
  }
  if (handler != SIG_IGN) signal(SIGPIPE, handler);
  return status;
}

/* Read data from PEER.
   Side effect: internal buffer get re-allocated as needed. */
static int read_client(client, pending)
     cmd_client_t     *client;    /* Connection to client. */
     size_t          pending; /* Number of bytes to receive. */
{
  ssize_t nbytes;
  char *address;

  /* Make sure the internal buffer is large enough to read the data
     (plus 1 extra byte for the final null, in case we are reading
     the script string). */
  if (pending >= client->bufsize &&
      alloc_client_buffer(client, pending + 1) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Read until all expected bytes are received or until error. */
  address = client->buf;
  for (;;) {
    /* Attemp to read data. */
    nbytes = read(client->fd, address, pending);
    if (nbytes == pending) return TCL_OK;
    if (nbytes > 0 && nbytes < pending) {
      /* Partial read: update number of pending bytes and buffer offset. */
      pending -= nbytes;
      address += nbytes;
    } else if (nbytes == 0) {
      /* The standard is that read() returns zero byte on end-of-file, which
	 only occurs when other side of socket was closed. */
      DEBUG_MSG("connection on \"%s\" closed by client\n",
		client->server->path);
      return TCL_ERROR;
    } else if (nbytes == -1) {
      /* An error has occured. */
      if (errno == EAGAIN) continue; /* only occur in non-blocking mode */
      DEBUG_MSG("read error on \"%s\"\n", client->server->path);
      return TCL_ERROR;
    } else {
      /* Unexpected return value! */
      DEBUG_MSG("unexpected read error on \"%s\"\n", client->server->path);
      return TCL_ERROR;
    }
  }
}

static int alloc_client_buffer(client, size)
     cmd_client_t *client;
     size_t size;
{
  size_t bufsize = 2;
  while (size > bufsize) bufsize <<= 1;
  if (client->buf) FREE(client->buf);
  client->buf = ALLOC(bufsize);
  client->bufsize = 0;
  if (! client->buf) {
    DEBUG_MSG("failed to allocate %u bytes for peer buffer on \"%s\".\n",
	      bufsize, client->server->path);
    return TCL_ERROR;
  }
  client->bufsize = bufsize;
  return TCL_OK;
}

/*---------------------------------------------------------------------------*/
/* PRIVATE ROUTINES */

static int PosixError(interp, begin, end)
     Tcl_Interp *interp;
     CONST char *begin;
     CONST char *end;
{
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp, begin, Tcl_PosixError(interp), end, NULL);
  return TCL_ERROR;
}

static int WrongArgs(interp, objCmd, syntax)
     Tcl_Interp *interp;
     Tcl_Obj *objCmd;
     CONST char *syntax;
{
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp, "wrong # args: should be \"", Tcl_GetString(objCmd),
		   " ", syntax, "\"", NULL);
  return TCL_ERROR;
}

/*---------------------------------------------------------------------------*/
