#include <math.h>
#include "../gpl/gpl_num.h"

#define DEBUG 1
#if DEBUG
# include <stdio.h>
# define DEBUGMSG fprintf(stderr, "%s\n", __FUNCTION__)
#else
# define DEBUGMSG
#endif 

#warning "unsigned integer is faster?"
#define integer_t long  /* that's for Yorick */

#if defined(SINGLE_PRECISION) && SINGLE_PRECISION
/* settings for single precision */
# define real_t    float
# define ZERO      0.0f
# define ONE       1.0f
# define TWO       2.0f
# define MAXITER  30
# define EIRS     ant_eirs_f
# define ssiev    ssiev_f
# define rs       rs_f
# define r1mach   gpl_num_r1mach
# if HAVE_HYPOTF
#  define HYPOT   hypotf
# else
#  define HYPOT   hypot
# endif
# if HAVE_SQRTF
#  define SQRT    sqrtf
# else
#  define SQRT    sqrt
# endif
# if HAVE_FABSF
#  define ABS     fabsf
# else
#  define ABS     fabs
# endif
#else
/* settings for double precision */
# define real_t    double
# define ZERO      0.0
# define ONE       1.0
# define TWO       2.0
# define MAXITER  30
# define EIRS     ant_eirs_d
# define ssiev    ssiev_d
# define rs       rs_d
# define r1mach   gpl_num_d1mach
# define HYPOT    hypot
# define SQRT     sqrt
# define ABS      fabs
#endif

/*-----Fortran intrinsics converted-----*/
#define SIGN(x,y) ((((x)<0)==((y)<0))?(x):-(x))
/*-----end of Fortran intrinsics-----*/

/*
 * FORTRAN intrinsics:
 *   sign(a,b) 
 *       x = (a >= 0 ? a : -a);
 *       return( b >= 0 ? x : -x);
 *
 *  a >= 0:       yes   no
 *  b >= 0: yes    a    -a
 *          no    -a     a
 *
 */



/** All eigenvalues and optionally eigenvectors of a real symmetric matrix
    (EISPACK driver routine). */
extern void rs(integer_t nm, integer_t n, real_t a[], real_t w[],
	       integer_t matz, real_t z[], real_t fv1[], real_t fv2[],
	       integer_t *ierr);

/** All eigenvalues and optionally eigenvectors of a real symmetric matrix
    (SLATEC driver routine). */
extern void ssiev(real_t a[], integer_t lda, integer_t n, real_t e[],
		  real_t work[], integer_t job, integer_t *info);

#if 0
/** IMTQL1 (LAPACK: SSTEQR or SSTERF) All eigenvalues of a symmetric
    tridiagonal matrix, by the implicit QL algorithm. */
static void imtql1(void);
#endif

/* TQL2 (EISPACK) = SSTEQR or SSTEDC (LAPACK): All eigenvalues/vectors
   of a symmetric tridiagonal matrix by the explicit QL algorithm. */
static void tql2 (integer_t nm, integer_t n, real_t d[], real_t e[],
		  real_t z[], integer_t *ierr);

/** All eigenvalues/vectors of a symmetric tridiagonal matrix, by the
    implicit QL algorithm. */
static void imtql2(integer_t nm, integer_t n, real_t d[], real_t e[], real_t z[],
		   integer_t *ierr);

/** TQLRAT (LAPACK: SSTERF) All eigenvalues of a symmetric tridiagonal
    matrix by a rational variant of the QL algorithm. */
static void tqlrat(integer_t n, real_t d[], real_t e2[], integer_t *ierr);

/** TRBAK1 (LAPACK: SORMTR) Backtransform eigenvectors of a real symmetric
    matrix after reduction by TRED1.

    TRBAK3 (LAPACK: SOPMTR) Backtransform eigenvectors of a real symmetric
    matrix after reduction by TRED3 (packed storage).

    TRED1(LAPACK: SSYTRD) Reduce a real symmetric matrix to tridiagonal form

    TRED2 (LAPACK: SSYTRD SORGTR) As TRED1, but also generating the
    orthogonal transformation matrix.

    TRED3 (LAPACK: SSPTRD) Reduce a real symmetric matrix to tridiagonal
    form (packed storage).

    TRIDIB (LAPACK: SSTEBZ) Eigenvalues between specified indices of a
    symmetric tridiagonal matrix.

    TSTURM (LAPACK: SSTEBZ SSTEIN) Eigenvalues in a specified interval of a
    symmetric tridiagonal matrix, and corresponding eigenvectors by inverse
    iteration.
*/

static void tred1(integer_t nm, integer_t n, real_t a[], real_t d[],
		  real_t e[], real_t e2[]);
static void tred2(integer_t nm, integer_t n, const real_t a[], real_t d[],
		  real_t e[], real_t z[]);
/*-----end of prototypes-----*/

/*---------------------------------------------------------------------------*/
/* DRIVER ROUTINES */


/* EI = eigenvalue/eigenvector
   RS = Real Symmetric matrix
   CH = Complex Hermitian matrix */
#define ANT_EI_LOWER    1 /* only lower part of A is instanciated */
#define ANT_EI_UPPER    2 /* only upper part of A is instanciated */
#define ANT_EI_IMPLICIT 4 /* use implicit QL algorithm */

/*
 *  A LDA-by-N array.  Left unchanged (except if Z=A or if only the lower or
 *     upper part of A is computed).
 *
 *  Z LDA-by-N array to store the eigenvectors, if Z is null only the eigenvalues
 *     are computed.  Z can be the same array as A.
 *
 *  FLAGS
 *    
 */
extern void EIRS(real_t a[], integer_t lda, integer_t n, real_t e[],
		 real_t z[], real_t work[], integer_t flags, integer_t *ierr);
void EIRS(real_t a[], integer_t lda, integer_t n, real_t e[],
	  real_t z[], real_t work[], integer_t flags, integer_t *ierr)
{
  integer_t i, j;

  /* check arguments */
  if (n > lda) {
    *ierr = -1;
    return;
  }
  if (n < 1) {
    *ierr = -2;
    return;
  }
  *ierr = 0;

  /* account for N=1 case */
  e[0] = a[0];
  if (n > 1) {
    if ((flags & (ANT_EI_LOWER|ANT_EI_UPPER)) == ANT_EI_UPPER) {
      /* only lower part - copy upper triangle to lower */
      for (j=1 ; j<n ; ++j) {
	for (i=0 ; i<j ; ++i) {
	  a[j + lda*i] = a[i + lda*j];
	}
      }
    } else if ((flags & (ANT_EI_LOWER|ANT_EI_UPPER)) == ANT_EI_LOWER) {
      /* only upper part - copy lower triangle to upper */
      for (j=1 ; j<n ; ++j) {
	for (i=0 ; i<j ; ++i) {
	   a[i + lda*j] = a[j + lda*i];
	}
      }
    }
#warning "only the lower triangle of the matrix need be supplied"
    if (z) {
      /* eigenvalues and eigenvectors */
      if (flags & ANT_EI_IMPLICIT) {
#if 0
	tred2(lda, n, a, e, work, a);
	imtql2(lda, n, e, work, a, ierr);
#endif
	tred2(lda, n, a, e, work, z);
	imtql2(lda, n, e, work, z, ierr);
      } else {
	tred2(lda, n, a, e, work, z);
	tql2(lda, n, e, work, z, ierr);
      }
    } else {
      /* eigenvalues only */
      tred1(lda, n, a, e, work, work + n);
#ifdef VAX /* tqlrat encounters catastrophic underflow on the Vax */
      tql1(n, e, work, ierr);
#else
      tqlrat(n, e, work + n, ierr);
#endif
    }
  }
}


/* BEGIN PROLOGUE  RS
 * PURPOSE  Compute the eigenvalues and, optionally, the eigenvectors
 *          of a real symmetric matrix.
 * LIBRARY   SLATEC (EISPACK)
 * CATEGORY  D4A1
 * TYPE      SINGLE PRECISION (RS-S, CH-C)
 * KEYWORDS  EIGENVALUES, EIGENVECTORS, EISPACK
 * AUTHOR  Smith, B. T., et al.
 * DESCRIPTION
 * 
 *   This subroutine calls the recommended sequence of subroutines from the
 *   eigensystem subroutine package (EISPACK) to find the eigenvalues and
 *   eigenvectors (if desired) of a REAL SYMMETRIC matrix.
 * 
 *   On Input
 * 
 *      NM must be set to the row dimension of the two-dimensional array
 *         parameters, A and Z, as declared in the calling program
 *         dimension statement.  NM is an INTEGER variable.
 * 
 *      N is the order of the matrix A.  N is an INTEGER variable.
 *        N must be less than or equal to NM.
 * 
 *      A contains the real symmetric matrix.  A is a two-dimensional
 *        REAL array, dimensioned A(NM,N).
 * 
 *      MATZ is an INTEGER variable set equal to zero if only
 *        eigenvalues are desired.  Otherwise, it is set to any
 *        non-zero integer for both eigenvalues and eigenvectors.
 * 
 *   On Output
 * 
 *      A is unaltered.
 * 
 *      W contains the eigenvalues in ascending order.  W is a one-
 *        dimensional REAL array, dimensioned W(N).
 * 
 *      Z contains the eigenvectors if MATZ is not zero.  The
 *        eigenvectors are orthonormal.  Z is a two-dimensional
 *        REAL array, dimensioned Z(NM,N).
 * 
 *      IERR is an INTEGER flag set to
 *        Zero       for normal return,
 *        10*N       if N is greater than NM,
 *        J          if the J-th eigenvalue has not been
 *                   determined after 30 iterations.
 *                   The eigenvalues, and eigenvectors if requested,
 *                   should be correct for indices 1, 2, ..., IERR-1.
 * 
 *      FV1 and FV2 are one-dimensional REAL arrays used for temporary
 *        storage, dimensioned FV1(N) and FV2(N).
 * 
 *   Questions and comments should be directed to B. S. Garbow,
 *   APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LABORATORY
 *   ------------------------------------------------------------------
 * 
 * REFERENCES  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
 *               Y. Ikebe, V. C. Klema and C. B. Moler, Matrix Eigen-
 *               system Routines - EISPACK Guide, Springer-Verlag,
 *               1976.
 * ROUTINES CALLED  TQL2, TQLRAT, TRED1, TRED2
 * REVISION HISTORY  (YYMMDD)
 *   760101  DATE WRITTEN
 *   890831  Modified array declarations.  (WRB)
 *   890831  REVISION DATE from Version 3.2
 *   891214  Prologue converted to Version 4.0 format.  (BAB)
 *   920501  Reformatted the REFERENCES section.  (WRB)
 *   040211  Conversion to C-language, single/double precision (Eric Thibaut).
 * END PROLOGUE  RS
 */
void rs(integer_t nm, integer_t n, real_t a[], real_t w[], integer_t matz,
	real_t z[], real_t fv1[], real_t fv2[], integer_t *ierr)
{
  if (n > nm) {
    *ierr = 10*n;
    return;
  }
  if (matz) {
    /* Find both eigenvalues and eigenvectors. */
    tred2(nm, n, a, w, fv1, z);
    tql2(nm, n, w, fv1, z, ierr);
  } else {
    /* Find eigenvalues only. */
    tred1(nm, n, a, w, fv1, fv2);
#ifdef VAX /* tqlrat encounters catastrophic underflow on the Vax */
    tql1(n, w, fv1, ierr);
#else
    tqlrat(n, w, fv2, ierr);
#endif
  }
}


/* BEGIN PROLOGUE  SSIEV
 * PURPOSE  Compute the eigenvalues and, optionally, the eigenvectors
 *          of a real symmetric matrix.
 * LIBRARY   SLATEC
 * CATEGORY  D4A1
 * TYPE      SINGLE PRECISION (SSIEV-S, CHIEV-C)
 * KEYWORDS  COMPLEX HERMITIAN, EIGENVALUES, EIGENVECTORS, MATRIX,
 *           SYMMETRIC
 * AUTHOR  Kahaner, D. K., (NBS)
 *         Moler, C. B., (U. of New Mexico)
 *         Stewart, G. W., (U. of Maryland)
 *
 * DESCRIPTION
 * 
 *   Abstract
 *    SSIEV computes the eigenvalues and, optionally, the eigenvectors
 *    of a real symmetric matrix.
 * 
 *   Call Sequence Parameters-
 *     (The values of parameters marked with * (star) will be  changed
 *      by SSIEV.)
 * 
 *     A*      REAL (LDA,N)
 *             real symmetric input matrix.
 *             Only the diagonal and upper triangle of A must be input, as
 *             SSIEV copies the upper triangle to the lower.  That is, the
 *             user must define A(I,J), for I=1,...,N and J=I,...,N.
 *             On return from SSIEV, if the user has set JOB
 *             = 0        the lower triangle of A has been altered.
 *             = non-zero the N eigenvectors of A are stored in its
 *                        first N columns.  See also INFO below.
 * 
 *     LDA     INTEGER
 *             set by the user to the leading dimension of the array A.
 * 
 *     N       INTEGER
 *             set by the user to the order of the matrix A and
 *             the number of elements in E.
 * 
 *     E*      REAL (N)
 *             on return from SSIEV, E contains the N eigenvalues of A.
 *             See also INFO below.
 * 
 *     WORK*   REAL (2*N)
 *             temporary storage vector.  Contents changed by SSIEV.
 * 
 *     JOB     INTEGER
 *             set by user on input
 *             = 0         only calculate eigenvalues of A.
 *             = nonzero   calculate eigenvalues and eigenvectors of A.
 * 
 *     INFO*   INTEGER
 *             on return from SSIEV, the value of INFO is
 *             = 0 for normal return.
 *             = K (with K>0) if the eigenvalue iteration fails to converge.
 *                 eigenvalues and vectors 1 through K-1 are correct.
 *             = -1 if N is greater than LDA (A, E and WORK left unchanged)
 *             = -2 if N is less than one (A, E and WORK left unchanged)
 * 
 * REFERENCES  (NONE)
 * ROUTINES CALLED  IMTQL2, TQLRAT, TRED1, TRED2, XERMSG
 * REVISION HISTORY  (YYMMDD)
 *   800808  DATE WRITTEN
 *   890831  Modified array declarations.  (WRB)
 *   890831  REVISION DATE from Version 3.2
 *   891214  Prologue converted to Version 4.0 format.  (BAB)
 *   900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
 *   900326  Removed duplicate information from DESCRIPTION section.
 *           (WRB)
 *   040129  Conversion to C-language, single/double precision (Eric Thibaut).
 * END PROLOGUE  SSIEV
 */

void ssiev(real_t a[], integer_t lda, integer_t n, real_t e[], real_t work[],
	   integer_t job, integer_t *info)
{
  integer_t i, j;

  /* check arguments */
  if (n > lda) {
    *info = -1;
    return;
  }
  if (n < 1) {
    *info = -2;
    return;
  }
  *info = 0;

  /* account for N=1 case */
  e[0] = a[0];
  if (n > 1) {  
    /* copy upper triangle to lower */
    for (j=1 ; j<n ; ++j) {
      for (i=0 ; i<j ; ++i) {
	a[j + lda*i] = a[i + lda*j];
      }
    }

    if (job) {
      /* eigenvalues and eigenvectors */
      tred2(lda, n, a, e, work, a);
      imtql2(lda, n, e, work, a, info);
    } else {
      /* eigenvalues only */
      tred1(lda, n, a, e, work, work + n);
      tqlrat(n, e, work + n, info);
    }
  }
}

/* BEGIN PROLOGUE  TQL2
 * PURPOSE  Compute the eigenvalues and eigenvectors of symmetric
 *          tridiagonal matrix.
 * LIBRARY   SLATEC (EISPACK)
 * CATEGORY  D4A5, D4C2A
 * TYPE      SINGLE PRECISION (TQL2-S)
 * KEYWORDS  EIGENVALUES, EIGENVECTORS, EISPACK
 * AUTHOR  Smith, B. T., et al.
 * DESCRIPTION
 * 
 *   This subroutine is a translation of the ALGOL procedure TQL2, NUM.
 *   MATH. 11, 293-306(1968) by Bowdler, Martin, Reinsch, and Wilkinson.
 *   HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 227-240(1971).
 * 
 *   This subroutine finds the eigenvalues and eigenvectors of a SYMMETRIC
 *   TRIDIAGONAL matrix by the QL method.  The eigenvectors of a FULL
 *   SYMMETRIC matrix can also be found if TRED2 has been used to reduce
 *   this full matrix to tridiagonal form.
 * 
 *   On Input
 * 
 *      NM must be set to the row dimension of the two-dimensional array
 *        parameter, Z, as declared in the calling program dimension
 *        statement.  NM is an INTEGER variable.
 * 
 *      N is the order of the matrix.  N is an INTEGER variable.
 *        N must be less than or equal to NM.
 * 
 *      D contains the diagonal elements of the symmetric tridiagonal
 *        matrix.  D is a one-dimensional REAL array, dimensioned D(N).
 * 
 *      E contains the subdiagonal elements of the symmetric tridiagonal
 *        matrix in its last N-1 positions.  E(1) is arbitrary.  E is a
 *        one-dimensional REAL array, dimensioned E(N).
 * 
 *      Z contains the transformation matrix produced in the reduction by
 *        TRED2, if performed.  If the eigenvectors of the tridiagonal
 *        matrix are desired, Z must contain the identity matrix.  Z is a
 *        two-dimensional REAL array, dimensioned Z(NM,N).
 * 
 *    On Output
 * 
 *      D contains the eigenvalues in ascending order.  If an error exit is
 *        made, the eigenvalues are correct but unordered for indices 1, 2,
 *        ..., IERR-1.
 * 
 *      E has been destroyed.
 * 
 *      Z contains orthonormal eigenvectors of the symmetric tridiagonal
 *        (or full) matrix.  If an error exit is made, Z contains the
 *        eigenvectors associated with the stored eigenvalues.
 * 
 *      IERR is an INTEGER flag set to
 *        Zero    for normal return,
 *        J       if the J-th eigenvalue has not been determined after
 *                30 iterations.
 * 
 *   Calls PYTHAG(A,B) for sqrt(A**2 + B**2).
 * 
 *   Questions and comments should be directed to B. S. Garbow,
 *   APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LABORATORY
 *   ------------------------------------------------------------------
 * 
 * REFERENCES  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
 *               Y. Ikebe, V. C. Klema and C. B. Moler, Matrix Eigen-
 *               system Routines - EISPACK Guide, Springer-Verlag,
 *               1976.
 * ROUTINES CALLED  PYTHAG
 * REVISION HISTORY  (YYMMDD)
 *   760101  DATE WRITTEN
 *   890831  Modified array declarations.  (WRB)
 *   890831  REVISION DATE from Version 3.2
 *   891214  Prologue converted to Version 4.0 format.  (BAB)
 *   920501  Reformatted the REFERENCES section.  (WRB)
 *   040211  Conversion to C-language, single/double precision (Eric Thibaut).
 * END PROLOGUE  TQL2
 */
void tql2(integer_t nm, integer_t n, real_t *d_ptr, real_t *e_ptr,
	  real_t *z_ptr, integer_t *ierr)
{
  integer_t i,j,k,l,m,l1,l2;
  real_t b,c,c2,c3,dl1,el1,f,g,h,p,r,s,s2;

  DEBUGMSG;
  *ierr = 0;
  if (n == 1) return;

  /* define macros and adjust array addresses to use FORTRAN-like indexing */
#undef z
#define z(a1,a2) z_ptr[(a2)*nm + (a1)]
#undef e
#define e(a1)    e_ptr[a1]
#undef d
#define d(a1)    d_ptr[a1]
  z_ptr -= 1 + nm;
  --e_ptr;
  --d_ptr;

  for (i=2 ; i<=n ; ++i) {
    e(i-1) = e(i);
  }

  f = ZERO;
  b = ZERO;
  e(n) = ZERO;

  c3 = ZERO; /* avoids spurious compiler warning */
  s2 = ZERO; /* avoids spurious compiler warning */

  for (l=1 ; l<=n ; ++l) {
    j = 0;
    h = ABS(d(l)) + ABS(e(l));
    if (b < h) b = h;
    /* Look for small sub-diagonal element. */
    for (m=l ; m<=n ; ++m) {
#warning "extra precision below?  ABS(e(m)) < EPSILON*ABS(b)"
      if (b + ABS(e(m)) == b) break;
      /* E(N) is always zero, so there is no exit
	 through the bottom of the loop. */
    }
    if (m > l) {
      do {
	if (++j > MAXITER) {
	  /* Set error -- no convergence to an eigenvalue after MAXITER
	     iterations. */
	  *ierr = l;
	  return;
	}
	/* Form shift. */
	l1 = l + 1;
	l2 = l1 + 1;
	g = d(l);
	p = (d(l1) - g)/(TWO*e(l));
	r = HYPOT(p, ONE);
	h = (p >= ZERO ? p + r : p - r);
	d(l) = e(l)/h;
	d(l1) = e(l)*h;
	dl1 = d(l1);
	h = g - d(l);
	for (i=l2 ; i<=n ; ++i) {
	  d(i) -= h;
	}
	f += h;
	/* QL transformation. */
	p = d(m);
	c = ONE;
	c2 = c;
	el1 = e(l1);
	s = ZERO;
	for (i=m-1 ; i>=l ; --i) {
	  c3 = c2;
	  c2 = c;
	  s2 = s;
	  g = c*e(i);
	  h = c*p;
	  if (ABS(p) < ABS(e(i))) {
	    c = p/e(i);
	    r = SQRT(c*c + ONE);
	    e(i+1) = s*e(i)*r;
	    s = ONE/r;
	    c *= s;
	  } else {
	    c = e(i)/p;
	    r = SQRT(c*c + ONE);
	    e(i+1) = s*p*r;
	    s = c/r;
	    c = ONE/r;
	  }
	  p = c*d(i) - s*g;
	  d(i+1) = h + s*(c*g + s*d(i));
	  /* Form vector. */
	  for (k=1 ; k<=n ; ++k) {
	    h = z(k,i+1);
	    z(k,i+1) = s*z(k,i) + c*h;
	    z(k,i) = c*z(k,i) - s*h;
	  }
	}
	p = -s*s2*c3*el1*e(l)/dl1;
	e(l) = s*p;
	d(l) = c*p;
#warning "extra precision below?"
      } while (b + ABS(e(l)) > b);
    }
    d(l) += f;
  }

  /* order eigenvalues and eigenvectors */
  for (i=1 ; i<n ; ++i) {
    k = i;
    p = d(i);
    for (j=i+1 ; j<=n ; ++j) {
      if (d(j) < p) {
	p = d(j);
	k = j;
      }
    }
    if (k != i) {
      d(k) = d(i);
      d(i) = p;
      for (j=1 ; j<=n ; ++j) {
	p = z(j,i);
	z(j,i) = z(j,k);
	z(j,k) = p;
      }
    }
  }
}

/* BEGIN PROLOGUE  IMTQL2
 * PURPOSE  Compute the eigenvalues and eigenvectors of a symmetric
 *          tridiagonal matrix using the implicit QL method.
 * LIBRARY   SLATEC (EISPACK)
 * CATEGORY  D4A5, D4C2A
 * TYPE      SINGLE PRECISION (IMTQL2-S)
 * KEYWORDS  EIGENVALUES, EIGENVECTORS, EISPACK
 * AUTHOR  Smith, B. T., et al.
 * DESCRIPTION
 * 
 *   This subroutine is a translation of the ALGOL procedure IMTQL2,
 *   NUM. MATH. 12, 377-383(1968) by Martin and Wilkinson,
 *   as modified in NUM. MATH. 15, 450(1970) by Dubrulle.
 *   HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 241-248(1971).
 * 
 *   This subroutine finds the eigenvalues and eigenvectors of a SYMMETRIC
 *   TRIDIAGONAL matrix by the implicit QL method.  The eigenvectors of a
 *   FULL SYMMETRIC matrix can also be found if TRED2 has been used to
 *   reduce this full matrix to tridiagonal form.
 * 
 *   On INPUT
 * 
 *      NM must be set to the row dimension of the two-dimensional
 *        array parameter, Z, as declared in the calling program
 *        dimension statement.  NM is an INTEGER variable.
 * 
 *      N is the order of the matrix.  N is an INTEGER variable.
 *        N must be less than or equal to NM.
 * 
 *      D contains the diagonal elements of the symmetric tridiagonal
 *        matrix.  D is a one-dimensional REAL array, dimensioned D(N).
 * 
 *      E contains the subdiagonal elements of the symmetric
 *        tridiagonal matrix in its last N-1 positions.  E(1) is
 *        arbitrary.  E is a one-dimensional REAL array, dimensioned
 *        E(N).
 * 
 *      Z contains the transformation matrix produced in the reduction
 *        by  TRED2,  if performed.  This transformation matrix is
 *        necessary if you want to obtain the eigenvectors of the full
 *        symmetric matrix.  If the eigenvectors of the symmetric
 *        tridiagonal matrix are desired, Z must contain the identity
 *        matrix.  Z is a two-dimensional REAL array, dimensioned
 *        Z(NM,N).
 * 
 *    On OUTPUT
 * 
 *      D contains the eigenvalues in ascending order.  If an
 *        error exit is made, the eigenvalues are correct but
 *        unordered for indices 1, 2, ..., IERR-1.
 * 
 *      E has been destroyed.
 * 
 *      Z contains orthonormal eigenvectors of the full symmetric
 *        or symmetric tridiagonal matrix, depending on what it
 *        contained on input.  If an error exit is made,  Z contains
 *        the eigenvectors associated with the stored eigenvalues.
 * 
 *      IERR is an INTEGER flag set to
 *        Zero       for normal return,
 *        J          if the J-th eigenvalue has not been
 *                   determined after MAXITER (30) iterations.
 *                   The eigenvalues and eigenvectors should be correct
 *                   for indices 1, 2, ..., IERR-1, but the eigenvalues
 *                   are not ordered.
 * 
 *   Calls PYTHAG(A,B) for sqrt(A**2 + B**2).
 * 
 *   Questions and comments should be directed to B. S. Garbow,
 *   APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LABORATORY
 *   ------------------------------------------------------------------
 * 
 * REFERENCES  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
 *               Y. Ikebe, V. C. Klema and C. B. Moler, Matrix Eigen-
 *               system Routines - EISPACK Guide, Springer-Verlag,
 *               1976.
 * ROUTINES CALLED  PYTHAG
 * REVISION HISTORY  (YYMMDD)
 *   760101  DATE WRITTEN
 *   890831  Modified array declarations.  (WRB)
 *   890831  REVISION DATE from Version 3.2
 *   891214  Prologue converted to Version 4.0 format.  (BAB)
 *   920501  Reformatted the REFERENCES section.  (WRB)
 *   040211  Conversion to C-language, single/double precision (Eric Thibaut).
 * END PROLOGUE  IMTQL2
 */
static void imtql2(integer_t nm, integer_t n, real_t *d_ptr, real_t *e_ptr,
		   real_t *z_ptr, integer_t *ierr)
{
  integer_t i,j,k,l,m;
  real_t b,c,f,g,p,r,s,s1,s2;

  DEBUGMSG;
  *ierr = 0;
  if (n == 1) return;

  /* define macros and adjust array addresses to use FORTRAN-like indexing */
#undef z
#define z(a1,a2) z_ptr[(a2)*nm + (a1)]
#undef e
#define e(a1)    e_ptr[a1]
#undef d
#define d(a1)    d_ptr[a1]
  z_ptr -= 1 + nm;
  --e_ptr;
  --d_ptr;

  for (i=2 ; i<=n ; ++i) {
    e(i-1) = e(i);
  }

  e(n) = ZERO;
  for (l=1 ; l<=n ; ++l) {
    for (j=0 ; ; ) {
      /* look for small sub-diagonal element */
      for (m=l ; m<=n ; ++m) {
	s1 = ABS(d(m)) + ABS(d(m+1));
	s2 = s1 + ABS(e(m));
#warning "FIXME: extra precision in s1 == s2  ABS(e(m)) < EPSILON*(ABS(d(m)) + ABS(d(m+1)))"
	if (s2 == s1) break;
      }
      if (m == l) break; /* convergence */
      if (++j > MAXITER) {
	/* Set error -- no convergence to an eigenvalue after MAXITER
	   iterations. */
	*ierr = l;
	return;
      }
    
      /* form shift */
      p = d(l);
      g = (d(l+1) - p)/(TWO*e(l));
      r = HYPOT(g, ONE);
      g = d(m) - p + e(l)/(g >= ZERO ? g + r : g - r);
      s = ONE;
      c = ONE;
      p = ZERO;
      for (i=m-1 ; i>=l ; --i) {
	f = s*e(i);
	b = c*e(i);      
	if (ABS(f) < ABS(g)) {
	  s = f/g;
	  r = SQRT(s*s + ONE);
	  e(i+1) = g*r;
	  c = ONE/r;
	  s *= c;
	} else {
	  c = g/f;
	  r = SQRT(c*c + ONE);
	  e(i+1) = f*r;
	  s = ONE/r;
	  c *= s;
	}
	g = d(i+1) - p;
	r = (d(i) - g)*s + TWO*c*b;
	p = s*r;
	d(i+1) = g + p;
	g = c*r - b;
	/* form vector */
	for (k=1 ; k<=n ; ++k) {
	  r = z(k,i);
	  f = z(k,i+1);
	  z(k,i) = c*r - s*f;
	  z(k,i+1) = s*r + c*f;
	}
      }
      d(l) -= p;
      e(l) = g;
      e(m) = ZERO;
    }
  }

  /* order eigenvalues and eigenvectors */
  for (i=1 ; i<n ; ++i) {
    k = i;
    p = d(i);
    for (j=i+1 ; j<=n ; ++j) {
      if (d(j) < p) {
	p = d(j);
	k = j;
      }
    }
    if (k != i) {
      d(k) = d(i);
      d(i) = p;
      for (j=1 ; j<=n ; ++j) {
	p = z(j,i);
	z(j,i) = z(j,k);
	z(j,k) = p;
      }
    }
  }
}

/* BEGIN PROLOGUE  TQLRAT
 * PURPOSE  Compute the eigenvalues of symmetric tridiagonal matrix
 *          using a rational variant of the QL method.
 * LIBRARY   SLATEC (EISPACK)
 * CATEGORY  D4A5, D4C2A
 * TYPE      SINGLE PRECISION (TQLRAT-S)
 * KEYWORDS  EIGENVALUES OF A SYMMETRIC TRIDIAGONAL MATRIX, EISPACK,
 *           QL METHOD
 * AUTHOR  Smith, B. T., et al.
 * DESCRIPTION
 * 
 *   This subroutine is a translation of the ALGOL procedure TQLRAT.
 * 
 *   This subroutine finds the eigenvalues of a SYMMETRIC
 *   TRIDIAGONAL matrix by the rational QL method.
 * 
 *   On Input
 * 
 *      N is the order of the matrix.  N is an INTEGER variable.
 * 
 *      D contains the diagonal elements of the symmetric tridiagonal
 *        matrix.  D is a one-dimensional REAL array, dimensioned D(N).
 * 
 *      E2 contains the squares of the subdiagonal elements of the
 *        symmetric tridiagonal matrix in its last N-1 positions.
 *        E2(1) is arbitrary.  E2 is a one-dimensional REAL array,
 *        dimensioned E2(N).
 * 
 *    On Output
 * 
 *      D contains the eigenvalues in ascending order.  If an
 *        error exit is made, the eigenvalues are correct and
 *        ordered for indices 1, 2, ..., IERR-1, but may not be
 *        the smallest eigenvalues.
 * 
 *      E2 has been destroyed.
 * 
 *      IERR is an INTEGER flag set to
 *        Zero       for normal return,
 *        J          if the J-th eigenvalue has not been
 *                   determined after MAXITER iterations.
 * 
 *   Calls PYTHAG(A,B) for sqrt(A**2 + B**2).
 * 
 *   Questions and comments should be directed to B. S. Garbow,
 *   APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LABORATORY
 *   ------------------------------------------------------------------
 * 
 * REFERENCES  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
 *               Y. Ikebe, V. C. Klema and C. B. Moler, Matrix Eigen-
 *               system Routines - EISPACK Guide, Springer-Verlag,
 *               1976.
 *             C. H. Reinsch, Eigenvalues of a real, symmetric, tri-
 *               diagonal matrix, Algorithm 464, Communications of the
 *               ACM 16, 11 (November 1973), pp. 689.
 * ROUTINES CALLED  PYTHAG, R1MACH
 * REVISION HISTORY  (YYMMDD)
 *   760101  DATE WRITTEN
 *   890831  Modified array declarations.  (WRB)
 *   890831  REVISION DATE from Version 3.2
 *   891214  Prologue converted to Version 4.0 format.  (BAB)
 *   920501  Reformatted the REFERENCES section.  (WRB)
 *   040129  Conversion to C-language, single/double precision (Eric Thibaut).
 * END PROLOGUE  TQLRAT
 */
static void tqlrat(integer_t n, real_t *d_ptr, real_t *e2_ptr, integer_t *ierr)
{
  integer_t i,j,l,m;
  real_t b,c,f,g,h,p,r,s;
  static real_t machep=ZERO;

  DEBUGMSG;
  if (! machep) machep = r1mach(4);

  *ierr = 0;
  if (n == 1) return;

  /* define macros and adjust array addresses to use FORTRAN-like indexing */
#undef e2
#define e2(a1)    e2_ptr[a1]
#undef d
#define d(a1)    d_ptr[a1]
  --e2_ptr;
  --d_ptr;

  for (i=2 ; i<=n ; ++i) {
    e2(i-1) = e2(i);
  }

  e2(n) = c = f = b = ZERO;

  for (l=1 ; l<=n ; ++l) {
    j = 0;
    h = machep*(ABS(d(l)) + SQRT(e2(l)));
    if (b <= h) {
      b = h;
      c = b*b;
    }
    /* look for small squared sub-diagonal element */
    for (m=l ; m<=n ; ++m) {
      if (e2(m) <= c) break;
      /* E2(N) is always zero, so there is no exit through the bottom of
	 the loop. */
    }

    if (m > l) {
      do {
	if (++j > MAXITER) {
	  /* Set error -- no convergence to an eigenvalue after MAXITER
	     iterations. */
	  *ierr = l;
	  return;
	}
      
	/* form shift */
	s = SQRT(e2(l));
	g = d(l);
	p = (d(l+1) - g)/(s + s);
	r = HYPOT(p, ONE);
	d(l) = s/(p >= 0 ? p + r : p - r);
	h = g - d(l);
      
	for (i=l+1 ; i<=n ; ++i) {
	  d(i) -= h;
	}
      
	f += h;
	/* rational QL transformation */
	g = d(m);
	if (g == ZERO) g = b;
	h = g;
	s = ZERO;
	for (i=m-1 ; i>=l ; --i) {
	  p = g*h;
	  r = p + e2(i);
	  e2(i+1) = s*r;
	  s = e2(i)/r;
	  d(i+1) = h + s*(h + d(i));
	  g = d(i) - e2(i)/g;
	  if (g == ZERO) g = b;
	  h = g*p/r;
	}
      
	e2(l) = s*g;
	d(l) = h;
	/* guard against underflow in convergence test */
	if (h == ZERO || ABS(e2(l)) <= ABS(c/h)) break;
	e2(l) *= h;
      } while (e2(l));
    }
    /* order eigenvalues */
    p = d(l) + f;
    for (i=l ; i>=2 ; --i) {
      if (p >= d(i-1)) break;
      d(i) = d(i-1);
    }
    d(i) = p;
  }
}

/* BEGIN PROLOGUE  TRED1
 * PURPOSE  Reduce a real symmetric matrix to symmetric tridiagonal
 *          matrix using orthogonal similarity transformations.
 * LIBRARY   SLATEC (EISPACK)
 * CATEGORY  D4C1B1
 * TYPE      SINGLE PRECISION (TRED1-S)
 * KEYWORDS  EIGENVALUES, EIGENVECTORS, EISPACK
 * AUTHOR  Smith, B. T., et al.
 * DESCRIPTION
 * 
 *   This subroutine is a translation of the ALGOL procedure TRED1,
 *   NUM. MATH. 11, 181-195(1968) by Martin, Reinsch, and Wilkinson.
 *   HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 212-226(1971).
 * 
 *   This subroutine reduces a REAL SYMMETRIC matrix
 *   to a symmetric tridiagonal matrix using
 *   orthogonal similarity transformations.
 * 
 *   On Input
 * 
 *      NM must be set to the row dimension of the two-dimensional
 *        array parameter, A, as declared in the calling program
 *        dimension statement.  NM is an INTEGER variable.
 * 
 *      N is the order of the matrix A.  N is an INTEGER variable.
 *        N must be less than or equal to NM.
 * 
 *      A contains the real symmetric input matrix.  Only the lower
 *        triangle of the matrix need be supplied.  A is a two-
 *        dimensional REAL array, dimensioned A(NM,N).
 * 
 *   On Output
 * 
 *      A contains information about the orthogonal transformations
 *        used in the reduction in its strict lower triangle.  The
 *        full upper triangle of A is unaltered.
 * 
 *      D contains the diagonal elements of the symmetric tridiagonal
 *        matrix.  D is a one-dimensional REAL array, dimensioned D(N).
 * 
 *      E contains the subdiagonal elements of the symmetric
 *        tridiagonal matrix in its last N-1 positions.  E(1) is set
 *        to zero.  E is a one-dimensional REAL array, dimensioned
 *        E(N).
 * 
 *      E2 contains the squares of the corresponding elements of E.
 *        E2 may coincide with E if the squares are not needed.
 *        E2 is a one-dimensional REAL array, dimensioned E2(N).
 * 
 *   Questions and comments should be directed to B. S. Garbow,
 *   APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LABORATORY
 *   ------------------------------------------------------------------
 * 
 * REFERENCES  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
 *               Y. Ikebe, V. C. Klema and C. B. Moler, Matrix Eigen-
 *               system Routines - EISPACK Guide, Springer-Verlag,
 *               1976.
 * ROUTINES CALLED  (NONE)
 * REVISION HISTORY  (YYMMDD)
 *   760101  DATE WRITTEN
 *   890831  Modified array declarations.  (WRB)
 *   890831  REVISION DATE from Version 3.2
 *   891214  Prologue converted to Version 4.0 format.  (BAB)
 *   920501  Reformatted the REFERENCES section.  (WRB)
 *   040129  Conversion to C-language, single/double precision (Eric Thibaut).
 * END PROLOGUE  TRED1
 */
static void tred1(integer_t nm, integer_t n, real_t *a_ptr,
		  real_t *d_ptr, real_t *e_ptr, real_t *e2_ptr)
{
  real_t f,g,h,scale;
  integer_t i,j,k,l;

  DEBUGMSG;

  /* define macros and adjust array addresses to use FORTRAN-like indexing */
#undef a
#define a(a1,a2) a_ptr[(a2)*nm + (a1)]
  a_ptr -= 1 + nm;
#undef d
#define d(a1) d_ptr[a1]
  --d_ptr;
#undef e
#define e(a1) e_ptr[a1]
  --e_ptr;
#undef e2
#define e2(a1) e2_ptr[a1]
  --e2_ptr;

  for (i=1 ; i<=n ; ++i) {
    d(i) = a(i,i);
  }
  for (i=n ; i>=1 ; --i) {
    l = i - 1;
    h = ZERO;

    /* scale row (algol tol then not needed) */
    scale = ZERO;
    for (k=1 ; k<=l ; ++k) {
      scale += ABS(a(i,k));
    }
    if (scale == ZERO) {
      e(i) = ZERO;
      e2(i) = ZERO;
    } else {
      for (k=1 ; k<=l ; ++k) {
	a(i,k) /= scale;
	h += a(i,k)*a(i,k);
      }
      e2(i) = scale*scale*h;
      f = a(i,l);
      g = (f >= 0 ? -SQRT(h) : SQRT(h));
      e(i) = scale*g;
      h -= f*g;
      a(i,l) = f - g;
      if (l != 1) {
	f = ZERO;
	for (j=1 ; j<=l ; ++j) {
	  /* form element of A*U */
	  g = ZERO;
	  for (k=1 ; k<=j ; ++k) g += a(j,k)*a(i,k);
	  for (k=j+1 ; k<=l ; ++k) g += a(k,j)*a(i,k);
	
	  /* form element of P */
	  e(j) = g/h;
	  f += e(j)*a(i,j);
	}
	h = f/(h + h);
	/* form reduced A */
	for (j=1 ; j<=l ; ++j) {
	  f = a(i,j);
	  g = e(j) - h*f;
	  e(j) = g;
	
	  for (k=1 ; k<=j ; ++k) {
	    a(j,k) -= f*e(k) + g*a(i,k);
	  }
	}
      }
      for (k=1 ; k<=l ; ++k) {
	a(i,k) *= scale;
      }
    }
    h = d(i);
    d(i) = a(i,i);
    a(i,i) = h;
  }
}

/* BEGIN PROLOGUE  TRED2
 * PURPOSE  Reduce a real symmetric matrix to a symmetric tridiagonal
 *          matrix using and accumulating orthogonal transformations.
 * LIBRARY   SLATEC (EISPACK)
 * CATEGORY  D4C1B1
 * TYPE      SINGLE PRECISION (TRED2-S)
 * KEYWORDS  EIGENVALUES, EIGENVECTORS, EISPACK
 * AUTHOR  Smith, B. T., et al.
 * DESCRIPTION
 * 
 *   This subroutine is a translation of the ALGOL procedure TRED2,
 *   NUM. MATH. 11, 181-195(1968) by Martin, Reinsch, and Wilkinson.
 *   HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 212-226(1971).
 * 
 *   This subroutine reduces a REAL SYMMETRIC matrix to a
 *   symmetric tridiagonal matrix using and accumulating
 *   orthogonal similarity transformations.
 * 
 *   On Input
 * 
 *      NM must be set to the row dimension of the two-dimensional
 *        array parameters, A and Z, as declared in the calling
 *        program dimension statement.  NM is an INTEGER variable.
 * 
 *      N is the order of the matrix A.  N is an INTEGER variable.
 *        N must be less than or equal to NM.
 * 
 *      A contains the real symmetric input matrix.  Only the lower
 *        triangle of the matrix need be supplied.  A is a two-
 *        dimensional REAL array, dimensioned A(NM,N).
 * 
 *   On Output
 * 
 *      D contains the diagonal elements of the symmetric tridiagonal
 *        matrix.  D is a one-dimensional REAL array, dimensioned D(N).
 * 
 *      E contains the subdiagonal elements of the symmetric
 *        tridiagonal matrix in its last N-1 positions.  E(1) is set
 *        to zero.  E is a one-dimensional REAL array, dimensioned
 *        E(N).
 * 
 *      Z contains the orthogonal transformation matrix produced in
 *        the reduction.  Z is a two-dimensional REAL array,
 *        dimensioned Z(NM,N).
 * 
 *      A and Z may coincide.  If distinct, A is unaltered.
 * 
 *   Questions and comments should be directed to B. S. Garbow,
 *   APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LABORATORY
 *   ------------------------------------------------------------------
 * 
 * REFERENCES  B. T. Smith, J. M. Boyle, J. J. Dongarra, B. S. Garbow,
 *               Y. Ikebe, V. C. Klema and C. B. Moler, Matrix Eigen-
 *               system Routines - EISPACK Guide, Springer-Verlag,
 *               1976.
 * ROUTINES CALLED  (NONE)
 * REVISION HISTORY  (YYMMDD)
 *   760101  DATE WRITTEN
 *   890831  Modified array declarations.  (WRB)
 *   890831  REVISION DATE from Version 3.2
 *   891214  Prologue converted to Version 4.0 format.  (BAB)
 *   920501  Reformatted the REFERENCES section.  (WRB)
 *   040129  Conversion to C-language, single/double precision (Eric Thibaut).
 * END PROLOGUE  TRED2
 */
static void tred2(integer_t nm, integer_t n, const real_t *a_ptr,
		  real_t *d_ptr, real_t *e_ptr, real_t *z_ptr)
{
  integer_t i,j,k,l;
  real_t f,g,h,hh,scale;

  DEBUGMSG;

  /* define macros and adjust array addresses to use FORTRAN-like indexing */
#undef a
#define a(a1,a2) a_ptr[(a2)*nm + (a1)]
  a_ptr -= 1 + nm;
#undef d
#define d(a1) d_ptr[a1]
  --d_ptr;
#undef e
#define e(a1) e_ptr[a1]
  --e_ptr;
#undef z
#define z(a1,a2) z_ptr[(a2)*nm + (a1)]
  z_ptr -= 1 + nm;

#warning "FIXME: re-order indices"
  for (i=1 ; i<=n ; ++i) {
    for (j=1 ; j<=i ; ++j) {
      z(i,j) = a(i,j);
    }
  }

  for (i=n ; i>=2 ; --i) {
    l = i - 1;
    h = ZERO;

    /* scale row (algol TOL then not needed) */
    scale = ZERO;
    if (l >= 2) {
      for (k=1 ; k<=l ; ++k) {
	scale += ABS(z(i,k));
      }
    }
    if (scale) {
#warning "FIXME: division by SCALE, use temp var for z(i,k)"
      for (k=1 ; k<=l ; ++k) {
	z(i,k) = z(i,k)/scale;
	h += z(i,k)*z(i,k);
      }

      f = z(i,l);
      g = (f >= 0 ? -SQRT(h) : SQRT(h));
      e(i) = scale*g;
      h -= f*g;
      z(i,l) = f - g;
      f = ZERO;

      for (j=1 ; j<=l ; ++j) {
	z(j,i) = z(i,j)/h;
	g = ZERO;
	/* form element of A*U */
	for (k=1 ; k<=j ; ++k) {
	  g += z(j,k)*z(i,k);
	}
	for (k=j+1 ; k<=l ; ++k) {
	  g += z(k,j)*z(i,k);
	}
	/* form element of P */
	e(j) = g/h;
	f += e(j)*z(i,j);
      }

      hh = f/(h + h);
      /* form reduced A */
      for (j=1 ; j<=l ; ++j) {
	f = z(i,j);
	g = e(j) - hh*f;
	e(j) = g;
	for (k=1 ; k<=j ; ++k) {
	  z(j,k) -= f*e(k) + g*z(i,k);
	}
      }
    } else {
      e(i) = z(i,l);
    }
    d(i) = h;
  }

  d(1) = ZERO;
  e(1) = ZERO;
  /* accumulation of transformation matrices */
  for (i=1 ; i<=n ; ++i) {
#warning "FIXME: remove variable l, remove ZERO below"
    l = i - 1;
    if (d(i)) {
      for (j=1 ; j<=l ; ++j) {
	g = ZERO;
	for (k=1 ; k<=l ; ++k) {
	  g += z(i,k)*z(k,j);
	}
	for (k=1 ; k<=l ; ++k) {
	  z(k,j) -= g*z(k,i);
	}
      }
    }
    d(i) = z(i,i);
    z(i,i) = ONE;
    for (j=1 ; j<=l ; ++j) {
      z(i,j) = ZERO;
      z(j,i) = ZERO;
    }
  }
}
