/*
 * yeti_heapsort.c --
 *
 *	Implement heap-sort function in Yeti.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2001-2002 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_heapsort.c,v 1.2 2004/12/02 08:05:22 eric Exp $
 *	$Log: yeti_heapsort.c,v $
 *	Revision 1.2  2004/12/02 08:05:22  eric
 *	 - Minor mostly cosmetic changes.
 *
 *	Revision 1.1  2004/11/09 12:02:08  eric
 *	Initial revision
 */

#ifndef _YETI_HEAPSORT_C
#define _YETI_HEAPSORT_C 1

#include "yeti.h"
#if (YORICK_VERSION_MAJOR == 1) && (YORICK_VERSION_MINOR <= 4)
# include <stdio.h>
#else
# include "pstdio.h"
#endif
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "yio.h"

#define index_t long

#define value_t unsigned char
#define HEAPSORT  heapsort_b
#define HEAPSORT1 heapsort1_b
#include __FILE__

#define value_t short
#define HEAPSORT  heapsort_s
#define HEAPSORT1 heapsort1_s
#include __FILE__

#define value_t int
#define HEAPSORT  heapsort_i
#define HEAPSORT1 heapsort1_i
#include __FILE__

#define value_t long
#define HEAPSORT  heapsort_l
#define HEAPSORT1 heapsort1_l
#include __FILE__

#define value_t float
#define HEAPSORT  heapsort_f
#define HEAPSORT1 heapsort1_f
#include __FILE__

#define value_t double
#define HEAPSORT  heapsort_d
#define HEAPSORT1 heapsort1_d
#include __FILE__

extern BuiltIn Y_heapsort;

void Y_heapsort(int argc)
{
  Operand op;
  index_t *index = NULL, number;

  if (argc != 1) YError("heapsort takes exactly one argument");
  if (! sp->ops) YError("unexpected keyword");
  sp->ops->FormOperand(sp, &op);
  number = op.type.number;
  if (CalledAsSubroutine()) {
    switch (op.ops->typeID) {
    case T_CHAR:
      heapsort_b(op.value, number);
      return;
    case T_SHORT:
      heapsort_s(op.value, number);
      return;
    case T_INT:
      heapsort_i(op.value, number);
      return;
    case T_LONG:
      heapsort_l(op.value, number);
      return;
    case T_FLOAT:
      heapsort_f(op.value, number);
      return;
    case T_DOUBLE:
      heapsort_d(op.value, number);
      return;
    }
  } else {
    switch (op.ops->typeID) {
    case T_CHAR:
    case T_SHORT:
    case T_INT:
    case T_LONG:
    case T_FLOAT:
    case T_DOUBLE:
      index = yeti_push_new_long_vector_base(number);
      switch (op.ops->typeID) {
      case T_CHAR:
	heapsort1_b(index, op.value, number);
	break;
      case T_SHORT:
	heapsort1_s(index, op.value, number);
	break;
      case T_INT:
	heapsort1_i(index, op.value, number);
	break;
      case T_LONG:
	heapsort1_l(index, op.value, number);
	break;
      case T_FLOAT:
	heapsort1_f(index, op.value, number);
	break;
      default:
	heapsort1_d(index, op.value, number);
	break;
      }
      return;
    }
  }
  YError("bad data type");
}

#else /* _YETI_HEAPSORT_C */

#ifdef HEAPSORT
/* HEAPSORT - in-place sorting of array */
static void HEAPSORT(value_t a[], const index_t n)
{
  index_t i,j,k,l;
  value_t asave;

  if (n < 2) return;
  k = n/2;
  l = n - 1;
  for (;;) {
    if (k > 0) {
      asave = a[--k];
    } else {
      asave = a[l];
      a[l] = a[0];
      if (--l == 0) {
        a[0] = asave;
        return;
      }
    }
    i = k;
    while ((j = 2*i + 1) <= l) {
      if (j < l && a[j] < a[j + 1]) ++j;
      if (a[j] <= asave) break;
      a[i] = a[j];
      i = j;
    }
    a[i] = asave;
  }
}
#endif /* HEAPSORT */

#ifdef HEAPSORT0
/* HEAPSORT0 - indirect sorting of an array, with C-indexing (starting at 0) */
static void HEAPSORT0(index_t index[], const value_t a[], const index_t n)
{
  index_t i,j,k,l,isave;
  value_t asave;

  for (i=0 ; i<n ; ++i) index[i] = i;
  if (n < 2) return;
  k = n/2;
  l = n - 1;
  for (;;) {
    if (k > 0) {
      isave = index[--k];
    } else {
      isave = index[l];
      index[l] = index[0];
      if (--l == 0) {
	index[0] = isave;
        return;
      }
    }
    asave = a[isave];
    i = k;
    while ((j = 2*i + 1) <= l) {
      if (j < l && a[index[j]] < a[index[j + 1]]) ++j;
      if (a[index[j]] <= asave) break;
      index[i] = index[j];
      i = j;
    }
    index[i] = isave;
  }
}
#endif /* HEAPSORT0 */

#ifdef HEAPSORT1
/* HEAPSORT1 - indirect sorting of an array, with Yorick/FORTRAN indexing
   (starting at 1) */
static void HEAPSORT1(index_t index[], const value_t *a, index_t n)
{
  index_t i,j,k,l,isave;
  value_t asave;

  for (i=0 ; i<n ; ++i) index[i] = i + 1;
  if (n < 2) return;
  --a;
  k = n/2;
  l = n - 1;
  for (;;) {
    if (k > 0) {
      isave = index[--k];
    } else {
      isave = index[l];
      index[l] = index[0];
      if (--l == 0) {
	index[0] = isave;
        return;
      }
    }
    asave = a[isave];
    i = k;
    while ((j = 2*i + 1) <= l) {
      if (j < l && a[index[j]] < a[index[j + 1]]) ++j;
      if (a[index[j]] <= asave) break;
      index[i] = index[j];
      i = j;
    }
    index[i] = isave;
  }
}
#endif /* HEAPSORT1 */

#undef HEAPSORT
#undef HEAPSORT0
#undef HEAPSORT1
#undef value_t

#endif /* _YETI_HEAPSORT_C */
