/*
 * yeti_img.c --
 *
 *	Implement image manipulation routines in Yorick.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2004 Eric Thibaut.
 *
 *	This file is part of Yeti.
 *
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

/* The trick here is include this file several times with different
   definitions of the pixel type.  The code is separated in two parts
   depending whether or not the following macro is defined. */
#ifndef _YETI_IMG_C
#define _YETI_IMG_C 1

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include "play.h"
#include "pstdlib.h"
#include "ydata.h"
#include "yio.h"

/*---------------------------------------------------------------------------*/
/* PUBLIC ROUTINES */
extern BuiltIn Y_tiff_open;
extern BuiltIn Y_tiff_read_directory;
extern BuiltIn Y_tiff_read_image;
extern BuiltIn Y_tiff_read_pixels;
extern BuiltIn Y_tiff_debug;

#define pixel_t unsigned char
#define BINARIZE img_binarize_b
#define EROSION  img_erosion_b
#define DILATION img_dilation_b
#include __FILE__

#define pixel_t  short
#define BINARIZE img_binarize_s
#define EROSION  img_erosion_s
#define DILATION img_dilation_s
#include __FILE__


static int binarize(byte_t dst[], int dst_pitch,
		     const pixel_t img[],
		     int img_width, int img_height, int img_pitch,
		     int off_x, int off_y,
		     const mask_t mask[],
		     int msk_width, int msk_height, int msk_pitch);
static int erosion(pixel_t dst[], int dst_pitch,
		   const pixel_t img[],
		   int img_width, int img_height, int img_pitch,
		   int off_x, int off_y,
		   const mask_t msk[],
		   int msk_width, int msk_height, int msk_pitch);
static int dilation(pixel_t dst[], int dst_pitch,
		    const pixel_t img[],
		    int img_width, int img_height, int img_pitch,
		    int off_x, int off_y,
		    const mask_t msk[],
		    int msk_width, int msk_height, int msk_pitch);

#else /* _YETI_IMG_C is defined ---------------------------------------------*/


static int binarize(mask_t dst[], int dst_pitch,
		    const pixel_t img[],
		    int img_width, int img_height, int img_pitch,
		    int off_x, int off_y,
		    const mask_t msk[],
		    int msk_width, int msk_height, int msk_pitch)
{

  /* NOTE: the following variables must all be signed integers */
  int img_x, img_y; /* coordinates in source/destination image */
  int ovl_x, ovl_y; /* coordinates in overlapping region */
  int ovl_width, ovl_height; /* dimensions of overlapping region */
  int msk_x0, msk_y0; /* origin of overlapping region in mask frame */
  int img_x0, img_y0; /* origin of overlapping region in image frame */
  pixel_t pmin, pmax, pval;
  const pixel_t *img_ptr;
  const mask_t *msk_ptr;

  for (img_y = 0 ; img_y < img_height ; ++img_y) {
    /* Compute origins and size of overlapping region in vertical
       direction. */
    if (img_y >= off_y) {
      img_y0 = img_y - off_y;
      msk_y0 = 0;
      if ((ovl_height = img_height - img_y0) > msk_height) {
	ovl_height = msk_height;
      }
    } else {
      img_y0 = 0;
      msk_y0 = off_y - img_y;
      if ((ovl_height = msk_height - msk_y0) > img_height) {
	ovl_height = img_height;
      }
    }

    for (img_x = 0 ; img_x < img_width ; ++img_x) {
      /* Compute origins and size of overlapping region in horizontal
	 direction.  Note: I've tried to avoid identical calculations by
	 pre-computing the values of IMG_X0, MSK_X0 and OVL_WIDTH in
	 temporary array for all IMG_X coordinates but this does not yield
	 a faster code. */
      if (img_x >= off_x) {
	img_x0 = img_x - off_x;
	msk_x0 = 0;
	if ((ovl_width = img_width - img_x0) > msk_width) {
	  ovl_width = msk_width;
	}
      } else {
	img_x0 = 0;
	msk_x0 = off_x - img_x;
	if ((ovl_width = msk_width - msk_x0) > img_width) {
	  ovl_width = img_width;
	}
      }

      img_ptr = img + (img_x0 + img_y0*img_pitch);
      msk_ptr = msk + (msk_x0 + msk_y0*msk_pitch);
      pmin = PIXEL_MAX;
      pmax = PIXEL_MIN;
      for (ovl_y = 0 ; ovl_y < ovl_height ; ++ovl_y,
	     img_ptr += img_pitch, msk_ptr += msk_pitch) {
	for (ovl_x = 0 ; ovl_x < ovl_width ; ++ovl_x) {	  
	  if (msk_ptr[ovl_x]) {
	    if (img_ptr[ovl_x] < pmin) pmin = img_ptr[ovl_x];
	    if (img_ptr[ovl_x] > pmax) pmax = img_ptr[ovl_x];
	  }
	}
      }

      /* slightly biased toward white */
      pval = img[img_x + img_y*img_pitch];
      dst[img_x + img_y*dst_pitch] = (pval - pmin >= pmax - pval);
    }
  }
  return 0;
}

static int erosion(byte_t dst[], int dst_pitch,
		   const pixel_t img[],
		   int img_width, int img_height, int img_pitch,
		   int off_x, int off_y,
		   const mask_t msk[],
		   int msk_width, int msk_height, int msk_pitch)
{

  /* NOTE: the following variables must all be signed integers */
  int img_x, img_y; /* coordinates in source/destination image */
  int ovl_x, ovl_y; /* coordinates in overlapping region */
  int ovl_width, ovl_height; /* dimensions of overlapping region */
  int msk_x0, msk_y0; /* origin of overlapping region in mask frame */
  int img_x0, img_y0; /* origin of overlapping region in image frame */
  pixel_t pmin;
  const pixel_t *img_ptr;
  const mask_t *msk_ptr;

  for (img_y = 0 ; img_y < img_height ; ++img_y) {
    /* Compute origins and size of overlapping region in vertical
       direction. */
    if (img_y >= off_y) {
      img_y0 = img_y - off_y;
      msk_y0 = 0;
      if ((ovl_height = img_height - img_y0) > msk_height) {
	ovl_height = msk_height;
      }
    } else {
      img_y0 = 0;
      msk_y0 = off_y - img_y;
      if ((ovl_height = msk_height - msk_y0) > img_height) {
	ovl_height = img_height;
      }
    }

    for (img_x = 0 ; img_x < img_width ; ++img_x) {
      /* Compute origins and size of overlapping region in horizontal
	 direction.  Note: I've tried to avoid identical calculations by
	 pre-computing the values of IMG_X0, MSK_X0 and OVL_WIDTH in
	 temporary array for all IMG_X coordinates but this does not yield
	 a faster code. */
      if (img_x >= off_x) {
	img_x0 = img_x - off_x;
	msk_x0 = 0;
	if ((ovl_width = img_width - img_x0) > msk_width) {
	  ovl_width = msk_width;
	}
      } else {
	img_x0 = 0;
	msk_x0 = off_x - img_x;
	if ((ovl_width = msk_width - msk_x0) > img_width) {
	  ovl_width = img_width;
	}
      }

      img_ptr = img + (img_x0 + img_y0*img_pitch);
      msk_ptr = msk + (msk_x0 + msk_y0*msk_pitch);
      pmin = PIXEL_MAX;
      for (ovl_y = 0 ; ovl_y < ovl_height ; ++ovl_y,
	     img_ptr += img_pitch, msk_ptr += msk_pitch) {
	for (ovl_x = 0 ; ovl_x < ovl_width ; ++ovl_x) {	  
	  if (msk_ptr[ovl_x]) {
	    if (img_ptr[ovl_x] < pmin) pmin = img_ptr[ovl_x];
	  }
	}
      }
      dst[img_x + img_y*dst_pitch] = pmin;
    }
  }
  return 0;
}

static int dilation(byte_t dst[], int dst_pitch,
		    const pixel_t img[],
		    int img_width, int img_height, int img_pitch,
		    int off_x, int off_y,
		    const mask_t msk[],
		    int msk_width, int msk_height, int msk_pitch)
{

  /* NOTE: the following variables must all be signed integers */
  int img_x, img_y; /* coordinates in source/destination image */
  int ovl_x, ovl_y; /* coordinates in overlapping region */
  int ovl_width, ovl_height; /* dimensions of overlapping region */
  int msk_x0, msk_y0; /* origin of overlapping region in mask frame */
  int img_x0, img_y0; /* origin of overlapping region in image frame */
  pixel_t pmax;
  const pixel_t *img_ptr;
  const mask_t *msk_ptr;

  for (img_y = 0 ; img_y < img_height ; ++img_y) {
    /* Compute origins and size of overlapping region in vertical
       direction. */
    if (img_y >= off_y) {
      img_y0 = img_y - off_y;
      msk_y0 = 0;
      if ((ovl_height = img_height - img_y0) > msk_height) {
	ovl_height = msk_height;
      }
    } else {
      img_y0 = 0;
      msk_y0 = off_y - img_y;
      if ((ovl_height = msk_height - msk_y0) > img_height) {
	ovl_height = img_height;
      }
    }

    for (img_x = 0 ; img_x < img_width ; ++img_x) {
      /* Compute origins and size of overlapping region in horizontal
	 direction.  Note: I've tried to avoid identical calculations by
	 pre-computing the values of IMG_X0, MSK_X0 and OVL_WIDTH in
	 temporary array for all IMG_X coordinates but this does not yield
	 a faster code. */
      if (img_x >= off_x) {
	img_x0 = img_x - off_x;
	msk_x0 = 0;
	if ((ovl_width = img_width - img_x0) > msk_width) {
	  ovl_width = msk_width;
	}
      } else {
	img_x0 = 0;
	msk_x0 = off_x - img_x;
	if ((ovl_width = msk_width - msk_x0) > img_width) {
	  ovl_width = img_width;
	}
      }

      img_ptr = img + (img_x0 + img_y0*img_pitch);
      msk_ptr = msk + (msk_x0 + msk_y0*msk_pitch);
      pmax = PIXEL_MIN;
      for (ovl_y = 0 ; ovl_y < ovl_height ; ++ovl_y,
	     img_ptr += img_pitch, msk_ptr += msk_pitch) {
	for (ovl_x = 0 ; ovl_x < ovl_width ; ++ovl_x) {	  
	  if (msk_ptr[ovl_x]) {
	    if (img_ptr[ovl_x] > pmax) pmax = img_ptr[ovl_x];
	  }
	}
      }
      dst[img_x + img_y*dst_pitch] = pmax;
    }
  }
  return 0;
}


#undef pixel_t
#undef BINARIZE
#undef EROSION
#undef DILATION

#endif /* _YETI_IMG_C -------------------------------------------------------*/
