/*
 * yeti_misc.c --
 *
 *	Implement miscellaneous builtin functions in Yeti.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2001-2002 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *	
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *	
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *	
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_misc.c,v 1.1 2005/05/24 13:24:44 eric Exp $
 *	$Log: yeti_misc.c,v $
 *	Revision 1.1  2005/05/24 13:24:44  eric
 *	Initial revision
 *
 */

#include "yeti.h"
#if (YORICK_VERSION_MAJOR == 1) && (YORICK_VERSION_MINOR <= 4)
# include <stdio.h>
#else
# include <pstdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <yio.h>

/* Shall we use faster complex division? (depends on compiler and machine) */
#define USE_FASTER_DIVIDE_Z 1
#ifndef USE_FASTER_DIVIDE_Z
# define USE_FASTER_DIVIDE_Z 0
#endif


/* Built-in functions defined in this file: */
extern BuiltIn Y__yeti_init;
extern BuiltIn Y_mem_base, Y_mem_copy, Y_mem_peek;
extern BuiltIn Y_window_geometry;
extern BuiltIn Y_strtrim, Y_strlower, Y_strupper;
extern BuiltIn Y_get_encoding;
extern BuiltIn Y_symbol_exists, Y_symbol_names;
extern BuiltIn Y_unref, Y_swap;
extern BuiltIn Y_get_path, Y_expand_path, Y_get_includes, Y_current_include;
extern BuiltIn Y_typeIDof, Y_nrefsof;
extern BuiltIn Y_smooth3;

/* Yorick functions defined elsewhere: */
extern void BuildDimList(Symbol *stack, int argc);  /* ops3.c */
extern DataBlock *ForceToDB(Symbol *s);             /* ops3.c */


/*---------------------------------------------------------------------------*/
/* INITIALIZATION OF YETI */

/* The order of parsing of startup files is as follows:
 *   1. Yorick startup scripts: paths.i, std.i, graph.i, matrix.i, fft.i;
 *   2. Package(s) startup scripts: yeti.i, ...;
 *   3. Yorick post-initialization: stdx.i  (just call 'set_path').
 *
 * It is therefore possible to fool Yorick post-initialization by
 * changing builtin function 'set_path' to something else.
 *
 * Until step 3, search path include the launch directory.
 * Built-in 'set_site' function is called at statup by 'std.i' to
 * define global variables:
 *   Y_LAUNCH    the directory containing the Yorick executable
 *   Y_VERSION   Yorick's version as "MAJOR.MINOR.MICRO"
 *   Y_HOME      Yorick's "site directory" with machine dependent files
 *   Y_SITE      Yorick's "site directory" with machine independent files
 */

/* Symbols defined in std0.c: */
extern char *yLaunchDir;
extern int yBatchMode;

/* Symbols defined in ops0.c: */
extern void *BuildResult2(Operand *l, Operand *r);

/* Symbols defined in ycode.c: */
extern char *yHomeDir;  /* e.g., "/usr/local/lib/yorick/1.5"   */
extern char *ySiteDir;  /* e.g., "/usr/local/share/yorick/1.5" */
extern char *yUserPath; /* e.g., ".:~/yorick:~/Yorick"         */

/* Yeti default's: */
static char *default_path = (char *)0;
static char *yeti_home = YETI_HOME;
static char *yeti_version = YETI_VERSION;

static void globalize_string(const char *name, const char *value);
static void globalize_function(const char *name, BuiltIn *value, int quiet);
static void append_to_default_path(const char *part1, const char *part2);
static void set_path(int argc);
#if USE_FASTER_DIVIDE_Z
static void fast_DivideZ(Operand *l, Operand *r);
#endif /* USE_FASTER_DIVIDE_Z */

void Y__yeti_init(int argc)
{
  if (! default_path) {
    /* This is the first time _yeti_init has been called. */

    /* Setup default path (for Yeti + Yorick); note that Yorick's
       "contrib" directory is always left at the end. */
    default_path = (yUserPath ? p_strcpy(yUserPath) : (char *)0);
    append_to_default_path(ySiteDir, "i");
    if (yLaunchDir && yLaunchDir[0]) {
      int offset, length = strlen(yLaunchDir);
      char *pathtmp = p_malloc(length + 1);
      memcpy(pathtmp, yLaunchDir, length + 1);
      while (length > 0 && pathtmp[length - 1] == '/') {
	--length;
      }
      pathtmp[length] = 0;
      offset = length;
      while (offset > 0 && pathtmp[offset - 1] != '/') {
	--offset;
      }
      if (strcmp(pathtmp + offset, "launch")) {
	/* Tail of launch directory does not match "launch": assume
	   execution from source directory and append launch directory
	   to list of search paths. */
	append_to_default_path(pathtmp, (char *)0);
      
      } else {
	/* Tail of launch directory does match "launch": assume standard
	   installation. */
	pathtmp[offset] = 0;
	append_to_default_path(pathtmp, "plugins");
	append_to_default_path(pathtmp, "i");
      }
      p_free(pathtmp);
    } else {
      append_to_default_path(yeti_home, "plugins");
      append_to_default_path(yeti_home, "i");
    }
    append_to_default_path(ySiteDir, "contrib");

    /* Replace built-in function 'set_path' by Yeti's version. */
    globalize_function("set_path", set_path, 1);

#if USE_FASTER_DIVIDE_Z
    /* Replace complex division by faster code. */
    complexOps.Divide = fast_DivideZ;
#endif /* USE_FASTER_DIVIDE_Z */

    /* Print out Yeti's greetings unless in batch mode. */
    if (! yBatchMode) {
      char buf[128];
      sprintf(buf,
	      " Yeti %d.%d.%d ready.  Copyright (c) 1996-2005, Eric Thibaut.",
	      YETI_VERSION_MAJOR, YETI_VERSION_MINOR, YETI_VERSION_MICRO);
      YputsOut(buf);
    }
  }

  /* Restore global variables. */
  globalize_string("YETI_HOME", yeti_home);
  globalize_string("YETI_VERSION", yeti_version);
  if (! CalledAsSubroutine()) {
    yeti_push_string_value(yeti_version);
  }
}

static void set_path(int argc)
{
  char *path;
  if (argc < 1) path = default_path;
  else if (argc == 1) path = YGetString(sp);
  else { YError("set_path takes at most one argument"); path = 0; }
  YpSetPaths(path);
}

static void append_to_default_path(const char *part1, const char *part2)
{
  char *ptr, *path;
  int len;
#define STRLEN(str) ((str) && (str)[0] ? strlen(str) : 0)
  int len0 = STRLEN(default_path);
  int len1 = STRLEN(part1);
  int len2 = STRLEN(part2);
#undef STRLEN

  /* Remove trailing '/' from PART1 and PART2 */
  while (len1 > 0 && part1[len1 - 1] == '/') --len1;
  while (len2 > 0 && part2[len2 - 1] == '/') --len2;

  /* Remove leading '/' from PART2. */
  while (len2 > 0 && part2[0] == '/') {
    --len2;
    ++part2;
  }

  if (len1 <= 0 && len2 <= 0) return; /* nothing to do */

  /* Append directory path to list of search pathes. */
  len = (len0 > 0 ? len0 + 1 : 0); /* plus 1 for the ':' separator */
  if (len1 > 0) len += len1 + 1; /* plus 1 for the '/' separator */
  if (len2 > 0) len += len2 + 1; /* plus 1 for the '/' separator */
  path = p_malloc(len + 1);
  ptr = path;
  if (len0 > 0) {
    memcpy(ptr, default_path, len0);
    ptr[len0] = ':';
    ptr += len0 + 1;
  }
  if (len1 > 0) {
    memcpy(ptr, part1, len1);
    ptr[len1] = '/';
    ptr += len1 + 1;
  }
  if (len2 > 0) {
    memcpy(ptr, part2, len2);
    ptr[len2] = '/';
    ptr += len2 + 1;
  }
  path[len] = 0;
  if (default_path) p_free(default_path);
  default_path = path;  
}

static void globalize_string(const char *name, const char *value)
{
  long index = Globalize(name, 0L);
  DataBlock *old = (globTab[index].ops == &dataBlockSym ?
		    globTab[index].value.db : 0);
  Array *obj = NewArray(&stringStruct, (Dimension *)0);
  globTab[index].ops = &intScalar; /* in case of interrupt */
  globTab[index].value.db = (DataBlock *)obj;
  globTab[index].ops = &dataBlockSym;
  Unref(old);
  obj->value.q[0] = p_strcpy(value);
}

static void globalize_function(const char *name, BuiltIn *value, int quiet)
{
  long index = Globalize(name, 0L);
  DataBlock *old;
  if (globTab[index].ops == &dataBlockSym) {
    /* Symbol currently stored in global table is a data block. */
    old = globTab[index].value.db;
    if (old->ops == &builtinOps) {
      /* Previous symbol is a builtin function. */
      if (((BIFunction *)old)->function == value) {
	/* Nothing to do. */
	return;
      } else if (old->references == 0) {
	/* Just overwrite pointer to function. */
	((BIFunction *)old)->function = value;
	((BIFunction *)old)->index = index;
	return;
      } else if (quiet) {
	/* This is a hack to avoid printing of a warning when
	   unreferencing a builtin function; the cost is that the memory
	   block allocated for the builtin function is lost forever (but
	   that's 16 bytes on a 32-bit machine). */
	globTab[index].value.db = (DataBlock *)NewBIFunction(value, index);
	return;
      }
    }
  } else {
    /* Symbol currently stored in global table is not a data block. */
    old = (DataBlock *)0;
  }
  globTab[index].ops = &intScalar; /* in case of interrupt */
  globTab[index].value.db = (DataBlock *)NewBIFunction(value, index);
  globTab[index].ops = &dataBlockSym;
  Unref(old);
}

#if USE_FASTER_DIVIDE_Z
/* Faster code for complex division (save 1 division out of 3 with
   respect to original Yorick DivideZ code resulting in ~33% faster
   code). */
static void fast_DivideZ(Operand *l, Operand *r)
{
  const double one=1.0;
  double lr, li, rr, ri;          /* watch out for dst==lv or rv */
  double *lv, *rv, *dst;
  size_t i, n;

  dst = BuildResult2(l, r);
  if (! dst) YError("operands not conformable in binary /");
  n = l->type.number;
  lv = l->value;
  rv = r->value;
  for (i=0 ; i<n ; ++i) {
    lr= lv[2*i];  li= lv[2*i+1];
    rr= rv[2*i];  ri= rv[2*i+1];
    if ((rr>0?rr:-rr)>(ri>0?ri:-ri)) { /* be careful about overflow... */
      ri /= rr;
      rr = one/((one + ri*ri)*rr);
      dst[2*i] = (lr + li*ri)*rr;
      dst[2*i+1] = (li - lr*ri)*rr;
    } else {
      rr /= ri; /* do not care of division by zero here, since Yorick
		   catches floating point exceptions */
      ri = one/((one + rr*rr)*ri);
      dst[2*i] = (lr*rr + li)*ri;
      dst[2*i+1] = (li*rr - lr)*ri;
    }
  }
  PopTo(l->owner);
}
#endif /* USE_FASTER_DIVIDE_Z */

/*---------------------------------------------------------------------------*/
/* MEMORY HACKING ROUTINES */

static void *get_address(Symbol *s);

void Y_mem_base(int argc)
{
  Array *array;
  Symbol *glob;
  long value;

  if (argc != 1) YError("mem_base takes exactly 1 argument");

  /*** based on Address() in ops3.c ***/

  /* Taking the address of a variable X, where X is a scalar constant,
     causes X to be replaced by an Array.  This is obscure, but there is no
     other obvious way to get both the efficiency of the scalar Symbols,
     AND the reference-count safety of Yorick pointers.  Notice that if the
     address of a scalar is taken, the efficient representation is lost.  */
  if (sp->ops != &referenceSym) {
  bad_arg:
    YError("expected a reference to an array object");
  }
  glob = &globTab[sp->index];
  if (glob->ops == &dataBlockSym) array = (Array *)glob->value.db;
  else                            array = (Array *)ForceToDB(glob);
  if (! array->ops->isArray) goto bad_arg;
  value = (long)array->value.c;
  Drop(2);
  PushLongValue(value);
}

void Y_mem_copy(int argc)
{
  void *address;
  Symbol *s;

  if (argc != 2) YError("mem_copy takes exactly 2 arguments");
  address = get_address(sp-1);
  s = (sp->ops == &referenceSym) ? &globTab[sp->index] : sp;
  if (s->ops == &doubleScalar) {
    (void)memcpy(address, &(s->value.d), sizeof(double));
  } else if (s->ops == &longScalar) {
    (void)memcpy(address, &(s->value.l), sizeof(long));
  } else if (s->ops == &intScalar) {
    (void)memcpy(address, &(s->value.i), sizeof(int));
  } else if (s->ops == &dataBlockSym && s->value.db->ops->isArray) {
    Array *array = (Array *)s->value.db;
    (void)memcpy(address, array->value.c,
		 array->type.number*array->type.base->size);
  } else {
    YError("unexpected non-array data");
  }
}

void Y_mem_peek(int argc)
{
  Symbol *s, *stack = sp-argc+1;
  StructDef *base;
  Array *array;
  void *address;

  if (argc < 2) YError("mem_peek takes at least 2 arguments");
  address = get_address(stack);
  s = stack+1;
  if (s->ops == &referenceSym) s = &globTab[s->index];
  if (s->ops != &dataBlockSym || s->value.db->ops != &structDefOps)
    YError("expected type definition as second argument");
  base = (StructDef *)s->value.db;
  if (base->dataOps->typeID < T_CHAR || base->dataOps->typeID > T_COMPLEX)
    YError("only basic data types are supported");
  BuildDimList(stack+2, argc-2);
  array = PushDataBlock(NewArray(base, tmpDims));
  memcpy(array->value.c, address, array->type.number*array->type.base->size);
}

static void *get_address(Symbol *s)
{
  if (s->ops == &referenceSym) s = &globTab[s->index];
#if 0 /* sizeof(int) != sizeof(size_t) on 64-bit machine */
  if (s->ops == &intScalar) return (void *)s->value.i;
#endif
  if (s->ops == &longScalar) return (void *)s->value.l;
  if (s->ops == &dataBlockSym) {
    Array *array = (Array *)s->value.db;
    if (array->ops->isArray && ! array->type.dims) {
#if 0 /* sizeof(int) != sizeof(size_t) on 64-bit machine */
      if (array->ops == &intOps) return (void *)array->value.i[0];
#endif
      if (array->ops == &longOps) return (void *)array->value.l[0];
      if (array->ops == &pointerOps) return array->value.p[0];
    }
  }
  YError("bad address (expecting integer or pointer scalar)");
  return 0; /* avoid compiler warning */
}

/*---------------------------------------------------------------------------*/
/* GIST WINDOW */
#include "hlevel.h"
#include "xfancy.h"
#ifndef GH_NDEVS
# define GH_NDEVS 8 /* max number of graphic windows in old Yorick */
#endif

extern BuiltIn Y_window_geometry;
void Y_window_geometry(int argc)
{
  int win;
  double *geom;
  XEngine *engine;
  GpXYMap *map;

  GpTransform *transform;
  double one_pixel, dpi, xbias, ybias, width, height;

  if (argc == 1) {
    if (YNotNil(sp)) {
      win = (int)YGetInteger(sp);
      if (win < 0 || win >= GH_NDEVS) YError("bad graphic window number");
    } else if ((win = GhGetPlotter()) < 0 || win >= GH_NDEVS) {
      PushDataBlock(RefNC(&nilDB));
      return;
    }
  } else {
    YError("window_geometry takes exactly one, possibly nil, argument");
  }

  /* NDC -> pixel coordinate transform:
   *   XPIX = (int)(XSCALE*XNDC + XOFFSET)
   *   YPIX = (int)(YSCALE*YNDC + YOFFSET)
   * with:
   *   XSCALE = ENGINE->map.x.scale    XOFFSET = ENGINE->map.x.offset - margin
   *   YSCALE = ENGINE->map.y.scale    YOFFSET = ENGINE->map.y.offset - margin
   * assuming:
   *   (XSCALE*XNDC + XOFFSET) >= 0
   *   (YSCALE*YNDC + YOFFSET) >= 0
   * the reverse transform is:
   *   XPIX <= XSCALE*XNDC + XOFFSET < XPIX + 1
   *   YPIX <= YSCALE*YNDC + YOFFSET < YPIX + 1
   * to avoid rounding errors we choose the middle of the interval:
   *   XNDC  =  (XPIX - XOFFSET + 0.5)/XSCALE  =  XBIAS + XPIX*ONE_PIXEL
   *   YNDC  =  (YPIX - YOFFSET + 0.5)/YSCALE  =  YBIAS - YPIX*ONE_PIXEL
   * with:
   *   ONE_PIXEL = 1.0/XSCALE = -1.0/YSCALE
   *       XBIAS = (0.5 - XOFFSET)*ONE_PIXEL
   *       XBIAS = (YOFFSET - 0.5)*ONE_PIXEL
   */

  engine = (XEngine *)ghDevices[win].display;
  if (engine) {
    map= &engine->e.map;
    transform = &(engine->e.transform);
    dpi = engine->dpi;
    one_pixel = 2.0/(map->x.scale - map->y.scale);
#define MARGIN(SIDE) (engine->SIDE##Margin)
    xbias = (MARGIN(left) - map->x.offset + 0.5)/map->x.scale;
    ybias = (MARGIN(top)  - map->y.offset + 0.5)/map->y.scale;
#undef MARGIN
#if (YORICK_VERSION_MAJOR == 1) && (YORICK_VERSION_MINOR <= 4)
    fprintf(stderr, "/* XEngine */ width=%d; height=%d;\nleftMargin=%d; topMargin=%d;\nx=%d; y=%d; \n",
	    engine->width, engine->height,
	    engine->leftMargin, engine->topMargin,
	    engine->x, engine->y);
#else
    width = engine->wtop;
    height = engine->htop;
#endif
  } else {
    dpi = one_pixel = xbias = ybias = width = height = 0.0;
  }

  /* Build result array: [DPI, ONE_PIXEL, XBIAS, YBIAS, WIDTH, HEIGHT] */
  geom = yeti_push_new_double_array_base(yeti_first_dimension(6, 1));
  geom[0] = dpi;
  geom[1] = one_pixel;
  geom[2] = xbias;
  geom[3] = ybias;
  geom[4] = width;
  geom[5] = height;
}

/*---------------------------------------------------------------------------*/
/* STRING ROUTINES */

/* Note: ctype functions require values equal to EOF (-1) or in the range
         0-255 we therefore use "unsigned char *" instead of "char *" to
	 represent a "regular" string. */
typedef unsigned char *string_t;

#define STRTRIM_LEFT  1
#define STRTRIM_RIGHT 2
#define STRTRIM_BOTH  (STRTRIM_LEFT|STRTRIM_RIGHT)

extern BuiltIn Y_strtrim;
void Y_strtrim(int argc)
{
  Dimension *dims;
  unsigned char **input, **output, *src, *dst;
  int flags;
  size_t i, j, len, number;
  Symbol *stack = sp;

  if (argc == 2) flags = yeti_get_flag(stack--, STRTRIM_BOTH);
  else if (argc==1) flags = STRTRIM_BOTH;
  else YError("strtrim takes one or two arguments");
  input = (unsigned char **)YGet_Q(stack, 0, &dims);
  if (! input) {
    PushDataBlock(RefNC(&nilDB));
  } else {
    number = (size_t)TotalNumber(dims);
    output = (unsigned char **)yeti_push_new_string_array(dims)->value.q;
    for (j=0 ; j<number ; ++j) {
      if (! (src = input[j])) {
	output[j] = 0;
      } else {
	if (flags&STRTRIM_LEFT) {
	  /* Skip leading spaces. */
	  while (isspace(src[0])) ++src;
	}
	len = strlen((char *)src);
	if ((flags&STRTRIM_RIGHT) && len>=1) {
	  /* Discard trailing spaces. */
	  for (i=len-1 ; i>=0 && isspace(src[i]) ; --i)
	    ;
	  len = i+1;
	}
	dst = (unsigned char *)p_stralloc(len);
	dst[len] = 0;
	output[j] = dst;
	memcpy(dst, src, len);
      }
    }
  }
}

extern BuiltIn Y_strlower, Y_strupper;
#define FUNCTION(op)							      \
void Y_str##op(int argc)						      \
{									      \
  Dimension *dims;							      \
  unsigned char **input, **output, *src;				      \
  if (argc != 1) YError("str"#op" takes exactly one argument");		      \
  if (! (input = (unsigned char **)YGet_Q(sp, 0, &dims))) {		      \
    PushDataBlock(RefNC(&nilDB));					      \
  } else {								      \
    int j, number = TotalNumber(dims);					      \
    output = (unsigned char **)yeti_push_new_string_array(dims)->value.q;     \
    for (j=0 ; j<number ; ++j) {					      \
      if (! (src = input[j])) {						      \
	output[j] = 0;							      \
      } else {								      \
	int k, len = strlen((char *)src);				      \
	unsigned char *dst = (unsigned char *)p_stralloc(len);		      \
	dst[len] = 0;							      \
	output[j] = dst;						      \
	for (k=0 ; k<len ; ++k) dst[k] = to##op(src[k]);		      \
      }									      \
    }									      \
  }									      \
}
FUNCTION(lower)
FUNCTION(upper)
#undef FUNCTION

/*---------------------------------------------------------------------------*/
/* DATA ENCODING */

#include "prmtyp.h"

void Y_get_encoding(int argc)
{
  const char *name;
  static struct {
    const char *name;
    long        layout[32];
  } db[] = {
    {"alpha", {1,1,-1, 2,2,-1, 4,4,-1, 8,8,-1, 4,4,-1, 8,8,-1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"cray",  {1,1,1, 8,8,1, 8,8,1, 8,8,1, 8,8,1, 8,8,1,
	       0,1,15,16,48,1,16384, 0,1,15,16,48,1,16384}},
    {"dec",   {1,1,-1, 2,2,-1, 4,4,-1, 4,4,-1, 4,4,-1, 8,8,-1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"i86",   {1,1,-1, 2,2,-1, 4,4,-1, 4,4,-1, 4,4,-1, 8,4,-1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"ibmpc", {1,1,-1, 2,2,-1, 2,2,-1, 4,2,-1, 4,2,-1, 8,2,-1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"mac",   {1,1,1, 2,2,1, 2,2,1, 4,2,1, 4,2,1, 8,2,1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"macl",  {1,1,1, 2,2,1, 2,2,1, 4,2,1, 4,2,1, 12,2,1,
	       0,1,8,9,23,0,127, 0,1,15,32,64,1,16382}},
    {"sgi64", {1,1,1, 2,2,1, 4,4,1, 8,8,1, 4,4,1, 8,8,1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"sun",   {1,1,1, 2,2,1, 4,4,1, 4,4,1, 4,4,1, 8,8,1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"sun3",  {1,1,1, 2,2,1, 4,2,1, 4,2,1, 4,2,1, 8,2,1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"vax",   {1,1,-1, 2,1,-1, 4,1,-1, 4,1,-1, 4,1,2, 8,1,2,
	       0,1,8,9,23,0,129, 0,1,8,9,55,0,129}},
    {"vaxg",  {1,1,-1, 2,1,-1, 4,1,-1, 4,1,-1, 4,1,2, 8,1,2,
	       0,1,8,9,23,0,129, 0,1,11,12,52,0,1025}},
    {"xdr",   {1,1,1, 2,2,1, 4,4,1, 4,4,1, 4,4,1, 8,4,1,
	       0,1,8,9,23,0,127, 0,1,11,12,52,0,1023}},
    {"native", {sizeof(char),   P_STRUCT_ALIGN, 0,
                sizeof(short),  P_SHORT_ALIGN,  P_SHORT_ORDER,
		sizeof(int),    P_INT_ALIGN,    P_INT_ORDER,
		sizeof(long),   P_LONG_ALIGN,   P_LONG_ORDER,
		sizeof(float),  P_FLOAT_ALIGN,  P_FLOAT_ORDER,
		sizeof(double), P_DOUBLE_ALIGN, P_DOUBLE_ORDER,
		P_FLOAT_LAYOUT, P_DOUBLE_LAYOUT}}
  };
  const int ndb = sizeof(db)/sizeof(db[0]);

  if (argc!=1) YError("get_encoding takes exactly one argument");
  name = YGetString(sp);
  if (name) {
    long *result = yeti_push_new_long_array_base(yeti_first_dimension(32,1));
    int i, c = name[0];
    for (i=0 ; i<ndb ; ++i) {
      if (c==db[i].name[0] && ! strcmp(name, db[i].name)) {
	long *layout = db[i].layout;
	for (i=0 ; i<32 ; ++i) result[i] = layout[i];
	return;
      }
    }
  }
  YError("unknown encoding name");
}

/*---------------------------------------------------------------------------*/
/* SYMBOLS */

void Y_typeIDof(int argc)
{
  Symbol *s;
  long typeID;
  if (argc != 1) YError("typeIDof takes exactly one argument");
  s = YETI_DEREF_SYMBOL(sp);
  if (s->ops==&doubleScalar) {
    typeID = T_DOUBLE;
  } else if (s->ops==&longScalar) {
    typeID = T_LONG;
  } else if (s->ops==&intScalar) {
    typeID = T_INT;
  } else if (s->ops==&dataBlockSym) {
    typeID = s->value.db->ops->typeID;
  } else {
    YError("unexpected argument");
  }
  PushLongValue(typeID);
}

void Y_nrefsof(int argc)
{
  Operand op;
  if (argc != 1) YError("nrefsof takes exactly one argument");
  if (! sp->ops) YError("unexpected keyword argument");
  PushLongValue(sp->ops->FormOperand(sp, &op)->references);
}

#if 0
void Y_object_info(int argc)
{
  char line[256];
  for (i=1 ; i<=argc ; ++i) {
    Symbol *s = sp - argc + i;
#define _(x) sprintf x; PrintFunc(line); PrintFunc(line)
    _((line, "arg(%d) = (Symbol *)0x%x", i, (unsigned)s));
    if (s == NULL) continue;
    if (s->ops==&doubleScalar) {
      _((line, "arg(%d)->ops = &doubleScalar", i));
      _((line, "arg(%d)->index = (long)%ld", i, s->index));
      _((line, "arg(%d)->value.d = (double)%g", i, s->value.d));
    } else if (s->ops==&longScalar) {
      _((line, "arg(%d)->ops = &longScalar", i));
      _((line, "arg(%d)->index = (long)%ld", i, s->index));
      _((line, "arg(%d)->value.l = (long)%ld", i, s->value.l));
    } else if (s->ops==&intScalar) {
      _((line, "arg(%d)->ops = &intScalar"));
      _((line, "arg(%d)->index = (long)%ld", i, s->index));
      _((line, "arg(%d)->value.i = (int)%d", i, s->value.i));
    } else if (s->ops==&dataBlockSym) {
      _((line, "arg(%d)->ops = &dataBlockSym", i));
      _((line, "arg(%d)->index = (long)%ld", i, s->index));
      _((line, "arg(%d)->value.db = (DataBlock*)0x%x", i,
	 (unsigned)s->value.db));
      _((line, "arg(%d)->value.db->references = %d", i,
	 (unsigned)s->value.db->references));

      
    } else if (s->ops==&referenceSym) {
      /* In `ydata.h': referenceSym is not a "complete" Symbol in the sense
	 that the parser ensures it never appears in a binary operation
           ops          - &referenceSym
	   index        - to globTab entry
	   value.offset - stack offset (for Return only)  */
      _((line, "arg(%d)->ops = &referenceSym", i));
      _((line, "arg(%d)->index = (long)%ld", i, s->index));
      _((line, "arg(%d)->value.offset = (long)%ld", i, s->value.offset));
    } else if (s->ops==&returnSym) {
      /* In `ydata.h': returnSym is not a "complete" Symbol in the sense
	 that the parser ensures it never appears in a binary operation
           ops      - &returnSym
	   index    - (unused)
	   value.pc - VM program counter (for Return only)  */
      _((line, "arg(%d)->ops = &returnSym", i));
      _((line, "arg(%d)->value.pc = (Instruction *)0x%x", i,
	 (unsigned)s->value.pc));
    } else if (s->ops==NULL) {
      /* Keywords may also appear on the program stack--
	 these are marked by ops==0.  */
      _((line, "arg(%d)->ops= (OpTable*)NULL (KEYWORD)", i));
    } else {
      _((line, "arg(%d)->ops= (OpTable*)0x%x", i, (unsigned)s->ops));
    }
  }
#undef _
  Drop(argc);
}
#endif

void Y_symbol_exists(int argc)
{
  if (argc!=1) YError("symbol_exists takes exactly one argument");
  PushIntValue(HashFind(&globalTable, YGetString(sp), 0L));
}

#define GET_ARRAY       1
#define GET_STRUCT      2
#define GET_RANGE       4
#define GET_VOID        8
#define GET_FUNCTION   16
#define GET_BUILTIN    32
#define GET_STRUCTDEF  64
#define GET_STREAM    128
#define GET_OPAQUE    256
void Y_symbol_names(int argc)
{
  int flags;
  char **ret;
  long i, nitems = globalTable.nItems;
  int match[T_OPAQUE+1];

  if (argc != 1) YError("symbol_list takes exactly one argument");
  flags = yeti_get_flag(sp, (GET_ARRAY|GET_STRUCT|GET_RANGE|GET_FUNCTION|
			     GET_BUILTIN|GET_STRUCTDEF|GET_STREAM|GET_OPAQUE));
  /*Drop(argc+1);*/
  if (nitems > 0) {
    if (flags == -1) {
      ret = yeti_push_new_string_vector_base(nitems);
      for (i=0 ; i<nitems ; ++i) ret[i] = p_strcpy(globalTable.names[i]);
      return;
    } else {
      unsigned int typeID;
      long number = 0;
      for (i=0 ; i<=T_OPAQUE ; ++i) match[i] = 0;
      match[T_CHAR]      = (flags&GET_ARRAY);
      match[T_SHORT]     = (flags&GET_ARRAY);
      match[T_INT]       = (flags&GET_ARRAY);
      match[T_LONG]      = (flags&GET_ARRAY);
      match[T_FLOAT]     = (flags&GET_ARRAY);
      match[T_DOUBLE]    = (flags&GET_ARRAY);
      match[T_COMPLEX]   = (flags&GET_ARRAY);
      match[T_STRING]    = (flags&GET_ARRAY);
      match[T_POINTER]   = (flags&GET_ARRAY);
      match[T_STRUCT]    = (flags&GET_STRUCT);
      match[T_RANGE]     = (flags&GET_RANGE);
#ifdef GET_LVALUE
      match[T_LVALUE]    = (flags&GET_LVALUE);
#endif
      match[T_VOID]      = (flags&GET_VOID);
      match[T_FUNCTION]  = (flags&GET_FUNCTION);
      match[T_BUILTIN]   = (flags&GET_BUILTIN);
      match[T_STRUCTDEF] = (flags&GET_STRUCTDEF);
      match[T_STREAM]    = (flags&GET_STREAM);
      match[T_OPAQUE]    = (flags&GET_OPAQUE);

#define MATCH(i) (globTab[i].ops == &dataBlockSym \
                  && (typeID = globTab[i].value.db->ops->typeID) <= T_OPAQUE \
	          && match[typeID])
      for (i=0 ; i<nitems ; ++i) {
	if (MATCH(i)) ++number;
      }
      if (number) {
	ret = yeti_push_new_string_vector_base(number);
	for (i=0 ; i<nitems ; ++i) {
	  if (MATCH(i)) *ret++ = p_strcpy(globalTable.names[i]);
	}
	return;
      }
#undef MATCH
    }
  }
  PushDataBlock(RefNC(&nilDB));
}
#undef GET_ARRAY
#undef GET_STRUCT
#undef GET_RANGE
#undef GET_VOID
#undef GET_FUNCTION
#undef GET_BUILTIN
#undef GET_STRUCTDEF
#undef GET_STREAM
#undef GET_OPAQUE

/*---------------------------------------------------------------------------*/
/* VARIABLES */

void Y_unref(int argc)
{
  if (argc!=1) YError("unref takes exactly one argument");
  if (sp->ops == &referenceSym) {
    /* Replace reference without augmenting the number of references of
       the data block object if it is an array. */
    Symbol *ref = &globTab[sp->index];
    OpTable *ops = ref->ops;
    if (ops == &dataBlockSym) {
      DataBlock *db = ref->value.db;
      if (db && db->ops->isArray) {
	/* Replace symbol in global table by nil. */
	ref->value.db = RefNC(&nilDB);
	sp->value.db = db; /* no Ref */
      } else {
	sp->value.db = Ref(db);
      }
    } else {
      sp->value = ref->value;
    }
    sp->ops = ops; /* change ops only AFTER value updated */
  } /* NOT_NEEDED: else if (!sp->ops) YError("unexpected keyword argument"); */
}

void Y_swap(int argc)
{
  SymbolValue a_val, b_val;
  OpTable *a_ops, *b_ops;
  volatile Symbol *a_sym, *b_sym;
  if (argc!=2) YError("swap takes exactly 2 arguments");
  a_sym = sp;
  b_sym = sp-1;
  if (a_sym->ops != &referenceSym || b_sym->ops != &referenceSym)
    YError("arguments must be simple variable references");
  a_sym = &globTab[a_sym->index];
  a_ops = a_sym->ops;
  a_val = a_sym->value;
  a_sym->ops = &intScalar;
  b_sym = &globTab[b_sym->index];
  b_ops = b_sym->ops;
  b_val = b_sym->value;
  b_sym->ops = &intScalar;
  b_sym->value = a_val;
  a_sym->value = b_val;
  Drop(2);
  b_sym->ops = a_ops;
  a_sym->ops = b_ops;
}

/*---------------------------------------------------------------------------*/
/* PATHS */

void Y_current_include(int argc)
{
  if (argc != 1 || YNotNil(sp))
    YError("current_include takes exactly one nil argument");
  if (nYpIncludes > 0 && ypIncludes[nYpIncludes-1].filename) {
    yeti_push_string_value(ypIncludes[nYpIncludes-1].filename);
  } else {
    PushDataBlock(RefNC(&nilDB));
  }
}

void Y_get_includes(int argc)
{
  if (argc != 1 || YNotNil(sp))
    YError("get_includes takes exactly one nil argument");
  if (sourceTab.nItems > 0) {
    int i;
    char **s = yeti_push_new_string_vector_base(sourceTab.nItems);
    for (i=0 ; i<sourceTab.nItems ; ++i) s[i] = p_strcpy(sourceTab.names[i]);
  } else {
    PushDataBlock(RefNC(&nilDB));
  }
}

void Y_get_path(int argc)
{
  if (argc != 1 || YNotNil(sp))
    YError("get_path takes exactly one nil argument");
  if (nYpPrefixes > 0) {
    int i;
    char **s = yeti_push_new_string_vector_base(nYpPrefixes);
    for (i=0 ; i<nYpPrefixes ; ++i) s[i] = p_strcpy(ypPrefixes[i]);
  } else {
    PushDataBlock(RefNC(&nilDB));
  }
}

void Y_expand_path(int argc)
{
  Dimension *dims;
  char **inp;
  if (argc != 1) YError("expand_path takes exactly one argument");
  inp = YGet_Q(sp, 1, &dims);
  if (inp) {
    long i, number = TotalNumber(dims);
    char **out = yeti_push_new_string_array_base(dims);
    for (i=0 ; i<number ; ++i) {
      if (inp[i]) out[i] = YExpandName(inp[i]);
    }
  } else {
    PushDataBlock(RefNC(&nilDB));
  }
}

/*---------------------------------------------------------------------------*/
/* SMOOTHING */

static void smooth_single(double *x, double p25, double p50, double p75,
			  long n1, long n2, long n3);

void Y_smooth3(int argc)
{
  Operand op;
  double *x = NULL;
  long n1, n2, n3;
  int single = 0, is_complex;
  long which = 0; /* avoid compiler warning */
  Symbol *stack;
  Dimension *dims;
  int nparsed=0;
  double p25=0.25, p50=0.50, p75=0.75;

  for (stack=sp-argc+1 ; stack<=sp ; ++stack) {
    if (stack->ops) {
      /* non-keyword argument */
      if (++nparsed == 1) {
	stack->ops->FormOperand(stack, &op);
      } else {
	YError("too many arguments");
      }
    } else {
      /* keyword argument */
      const char *keyword = globalTable.names[stack->index];
      ++stack;
      if (keyword[0] == 'c' && keyword[1] == 0) {
	if (YNotNil(stack)) {
	  p50 = YGetReal(stack);
	  p25 = 0.5*(1.0 - p50);
	  p75 = 0.5*(1.0 + p50);
	}
      } else if (keyword[0] == 'w' && ! strcmp(keyword, "which")) {
	if (YNotNil(stack)) {
	  which = YGetInteger(stack);
	  single = 1;
	}
      } else {
	YError("unknown keyword");
      }
    }
  }
  if (nparsed != 1) YError("bad number of arguments");

  /* Get input array. */
  is_complex = (op.ops->typeID == T_COMPLEX);
  n1 = (is_complex ? 2*op.type.number : op.type.number);
  stack = op.owner;
  switch (op.ops->typeID) {
  case T_CHAR:
  case T_SHORT:
  case T_INT:
  case T_LONG:
  case T_FLOAT:
    /* Convert input in a new array of double's. */
    op.ops->ToDouble(&op);
    x = op.value;
    dims = op.type.dims;
    break;

  case T_DOUBLE:
  case T_COMPLEX:
    /* If input array has references (is not temporary), make a new copy. */
    if (op.references) {
      Array *array = NewArray((is_complex ? &complexStruct : &doubleStruct),
			      op.type.dims);
      PushDataBlock(array);
      x = array->value.d;
      dims = array->type.dims;
      memcpy(x, op.value, n1*sizeof(double));
      PopTo(stack);
    } else {
      x = op.value;
      dims = op.type.dims;
    }
    break;

  default:
    YError("bad data type for input array");
  }
  while (sp != stack) Drop(1);  /* left result on top of the stack */

  /* Apply operator. */
  n3 = 1; /* product of dimensions after current one */
  if (single) {
    /* Apply operator along a single dimension. */
    Dimension *tmp = dims;
    long rank=0;
    while (tmp) {
      ++rank;
      tmp = tmp->next;
    }
    if (which <= 0) which += rank;
    if (which <= 0 || which > rank) YError("WHICH is out of range");
    while (dims) {
      n2 = dims->number;
      n1 /= n2;
      if (rank-- == which) {
	smooth_single(x, p25, p50, p75, n1, n2, n3);
	break;
      }
      n3 *= n2;
      dims = dims->next;
    }
  } else {
    /* Apply operator to every dimensions. */
    while (dims) {
      n2 = dims->number;
      n1 /= n2;
      smooth_single(x, p25, p50, p75, n1, n2, n3);
      n3 *= n2;
      dims = dims->next;
    }
  }
}

static void smooth_single(double *x, double p25, double p50, double p75,
			  long n1, long n2, long n3)
{
  if (n2 >= 2) {
    long i, stride = n1, n = n1*n2;
    double x1, x2, x3;
    if (stride == 1) {
      for ( ; --n3>=0 ; x+=n) {
	x2 = x[0];
	x3 = x[1];
	x[0] = p75*x2 + p25*x3;
	for (i=2 ; i<n ; ++i) {
	  x1 = x2;
	  x2 = x3;
	  x3 = x[i];
	  x[i - 1] = p50*x2 + p25*(x1 + x3);
	}
	x[n - 1] = p75*x3 + p25*x2;
      }
    } else {
      long p = n - stride;
      for ( ; --n3>=0 ; x+=p) {
	for (n1=stride ; --n1>=0 ; ++x) {
	  x2 = x[0];
	  x3 = x[stride];
	  x[0] = p75*x2 + p25*x3;
	  for (i=2*stride ; i<n ; i+=stride) {
	    x1 = x2;
	    x2 = x3;
	    x3 = x[i];
	    x[i - stride] = p50*x2 + p25*(x1 + x3);
	  }
	  x[n - stride] = p75*x3 + p25*x2;
	}
      }
    }
  }
}

/*---------------------------------------------------------------------------*/
