/*
 * yeti_play.c --
 *
 *	Make some routines of PLAY (Portable LAYer) available in
 *	Yorick interpreter.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2003-2005, Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_play.c,v 1.1 2005/05/24 13:20:29 eric Exp $
 *	$Log: yeti_play.c,v $
 *	Revision 1.1  2005/05/24 13:20:29  eric
 *	Initial revision
 *
 */

#include "play.h"
#include "pstdlib.h"
#include "ydata.h"

extern BuiltIn Y_set_alarm;

static void check_symbol_name(const char *name);

typedef struct alarm_context alarm_context_t;
struct alarm_context {
  alarm_context_t *next;
  Function *task;  /* pointer to function or NULL */
  long      index; /* index in globTab[] or -1 */
  double    time;
};

static alarm_context_t *alarm_next = 0;
static alarm_context_t *alarm_free = 0;

static void alarm_callback(void *context);

static void alarm_free_context(alarm_context_t *this)
{
  Function *task = this->task;
  alarm_next = this->next;
  this->task = 0; /* fix alloc below */
  this->next = alarm_free;
  /*if (task)*/ Unref(task);
}

static void alarm_callback(void *context)
{
  Function *task;
  alarm_context_t *this = (alarm_context_t *)context;
  
  if (! (task = this->task) && this->index >= 0) {
    Symbol *sym = &globTab[this->index];
    if (sym->ops == &dataBlockSym) {
      int typeID = sym->value.db->ops->typeID;
      if (typeID == T_FUNCTION || typeID == T_BUILTIN)
	task = (Function *)sym->value.db;
    }
  }
  this->task = 0;
  alarm_free_context(this);
  if (task) RunTaskNow(task); /* unref? */
}

static void check_symbol_name(const char *name)
{
  static int char_type[256], first_time = 1;
  int i;
  if (first_time) {
    char_type[0] = 0;
    for (i=1 ; i<256 ; ++i) char_type[i] = -1;
    char_type['_'] = 1;
    for (i='a' ; i<='z' ; ++i) char_type[i] = 1;
    for (i='A' ; i<='Z' ; ++i) char_type[i] = 1;
    for (i='0' ; i<='9' ; ++i) char_type[i] = 2;
    first_time = 0;
  }
  if (name) {
    unsigned char *s = (unsigned char *)name;
    if (char_type[*s++] == 1) {
      for (;;) {
	if ((i = char_type[*s++]) < 0) break;
	if (i == 0) return;
      }
    }
  }
  YError("invalid symbol's name");
}

void Y_set_alarm(int nargs)
{
  alarm_context_t  *this;
  alarm_context_t  *next = alarm_next;
  alarm_context_t **prev = &alarm_next;
  double secs, time;
  Function *task = 0;
  Symbol *s;
  Operand op;
  int typeID;
  long index = -1;
  char *name;


  if (nargs != 2) YError("set_alarm takes exactly 2 arguments");
  secs = YGetReal(sp - nargs + 1);
  time = p_wall_secs() + secs;
  s = sp - nargs + 2;
  if (! s->ops) YError("unexpected keyword argument");
  typeID = s->ops->FormOperand(s, &op)->ops->typeID;
  if (typeID == T_STRING) {
    if (op.type.dims) YError("expecting scalar string argument");
    name = *(char **)op.value;
    check_symbol_name(name);
    index = Globalize(name, 0L);
  } else if (typeID == T_FUNCTION || typeID == T_BUILTIN) {
    task = (Function *)s->value.db;
  } else {
    YError("expecting function or function name");
  }
 
  if (! alarm_free) {
    int i, n = 8;
    alarm_context_t *new = p_malloc(n*sizeof(alarm_context_t));
    new[--n].next = 0;
    for (i=0 ; i<n ; ++i) new[i].next = &new[i + 1];
    alarm_free = new;
  }
  this = alarm_free;
  this->index = index;
  this->task = 0;
  this->time = time;
  /* insert THIS into alarm_next list, kept in order of time */
  while (next && next->time <= time) {
    prev = &next->next;
    next = next->next;
  }
  alarm_free = alarm_free->next;
  this->next = next;
  *prev = this;

  if (task) this->task = Ref(task);
  
  p_set_alarm(secs, alarm_callback, this);
  PushDoubleValue(time); 
}
