/*
 * yeti_tcl.c --
 *
 *	Implement client of Tcl/Tk server in Yeti.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2001-2002 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *	
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *	
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *	
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_tcl.c,v 1.1 2005/05/24 13:26:09 eric Exp $
 *	$Log: yeti_tcl.c,v $
 *	Revision 1.1  2005/05/24 13:26:09  eric
 *	Initial revision
 *
 */

#include "yeti.h"
#if (YORICK_VERSION_MAJOR == 1) && (YORICK_VERSION_MINOR <= 4)
# include <stdio.h>
#else
# include "pstdio.h"
#endif
#include <ctype.h>
#include "yio.h"
#include "tclsrv.h"

/*--------------------------------------------------------------------------*/
/* DEFINITIONS */

/* Define the following macro to non-zero value if you want get_client()
   to replace reference symbol onto the stack. */
#define REPLACE_REFERENCE 0

extern BuiltIn Y_tcl_create, Y_tcl_connect, Y_tcl_connected, Y_tcl_disconnect;
extern BuiltIn Y_tcl_path, Y_tcl_status, Y_tcl_result, Y_tcl_send;
extern BuiltIn Y_tcl_timeout;

typedef struct _client_t client_t;

static client_t *get_client(Symbol *stack);
/*----- Returns i/o client table stored by symbol STACK.  STACK get
	replaced by the referenced object if it is a reference symbol. */

struct _client_t {
  int references;         /* reference counter */
  Operations *ops;        /* virtual function table */
  long ms;                /* timeout for (re)connection in milliseconds */
  tclsrv_client_t *client;
};

/*--------------------------------------------------------------------------*/
/* IMPLEMENTATION OF TCL-CLIENTS AS OPAQUE YORICK OBJECTS */

extern PromoteOp PromXX;
extern UnaryOp ToAnyX, NegateX, ComplementX, NotX, TrueX;
extern BinaryOp AddX, SubtractX, MultiplyX, DivideX, ModuloX, PowerX;
extern BinaryOp EqualX, NotEqualX, GreaterX, GreaterEQX;
extern BinaryOp ShiftLX, ShiftRX, OrX, AndX, XorX;
extern BinaryOp AssignX, MatMultX;
extern UnaryOp EvalX, SetupX, PrintX;
extern MemberOp GetMemberX;
static UnaryOp PrintClient;
static void FreeClient(void *addr);  /* ******* Use Unref(obj) ******* */

Operations tclclientOps = {
  &FreeClient, T_OPAQUE, 0, T_STRING, "tcl_client",
  {&PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX},
  &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX,
  &NegateX, &ComplementX, &NotX, &TrueX,
  &AddX, &SubtractX, &MultiplyX, &DivideX, &ModuloX, &PowerX,
  &EqualX, &NotEqualX, &GreaterX, &GreaterEQX,
  &ShiftLX, &ShiftRX, &OrX, &AndX, &XorX,
  &AssignX, &EvalX, &SetupX, &GetMemberX, &MatMultX, &PrintClient
};

/* FreeClient is automatically called by Yorick to delete an object instance
   that is no longer referenced. */
static void FreeClient(void *addr)
{
  client_t *this = (client_t *)addr;
  tclsrv_destroy(this->client);
  p_free(this);
}

/* PrintClient is used by Yorick's info command. */
static void PrintClient(Operand *op)
{
  client_t *this = (client_t *)op->value;
  char line[80];
  ForceNewline();
  PrintFunc("Object of type: ");
  PrintFunc(this->ops->typeName);
  sprintf(line, " (references=%d, status=%d, ms=%ld, path=\"",
	  this->references, tclsrv_status(this->client), this->ms);
  PrintFunc(line);
  PrintFunc(tclsrv_path(this->client));
  PrintFunc("\")");
  ForceNewline();
}

/*--------------------------------------------------------------------------*/
/* BUILTIN FUNCTIONS */

void Y_tcl_create(int argc)
{
  const char *path;
  client_t *this;
  
  if (argc != 1) YError("tcl_create takes exactly one argument");
  path = YGetString(sp);
  this = p_malloc(sizeof(client_t));
  this->references = 0;
  this->ops = &tclclientOps;
  this->ms = 1000; /* default timeout of 1 sec. for (re)connection */
  this->client = NULL;
  PushDataBlock(this);
  this->client = tclsrv_create(path);
  if (! this->client) YError("insufficient memory or invalid socket name");
}

void Y_tcl_connect(int argc)
{
  client_t *this;
  if (argc < 1 || argc > 2) YError("tcl_connect takes one or two arguments");
  this = get_client(sp-argc+1);
  if (argc >= 2) {
    this->ms = YGetInteger(sp-argc+2);
    Drop(1);
  }
  PushLongValue((long)tclsrv_connect(this->client, this->ms));
}

void Y_tcl_disconnect(int argc)
{
  client_t *this;
  if (argc != 1) YError("tcl_disconnect takes exactly one argument");
  this = get_client(sp);
  PushLongValue((long)tclsrv_disconnect(this->client));
}

void Y_tcl_connected(int argc)
{
  if (argc != 1) YError("tcl_connected takes exactly one argument");
  PushIntValue(tclsrv_connected(get_client(sp)->client));
}

void Y_tcl_status(int argc)
{
  if (argc != 1) YError("tcl_status takes exactly one argument");
  PushLongValue((long)tclsrv_status(get_client(sp)->client));
}

void Y_tcl_timeout(int argc)
{
  long result;
  client_t *this;
  if (argc < 1 || argc > 2) YError("tcl_timeout takes 1 or 2 arguments");
  this = get_client(sp-argc+1);
  result = this->ms;
  if (argc >= 2) this->ms = YGetInteger(sp-argc+2);
  PushLongValue(result);
}

void Y_tcl_path(int argc)
{
  if (argc != 1) YError("tcl_path takes exactly one argument");
  yeti_push_string_value(tclsrv_path(get_client(sp)->client));
}

void Y_tcl_result(int argc)
{
  if (argc != 1) YError("tcl_connected takes exactly one argument");
  yeti_push_string_value(tclsrv_result(get_client(sp)->client));
}

void Y_tcl_send(int argc)
{
  long ms;
  client_t *this;
  const char *script;
  if (argc < 2 || argc > 3) YError("tcl_send takes 2 or 3 arguments");
  this = get_client(sp-argc+1);
  script = YGetString(sp-argc+2);
  if (argc >= 3) {
    ms = YGetInteger(sp-argc+3);
    Drop(1);
  } else {
    ms = -1;
  }
#if 1
  PushLongValue((long)tclsrv_send(this->client, ms, script));
#else
  if (! tclsrv_connected(this->client)) YError("not connected");
  if (tclsrv_send(this->client, ms, script) != TCLSRV_OK) {
    YError(tclsrv_result(this->client));
  }
  yeti_push_string_value(tclsrv_result(this->client));
#endif
}

/*---------------------------------------------------------------------------*/
/* PRIVATE UTILITY ROUTINES */

static client_t *get_client(Symbol *stack)
{
  DataBlock *db;
  Symbol *sym = (stack->ops == &referenceSym) ? &globTab[stack->index] : stack;
  if (sym->ops != &dataBlockSym || sym->value.db->ops != &tclclientOps)
    YError("expected Tcl/Tk client object");
  db = sym->value.db;
#if REPLACE_REFERENCE
  if (sym != stack) {
    /* Replace reference onto the stack (equivalent to the statement
       ReplaceRef(s); see ydata.c for actual code of this routine). */
    stack->value.db = Ref(db);
    stack->ops = &dataBlockSym;     /* change ops only AFTER value updated */
  }
#endif
  return (client_t *)db;
}

/*---------------------------------------------------------------------------*/
