/*
 * yeti_tcl.i --
 *
 *	Invoke Tcl/Tk interpreter to have some kind of widget server.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2001-2002 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *	
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *	
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *	
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_tcl.i,v 1.1 2005/05/24 13:26:15 eric Exp $
 *	$Log: yeti_tcl.i,v $
 *	Revision 1.1  2005/05/24 13:26:15  eric
 *	Initial revision
 *
 */

/* ALTERNATIVES TO TCL/TK SERVER:
 *   Commands are send to TCL interpreter via a pipe (popen) and the result
 *   of the command is read from a FIFO.  Thus the tcl_send routine will
 *   wait until a response is available.  It is however possible to detect
 *   that TCL has closed its side of the FIFO (rdline returns nil) and to
 *   re-open it as needed.  It seems that the only way to realize that
 *   TCL has exited is to attempt to write to the pipe and flush it but this
 *   raise an SIGIO signal so I need to use catch(0x02).
 *   Synchronisation is the hard part...
 *
 *   Another possibility is to implement a "spawn" opaque object which can
 *   be read/write by Yorick.  The advantages are that it applies to other
 *   Unix commands.  It would be however very difficult to have more
 *   sophisticated input/output (as Yorick text streams for instance) than
 *   just reading/writing a string to/from this object.  The other drawback
 *   is that is difficult to have a non-blocking communication channel
 *   and to make sure that the returned result is complete...
 *
 * TO DO LIST:
 *   - Find a way to kill the server upon Yorick quitting.
 *   - Improve the management of already running servers.
 *   - Write more wrappers to common Tcl/Tk facilities.
 *   - Tk server should do a "wm withdraw ." automatically.
 */

/*---------------------------------------------------------------------------*/
/* AUTOMATIC CONNECTION TO A SINGLE SERVER */

local __tcl_client;
local __tcl_tmpdir;
__tcl_tmpdir = "/tmp";
func tcl_eval(..)
/* DOCUMENT tcl_eval(..)
     Evaluate Tcl script made by  concatenating all arguments with a single
     space  separator into  a scalar  string.   The returned  value is  the
     result string returned  by the server.  This function  is a high level
     wrapper which is able to automatically launch a Tcl/Tk server.
     
   SEE ALSO tcl_send, tcl_start. */
{
  /* Build script. */
  glue = script = string(0);
  while (more_args()) {
    arg = next_arg();
    if ((s = structof(arg)) != string) {
      if (s==long || s==int || s==short || s==char) {
        arg = swrite(format="%d", arg);
      } else if (s==double || s==float) {
        arg = swrite(format="%g", arg);
        n = 0;
        c = string(0);
        if (sread(arg, format="%d%[.eE]", n, c) != 2) arg += ".0";
      } else if (! is_void(arg)) {
        error, "unexpected data type";
      }
    }
    if (script) {
      script += glue + arg;
    } else {
      eq_nocopy, script, arg;
      glue = " ";
    }
  }
  if (is_void(__tcl_client)) tcl_start;
  if (tcl_send(__tcl_client, script)) error, tcl_result(__tcl_client);
  return tcl_result(__tcl_client);
}

func tcl_start(ms)
/* DOCUMENT tcl_start;
       -or- tcl_start, ms;
     (Re)starts cached Tcl/Tk client used by tcl_eval.  Returns the client
     object when called as a function (this object is cached in global
     variable __tcl_client).

   SEE ALSO: tcl_create, tcl_connect, tcl_eval, tcl_finish. */
{
  /* fool codger */ extern __tcl_client, __tcl_tmpdir;

  /* If client already exists, try to use it, possibly after
     re-connection. */
  if (is_void(ms)) ms = 10000;
  if (! is_void(__tcl_client)) {
    if (tcl_connected(__tcl_client) || ! tcl_connect(__tcl_client, ms))
      return __tcl_client;
    __tcl_client = [];
  }

  /* Try to connect to existing servers. */
  ident = 0;
  dummy = string(0);
  list = lsdir(__tcl_tmpdir);
  fmt1 = __tcl_tmpdir + "/tksrv-" + get_env("USER") + ":%d";
  fmt2 = fmt1 + "%[^0-9]";
  n = numberof(list);
  for (i=1 ; i<=n ; ++i) {
    path = list(i);
    if (sread(format=fmt2, path, ident, dummy) == 1) {
      __tcl_client = tcl_create(path);
      if (! tcl_connect(__tcl_client, ms)) {
        tcl_send, __tcl_client, "wm widthdraw .";
        return __tcl_client;
      }
    }
  }

  /* Launch a new server. */
  tmax = 10000;  /* maximum delay (milliseconds) */
  dt = 10;       /* delay step (milliseconds) */
  ident = 0;
  for (;;) {
    path = swrite(format=fmt1, ident++);
    if (noneof(path == lsdir(__tcl_tmpdir))) {
      system, "tksrv -server " + path + " &";
      __tcl_client = tcl_create(path);
      for (t=0 ; t<=tmax ; t+=dt) {
        if (! tcl_connect(__tcl_client, ms)) {
          tcl_send, __tcl_client, "wm widthdraw .";
          return __tcl_client;
        }
        pause, dt;
      }
      error, swrite(format="timeout (%.3f sec) while waiting for server",
                    tmax*1e-3);
    }
  }
}

func tcl_finish
/* DOCUMENT tcl_finish
     Shutdown Tcl/Tk server started by tcl_start.

   SEE ALSO: tcl_start, tcl_eval. */
{
  if (! is_void(__tcl_client)) {
    tcl_send, __tcl_client, "exit";
    __tcl_client = [];
  }
}

func __tcl_get_file(op)
/* DOCUMENT __tcl_get_file(op);
     Worker routine for tcl_get_open_file and tcl_get_save_file.

   SEE ALSO  tcl_get_open_file, tcl_get_save_file. */
{
  /* Declare all keywords of the caller as 'external' (not really needed,
     but just to be clean). */
  /* fool codger */ extern initialdir, defaultextension, filetypes,
                      initialfile, multiple, parent, title;

  /* Build Tcl script. */
  script = swrite(format="%s -initialdir {%s}", op,
                  (is_void(initialdir) ? get_cwd() : initialdir));
  if (multiple) script += " -multiple";
  if (! is_void(defaultextension))
    script += " -defaultextension {" + defaultextension + "}";
  if (! is_void(filetypes))   script += " -filetypes {"   + filetypes   + "}";
  if (! is_void(initialfile)) script += " -initialfile {" + initialfile + "}";
  if (! is_void(parent))      script += " -parent {"      + parent      + "}";
  if (! is_void(title))       script += " -title {"       + title       + "}";
  return tcl_eval(script);
}

func tcl_get_open_file(dummy, defaultextension=, filetypes=, initialdir=,
                       initialfile=, multiple=, parent=, title=)
{ return __tcl_get_file("tk_getOpenFile"); }
func tcl_get_save_file(dummy, defaultextension=, filetypes=, initialdir=,
                       initialfile=, multiple=, parent=, title=)
{ return __tcl_get_file("tk_getSaveFile"); }
/* DOCUMENT tcl_get_open_file();
       -or- tcl_get_save_file();
     Pops up a file dialog box for the user to select a file to open or
     save and returns the user selection (which can be empty if the
     "cancel" button is clicked).

     The following keywords are 
       DEFAULTEXTENSION = String  that will be appended to  the filename if
           the  user enters  a filename  without an  extension.  The defaut
           value  is the  empty string,  which means  no extension  will be
           appended to the filename in any 
       FILETYPES = A list of file patterns, each pattern is of the form:
             TYPE {EXTENSION [..]} [{MACTYPE [...]}]
           e.g.:
             "{{{Text Files}      {.txt}        }
              {{TCL Scripts}      {.tcl}        }
              {{C Source Files}   {.c}      TEXT}
              {{GIF Files}        {.gif}        }
              {{GIF Files}        {}        GIFF}
              {{All Files}        *             }}"
       INITIALDIR = Name of starting directory when the dialog pops up.  By
           default,  the  files  in   the  current  working  directory  are
           displayed.
       INITIALFILE = Default filename to be displayed in the dialog when it
           pops up.
       MULTIPLE = If true (non-zero and non-nil), allows the user to choose
           multiple files from the Open dialog.           
       PARENT  = Name  of logical  parent window  of the  dialog  box.  The
           dialog box is displayed on top of its parent window.
       TITLE =  String to display as the  title of the dialog  box. If this
           option is not specified, then a default title is displayed.

   SEE ALSO: tcl_eval. */

func tcl_choose_directory(dummy, initialdir=, parent=, title=, mustexist=)
/* DOCUMENT tcl_choose_directory();
*/
{
  script = swrite(format="tk_chooseDirectory -mustexist %d -initialdir {%s}",
                  (mustexist ? 1 : 0),
                  (is_void(initialdir) ? get_cwd() : initialdir));
  if (! is_void(parent)) script += " -parent {"  + parent  + "}";
  if (! is_void(title))  script += " -title {"   + title   + "}";
  return tcl_eval(script);
}

func tcl_message_box(message, default=, icon=, parent=, title=, type=)
/* DOCUMENT tcl_message_box(msg)
     Pops up a message window and returns for user response.  MSG is the
     message to display in the message box.  The returned value is the
     symbolic name of the selected button.

     The following keywords are available:
       DEFAULT = Symbolic name of the default button for this message box
              ("ok", "cancel", and so on).  See keyword TYPE for a list of
              the symbolic names.  If this keyword is not specified, the
              first button in the dialog will be made the default.
       ICON = Name of icon to display. The value must be one of: "error",
              "info", "question" or "warning".  If this option is not
              specified, then the info icon will be displayed.
       PARENT = Name of logical parent window of the message box.  The
              message box is displayed on top of its parent window.
       TITLE = Text string to display as the title of the message box. The
              default value is an empty string.
       TYPE = Message box type, one of:
              Type:               Buttons:
              ------------------  -------------------------
              "abortretryignore"  "abort" "retry"  "ignore"
              "ok"                "ok"
              "okcancel"          "ok"    "cancel"
              "retrycancel"       "retry" "cancel"
              "yesno"             "yes"   "no"
              "yesnocancel"       "yes"   "no"     "cancel"

 SEE ALSO: tcl_send, tcl_dialog. */
{
  script = "tk_messageBox";
  if (! is_void(message)) script += " -message {" + message + "}";
  if (! is_void(default)) script += " -default {" + default + "}";
  if (! is_void(icon))    script += " -icon {"    + icon    + "}";
  if (! is_void(parent))  script += " -parent {"  + parent  + "}";
  if (! is_void(title))   script += " -title {"   + title   + "}";
  if (! is_void(type))    script += " -type {"    + type    + "}";
  return tcl_eval(script);
}

func tcl_strip_braces(s)
{
  strip = (strpart(s, 1:1) == "{")&(strpart(s, 0:0) == "}");
  i = where(strip);
  if (! is_array(i)) return s;
  if (numberof(i) == numberof(s)) return strpart(s, 2:-1);
  sp = array(string, dimsof(s));
  sp(i) = strpart(s(i), 2:-1);
  i = where(! strip);
  sp(i) = s(i);
  return sp;
}

func tcl_list(..)
/* DOCUMENT tcl_list(..)
     Returns a scalar string which is suitable to be used as a list for
     a Tcl/Tk command.  Any argument is converted to a string (if possible)
     and braces are used to protect elements.  For instance:
       tcl_list("hello","peoples","","of this",["fine","World"])
     yields:
       "{{hello} {peoples} {} {of this} {{fine} {World}}}"

     IMPORTANT: any curly brace in the arguments must be protected by
     a backslash.
     
   SEE ALSO: tcl_eval. */
{
  list = "{";
  glue = "";
  while (more_args()) {
    arg = next_arg();
    if (is_void(arg)) {
      list += glue + "{}";
      glue = " ";
    } else {
      if ((s = structof(arg)) == string) {
        arg = "{" + arg + "}";
      } else if (s==long || s==int || s==short || s==char) {
        arg = swrite(format="%d", arg);
      } else if (s==double || s==float) {
        arg = swrite(format="%g", arg);
        i = where(!(strmatch(arg, ".")|strmatch(arg, "e")|strmatch(arg, "E")));
        if (is_array(i)) arg(i) += ".0";
      } else {
        error, "unexpected data type";
      }
      number = numberof(arg);
      if (number == 1) {
        list += glue + arg(i);
        glue = " ";
      } else {
        /* Build up tail list to avoid growing the large list */
        tail = glue + "{";
        glue = "";
        for (i=1 ; i<=number ; ++i) { 
          tail += glue + arg(i);
          glue = " ";
        }
        list += tail + "}";
      }
    }
  }
  return list+"}";
}

/*---------------------------------------------------------------------------*/
/* LOW-LEVEL LAYER */

/*   This layer should trigger no Yorick errors (except if invalid arguments)
     but most functions return Tcl/Tk status.  These functions are able to
     deal with any number of servers (one client for each server). */

extern tcl_create;
/* DOCUMENT: tcl_create(path)
     Create a new (unconnected) Tcl/Tk client; PATH is the name of the
     local socket used by the server to listen for connections.  This
     function returns an opaque object; when this object is no longer
     referenced, the object get destroyed and the connection is closed.
     You must use tcl_connect to effectively attempt the connection with
     the server.  Note that after its creatiuon the timeout for CLIENT for
     further connect/send operation is set to 1 second.  This can be
     modified by, tcl_connect, tcl_send,  and/or tcl_timeout.

   SEE ALSO: tcl_connect, tcl_connected, tcl_disconnect, tcl_path, tcl_send,
             tcl_status. */

extern tcl_timeout;
/* DOCUMENT tcl_timeout(client)
       -or- tcl_timeout(client, ms)
     Returns the timeout value of CLIENT.  In the second form set the
     timeout for further interaction with the server and returns the
     timeout value prior to the change.
     
       if MS < 0, wait forever until connection accepted or refused;
       if MS = 0, asynchronous connection: connection will be completed
                  later (prior to the first successful tclsrv_send);
       if MS > 0, maximum number of milliseconds to wait for connection
                  to complete.

   SEE ALSO: tcl_connect. */

extern tcl_connect;
extern tcl_disconnect;
/* DOCUMENT tcl_connect(client)
       -or- tcl_connect(client, ms)
       -or- tcl_disconnect(client)
     tcl_connect connects or re-connects CLIENT to its server; MS is a
     timeout in milliseconds for the connection (if MS is not specified,
     the previous timeout setting for this client is kept).  tcl_disconnect
     closes the connection which can be re-open later with tcl_connect.
     These functions return CLIENT status (TCL_OK or TCL_ERROR) after the
     operation.  Since no Yorick error is raised, you'll have to check this
     status or call tcl_connected to make sure the connection was
     effective.

   SEE ALSO: tcl_create, tcl_connected, tcl_path,
             tcl_send, tcl_status, tcl_timeout. */

extern tcl_connected;
/* DOCUMENT tcl_connected(client)
     Returns 1 if CLIENT is connected; 0 otherwise.
   SEE ALSO: tcl_connect. */

extern tcl_path;
/* DOCUMENT tcl_path(client)
     Returns the path of the Unix socket attached to CLIENT.
   SEE ALSO: tcl_create. */

extern tcl_send;
/* DOCUMENT tcl_send(script)
       -or- tcl_send(script, ms)       
     Sends scalar string SCRIPT to Tcl/Tk interpreter and returns the
     status returned by the server (see tcl_status for possible returned
     values).  If MS is specified, a timeout of MS milliseconds is applied;
     otherwise the function waits forever the answer of the server.  The
     string result of the script can be retrieved by tcl_result(CLIENT).
     Inputs SCRIPT must not have embedded newlines since they are used as
     command separator by the Tcl/Tk interpreter.

   SEE ALSO: tcl_result, tcl_status. */

extern tcl_result;
/* DOCUMENT tcl_result(client)
     Returns the textual result stored into CLIENT.
   SEE ALSO: tcl_send, tcl_status. */

local TCL_OK;
local TCL_ERROR;
local TCL_RETURN;
local TCL_BREAK;
local TCL_CONTINUE;
local TCL_INTERRUPTED;
extern tcl_status;
/* DOCUMENT tcl_status(client)
     Returns the current status of CLIENT, one of:
       TCL_OK          - status upon success;
       TCL_ERROR       - status upon failure or when script terminated
                         by an "error" command;
       TCL_RETURN      - script terminated by a "return" command;
       TCL_BREAK       - script terminated by a "break" command;
       TCL_CONTINUE    - script terminated by a "continue" command;
       TCL_INTERRUPTED - status when send/receive interrupted (e.g. by Ctrl-C).

   SEE ALSO: tcl_send, tcl_result. */
TCL_OK          =  0;
TCL_ERROR       =  1;
TCL_RETURN      =  2;
TCL_BREAK       =  3;
TCL_CONTINUE    =  4;
TCL_INTERRUPTED = -1;

/*---------------------------------------------------------------------------*/
